@@ -6,7 +6,7 @@
 -- or the test suite. (Eventually the test suite will be run in a worker process,
 -- so this wouldn't be a separate case to consider)
 
-local vim = vim or {}
+vim = vim or {}
 
 --- Returns a deep copy of the given object. Non-table objects are copied as
 --- in a typical Lua assignment, whereas table objects are copied recursively.
@@ -14,9 +14,10 @@ local vim = vim or {}
 --- same functions as those in the input table. Userdata and threads are not
 --- copied and will throw an error.
 ---
----@param orig table Table to copy
----@returns New table of copied keys and (nested) values.
-function vim.deepcopy(orig) end  -- luacheck: no unused
+---@generic T: table
+---@param orig T Table to copy
+---@return T Table of copied keys and (nested) values.
+function vim.deepcopy(orig) end -- luacheck: no unused
 vim.deepcopy = (function()
   local function _id(v)
     return v
@@ -24,7 +25,9 @@ vim.deepcopy = (function()
 
   local deepcopy_funcs = {
     table = function(orig, cache)
-      if cache[orig] then return cache[orig] end
+      if cache[orig] then
+        return cache[orig]
+      end
       local copy = {}
 
       cache[orig] = copy
@@ -46,30 +49,64 @@ vim.deepcopy = (function()
     if f then
       return f(orig, cache or {})
     else
-      error("Cannot deepcopy object of type "..type(orig))
+      if type(orig) == 'userdata' and orig == vim.NIL then
+        return vim.NIL
+      end
+      error('Cannot deepcopy object of type ' .. type(orig))
     end
   end
 end)()
 
 --- Splits a string at each instance of a separator.
 ---
----@see |vim.split()|
----@see https://www.lua.org/pil/20.2.html
----@see http://lua-users.org/wiki/StringLibraryTutorial
+--- Example:
+---   <pre>lua
+---   for s in vim.gsplit(':aa::b:', ':', {plain=true}) do
+---     print(s)
+---   end
+---   </pre>
 ---
----@param s String to split
----@param sep Separator string or pattern
----@param plain If `true` use `sep` literally (passed to String.find)
----@returns Iterator over the split components
-function vim.gsplit(s, sep, plain)
-  vim.validate{s={s,'s'},sep={sep,'s'},plain={plain,'b',true}}
+--- If you want to also inspect the separator itself (instead of discarding it), use
+--- |string.gmatch()|. Example:
+---   <pre>lua
+---   for word, num in ('foo111bar222'):gmatch('([^0-9]*)(%d*)') do
+---     print(('word: %s num: %s'):format(word, num))
+---   end
+---   </pre>
+---
+--- @see |string.gmatch()|
+--- @see |vim.split()|
+--- @see |luaref-patterns|
+--- @see https://www.lua.org/pil/20.2.html
+--- @see http://lua-users.org/wiki/StringLibraryTutorial
+---
+--- @param s string String to split
+--- @param sep string Separator or pattern
+--- @param opts (table|nil) Keyword arguments |kwargs|:
+---       - plain: (boolean) Use `sep` literally (as in string.find).
+---       - trimempty: (boolean) Discard empty segments at start and end of the sequence.
+---@return fun():string|nil (function) Iterator over the split components
+function vim.gsplit(s, sep, opts)
+  local plain
+  local trimempty = false
+  if type(opts) == 'boolean' then
+    plain = opts -- For backwards compatibility.
+  else
+    vim.validate({ s = { s, 's' }, sep = { sep, 's' }, opts = { opts, 't', true } })
+    opts = opts or {}
+    plain, trimempty = opts.plain, opts.trimempty
+  end
 
   local start = 1
   local done = false
 
+  -- For `trimempty`: queue of collected segments, to be emitted at next pass.
+  local segs = {}
+  local empty_start = true -- Only empty segments seen so far.
+
   local function _pass(i, j, ...)
     if i then
-      assert(j+1 > start, "Infinite loop detected")
+      assert(j + 1 > start, 'Infinite loop detected')
       local seg = s:sub(start, i - 1)
       start = j + 1
       return seg, ...
@@ -80,72 +117,67 @@ function vim.gsplit(s, sep, plain)
   end
 
   return function()
-    if done or (s == '' and sep == '') then
-      return
-    end
-    if sep == '' then
+    if trimempty and #segs > 0 then
+      -- trimempty: Pop the collected segments.
+      return table.remove(segs)
+    elseif done or (s == '' and sep == '') then
+      return nil
+    elseif sep == '' then
       if start == #s then
         done = true
       end
-      return _pass(start+1, start)
+      return _pass(start + 1, start)
     end
-    return _pass(s:find(sep, start, plain))
+
+    local seg = _pass(s:find(sep, start, plain))
+
+    -- Trim empty segments from start/end.
+    if trimempty and seg ~= '' then
+      empty_start = false
+    elseif trimempty and seg == '' then
+      while not done and seg == '' do
+        table.insert(segs, 1, '')
+        seg = _pass(s:find(sep, start, plain))
+      end
+      if done and seg == '' then
+        return nil
+      elseif empty_start then
+        empty_start = false
+        segs = {}
+        return seg
+      end
+      if seg ~= '' then
+        table.insert(segs, 1, seg)
+      end
+      return table.remove(segs)
+    end
+
+    return seg
   end
 end
 
 --- Splits a string at each instance of a separator.
 ---
 --- Examples:
---- <pre>
----  split(":aa::b:", ":")     --> {'','aa','','b',''}
----  split("axaby", "ab?")     --> {'','x','y'}
----  split("x*yz*o", "*", {plain=true})  --> {'x','yz','o'}
+--- <pre>lua
+---  split(":aa::b:", ":")                   --> {'','aa','','b',''}
+---  split("axaby", "ab?")                   --> {'','x','y'}
+---  split("x*yz*o", "*", {plain=true})      --> {'x','yz','o'}
 ---  split("|x|y|z|", "|", {trimempty=true}) --> {'x', 'y', 'z'}
 --- </pre>
 ---
 ---@see |vim.gsplit()|
+---@see |string.gmatch()|
 ---
----@param s String to split
----@param sep Separator string or pattern
----@param kwargs Keyword arguments:
----       - plain: (boolean) If `true` use `sep` literally (passed to string.find)
----       - trimempty: (boolean) If `true` remove empty items from the front
----         and back of the list
----@returns List-like table of the split components.
-function vim.split(s, sep, kwargs)
-  local plain
-  local trimempty = false
-  if type(kwargs) == 'boolean' then
-    -- Support old signature for backward compatibility
-    plain = kwargs
-  else
-    vim.validate { kwargs = {kwargs, 't', true} }
-    kwargs = kwargs or {}
-    plain = kwargs.plain
-    trimempty = kwargs.trimempty
-  end
-
+---@param s string String to split
+---@param sep string Separator or pattern
+---@param opts (table|nil) Keyword arguments |kwargs| accepted by |vim.gsplit()|
+---@return string[] List of split components
+function vim.split(s, sep, opts)
   local t = {}
-  local skip = trimempty
-  for c in vim.gsplit(s, sep, plain) do
-    if c ~= "" then
-      skip = false
-    end
-
-    if not skip then
-      table.insert(t, c)
-    end
+  for c in vim.gsplit(s, sep, opts) do
+    table.insert(t, c)
   end
-
-  if trimempty then
-    for i = #t, 1, -1 do
-      if t[i] ~= "" then
-        break
-      end
-      table.remove(t, i)
-    end
-  end
-
   return t
 end
 
@@ -154,10 +186,11 @@ end
 ---
 ---@see From https://github.com/premake/premake-core/blob/master/src/base/table.lua
 ---
----@param t Table
----@returns list of keys
+---@generic T: table
+---@param t table<T, any> (table) Table
+---@return T[] (list) List of keys
 function vim.tbl_keys(t)
-  assert(type(t) == 'table', string.format("Expected table, got %s", type(t)))
+  assert(type(t) == 'table', string.format('Expected table, got %s', type(t)))
 
   local keys = {}
   for k, _ in pairs(t) do
@@ -169,10 +202,11 @@ end
 --- Return a list of all values used in a table.
 --- However, the order of the return table of values is not guaranteed.
 ---
----@param t Table
----@returns list of values
+---@generic T
+---@param t table<any, T> (table) Table
+---@return T[] (list) List of values
 function vim.tbl_values(t)
-  assert(type(t) == 'table', string.format("Expected table, got %s", type(t)))
+  assert(type(t) == 'table', string.format('Expected table, got %s', type(t)))
 
   local values = {}
   for _, v in pairs(t) do
@@ -183,10 +217,12 @@ end
 
 --- Apply a function to all values of a table.
 ---
----@param func function or callable table
----@param t table
+---@generic T
+---@param func fun(value: T): any (function) Function
+---@param t table<any, T> (table) Table
+---@return table Table of transformed values
 function vim.tbl_map(func, t)
-  vim.validate{func={func,'c'},t={t,'t'}}
+  vim.validate({ func = { func, 'c' }, t = { t, 't' } })
 
   local rettab = {}
   for k, v in pairs(t) do
@@ -197,10 +233,12 @@ end
 
 --- Filter a table using a predicate function
 ---
----@param func function or callable table
----@param t table
+---@generic T
+---@param func fun(value: T): boolean (function) Function
+---@param t table<any, T> (table) Table
+---@return T[] (table) Table of filtered values
 function vim.tbl_filter(func, t)
-  vim.validate{func={func,'c'},t={t,'t'}}
+  vim.validate({ func = { func, 'c' }, t = { t, 't' } })
 
   local rettab = {}
   for _, entry in pairs(t) do
@@ -213,13 +251,13 @@ end
 
 --- Checks if a list-like (vector) table contains `value`.
 ---
----@param t Table to check
----@param value Value to compare
----@returns true if `t` contains `value`
+---@param t table Table to check
+---@param value any Value to compare
+---@return boolean `true` if `t` contains `value`
 function vim.tbl_contains(t, value)
-  vim.validate{t={t,'t'}}
+  vim.validate({ t = { t, 't' } })
 
-  for _,v in ipairs(t) do
+  for _, v in ipairs(t) do
     if v == value then
       return true
     end
@@ -231,25 +269,30 @@ end
 ---
 ---@see https://github.com/premake/premake-core/blob/master/src/base/table.lua
 ---
----@param t Table to check
+---@param t table Table to check
+---@return boolean `true` if `t` is empty
 function vim.tbl_isempty(t)
-  assert(type(t) == 'table', string.format("Expected table, got %s", type(t)))
+  assert(type(t) == 'table', string.format('Expected table, got %s', type(t)))
   return next(t) == nil
 end
 
---- we only merge empty tables or tables that are not a list
+--- We only merge empty tables or tables that are not a list
 ---@private
 local function can_merge(v)
-  return type(v) == "table" and (vim.tbl_isempty(v) or not vim.tbl_islist(v))
+  return type(v) == 'table' and (vim.tbl_isempty(v) or not vim.tbl_islist(v))
 end
 
 local function tbl_extend(behavior, deep_extend, ...)
-  if (behavior ~= 'error' and behavior ~= 'keep' and behavior ~= 'force') then
-    error('invalid "behavior": '..tostring(behavior))
+  if behavior ~= 'error' and behavior ~= 'keep' and behavior ~= 'force' then
+    error('invalid "behavior": ' .. tostring(behavior))
   end
 
   if select('#', ...) < 2 then
-    error('wrong number of arguments (given '..tostring(1 + select('#', ...))..', expected at least 3)')
+    error(
+      'wrong number of arguments (given '
+        .. tostring(1 + select('#', ...))
+        .. ', expected at least 3)'
+    )
   end
 
   local ret = {}
@@ -259,15 +302,15 @@ local function tbl_extend(behavior, deep_extend, ...)
 
   for i = 1, select('#', ...) do
     local tbl = select(i, ...)
-    vim.validate{["after the second argument"] = {tbl,'t'}}
+    vim.validate({ ['after the second argument'] = { tbl, 't' } })
     if tbl then
       for k, v in pairs(tbl) do
         if deep_extend and can_merge(v) and can_merge(ret[k]) then
           ret[k] = tbl_extend(behavior, true, ret[k], v)
         elseif behavior ~= 'force' and ret[k] ~= nil then
           if behavior == 'error' then
-            error('key found in more than one map: '..k)
-          end  -- Else behavior is "keep".
+            error('key found in more than one map: ' .. k)
+          end -- Else behavior is "keep".
         else
           ret[k] = v
         end
@@ -281,38 +324,46 @@ end
 ---
 ---@see |extend()|
 ---
----@param behavior Decides what to do if a key is found in more than one map:
+---@param behavior string Decides what to do if a key is found in more than one map:
 ---      - "error": raise an error
 ---      - "keep":  use value from the leftmost map
 ---      - "force": use value from the rightmost map
----@param ... Two or more map-like tables.
+---@param ... table Two or more map-like tables
+---@return table Merged table
 function vim.tbl_extend(behavior, ...)
   return tbl_extend(behavior, false, ...)
 end
 
 --- Merges recursively two or more map-like tables.
 ---
----@see |tbl_extend()|
+---@see |vim.tbl_extend()|
 ---
----@param behavior Decides what to do if a key is found in more than one map:
+---@generic T1: table
+---@generic T2: table
+---@param behavior "error"|"keep"|"force" (string) Decides what to do if a key is found in more than one map:
 ---      - "error": raise an error
 ---      - "keep":  use value from the leftmost map
 ---      - "force": use value from the rightmost map
----@param ... Two or more map-like tables.
+---@param ... T2 Two or more map-like tables
+---@return T1|T2 (table) Merged table
 function vim.tbl_deep_extend(behavior, ...)
   return tbl_extend(behavior, true, ...)
 end
 
 --- Deep compare values for equality
 ---
---- Tables are compared recursively unless they both provide the `eq` methamethod.
+--- Tables are compared recursively unless they both provide the `eq` metamethod.
 --- All other types are compared using the equality `==` operator.
----@param a first value
----@param b second value
----@returns `true` if values are equals, else `false`.
+---@param a any First value
+---@param b any Second value
+---@return boolean `true` if values are equals, else `false`
 function vim.deep_equal(a, b)
-  if a == b then return true end
-  if type(a) ~= type(b) then return false end
+  if a == b then
+    return true
+  end
+  if type(a) ~= type(b) then
+    return false
+  end
   if type(a) == 'table' then
     for k, v in pairs(a) do
       if not vim.deep_equal(v, b[k]) then
@@ -332,15 +383,22 @@ end
 --- Add the reverse lookup values to an existing table.
 --- For example:
 --- ``tbl_add_reverse_lookup { A = 1 } == { [1] = 'A', A = 1 }``
---
---Do note that it *modifies* the input.
----@param o table The table to add the reverse to.
+---
+--- Note that this *modifies* the input.
+---@param o table Table to add the reverse to
+---@return table o
 function vim.tbl_add_reverse_lookup(o)
   local keys = vim.tbl_keys(o)
   for _, k in ipairs(keys) do
     local v = o[k]
     if o[v] then
-      error(string.format("The reverse lookup found an existing value for %q while processing key %q", tostring(v), tostring(k)))
+      error(
+        string.format(
+          'The reverse lookup found an existing value for %q while processing key %q',
+          tostring(v),
+          tostring(k)
+        )
+      )
     end
     o[v] = k
   end
@@ -349,29 +407,28 @@ end
 
 --- Index into a table (first argument) via string keys passed as subsequent arguments.
 --- Return `nil` if the key does not exist.
---_
+---
 --- Examples:
---- <pre>
+--- <pre>lua
 ---  vim.tbl_get({ key = { nested_key = true }}, 'key', 'nested_key') == true
 ---  vim.tbl_get({ key = {}}, 'key', 'nested_key') == nil
 --- </pre>
 ---
----@param o Table to index
----@param ... Optional strings (0 or more, variadic) via which to index the table
+---@param o table Table to index
+---@param ... string Optional strings (0 or more, variadic) via which to index the table
 ---
----@returns nested value indexed by key if it exists, else nil
+---@return any Nested value indexed by key (if it exists), else nil
 function vim.tbl_get(o, ...)
-  local keys = {...}
+  local keys = { ... }
   if #keys == 0 then
-    return
+    return nil
   end
   for i, k in ipairs(keys) do
-    if type(o[k]) ~= 'table' and next(keys, i) then
-      return nil
-    end
     o = o[k]
     if o == nil then
-      return
+      return nil
+    elseif type(o) ~= 'table' and next(keys, i) then
+      return nil
     end
   end
   return o
@@ -383,18 +440,19 @@ end
 ---
 ---@see |vim.tbl_extend()|
 ---
----@param dst list which will be modified and appended to.
----@param src list from which values will be inserted.
----@param start Start index on src. defaults to 1
----@param finish Final index on src. defaults to #src
----@returns dst
+---@generic T: table
+---@param dst T List which will be modified and appended to
+---@param src table List from which values will be inserted
+---@param start (integer|nil) Start index on src. Defaults to 1
+---@param finish (integer|nil) Final index on src. Defaults to `#src`
+---@return T dst
 function vim.list_extend(dst, src, start, finish)
-  vim.validate {
-    dst = {dst, 't'};
-    src = {src, 't'};
-    start = {start, 'n', true};
-    finish = {finish, 'n', true};
-  }
+  vim.validate({
+    dst = { dst, 't' },
+    src = { src, 't' },
+    start = { start, 'n', true },
+    finish = { finish, 'n', true },
+  })
   for i = start or 1, finish or #src do
     table.insert(dst, src[i])
   end
@@ -406,15 +464,15 @@ end
 ---
 ---@see From https://github.com/premake/premake-core/blob/master/src/base/table.lua
 ---
----@param t List-like table
----@returns Flattened copy of the given list-like table.
+---@param t table List-like table
+---@return table Flattened copy of the given list-like table
 function vim.tbl_flatten(t)
   local result = {}
   local function _tbl_flatten(_t)
     local n = #_t
     for i = 1, n do
       local v = _t[i]
-      if type(v) == "table" then
+      if type(v) == 'table' then
         _tbl_flatten(v)
       elseif v then
         table.insert(result, v)
@@ -425,14 +483,41 @@ function vim.tbl_flatten(t)
   return result
 end
 
+--- Enumerate a table sorted by its keys.
+---
+---@see Based on https://github.com/premake/premake-core/blob/master/src/base/table.lua
+---
+---@param t table List-like table
+---@return iterator over sorted keys and their values
+function vim.spairs(t)
+  assert(type(t) == 'table', string.format('Expected table, got %s', type(t)))
+
+  -- collect the keys
+  local keys = {}
+  for k in pairs(t) do
+    table.insert(keys, k)
+  end
+  table.sort(keys)
+
+  -- Return the iterator function.
+  -- TODO(justinmk): Return "iterator function, table {t}, and nil", like pairs()?
+  local i = 0
+  return function()
+    i = i + 1
+    if keys[i] then
+      return keys[i], t[keys[i]]
+    end
+  end
+end
+
 --- Tests if a Lua table can be treated as an array.
 ---
 --- Empty table `{}` is assumed to be an array, unless it was created by
 --- |vim.empty_dict()| or returned as a dict-like |API| or Vimscript result,
 --- for example from |rpcrequest()| or |vim.fn|.
 ---
----@param t Table
----@returns `true` if array-like table, else `false`.
+---@param t table Table
+---@return boolean `true` if array-like table, else `false`
 function vim.tbl_islist(t)
   if type(t) ~= 'table' then
     return false
@@ -441,7 +526,7 @@ function vim.tbl_islist(t)
   local count = 0
 
   for k, _ in pairs(t) do
-    if type(k) == "number" then
+    if type(k) == 'number' then
       count = count + 1
     else
       return false
@@ -454,7 +539,7 @@ function vim.tbl_islist(t)
     -- TODO(bfredl): in the future, we will always be inside nvim
     -- then this check can be deleted.
     if vim._empty_dict_mt == nil then
-      return nil
+      return false
     end
     return getmetatable(t) ~= vim._empty_dict_mt
   end
@@ -462,80 +547,84 @@ end
 
 --- Counts the number of non-nil values in table `t`.
 ---
---- <pre>
---- vim.tbl_count({ a=1, b=2 }) => 2
---- vim.tbl_count({ 1, 2 }) => 2
+--- <pre>lua
+--- vim.tbl_count({ a=1, b=2 })  --> 2
+--- vim.tbl_count({ 1, 2 })      --> 2
 --- </pre>
 ---
 ---@see https://github.com/Tieske/Penlight/blob/master/lua/pl/tablex.lua
----@param t Table
----@returns Number that is the number of the value in table
+---@param t table Table
+---@return integer Number of non-nil values in table
 function vim.tbl_count(t)
-  vim.validate{t={t,'t'}}
+  vim.validate({ t = { t, 't' } })
 
   local count = 0
-  for _ in pairs(t) do count = count + 1 end
+  for _ in pairs(t) do
+    count = count + 1
+  end
   return count
 end
 
 --- Creates a copy of a table containing only elements from start to end (inclusive)
 ---
----@param list table table
----@param start integer Start range of slice
----@param finish integer End range of slice
----@returns Copy of table sliced from start to finish (inclusive)
+---@generic T
+---@param list T[] (list) Table
+---@param start integer|nil Start range of slice
+---@param finish integer|nil End range of slice
+---@return T[] (list) Copy of table sliced from start to finish (inclusive)
 function vim.list_slice(list, start, finish)
   local new_list = {}
   for i = start or 1, finish or #list do
-    new_list[#new_list+1] = list[i]
+    new_list[#new_list + 1] = list[i]
   end
   return new_list
 end
 
 --- Trim whitespace (Lua pattern "%s") from both sides of a string.
 ---
+---@see |luaref-patterns|
 ---@see https://www.lua.org/pil/20.2.html
----@param s String to trim
----@returns String with whitespace removed from its beginning and end
+---@param s string String to trim
+---@return string String with whitespace removed from its beginning and end
 function vim.trim(s)
-  vim.validate{s={s,'s'}}
+  vim.validate({ s = { s, 's' } })
   return s:match('^%s*(.*%S)') or ''
 end
 
---- Escapes magic chars in a Lua pattern.
+--- Escapes magic chars in |lua-patterns|.
 ---
 ---@see https://github.com/rxi/lume
----@param s  String to escape
----@returns  %-escaped pattern string
+---@param s string String to escape
+---@return string %-escaped pattern string
 function vim.pesc(s)
-  vim.validate{s={s,'s'}}
-  return s:gsub('[%(%)%.%%%+%-%*%?%[%]%^%$]', '%%%1')
+  vim.validate({ s = { s, 's' } })
+  return (s:gsub('[%(%)%.%%%+%-%*%?%[%]%^%$]', '%%%1'))
 end
 
 --- Tests if `s` starts with `prefix`.
 ---
----@param s (string) a string
----@param prefix (string) a prefix
----@return (boolean) true if `prefix` is a prefix of s
+---@param s string String
+---@param prefix string Prefix to match
+---@return boolean `true` if `prefix` is a prefix of `s`
 function vim.startswith(s, prefix)
-  vim.validate { s = {s, 's'}; prefix = {prefix, 's'}; }
+  vim.validate({ s = { s, 's' }, prefix = { prefix, 's' } })
   return s:sub(1, #prefix) == prefix
 end
 
 --- Tests if `s` ends with `suffix`.
 ---
----@param s (string) a string
----@param suffix (string) a suffix
----@return (boolean) true if `suffix` is a suffix of s
+---@param s string String
+---@param suffix string Suffix to match
+---@return boolean `true` if `suffix` is a suffix of `s`
 function vim.endswith(s, suffix)
-  vim.validate { s = {s, 's'}; suffix = {suffix, 's'}; }
+  vim.validate({ s = { s, 's' }, suffix = { suffix, 's' } })
   return #suffix == 0 or s:sub(-#suffix) == suffix
 end
 
 --- Validates a parameter specification (types and values).
 ---
 --- Usage example:
---- <pre>
+--- <pre>lua
 ---  function user.new(name, age, hobbies)
 ---    vim.validate{
 ---      name={name, 'string'},
@@ -547,28 +636,28 @@ end
 --- </pre>
 ---
 --- Examples with explicit argument values (can be run directly):
---- <pre>
+--- <pre>lua
 ---  vim.validate{arg1={{'foo'}, 'table'}, arg2={'foo', 'string'}}
----     => NOP (success)
+---     --> NOP (success)
 ---
 ---  vim.validate{arg1={1, 'table'}}
----     => error('arg1: expected table, got number')
+---     --> error('arg1: expected table, got number')
 ---
 ---  vim.validate{arg1={3, function(a) return (a % 2) == 0 end, 'even number'}}
----     => error('arg1: expected even number, got 3')
+---     --> error('arg1: expected even number, got 3')
 --- </pre>
 ---
 --- If multiple types are valid they can be given as a list.
---- <pre>
+--- <pre>lua
 ---  vim.validate{arg1={{'foo'}, {'table', 'string'}}, arg2={'foo', {'table', 'string'}}}
----     => NOP (success)
+---     --> NOP (success)
 ---
----  vim.validate{arg1={1, {'string', table'}}}
----     => error('arg1: expected string|table, got number')
+---  vim.validate{arg1={1, {'string', 'table'}}}
+---     --> error('arg1: expected string|table, got number')
 ---
 --- </pre>
 ---
----@param opt table of parameter names to validations. Each key is a parameter
+---@param opt table Names of parameters to validate. Each key is a parameter
 ---          name; each value is a tuple in one of these forms:
 ---          1. (arg_value, type_name, optional)
 ---             - arg_value: argument value
@@ -582,18 +671,24 @@ end
 ---               only if the argument is valid. Can optionally return an additional
 ---               informative error message as the second returned value.
 ---             - msg: (optional) error string if validation fails
-function vim.validate(opt) end  -- luacheck: no unused
+function vim.validate(opt) end -- luacheck: no unused
 
 do
   local type_names = {
-    ['table']    = 'table',    t = 'table',
-    ['string']   = 'string',   s = 'string',
-    ['number']   = 'number',   n = 'number',
-    ['boolean']  = 'boolean',  b = 'boolean',
-    ['function'] = 'function', f = 'function',
-    ['callable'] = 'callable', c = 'callable',
-    ['nil']      = 'nil',
-    ['thread']   = 'thread',
+    ['table'] = 'table',
+    t = 'table',
+    ['string'] = 'string',
+    s = 'string',
+    ['number'] = 'number',
+    n = 'number',
+    ['boolean'] = 'boolean',
+    b = 'boolean',
+    ['function'] = 'function',
+    f = 'function',
+    ['callable'] = 'callable',
+    c = 'callable',
+    ['nil'] = 'nil',
+    ['thread'] = 'thread',
     ['userdata'] = 'userdata',
   }
 
@@ -612,21 +707,22 @@ do
         return false, string.format('opt[%s]: expected table, got %s', param_name, type(spec))
       end
 
-      local val = spec[1]   -- Argument value.
-      local types = spec[2]     -- Type name, or callable.
+      local val = spec[1] -- Argument value
+      local types = spec[2] -- Type name, or callable
       local optional = (true == spec[3])
 
       if type(types) == 'string' then
-        types = {types}
+        types = { types }
       end
 
       if vim.is_callable(types) then
-        -- Check user-provided validation function.
+        -- Check user-provided validation function
         local valid, optional_message = types(val)
         if not valid then
-          local error_message = string.format("%s: expected %s, got %s", param_name, (spec[3] or '?'), tostring(val))
+          local error_message =
+            string.format('%s: expected %s, got %s', param_name, (spec[3] or '?'), tostring(val))
           if optional_message ~= nil then
-            error_message = error_message .. string.format(". Info: %s", optional_message)
+            error_message = error_message .. string.format('. Info: %s', optional_message)
           end
 
           return false, error_message
@@ -646,10 +742,16 @@ do
           end
         end
         if not success then
-          return false, string.format("%s: expected %s, got %s", param_name, table.concat(types, '|'), type(val))
+          return false,
+            string.format(
+              '%s: expected %s, got %s',
+              param_name,
+              table.concat(types, '|'),
+              type(val)
+            )
         end
       else
-        return false, string.format("invalid type name: %s", tostring(types))
+        return false, string.format('invalid type name: %s', tostring(types))
       end
     end
 
@@ -665,14 +767,45 @@ do
 end
 --- Returns true if object `f` can be called as a function.
 ---
----@param f Any object
----@return true if `f` is callable, else false
+---@param f any Any object
+---@return boolean `true` if `f` is callable, else `false`
 function vim.is_callable(f)
-  if type(f) == 'function' then return true end
+  if type(f) == 'function' then
+    return true
+  end
   local m = getmetatable(f)
-  if m == nil then return false end
+  if m == nil then
+    return false
+  end
   return type(m.__call) == 'function'
 end
 
+--- Creates a table whose members are automatically created when accessed, if they don't already
+--- exist.
+---
+--- They mimic defaultdict in python.
+---
+--- If {create} is `nil`, this will create a defaulttable whose constructor function is
+--- this function, effectively allowing to create nested tables on the fly:
+---
+--- <pre>lua
+--- local a = vim.defaulttable()
+--- a.b.c = 1
+--- </pre>
+---
+---@param create function?(key:any):any The function called to create a missing value.
+---@return table Empty table with metamethod
+function vim.defaulttable(create)
+  create = create or function(_)
+    return vim.defaulttable()
+  end
+  return setmetatable({}, {
+    __index = function(tbl, key)
+      rawset(tbl, key, create(key))
+      return rawget(tbl, key)
+    end,
+  })
+end
+
 return vim
 -- vim:sw=2 ts=2 et
