@@ -1,8 +1,20 @@
 #ifndef NVIM_AUTOCMD_H
 #define NVIM_AUTOCMD_H
 
+#include <stdbool.h>
+#include <stdint.h>
+
+#include "nvim/api/private/defs.h"
 #include "nvim/buffer_defs.h"
+#include "nvim/eval/typval_defs.h"
 #include "nvim/ex_cmds_defs.h"
+#include "nvim/macros.h"
+#include "nvim/regexp_defs.h"
+#include "nvim/types.h"
+
+struct AutoCmd_S;
+struct AutoPatCmd_S;
+struct AutoPat_S;
 
 // event_T definition
 #ifdef INCLUDE_GENERATED_DECLARATIONS
@@ -13,30 +25,33 @@
 // not the current buffer.
 typedef struct {
   buf_T *save_curbuf;             ///< saved curbuf
-  bool use_aucmd_win;             ///< using aucmd_win
+  int use_aucmd_win_idx;          ///< index in aucmd_win[] if >= 0
   handle_T save_curwin_handle;    ///< ID of saved curwin
   handle_T new_curwin_handle;     ///< ID of new curwin
   handle_T save_prevwin_handle;   ///< ID of saved prevwin
   bufref_T new_curbuf;            ///< new curbuf
-  char_u *globaldir;              ///< saved value of globaldir
+  char *globaldir;                ///< saved value of globaldir
   bool save_VIsual_active;        ///< saved VIsual_active
+  int save_State;                 ///< saved State
 } aco_save_T;
 
-typedef struct AutoCmd {
+typedef struct AutoCmd_S AutoCmd;
+struct AutoCmd_S {
   AucmdExecutable exec;
   bool once;                            // "One shot": removed after execution
   bool nested;                          // If autocommands nest here
   bool last;                            // last command in list
   int64_t id;                           // ID used for uniquely tracking an autocmd.
-  sctx_T script_ctx;                    // script context where defined
+  sctx_T script_ctx;                    // script context where it is defined
   char *desc;                           // Description for the autocmd.
-  struct AutoCmd *next;                 // Next AutoCmd in list
-} AutoCmd;
+  AutoCmd *next;                        // Next AutoCmd in list
+};
 
-typedef struct AutoPat {
-  struct AutoPat *next;                 // next AutoPat in AutoPat list; MUST
+typedef struct AutoPat_S AutoPat;
+struct AutoPat_S {
+  AutoPat *next;                        // next AutoPat in AutoPat list; MUST
                                         // be the first entry
-  char_u *pat;                          // pattern as typed (NULL when pattern
+  char *pat;                            // pattern as typed (NULL when pattern
                                         // has been removed)
   regprog_T *reg_prog;                  // compiled regprog for pattern
   AutoCmd *cmds;                        // list of commands to do
@@ -45,22 +60,23 @@ typedef struct AutoPat {
   int buflocal_nr;                      // !=0 for buffer-local AutoPat
   char allow_dirs;                      // Pattern may match whole path
   char last;                            // last pattern for apply_autocmds()
-} AutoPat;
+};
 
 /// Struct used to keep status while executing autocommands for an event.
-typedef struct AutoPatCmd {
+typedef struct AutoPatCmd_S AutoPatCmd;
+struct AutoPatCmd_S {
   AutoPat *curpat;          // next AutoPat to examine
   AutoCmd *nextcmd;         // next AutoCmd to execute
-  int group;                    // group being used
-  char_u *fname;           // fname to match with
-  char_u *sfname;          // sfname to match with
-  char_u *tail;            // tail of fname
-  event_T event;                // current event
-  int arg_bufnr;                // initially equal to <abuf>, set to zero when
-                                // buf is deleted
-  struct AutoPatCmd *next;    // chain of active apc-s for auto-invalidation
-} AutoPatCmd;
-
+  int group;                // group being used
+  char *fname;              // fname to match with
+  char *sfname;             // sfname to match with
+  char *tail;               // tail of fname
+  event_T event;            // current event
+  sctx_T script_ctx;        // script context where it is defined
+  int arg_bufnr;            // initially equal to <abuf>, set to zero when buf is deleted
+  Object *data;             // arbitrary data
+  AutoPatCmd *next;         // chain of active apc-s for auto-invalidation
+};
 
 // Set by the apply_autocmds_group function if the given event is equal to
 // EVENT_FILETYPE. Used by the readfile function in order to determine if
@@ -70,7 +86,6 @@ typedef struct AutoPatCmd {
 // apply_autocmds_group.
 EXTERN bool au_did_filetype INIT(= false);
 
-
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "autocmd.h.generated.h"
 #endif
