@@ -1,15 +1,19 @@
 // This is an open source non-commercial project. Dear PVS-Studio, please check
 // it. PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
 
+#include <assert.h>
+
 #include "nvim/buffer.h"
 #include "nvim/decoration.h"
+#include "nvim/drawscreen.h"
 #include "nvim/extmark.h"
+#include "nvim/fold.h"
 #include "nvim/highlight.h"
 #include "nvim/highlight_group.h"
-#include "nvim/lua/executor.h"
+#include "nvim/memory.h"
 #include "nvim/move.h"
-#include "nvim/screen.h"
-#include "nvim/vim.h"
+#include "nvim/pos.h"
+#include "nvim/sign_defs.h"
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "decoration.c.generated.h"
@@ -45,7 +49,7 @@ void bufhl_add_hl_pos_offset(buf_T *buf, int src_id, int hl_id, lpos_T pos_start
       // substituted text. But it would be more consistent to highlight
       // a space _after_ the previous line instead (like highlight EOL list
       // char)
-      hl_start = MAX(offset-1, 0);
+      hl_start = MAX(offset - 1, 0);
       end_off = 1;
       hl_end = 0;
     } else if (lnum == pos_start.lnum && lnum < pos_end.lnum) {
@@ -53,14 +57,14 @@ void bufhl_add_hl_pos_offset(buf_T *buf, int src_id, int hl_id, lpos_T pos_start
       end_off = 1;
       hl_end = 0;
     } else if (pos_start.lnum < lnum && lnum == pos_end.lnum) {
-      hl_start = MAX(offset-1, 0);
+      hl_start = MAX(offset - 1, 0);
       hl_end = pos_end.col + offset;
     } else if (pos_start.lnum == lnum && pos_end.lnum == lnum) {
       hl_start = pos_start.col + offset;
       hl_end = pos_end.col + offset;
     }
     extmark_set(buf, (uint32_t)src_id, NULL,
-                (int)lnum-1, hl_start, (int)lnum-1+end_off, hl_end,
+                (int)lnum - 1, hl_start, (int)lnum - 1 + end_off, hl_end,
                 &decor, true, false, kExtmarkNoUndo);
   }
 }
@@ -68,18 +72,22 @@ void bufhl_add_hl_pos_offset(buf_T *buf, int src_id, int hl_id, lpos_T pos_start
 void decor_redraw(buf_T *buf, int row1, int row2, Decoration *decor)
 {
   if (row2 >= row1) {
-    if (!decor || decor->hl_id || decor_has_sign(decor) || decor->conceal) {
-      redraw_buf_range_later(buf, row1+1, row2+1);
+    if (!decor
+        || decor->hl_id
+        || decor_has_sign(decor)
+        || decor->conceal
+        || decor->spell != kNone) {
+      redraw_buf_range_later(buf, row1 + 1, row2 + 1);
     }
   }
 
-  if (decor && kv_size(decor->virt_text)) {
-    redraw_buf_line_later(buf, row1+1);
+  if (decor && decor_virt_pos(*decor)) {
+    redraw_buf_line_later(buf, row1 + 1, false);
   }
 
   if (decor && kv_size(decor->virt_lines)) {
-    redraw_buf_line_later(buf, MIN(buf->b_ml.ml_line_count,
-                                   row1+1+(decor->virt_lines_above?0:1)));
+    redraw_buf_line_later(buf, row1 + 1 + (decor->virt_lines_above?0:1), true);
+    changed_line_display_buf(buf);
   }
 }
 
@@ -94,27 +102,41 @@ void decor_remove(buf_T *buf, int row, int row2, Decoration *decor)
     if (decor_has_sign(decor)) {
       assert(buf->b_signs > 0);
       buf->b_signs--;
-    }
-    if (row2 >= row && decor->sign_text) {
-      buf_signcols_del_check(buf, row+1, row2+1);
+      if (decor->sign_text) {
+        assert(buf->b_signs_with_text > 0);
+        buf->b_signs_with_text--;
+        if (row2 >= row) {
+          buf_signcols_del_check(buf, row + 1, row2 + 1);
+        }
+      }
     }
   }
   decor_free(decor);
 }
 
+void decor_clear(Decoration *decor)
+{
+  clear_virttext(&decor->virt_text);
+  for (size_t i = 0; i < kv_size(decor->virt_lines); i++) {
+    clear_virttext(&kv_A(decor->virt_lines, i).line);
+  }
+  kv_destroy(decor->virt_lines);
+  xfree(decor->sign_text);
+}
+
 void decor_free(Decoration *decor)
 {
   if (decor) {
-    clear_virttext(&decor->virt_text);
-    for (size_t i = 0; i < kv_size(decor->virt_lines); i++) {
-      clear_virttext(&kv_A(decor->virt_lines, i).line);
-    }
-    kv_destroy(decor->virt_lines);
-    xfree(decor->sign_text);
+    decor_clear(decor);
     xfree(decor);
   }
 }
 
+void decor_state_free(DecorState *state)
+{
+  xfree(state->active.items);
+}
+
 void clear_virttext(VirtText *text)
 {
   for (size_t i = 0; i < kv_size(*text); i++) {
@@ -146,10 +168,10 @@ next_mark:
   return NULL;
 }
 
-bool decor_redraw_reset(buf_T *buf, DecorState *state)
+bool decor_redraw_reset(win_T *wp, DecorState *state)
 {
   state->row = -1;
-  state->buf = buf;
+  state->win = wp;
   for (size_t i = 0; i < kv_size(state->active); i++) {
     DecorRange item = kv_A(state->active, i);
     if (item.virt_text_owned) {
@@ -157,25 +179,30 @@ bool decor_redraw_reset(buf_T *buf, DecorState *state)
     }
   }
   kv_size(state->active) = 0;
-  return buf->b_marktree->n_keys;
+  return wp->w_buffer->b_marktree->n_keys;
 }
 
 Decoration get_decor(mtkey_t mark)
 {
   if (mark.decor_full) {
     return *mark.decor_full;
-  } else {
-    Decoration fake = DECORATION_INIT;
-    fake.hl_id = mark.hl_id;
-    fake.priority = mark.priority;
-    fake.hl_eol = (mark.flags & MT_FLAG_HL_EOL);
-    return fake;
   }
+  Decoration fake = DECORATION_INIT;
+  fake.hl_id = mark.hl_id;
+  fake.priority = mark.priority;
+  fake.hl_eol = (mark.flags & MT_FLAG_HL_EOL);
+  return fake;
 }
 
-
-bool decor_redraw_start(buf_T *buf, int top_row, DecorState *state)
+/// @return true if decor has a virtual position (virtual text or ui_watched)
+static bool decor_virt_pos(Decoration decor)
 {
+  return kv_size(decor.virt_text) || decor.ui_watched;
+}
+
+bool decor_redraw_start(win_T *wp, int top_row, DecorState *state)
+{
+  buf_T *buf = wp->w_buffer;
   state->top_row = top_row;
   marktree_itr_get(buf->b_marktree, top_row, 0, state->itr);
   if (!state->itr->node) {
@@ -194,33 +221,25 @@ bool decor_redraw_start(buf_T *buf, int top_row, DecorState *state)
 
     Decoration decor = get_decor(mark);
 
-    // Exclude non-paired marks unless they contain virt_text or a sign
-    if (!mt_paired(mark)
-        && !kv_size(decor.virt_text)
-        && !decor_has_sign(&decor)) {
-      goto next_mark;
-    }
-
     mtpos_t altpos = marktree_get_altpos(buf->b_marktree, mark, NULL);
 
     // Exclude start marks if the end mark position is above the top row
     // Exclude end marks if we have already added the start mark
-    if ((mt_start(mark) && altpos.row < top_row
-         && !kv_size(decor.virt_text))
+    if ((mt_start(mark) && altpos.row < top_row && !decor_virt_pos(decor))
         || (mt_end(mark) && altpos.row >= top_row)) {
       goto next_mark;
     }
 
     if (mt_end(mark)) {
       decor_add(state, altpos.row, altpos.col, mark.pos.row, mark.pos.col,
-                &decor, false);
+                &decor, false, mark.ns, mark.id);
     } else {
       if (altpos.row == -1) {
         altpos.row = mark.pos.row;
         altpos.col = mark.pos.col;
       }
       decor_add(state, mark.pos.row, mark.pos.col, altpos.row, altpos.col,
-                &decor, false);
+                &decor, false, mark.ns, mark.id);
     }
 
 next_mark:
@@ -234,10 +253,10 @@ next_mark:
   return true;  // TODO(bfredl): check if available in the region
 }
 
-bool decor_redraw_line(buf_T *buf, int row, DecorState *state)
+bool decor_redraw_line(win_T *wp, int row, DecorState *state)
 {
   if (state->row == -1) {
-    decor_redraw_start(buf, row, state);
+    decor_redraw_start(wp, row, state);
   }
   state->row = row;
   state->col_until = -1;
@@ -246,28 +265,29 @@ bool decor_redraw_line(buf_T *buf, int row, DecorState *state)
 }
 
 static void decor_add(DecorState *state, int start_row, int start_col, int end_row, int end_col,
-                      Decoration *decor, bool owned)
+                      Decoration *decor, bool owned, uint64_t ns_id, uint64_t mark_id)
 {
   int attr_id = decor->hl_id > 0 ? syn_id2attr(decor->hl_id) : 0;
 
   DecorRange range = { start_row, start_col, end_row, end_col,
                        *decor, attr_id,
-                       kv_size(decor->virt_text) && owned, -1 };
+                       kv_size(decor->virt_text) && owned, -1, ns_id, mark_id };
 
   kv_pushp(state->active);
   size_t index;
-  for (index = kv_size(state->active)-1; index > 0; index--) {
-    DecorRange item = kv_A(state->active, index-1);
+  for (index = kv_size(state->active) - 1; index > 0; index--) {
+    DecorRange item = kv_A(state->active, index - 1);
     if (item.decor.priority <= range.decor.priority) {
       break;
     }
-    kv_A(state->active, index) = kv_A(state->active, index-1);
+    kv_A(state->active, index) = kv_A(state->active, index - 1);
   }
   kv_A(state->active, index) = range;
 }
 
-int decor_redraw_col(buf_T *buf, int col, int win_col, bool hidden, DecorState *state)
+int decor_redraw_col(win_T *wp, int col, int win_col, bool hidden, DecorState *state)
 {
+  buf_T *buf = wp->w_buffer;
   if (col <= state->col_until) {
     return state->current;
   }
@@ -279,7 +299,7 @@ int decor_redraw_col(buf_T *buf, int col, int win_col, bool hidden, DecorState *
     if (mark.pos.row < 0 || mark.pos.row > state->row) {
       break;
     } else if (mark.pos.row == state->row && mark.pos.col > col) {
-      state->col_until = mark.pos.col-1;
+      state->col_until = mark.pos.col - 1;
       break;
     }
 
@@ -296,15 +316,8 @@ int decor_redraw_col(buf_T *buf, int col, int win_col, bool hidden, DecorState *
       endpos = mark.pos;
     }
 
-    if (endpos.row < mark.pos.row
-        || (endpos.row == mark.pos.row && endpos.col <= mark.pos.col)) {
-      if (!kv_size(decor.virt_text)) {
-        goto next_mark;
-      }
-    }
-
     decor_add(state, mark.pos.row, mark.pos.col, endpos.row, endpos.col,
-              &decor, false);
+              &decor, false, mark.ns, mark.id);
 
 next_mark:
     marktree_itr_next(buf->b_marktree, state->itr);
@@ -315,13 +328,14 @@ next_mark:
   bool conceal = 0;
   int conceal_char = 0;
   int conceal_attr = 0;
+  TriState spell = kNone;
 
   for (size_t i = 0; i < kv_size(state->active); i++) {
     DecorRange item = kv_A(state->active, i);
     bool active = false, keep = true;
     if (item.end_row < state->row
         || (item.end_row == state->row && item.end_col <= col)) {
-      if (!(item.start_row >= state->row && kv_size(item.decor.virt_text))) {
+      if (!(item.start_row >= state->row && decor_virt_pos(item.decor))) {
         keep = false;
       }
     } else {
@@ -329,11 +343,11 @@ next_mark:
           || (item.start_row == state->row && item.start_col <= col)) {
         active = true;
         if (item.end_row == state->row && item.end_col > col) {
-          state->col_until = MIN(state->col_until, item.end_col-1);
+          state->col_until = MIN(state->col_until, item.end_col - 1);
         }
       } else {
         if (item.start_row == state->row) {
-          state->col_until = MIN(state->col_until, item.start_col-1);
+          state->col_until = MIN(state->col_until, item.start_col - 1);
         }
       }
     }
@@ -348,8 +362,11 @@ next_mark:
         conceal_attr = item.attr_id;
       }
     }
+    if (active && item.decor.spell != kNone) {
+      spell = item.decor.spell;
+    }
     if ((item.start_row == state->row && item.start_col <= col)
-        && kv_size(item.decor.virt_text)
+        && decor_virt_pos(item.decor)
         && item.decor.virt_text_pos == kVTOverlay && item.win_col == -1) {
       item.win_col = (item.decor.virt_text_hide && hidden) ? -2 : win_col;
     }
@@ -364,10 +381,12 @@ next_mark:
   state->conceal = conceal;
   state->conceal_char = conceal_char;
   state->conceal_attr = conceal_attr;
+  state->spell = spell;
   return attr;
 }
 
-void decor_redraw_signs(buf_T *buf, int row, int *num_signs, sign_attrs_T sattrs[])
+void decor_redraw_signs(buf_T *buf, int row, int *num_signs, SignTextAttrs sattrs[],
+                        HlPriId *num_id, HlPriId *line_id, HlPriId *cul_id)
 {
   if (!buf->b_signs) {
     return;
@@ -392,30 +411,39 @@ void decor_redraw_signs(buf_T *buf, int row, int *num_signs, sign_attrs_T sattrs
       goto next_mark;
     }
 
-    int j;
-    for (j = (*num_signs); j > 0; j--) {
-      if (sattrs[j].sat_prio <= decor->priority) {
-        break;
+    if (decor->sign_text) {
+      int j;
+      for (j = (*num_signs); j > 0; j--) {
+        if (sattrs[j - 1].priority >= decor->priority) {
+          break;
+        }
+        if (j < SIGN_SHOW_MAX) {
+          sattrs[j] = sattrs[j - 1];
+        }
+      }
+      if (j < SIGN_SHOW_MAX) {
+        sattrs[j] = (SignTextAttrs) {
+          .text = decor->sign_text,
+          .hl_id = decor->sign_hl_id,
+          .priority = decor->priority
+        };
+        (*num_signs)++;
       }
-      sattrs[j] = sattrs[j-1];
     }
-    if (j < SIGN_SHOW_MAX) {
-      memset(&sattrs[j], 0, sizeof(sign_attrs_T));
-      sattrs[j].sat_text = decor->sign_text;
-      if (decor->sign_hl_id != 0) {
-        sattrs[j].sat_texthl = syn_id2attr(decor->sign_hl_id);
+
+    struct { HlPriId *dest; int hl; } cattrs[] = {
+      { line_id, decor->line_hl_id        },
+      { num_id,  decor->number_hl_id      },
+      { cul_id,  decor->cursorline_hl_id  },
+      { NULL, -1 },
+    };
+    for (int i = 0; cattrs[i].dest; i++) {
+      if (cattrs[i].hl != 0 && decor->priority >= cattrs[i].dest->priority) {
+        *cattrs[i].dest = (HlPriId) {
+          .hl_id = cattrs[i].hl,
+          .priority = decor->priority
+        };
       }
-      if (decor->number_hl_id != 0) {
-        sattrs[j].sat_numhl = syn_id2attr(decor->number_hl_id);
-      }
-      if (decor->line_hl_id != 0) {
-        sattrs[j].sat_linehl = syn_id2attr(decor->line_hl_id);
-      }
-      if (decor->cursorline_hl_id != 0) {
-        sattrs[j].sat_culhl = syn_id2attr(decor->cursorline_hl_id);
-      }
-      sattrs[j].sat_prio = decor->priority;
-      (*num_signs)++;
     }
 
 next_mark:
@@ -432,11 +460,11 @@ int decor_signcols(buf_T *buf, DecorState *state, int row, int end_row, int max)
   int signcols = 0;      // highest value of count
   int currow = -1;       // current row
 
-  if (max <= 1 && buf->b_signs >= (size_t)max) {
+  if (max <= 1 && buf->b_signs_with_text >= (size_t)max) {
     return max;
   }
 
-  if (buf->b_signs == 0) {
+  if (buf->b_signs_with_text == 0) {
     return 0;
   }
 
@@ -507,17 +535,17 @@ next_mark:
 
 void decor_redraw_end(DecorState *state)
 {
-  state->buf = NULL;
+  state->win = NULL;
 }
 
-bool decor_redraw_eol(buf_T *buf, DecorState *state, int *eol_attr, int eol_col)
+bool decor_redraw_eol(win_T *wp, DecorState *state, int *eol_attr, int eol_col)
 {
-  decor_redraw_col(buf, MAXCOL, MAXCOL, false, state);
+  decor_redraw_col(wp, MAXCOL, MAXCOL, false, state);
   state->eol_col = eol_col;
   bool has_virttext = false;
   for (size_t i = 0; i < kv_size(state->active); i++) {
     DecorRange item = kv_A(state->active, i);
-    if (item.start_row == state->row && kv_size(item.decor.virt_text)) {
+    if (item.start_row == state->row && decor_virt_pos(item.decor)) {
       has_virttext = true;
     }
 
@@ -528,18 +556,18 @@ bool decor_redraw_eol(buf_T *buf, DecorState *state, int *eol_attr, int eol_col)
   return has_virttext;
 }
 
-void decor_add_ephemeral(int start_row, int start_col, int end_row, int end_col, Decoration *decor)
+void decor_add_ephemeral(int start_row, int start_col, int end_row, int end_col, Decoration *decor,
+                         uint64_t ns_id, uint64_t mark_id)
 {
   if (end_row == -1) {
     end_row = start_row;
     end_col = start_col;
   }
-  decor_add(&decor_state, start_row, start_col, end_row, end_col, decor, true);
+  decor_add(&decor_state, start_row, start_col, end_row, end_col, decor, true, ns_id, mark_id);
 }
 
-
-
-int decor_virt_lines(win_T *wp, linenr_T lnum, VirtLines *lines)
+/// @param has_fold  whether line "lnum" has a fold, or kNone when not calculated yet
+int decor_virt_lines(win_T *wp, linenr_T lnum, VirtLines *lines, TriState has_fold)
 {
   buf_T *buf = wp->w_buffer;
   if (!buf->b_virt_line_blocks) {
@@ -548,21 +576,34 @@ int decor_virt_lines(win_T *wp, linenr_T lnum, VirtLines *lines)
     return 0;
   }
 
+  assert(lnum > 0);
+  bool below_fold = lnum > 1 && hasFoldingWin(wp, lnum - 1, NULL, NULL, true, NULL);
+  if (has_fold == kNone) {
+    has_fold = hasFoldingWin(wp, lnum, NULL, NULL, true, NULL);
+  }
+
+  const int row = lnum - 1;
+  const int start_row = below_fold ? row : MAX(row - 1, 0);
+  const int end_row = has_fold ? row : row + 1;
+  if (start_row >= end_row) {
+    return 0;
+  }
+
   int virt_lines = 0;
-  int row = (int)MAX(lnum - 2, 0);
-  int end_row = (int)lnum;
   MarkTreeIter itr[1] = { 0 };
-  marktree_itr_get(buf->b_marktree, row, 0,  itr);
+  marktree_itr_get(buf->b_marktree, start_row, 0, itr);
   while (true) {
     mtkey_t mark = marktree_itr_current(itr);
     if (mark.pos.row < 0 || mark.pos.row >= end_row) {
       break;
-    } else if (marktree_decor_level(mark) < kDecorLevelVirtLine) {
+    } else if (mt_end(mark)
+               || marktree_decor_level(mark) < kDecorLevelVirtLine
+               || !mark.decor_full) {
       goto next_mark;
     }
-    bool above = mark.pos.row > (int)(lnum - 2);
-    Decoration *decor = mark.decor_full;
-    if (decor && decor->virt_lines_above == above) {
+    Decoration *const decor = mark.decor_full;
+    const int draw_row = mark.pos.row + (decor->virt_lines_above ? 0 : 1);
+    if (draw_row == row) {
       virt_lines += (int)kv_size(decor->virt_lines);
       if (lines) {
         kv_splice(*lines, decor->virt_lines);
