@@ -8,39 +8,51 @@
 #include <assert.h>
 #include <inttypes.h>
 #include <stdbool.h>
+#include <string.h>
 
 #include "nvim/ascii.h"
+#include "nvim/buffer_defs.h"
 #include "nvim/charset.h"
 #include "nvim/digraph.h"
+#include "nvim/drawscreen.h"
+#include "nvim/eval.h"
 #include "nvim/eval/typval.h"
-#include "nvim/ex_cmds2.h"
+#include "nvim/eval/typval_defs.h"
+#include "nvim/ex_cmds_defs.h"
 #include "nvim/ex_docmd.h"
 #include "nvim/ex_getln.h"
 #include "nvim/garray.h"
 #include "nvim/getchar.h"
+#include "nvim/gettext.h"
+#include "nvim/globals.h"
+#include "nvim/highlight_defs.h"
+#include "nvim/keycodes.h"
+#include "nvim/mapping.h"
 #include "nvim/mbyte.h"
 #include "nvim/memory.h"
 #include "nvim/message.h"
 #include "nvim/normal.h"
+#include "nvim/option_defs.h"
 #include "nvim/os/input.h"
-#include "nvim/screen.h"
+#include "nvim/runtime.h"
 #include "nvim/strings.h"
+#include "nvim/types.h"
 #include "nvim/vim.h"
 
 typedef int result_T;
 
 typedef struct digraph {
-  char_u char1;
-  char_u char2;
+  uint8_t char1;
+  uint8_t char2;
   result_T result;
 } digr_T;
 
-static char e_digraph_must_be_just_two_characters_str[]
-    = N_("E1214: Digraph must be just two characters: %s");
-static char e_digraph_argument_must_be_one_character_str[]
-    = N_("E1215: Digraph must be one character: %s");
-static char e_digraph_setlist_argument_must_be_list_of_lists_with_two_items[]
-    = N_("E1216: digraph_setlist() argument must be a list of lists with two items");
+static const char e_digraph_must_be_just_two_characters_str[]
+  = N_("E1214: Digraph must be just two characters: %s");
+static const char e_digraph_argument_must_be_one_character_str[]
+  = N_("E1215: Digraph must be one character: %s");
+static const char e_digraph_setlist_argument_must_be_list_of_lists_with_two_items[]
+  = N_("E1216: digraph_setlist() argument must be a list of lists with two items");
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "digraph.c.generated.h"
@@ -51,7 +63,6 @@ static garray_T user_digraphs = { 0, 0, (int)sizeof(digr_T), 10, NULL };
 /// Note: Characters marked with XX are not included literally, because some
 /// compilers cannot handle them (Amiga SAS/C is the most picky one).
 static digr_T digraphdefault[] =
-
 // digraphs for Unicode from RFC1345
 // (also work for ISO-8859-1 aka latin1)
 {
@@ -1479,11 +1490,11 @@ int do_digraph(int c)
 
 /// Find a digraph for "val".  If found return the string to display it.
 /// If not found return NULL.
-char_u *get_digraph_for_char(int val_arg)
+char *get_digraph_for_char(int val_arg)
 {
   const int val = val_arg;
   const digr_T *dp;
-  static char_u r[3];
+  static char r[3];
 
   for (int use_defaults = 0; use_defaults <= 1; use_defaults++) {
     if (use_defaults == 0) {
@@ -1493,8 +1504,8 @@ char_u *get_digraph_for_char(int val_arg)
     }
     for (int i = 0; use_defaults ? dp->char1 != NUL : i < user_digraphs.ga_len; i++) {
       if (dp->result == val) {
-        r[0] = dp->char1;
-        r[1] = dp->char2;
+        r[0] = (char)dp->char1;
+        r[1] = (char)dp->char2;
         r[2] = NUL;
         return r;
       }
@@ -1513,31 +1524,36 @@ char_u *get_digraph_for_char(int val_arg)
 int get_digraph(bool cmdline)
 {
   no_mapping++;
+  allow_keys++;
   int c = plain_vgetc();
   no_mapping--;
+  allow_keys--;
 
-  if (c != ESC) {
+  if (c == ESC) {  // ESC cancels CTRL-K
+    return NUL;
+  }
+
+  if (IS_SPECIAL(c)) {
+    // insert special key code
+    return c;
+  }
+
+  if (cmdline) {
+    if ((char2cells(c) == 1) && c < 128 && (cmdline_star == 0)) {
+      putcmdline((char)c, true);
+    }
+  } else {
+    add_to_showcmd(c);
+  }
+  no_mapping++;
+  allow_keys++;
+  int cc = plain_vgetc();
+  no_mapping--;
+  allow_keys--;
+
+  if (cc != ESC) {
     // ESC cancels CTRL-K
-    if (IS_SPECIAL(c)) {
-      // insert special key code
-      return c;
-    }
-
-    if (cmdline) {
-      if ((char2cells(c) == 1) && c < 128 && (cmdline_star == 0)) {
-        putcmdline((char)c, true);
-      }
-    } else {
-      add_to_showcmd(c);
-    }
-    no_mapping++;
-    int cc = plain_vgetc();
-    no_mapping--;
-
-    if (cc != ESC) {
-      // ESC cancels CTRL-K
-      return digraph_get(c, cc, true);
-    }
+    return digraph_get(c, cc, true);
   }
   return NUL;
 }
@@ -1551,6 +1567,7 @@ int get_digraph(bool cmdline)
 /// @return If no match, return "char2". If "meta_char" is true and "char1"
 //          is a space, return "char2" | 0x80.
 static int getexactdigraph(int char1, int char2, bool meta_char)
+  FUNC_ATTR_PURE
 {
   int retval = 0;
 
@@ -1572,7 +1589,7 @@ static int getexactdigraph(int char1, int char2, bool meta_char)
   if (retval == 0) {
     dp = digraphdefault;
 
-    for (int i = 0; dp->char1 != 0; i++) {
+    while (dp->char1 != 0) {
       if (((int)dp->char1 == char1) && ((int)dp->char2 == char2)) {
         retval = dp->result;
         break;
@@ -1601,6 +1618,7 @@ static int getexactdigraph(int char1, int char2, bool meta_char)
 ///
 /// @return The digraph.
 int digraph_get(int char1, int char2, bool meta_char)
+  FUNC_ATTR_PURE
 {
   int retval;
 
@@ -1628,8 +1646,8 @@ static void registerdigraph(int char1, int char2, int n)
 
   // Add a new digraph to the table.
   dp = GA_APPEND_VIA_PTR(digr_T, &user_digraphs);
-  dp->char1 = (char_u)char1;
-  dp->char2 = (char_u)char2;
+  dp->char1 = (uint8_t)char1;
+  dp->char2 = (uint8_t)char2;
   dp->result = n;
 }
 
@@ -1639,7 +1657,7 @@ static void registerdigraph(int char1, int char2, int n)
 bool check_digraph_chars_valid(int char1, int char2)
 {
   if (char2 == 0) {
-    char_u msg[MB_MAXBYTES + 1];
+    char msg[MB_MAXBYTES + 1];
     msg[utf_char2bytes(char1, msg)] = NUL;
     semsg(_(e_digraph_must_be_just_two_characters_str), msg);
     return false;
@@ -1655,7 +1673,7 @@ bool check_digraph_chars_valid(int char1, int char2)
 /// format: {c1}{c2} char {c1}{c2} char ...
 ///
 /// @param str
-void putdigraph(char_u *str)
+void putdigraph(char *str)
 {
   while (*str != NUL) {
     str = skipwhite(str);
@@ -1663,8 +1681,8 @@ void putdigraph(char_u *str)
     if (*str == NUL) {
       return;
     }
-    char_u char1 = *str++;
-    char_u char2 = *str++;
+    uint8_t char1 = (uint8_t)(*str++);
+    uint8_t char2 = (uint8_t)(*str++);
 
     if (!check_digraph_chars_valid(char1, char2)) {
       return;
@@ -1688,7 +1706,7 @@ static void digraph_header(const char *msg)
   if (msg_col > 0) {
     msg_putchar('\n');
   }
-  msg_outtrans_attr((const char_u *)msg, HL_ATTR(HLF_CM));
+  msg_outtrans_attr(msg, HL_ATTR(HLF_CM));
   msg_putchar('\n');
 }
 
@@ -1700,7 +1718,7 @@ void listdigraphs(bool use_headers)
 
   const digr_T *dp = digraphdefault;
 
-  for (int i = 0; dp->char1 != NUL && !got_int; i++) {
+  while (dp->char1 != NUL && !got_int) {
     digr_T tmp;
 
     // May need to convert the result to 'encoding'.
@@ -1732,16 +1750,16 @@ static void digraph_getlist_appendpair(const digr_T *dp, list_T *l)
   list_T *l2 = tv_list_alloc(2);
   tv_list_append_list(l, l2);
 
-  char_u buf[30];
-  buf[0] = dp->char1;
-  buf[1] = dp->char2;
+  char buf[30];
+  buf[0] = (char)dp->char1;
+  buf[1] = (char)dp->char2;
   buf[2] = NUL;
-  tv_list_append_string(l2, (char *)buf, -1);
+  tv_list_append_string(l2, buf, -1);
 
-  char_u *p = buf;
+  char *p = buf;
   p += utf_char2bytes(dp->result, p);
   *p = NUL;
-  tv_list_append_string(l2, (char *)buf, -1);
+  tv_list_append_string(l2, buf, -1);
 }
 
 void digraph_getlist_common(bool list_all, typval_T *rettv)
@@ -1752,7 +1770,7 @@ void digraph_getlist_common(bool list_all, typval_T *rettv)
 
   if (list_all) {
     dp = digraphdefault;
-    for (int i = 0; dp->char1 != NUL && !got_int; i++) {
+    while (dp->char1 != NUL && !got_int) {
       digr_T tmp;
       tmp.char1 = dp->char1;
       tmp.char2 = dp->char2;
@@ -1807,58 +1825,59 @@ struct dg_header_entry {
 static void printdigraph(const digr_T *dp, result_T *previous)
   FUNC_ATTR_NONNULL_ARG(1)
 {
-  char_u buf[30];
+  char buf[30];
   int list_width = 13;
 
-  if (dp->result != 0) {
-    if (previous != NULL) {
-      for (int i = 0; header_table[i].dg_header != NULL; i++) {
-        if (*previous < header_table[i].dg_start
-            && dp->result >= header_table[i].dg_start
-            && dp->result < header_table[i + 1].dg_start) {
-          digraph_header(_(header_table[i].dg_header));
-          break;
-        }
-      }
-      *previous = dp->result;
-    }
-    if (msg_col > Columns - list_width) {
-      msg_putchar('\n');
-    }
-
-
-    // Make msg_col a multiple of list_width by using spaces.
-    if (msg_col % list_width != 0) {
-      int spaces = (msg_col / list_width + 1) * list_width - msg_col;
-      while (spaces--) {
-        msg_putchar(' ');
-      }
-    }
-
-    char_u *p = &buf[0];
-    *p++ = dp->char1;
-    *p++ = dp->char2;
-    *p++ = ' ';
-    *p = NUL;
-    msg_outtrans(buf);
-    p = buf;
-
-    // add a space to draw a composing char on
-    if (utf_iscomposing(dp->result)) {
-      *p++ = ' ';
-    }
-    p += utf_char2bytes(dp->result, p);
-
-    *p = NUL;
-    msg_outtrans_attr(buf, HL_ATTR(HLF_8));
-    p = buf;
-    if (char2cells(dp->result) == 1) {
-      *p++ = ' ';
-    }
-    assert(p >= buf);
-    vim_snprintf((char *)p, sizeof(buf) - (size_t)(p - buf), " %3d", dp->result);
-    msg_outtrans(buf);
+  if (dp->result == 0) {
+    return;
   }
+
+  if (previous != NULL) {
+    for (int i = 0; header_table[i].dg_header != NULL; i++) {
+      if (*previous < header_table[i].dg_start
+          && dp->result >= header_table[i].dg_start
+          && dp->result < header_table[i + 1].dg_start) {
+        digraph_header(_(header_table[i].dg_header));
+        break;
+      }
+    }
+    *previous = dp->result;
+  }
+  if (msg_col > Columns - list_width) {
+    msg_putchar('\n');
+  }
+
+  // Make msg_col a multiple of list_width by using spaces.
+  if (msg_col % list_width != 0) {
+    int spaces = (msg_col / list_width + 1) * list_width - msg_col;
+    while (spaces--) {
+      msg_putchar(' ');
+    }
+  }
+
+  char *p = &buf[0];
+  *p++ = (char)dp->char1;
+  *p++ = (char)dp->char2;
+  *p++ = ' ';
+  *p = NUL;
+  msg_outtrans(buf);
+  p = buf;
+
+  // add a space to draw a composing char on
+  if (utf_iscomposing(dp->result)) {
+    *p++ = ' ';
+  }
+  p += utf_char2bytes(dp->result, p);
+
+  *p = NUL;
+  msg_outtrans_attr(buf, HL_ATTR(HLF_8));
+  p = buf;
+  if (char2cells(dp->result) == 1) {
+    *p++ = ' ';
+  }
+  assert(p >= buf);
+  vim_snprintf(p, sizeof(buf) - (size_t)(p - buf), " %3d", dp->result);
+  msg_outtrans(buf);
 }
 
 /// Get the two digraph characters from a typval.
@@ -1867,7 +1886,7 @@ static int get_digraph_chars(const typval_T *arg, int *char1, int *char2)
 {
   char buf_chars[NUMBUFLEN];
   const char *chars = tv_get_string_buf_chk(arg, buf_chars);
-  const char_u *p = (const char_u *)chars;
+  const char *p = chars;
 
   if (p != NULL) {
     if (*p != NUL) {
@@ -1899,7 +1918,7 @@ static bool digraph_set_common(const typval_T *argchars, const typval_T *argdigr
   if (digraph == NULL) {
     return false;
   }
-  const char_u *p = (const char_u *)digraph;
+  const char *p = digraph;
   int n = mb_cptr2char_adv(&p);
   if (*p != NUL) {
     semsg(_(e_digraph_argument_must_be_one_character_str), digraph);
@@ -1911,7 +1930,7 @@ static bool digraph_set_common(const typval_T *argchars, const typval_T *argdigr
 }
 
 /// "digraph_get()" function
-void f_digraph_get(typval_T *argvars, typval_T *rettv, FunPtr fptr)
+void f_digraph_get(typval_T *argvars, typval_T *rettv, EvalFuncData fptr)
 {
   rettv->v_type = VAR_STRING;
   rettv->vval.v_string = NULL;  // Return empty string for failure
@@ -1920,19 +1939,19 @@ void f_digraph_get(typval_T *argvars, typval_T *rettv, FunPtr fptr)
   if (digraphs == NULL) {
     return;
   }
-  if (STRLEN(digraphs) != 2) {
+  if (strlen(digraphs) != 2) {
     semsg(_(e_digraph_must_be_just_two_characters_str), digraphs);
     return;
   }
   int code = digraph_get(digraphs[0], digraphs[1], false);
 
-  char_u buf[NUMBUFLEN];
+  char buf[NUMBUFLEN];
   buf[utf_char2bytes(code, buf)] = NUL;
-  rettv->vval.v_string = vim_strsave(buf);
+  rettv->vval.v_string = xstrdup(buf);
 }
 
 /// "digraph_getlist()" function
-void f_digraph_getlist(typval_T *argvars, typval_T *rettv, FunPtr fptr)
+void f_digraph_getlist(typval_T *argvars, typval_T *rettv, EvalFuncData fptr)
 {
   bool flag_list_all;
 
@@ -1951,7 +1970,7 @@ void f_digraph_getlist(typval_T *argvars, typval_T *rettv, FunPtr fptr)
 }
 
 /// "digraph_set()" function
-void f_digraph_set(typval_T *argvars, typval_T *rettv, FunPtr fptr)
+void f_digraph_set(typval_T *argvars, typval_T *rettv, EvalFuncData fptr)
 {
   rettv->v_type = VAR_BOOL;
   rettv->vval.v_bool = kBoolVarFalse;
@@ -1964,7 +1983,7 @@ void f_digraph_set(typval_T *argvars, typval_T *rettv, FunPtr fptr)
 }
 
 /// "digraph_setlist()" function
-void f_digraph_setlist(typval_T *argvars, typval_T *rettv, FunPtr fptr)
+void f_digraph_setlist(typval_T *argvars, typval_T *rettv, EvalFuncData fptr)
 {
   rettv->v_type = VAR_BOOL;
   rettv->vval.v_bool = kBoolVarFalse;
@@ -2004,13 +2023,12 @@ void f_digraph_setlist(typval_T *argvars, typval_T *rettv, FunPtr fptr)
 
 /// structure used for b_kmap_ga.ga_data
 typedef struct {
-  char_u *from;
-  char_u *to;
+  char *from;
+  char *to;
 } kmap_T;
 
 #define KMAP_MAXLEN 20  // maximum length of "from" or "to"
 
-
 /// Set up key mapping tables for the 'keymap' option.
 ///
 /// @return NULL if OK, an error message for failure.  This only needs to be
@@ -2031,7 +2049,7 @@ char *keymap_init(void)
 
     // Source the keymap file.  It will contain a ":loadkeymap" command
     // which will call ex_loadkeymap() below.
-    buflen = STRLEN(curbuf->b_p_keymap) + STRLEN(p_enc) + 14;
+    buflen = strlen(curbuf->b_p_keymap) + strlen(p_enc) + 14;
     buf = xmalloc(buflen);
 
     // try finding "keymap/'keymap'_'encoding'.vim"  in 'runtimepath'
@@ -2059,11 +2077,11 @@ char *keymap_init(void)
 /// @param eap
 void ex_loadkeymap(exarg_T *eap)
 {
-  char_u *s;
+  char *s;
 
 #define KMAP_LLEN 200  // max length of "to" and "from" together
-  char_u buf[KMAP_LLEN + 11];
-  char_u *save_cpo = p_cpo;
+  char buf[KMAP_LLEN + 11];
+  char *save_cpo = p_cpo;
 
   if (!getline_equal(eap->getline, eap->cookie, getsourceline)) {
     emsg(_("E105: Using :loadkeymap not in a sourced file"));
@@ -2077,27 +2095,27 @@ void ex_loadkeymap(exarg_T *eap)
   ga_init(&curbuf->b_kmap_ga, (int)sizeof(kmap_T), 20);
 
   // Set 'cpoptions' to "C" to avoid line continuation.
-  p_cpo = (char_u *)"C";
+  p_cpo = "C";
 
   // Get each line of the sourced file, break at the end.
   for (;;) {
-    char_u *line = eap->getline(0, eap->cookie, 0, true);
+    char *line = eap->getline(0, eap->cookie, 0, true);
 
     if (line == NULL) {
       break;
     }
 
-    char_u *p = skipwhite(line);
+    char *p = skipwhite(line);
 
     if ((*p != '"') && (*p != NUL)) {
       kmap_T *kp = GA_APPEND_VIA_PTR(kmap_T, &curbuf->b_kmap_ga);
       s = skiptowhite(p);
-      kp->from = vim_strnsave(p, (size_t)(s - p));
+      kp->from = xstrnsave(p, (size_t)(s - p));
       p = skipwhite(s);
       s = skiptowhite(p);
-      kp->to = vim_strnsave(p, (size_t)(s - p));
+      kp->to = xstrnsave(p, (size_t)(s - p));
 
-      if ((STRLEN(kp->from) + STRLEN(kp->to) >= KMAP_LLEN)
+      if ((strlen(kp->from) + strlen(kp->to) >= KMAP_LLEN)
           || (*kp->from == NUL)
           || (*kp->to == NUL)) {
         if (*kp->to == NUL) {
@@ -2113,10 +2131,10 @@ void ex_loadkeymap(exarg_T *eap)
 
   // setup ":lmap" to map the keys
   for (int i = 0; i < curbuf->b_kmap_ga.ga_len; i++) {
-    vim_snprintf((char *)buf, sizeof(buf), "<buffer> %s %s",
+    vim_snprintf(buf, sizeof(buf), "<buffer> %s %s",
                  ((kmap_T *)curbuf->b_kmap_ga.ga_data)[i].from,
                  ((kmap_T *)curbuf->b_kmap_ga.ga_data)[i].to);
-    (void)do_map(0, buf, LANGMAP, false);
+    (void)do_map(MAPTYPE_MAP, buf, MODE_LANGMAP, false);
   }
 
   p_cpo = save_cpo;
@@ -2138,22 +2156,22 @@ void keymap_ga_clear(garray_T *kmap_ga)
 /// Stop using 'keymap'.
 static void keymap_unload(void)
 {
-  char_u buf[KMAP_MAXLEN + 10];
-  char_u *save_cpo = p_cpo;
+  char buf[KMAP_MAXLEN + 10];
+  char *save_cpo = p_cpo;
 
   if (!(curbuf->b_kmap_state & KEYMAP_LOADED)) {
     return;
   }
 
   // Set 'cpoptions' to "C" to avoid line continuation.
-  p_cpo = (char_u *)"C";
+  p_cpo = "C";
 
   // clear the ":lmap"s
   kmap_T *kp = (kmap_T *)curbuf->b_kmap_ga.ga_data;
 
   for (int i = 0; i < curbuf->b_kmap_ga.ga_len; i++) {
-    vim_snprintf((char *)buf, sizeof(buf), "<buffer> %s", kp[i].from);
-    (void)do_map(1, buf, LANGMAP, false);
+    vim_snprintf(buf, sizeof(buf), "<buffer> %s", kp[i].from);
+    (void)do_map(MAPTYPE_UNMAP, buf, MODE_LANGMAP, false);
   }
   keymap_ga_clear(&curbuf->b_kmap_ga);
 
@@ -2163,3 +2181,42 @@ static void keymap_unload(void)
   curbuf->b_kmap_state &= ~KEYMAP_LOADED;
   status_redraw_curbuf();
 }
+
+/// Get the value to show for the language mappings, active 'keymap'.
+///
+/// @param fmt  format string containing one %s item
+/// @param buf  buffer for the result
+/// @param len  length of buffer
+bool get_keymap_str(win_T *wp, char *fmt, char *buf, int len)
+{
+  char *p;
+
+  if (wp->w_buffer->b_p_iminsert != B_IMODE_LMAP) {
+    return false;
+  }
+
+  buf_T *old_curbuf = curbuf;
+  win_T *old_curwin = curwin;
+  char *s;
+
+  curbuf = wp->w_buffer;
+  curwin = wp;
+  STRCPY(buf, "b:keymap_name");       // must be writable
+  emsg_skip++;
+  s = p = eval_to_string(buf, NULL, false);
+  emsg_skip--;
+  curbuf = old_curbuf;
+  curwin = old_curwin;
+  if (p == NULL || *p == NUL) {
+    if (wp->w_buffer->b_kmap_state & KEYMAP_LOADED) {
+      p = wp->w_buffer->b_p_keymap;
+    } else {
+      p = "lang";
+    }
+  }
+  if (vim_snprintf(buf, (size_t)len, fmt, p) > len - 1) {
+    buf[0] = NUL;
+  }
+  xfree(s);
+  return buf[0] != NUL;
+}
