@@ -10,6 +10,7 @@
 --       Defaults to BASE_NONE (function cannot be used as a method).
 -- func  Name of the C function which implements the VimL function. Defaults to
 --       `f_{funcname}`.
+-- fast  Function can run in |api-fast| events. Defaults to false.
 
 local varargs = function(nr)
   return {nr}
@@ -22,22 +23,22 @@ end
 return {
   funcs={
     abs={args=1, base=1},
-    acos={args=1, base=1, func="float_op_wrapper", data="&acos"},  -- WJMc
+    acos={args=1, base=1, float_func="acos"},  -- WJMc
     add={args=2, base=1},
     ['and']={args=2, base=1},
-    api_info={},
+    api_info={fast=true},
     append={args=2, base=2},
     appendbufline={args=3, base=3},
     argc={args={0, 1}},
     argidx={},
     arglistid={args={0, 2}},
     argv={args={0, 2}},
-    asin={args=1, base=1, func="float_op_wrapper", data="&asin"},  -- WJMc
+    asin={args=1, base=1, float_func="asin"},  -- WJMc
     assert_beeps={args=1, base=1},
     assert_equal={args={2, 3}, base=2},
     assert_equalfile={args={2, 3}, base=1},
     assert_exception={args={1, 2}},
-    assert_fails={args={1, 3}, base=1},
+    assert_fails={args={1, 5}, base=1},
     assert_false={args={1, 2}, base=1},
     assert_inrange={args={3, 4}, base=3},
     assert_match={args={2, 3}, base=2},
@@ -46,8 +47,9 @@ return {
     assert_notmatch={args={2, 3}, base=2},
     assert_report={args=1, base=1},
     assert_true={args={1, 2}, base=1},
-    atan={args=1, base=1, func="float_op_wrapper", data="&atan"},
+    atan={args=1, base=1, float_func="atan"},
     atan2={args=2, base=1},
+    blob2list={args=1, base=1},
     browse={args=4},
     browsedir={args=2},
     bufadd={args=1, base=1},
@@ -63,30 +65,30 @@ return {
     bufwinid={args=1, base=1},
     bufwinnr={args=1, base=1},
     byte2line={args=1, base=1},
-    byteidx={args=2, base=1},
-    byteidxcomp={args=2, base=1},
+    byteidx={args=2, base=1, fast=true},
+    byteidxcomp={args=2, base=1, fast=true},
     call={args={2, 3}, base=1},
-    ceil={args=1, base=1, func="float_op_wrapper", data="&ceil"},
+    ceil={args=1, base=1, float_func="ceil"},
     changenr={},
     chanclose={args={1, 2}},
     chansend={args=2},
-    char2nr={args={1, 2}, base=1},
-    charcol={args=1, base=1},
+    char2nr={args={1, 2}, base=1, fast=true},
+    charclass={args=1, base=1},
+    charcol={args={1, 2}, base=1},
     charidx={args={2, 3}, base=1},
     chdir={args=1, base=1},
     cindent={args=1, base=1},
     clearmatches={args={0, 1}, base=1},
-    col={args=1, base=1},
+    col={args={1, 2}, base=1},
     complete={args=2, base=2},
     complete_add={args=1, base=1},
     complete_check={},
     complete_info={args={0, 1}, base=1},
     confirm={args={1, 4}, base=1},
     copy={args=1, base=1},
-    cos={args=1, base=1, func="float_op_wrapper", data="&cos"},
-    cosh={args=1, base=1, func="float_op_wrapper", data="&cosh"},
+    cos={args=1, base=1, float_func="cos"},
+    cosh={args=1, base=1, float_func="cosh"},
     count={args={2, 4}, base=1},
-    cscope_connection={args={0, 3}},
     ctxget={args={0, 1}},
     ctxpop={},
     ctxpush={args={0, 1}},
@@ -99,7 +101,7 @@ return {
     deletebufline={args={2,3}, base=1},
     dictwatcheradd={args=3},
     dictwatcherdel={args=3},
-    did_filetype={},
+    did_filetype={fast=true},
     diff_filler={args=1, base=1},
     diff_hlID={args=2, base=1},
     digraph_get={args=1, base=1},
@@ -107,31 +109,33 @@ return {
     digraph_set={args=2, base=1},
     digraph_setlist={args=1, base=1},
     empty={args=1, base=1},
-    environ={},
-    escape={args=2, base=1},
+    environ={fast=true},
+    escape={args=2, base=1, fast=true},
     eval={args=1, base=1},
     eventhandler={},
-    executable={args=1, base=1},
+    executable={args=1, base=1, fast=true},
     execute={args={1, 2}, base=1},
     exepath={args=1, base=1},
     exists={args=1, base=1},
-    exp={args=1, base=1, func="float_op_wrapper", data="&exp"},
+    exp={args=1, base=1, float_func="exp"},
     expand={args={1, 3}, base=1},
-    expandcmd={args=1, base=1},
+    expandcmd={args={1, 2}, base=1},
     extend={args={2, 3}, base=1},
+    extendnew={args={2, 3}, base=1},
     feedkeys={args={1, 2}, base=1},
     file_readable={args=1, base=1, func='f_filereadable'},  -- obsolete
-    filereadable={args=1, base=1},
-    filewritable={args=1, base=1},
+    filereadable={args=1, base=1, fast=true},
+    filewritable={args=1, base=1, fast=true},
     filter={args=2, base=1},
     finddir={args={1, 3}, base=1},
     findfile={args={1, 3}, base=1},
     flatten={args={1, 2}, base=1},
+    flattennew={args={1, 2}, base=1},
     float2nr={args=1, base=1},
-    floor={args=1, base=1, func="float_op_wrapper", data="&floor"},
+    floor={args=1, base=1, float_func="floor"},
     fmod={args=2, base=1},
-    fnameescape={args=1, base=1},
-    fnamemodify={args=2, base=1},
+    fnameescape={args=1, base=1, fast=true},
+    fnamemodify={args=2, base=1, fast=true},
     foldclosed={args=1, base=1},
     foldclosedend={args=1, base=1},
     foldlevel={args=1, base=1},
@@ -145,15 +149,19 @@ return {
     get={args={2, 3}, base=1},
     getbufinfo={args={0, 1}, base=1},
     getbufline={args={2, 3}, base=1},
+    getbufoneline={args=2, base=1},
     getbufvar={args={2, 3}, base=1},
+    getcellwidths={},
     getchangelist={args={0, 1}, base=1},
     getchar={args={0, 1}},
     getcharmod={},
     getcharpos={args=1, base=1},
     getcharsearch={},
     getcharstr={args={0, 1}},
+    getcmdcompltype={},
     getcmdline={},
     getcmdpos={},
+    getcmdscreenpos={},
     getcmdtype={},
     getcmdwintype={},
     getcompletion={args={2, 3}, base=1},
@@ -162,17 +170,17 @@ return {
     getcwd={args={0, 2}, base=1},
     getenv={args=1, base=1},
     getfontname={args={0, 1}},
-    getfperm={args=1, base=1},
-    getfsize={args=1, base=1},
-    getftime={args=1, base=1},
-    getftype={args=1, base=1},
+    getfperm={args=1, base=1, fast=true},
+    getfsize={args=1, base=1, fast=true},
+    getftime={args=1, base=1, fast=true},
+    getftype={args=1, base=1, fast=true},
     getjumplist={args={0, 2}, base=1},
     getline={args={1, 2}, base=1},
     getloclist={args={1, 2}},
     getmarklist={args={0, 1}, base=1},
     getmatches={args={0, 1}},
     getmousepos={},
-    getpid={},
+    getpid={fast=true},
     getpos={args=1, base=1},
     getqflist={args={0, 1}},
     getreg={args={0, 3}, base=1},
@@ -181,7 +189,9 @@ return {
     gettabinfo={args={0, 1}, base=1},
     gettabvar={args={2, 3}, base=1},
     gettabwinvar={args={3, 4}, base=1},
+    getscriptinfo={},
     gettagstack={args={0, 1}, base=1},
+    gettext={args=1, base=1},
     getwininfo={args={0, 1}, base=1},
     getwinpos={args={0, 1}, base=1},
     getwinposx={},
@@ -202,10 +212,11 @@ return {
     histnr={args=1, base=1},
     hlID={args=1, base=1},
     hlexists={args=1, base=1},
-    hostname={},
-    iconv={args=3, base=1},
+    hostname={fast=true},
+    iconv={args=3, base=1, fast=true},
     indent={args=1, base=1},
     index={args={2, 4}, base=1},
+    indexof={args={2, 3}, base=1},
     input={args={1, 3}, base=1},
     inputdialog={args={1, 3}, base=1},
     inputlist={args=1, base=1},
@@ -215,7 +226,7 @@ return {
     insert={args={2, 3}, base=1},
     interrupt={args=0},
     invert={args=1, base=1},
-    isdirectory={args=1, base=1},
+    isdirectory={args=1, base=1, fast=true},
     isinf={args=1, base=1},
     islocked={args=1, base=1},
     isnan={args=1, base=1},
@@ -232,6 +243,7 @@ return {
     json_decode={args=1, base=1},
     json_encode={args=1, base=1},
     keys={args=1, base=1},
+    keytrans={args=1, base=1},
     last_buffer_nr={},  -- obsolete
     len={args=1, base=1},
     libcall={args=3, base=3},
@@ -239,14 +251,16 @@ return {
     line={args={1, 2}, base=1},
     line2byte={args=1, base=1},
     lispindent={args=1, base=1},
+    list2blob={args=1, base=1},
     list2str={args={1, 2}, base=1},
     localtime={},
-    log={args=1, base=1, func="float_op_wrapper", data="&log"},
-    log10={args=1, base=1, func="float_op_wrapper", data="&log10"},
+    log={args=1, base=1, float_func="log"},
+    log10={args=1, base=1, float_func="log10"},
     luaeval={args={1, 2}, base=1},
     map={args=2, base=1},
     maparg={args={1, 4}, base=1},
     mapcheck={args={1, 3}, base=1},
+    mapset={args=3, base=1},
     match={args={2, 4}, base=1},
     matchadd={args={2, 5}, base=1},
     matchaddpos={args={2, 5}, base=1},
@@ -260,6 +274,7 @@ return {
     matchstrpos={args={2,4}, base=1},
     max={args=1, base=1},
     menu_get={args={1, 2}},
+    menu_info={args={1, 2}, base=1},
     min={args=1, base=1},
     mkdir={args={1, 3}, base=1},
     mode={args={0, 1}, base=1},
@@ -284,21 +299,22 @@ return {
     perleval={args=1, base=1},
     rand={args={0, 1}, base=1},
     range={args={1, 3}, base=1},
+    readblob={args={1, 3}, base=1},
     readdir={args={1, 2}, base=1},
     readfile={args={1, 3}, base=1},
     reduce={args={2, 3}, base=1},
     reg_executing={},
     reg_recording={},
     reg_recorded={},
-    reltime={args={0, 2}, base=1},
-    reltimefloat={args=1, base=1},
-    reltimestr={args=1, base=1},
+    reltime={args={0, 2}, base=1, fast=true},
+    reltimefloat={args=1, base=1, fast=true},
+    reltimestr={args=1, base=1, fast=true},
     remove={args={2, 3}, base=1},
     rename={args=2, base=1},
-    ['repeat']={args=2, base=1},
-    resolve={args=1, base=1},
+    ['repeat']={args=2, base=1, fast=true},
+    resolve={args=1, base=1, fast=true},
     reverse={args=1, base=1},
-    round={args=1, base=1, func="float_op_wrapper", data="&round"},
+    round={args=1, base=1, float_func="round"},
     rpcnotify={args=varargs(2)},
     rpcrequest={args=varargs(2)},
     rpcstart={args={1, 2}},
@@ -322,9 +338,11 @@ return {
     serverstop={args=1},
     setbufline={args=3, base=3},
     setbufvar={args=3, base=3},
+    setcellwidths={args=1, base=1},
     setcharpos={args=2, base=2},
     setcharsearch={args=1, base=1},
     setcmdpos={args=1, base=1},
+    setcmdline={args={1, 2}, base=1},
     setcursorcharpos={args={1, 3}, base=1},
     setenv={args=2, base=2},
     setfperm={args=2, base=1},
@@ -351,8 +369,8 @@ return {
     sign_unplace={args={1, 2}, base=1},
     sign_unplacelist={args=1, base=1},
     simplify={args=1, base=1},
-    sin={args=1, base=1, func="float_op_wrapper", data="&sin"},
-    sinh={args=1, base=1, func="float_op_wrapper", data="&sinh"},
+    sin={args=1, base=1, float_func="sin"},
+    sinh={args=1, base=1, float_func="sinh"},
     sockconnect={args={2,3}},
     sort={args={1, 3}, base=1},
     soundfold={args=1, base=1},
@@ -360,25 +378,26 @@ return {
     spellbadword={args={0, 1}, base=1},
     spellsuggest={args={1, 3}, base=1},
     split={args={1, 3}, base=1},
-    sqrt={args=1, base=1, func="float_op_wrapper", data="&sqrt"},
+    sqrt={args=1, base=1, float_func="sqrt"},
     srand={args={0, 1}, base=1},
-    stdpath={args=1},
+    stdpath={args=1, fast=true},
     str2float={args=1, base=1},
     str2list={args={1, 2}, base=1},
     str2nr={args={1, 3}, base=1},
-    strcharpart={args={2, 3}, base=1},
+    strcharlen={args=1, base=1},
+    strcharpart={args={2, 3}, base=1, fast=true},
     strchars={args={1, 2}, base=1},
     strdisplaywidth={args={1, 2}, base=1},
     strftime={args={1, 2}, base=1},
     strgetchar={args=2, base=1},
-    stridx={args={2, 3}, base=1},
+    stridx={args={2, 3}, base=1, fast=true},
     string={args=1, base=1},
     strlen={args=1, base=1},
-    strpart={args={2, 4}, base=1},
+    strpart={args={2, 4}, base=1, fast=true},
     strptime={args=2, base=1},
     strridx={args={2, 3}, base=1},
-    strtrans={args=1, base=1},
-    strwidth={args=1, base=1},
+    strtrans={args=1, base=1, fast=true},
+    strwidth={args=1, base=1, fast=true},
     submatch={args={1, 2}, base=1},
     substitute={args=4, base=1},
     swapinfo={args=1, base=1},
@@ -395,8 +414,8 @@ return {
     tabpagewinnr={args={1, 2}, base=1},
     tagfiles={},
     taglist={args={1, 2}, base=1},
-    tan={args=1, base=1, func="float_op_wrapper", data="&tan"},
-    tanh={args=1, base=1, func="float_op_wrapper", data="&tanh"},
+    tan={args=1, base=1, float_func="tan"},
+    tanh={args=1, base=1, float_func="tanh"},
     tempname={},
     termopen={args={1, 2}},
     test_garbagecollect_now={},
@@ -406,17 +425,18 @@ return {
     timer_start={args={2, 3}, base=1},
     timer_stop={args=1, base=1},
     timer_stopall={args=0},
-    tolower={args=1, base=1},
-    toupper={args=1, base=1},
+    tolower={args=1, base=1, fast=true},
+    toupper={args=1, base=1, fast=true},
     tr={args=3, base=1},
     trim={args={1, 3}, base=1},
-    trunc={args=1, base=1, func="float_op_wrapper", data="&trunc"},
-    type={args=1, base=1},
+    trunc={args=1, base=1, float_func="trunc"},
+    type={args=1, base=1, fast=true},
     undofile={args=1, base=1},
     undotree={},
     uniq={args={1, 3}, base=1},
     values={args=1, base=1},
     virtcol={args=1, base=1},
+    virtcol2col={args=3, base=1},
     visualmode={args={0, 1}},
     wait={args={2,3}},
     wildmenumode={},
@@ -433,7 +453,7 @@ return {
     win_splitmove={args={2, 3}, base=1},
     winbufnr={args=1, base=1},
     wincol={},
-    windowsversion={},
+    windowsversion={fast=true},
     winheight={args=1, base=1},
     winlayout={args={0, 1}, base=1},
     winline={},
