@@ -10,34 +10,37 @@
 #include <assert.h>
 #include <inttypes.h>
 #include <math.h>
-#include <msgpack.h>
+#include <stdbool.h>
 #include <stddef.h>
+#include <stdlib.h>
+#include <string.h>
 
+#include "klib/kvec.h"
+#include "msgpack/pack.h"
 #include "nvim/ascii.h"
-#include "nvim/buffer_defs.h"
-#include "nvim/charset.h"  // vim_isprintc()
 #include "nvim/eval.h"
 #include "nvim/eval/encode.h"
 #include "nvim/eval/typval.h"
 #include "nvim/eval/typval_encode.h"
 #include "nvim/garray.h"
-#include "nvim/lib/kvec.h"
+#include "nvim/gettext.h"
+#include "nvim/hashtab.h"
 #include "nvim/macros.h"
 #include "nvim/math.h"
 #include "nvim/mbyte.h"
 #include "nvim/memory.h"
 #include "nvim/message.h"
+#include "nvim/strings.h"
+#include "nvim/types.h"
 #include "nvim/vim.h"  // For _()
 
-#define utf_char2len(b) ((size_t)utf_char2len(b))
-
 const char *const encode_bool_var_names[] = {
-  [kBoolVarTrue] = "true",
-  [kBoolVarFalse] = "false",
+  [kBoolVarTrue] = "v:true",
+  [kBoolVarFalse] = "v:false",
 };
 
 const char *const encode_special_var_names[] = {
-  [kSpecialVarNull] = "null",
+  [kSpecialVarNull] = "v:null",
 };
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
@@ -69,10 +72,10 @@ int encode_list_write(void *const data, const char *const buf, const size_t len)
     line_end = xmemscan(buf, NL, len);
     if (line_end != buf) {
       const size_t line_length = (size_t)(line_end - buf);
-      char *str = (char *)TV_LIST_ITEM_TV(li)->vval.v_string;
+      char *str = TV_LIST_ITEM_TV(li)->vval.v_string;
       const size_t li_len = (str == NULL ? 0 : strlen(str));
       TV_LIST_ITEM_TV(li)->vval.v_string = xrealloc(str, li_len + line_length + 1);
-      str = (char *)TV_LIST_ITEM_TV(li)->vval.v_string + li_len;
+      str = TV_LIST_ITEM_TV(li)->vval.v_string + li_len;
       memcpy(str, buf, line_length);
       str[line_length] = 0;
       memchrsub(str, NUL, NL, line_length);
@@ -132,35 +135,36 @@ static int conv_error(const char *const msg, const MPConvStack *const mpstack,
     case kMPConvDict: {
       typval_T key_tv = {
         .v_type = VAR_STRING,
-        .vval = { .v_string = (v.data.d.hi == NULL
-                                   ? v.data.d.dict->dv_hashtab.ht_array
-                                   : (v.data.d.hi - 1))->hi_key },
+        .vval = { .v_string =
+                    (v.data.d.hi ==
+                     NULL ? v.data.d.dict->dv_hashtab.ht_array : (v.data.d.hi -
+                                                                  1))->hi_key },
       };
       char *const key = encode_tv2string(&key_tv, NULL);
-      vim_snprintf((char *)IObuff, IOSIZE, key_msg, key);
+      vim_snprintf(IObuff, IOSIZE, key_msg, key);
       xfree(key);
-      ga_concat(&msg_ga, (char *)IObuff);
+      ga_concat(&msg_ga, IObuff);
       break;
     }
     case kMPConvPairs:
     case kMPConvList: {
       const int idx = (v.data.l.li == tv_list_first(v.data.l.list)
-                         ? 0
-                         : (v.data.l.li == NULL
-                            ? tv_list_len(v.data.l.list) - 1
-                            : (int)tv_list_idx_of_item(v.data.l.list,
-                                                       TV_LIST_ITEM_PREV(v.data.l.list,
-                                                                         v.data.l.li))));
+                       ? 0
+                       : (v.data.l.li == NULL
+                          ? tv_list_len(v.data.l.list) - 1
+                          : (int)tv_list_idx_of_item(v.data.l.list,
+                                                     TV_LIST_ITEM_PREV(v.data.l.list,
+                                                                       v.data.l.li))));
       const listitem_T *const li = (v.data.l.li == NULL
-                                      ? tv_list_last(v.data.l.list)
-                                      : TV_LIST_ITEM_PREV(v.data.l.list,
-                                                          v.data.l.li));
+                                    ? tv_list_last(v.data.l.list)
+                                    : TV_LIST_ITEM_PREV(v.data.l.list,
+                                                        v.data.l.li));
       if (v.type == kMPConvList
           || li == NULL
           || (TV_LIST_ITEM_TV(li)->v_type != VAR_LIST
               && tv_list_len(TV_LIST_ITEM_TV(li)->vval.v_list) <= 0)) {
-        vim_snprintf((char *)IObuff, IOSIZE, idx_msg, idx);
-        ga_concat(&msg_ga, (char *)IObuff);
+        vim_snprintf(IObuff, IOSIZE, idx_msg, idx);
+        ga_concat(&msg_ga, IObuff);
       } else {
         assert(li != NULL);
         listitem_T *const first_item =
@@ -168,9 +172,9 @@ static int conv_error(const char *const msg, const MPConvStack *const mpstack,
         assert(first_item != NULL);
         typval_T key_tv = *TV_LIST_ITEM_TV(first_item);
         char *const key = encode_tv2echo(&key_tv, NULL);
-        vim_snprintf((char *)IObuff, IOSIZE, key_pair_msg, key, idx);
+        vim_snprintf(IObuff, IOSIZE, key_pair_msg, key, idx);
         xfree(key);
-        ga_concat(&msg_ga, (char *)IObuff);
+        ga_concat(&msg_ga, IObuff);
       }
       break;
     }
@@ -189,8 +193,8 @@ static int conv_error(const char *const msg, const MPConvStack *const mpstack,
       break;
     case kMPConvPartialList: {
       const int idx = (int)(v.data.a.arg - v.data.a.argv) - 1;
-      vim_snprintf((char *)IObuff, IOSIZE, partial_arg_i_msg, idx);
-      ga_concat(&msg_ga, (char *)IObuff);
+      vim_snprintf(IObuff, IOSIZE, partial_arg_i_msg, idx);
+      ga_concat(&msg_ga, IObuff);
       break;
     }
     }
@@ -220,7 +224,7 @@ bool encode_vim_list_to_buf(const list_T *const list, size_t *const ret_len, cha
     }
     len++;
     if (TV_LIST_ITEM_TV(li)->vval.v_string != NULL) {
-      len += STRLEN(TV_LIST_ITEM_TV(li)->vval.v_string);
+      len += strlen(TV_LIST_ITEM_TV(li)->vval.v_string);
     }
   });
   if (len) {
@@ -265,7 +269,7 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
            || TV_LIST_ITEM_TV(state->li)->vval.v_string != NULL);
     for (size_t i = state->offset; i < state->li_length && p < buf_end; i++) {
       assert(TV_LIST_ITEM_TV(state->li)->vval.v_string != NULL);
-      const char ch = (char)(TV_LIST_ITEM_TV(state->li)->vval.v_string[state->offset++]);
+      const char ch = TV_LIST_ITEM_TV(state->li)->vval.v_string[state->offset++];
       *p++ = (char)(ch == (char)NL ? (char)NUL : ch);
     }
     if (p < buf_end) {
@@ -282,7 +286,7 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
       state->offset = 0;
       state->li_length = (TV_LIST_ITEM_TV(state->li)->vval.v_string == NULL
                           ? 0
-                          : STRLEN(TV_LIST_ITEM_TV(state->li)->vval.v_string));
+                          : strlen(TV_LIST_ITEM_TV(state->li)->vval.v_string));
     }
   }
   *read_bytes = nbuf;
@@ -294,8 +298,8 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
 
 #define TYPVAL_ENCODE_CONV_STRING(tv, buf, len) \
   do { \
-    const char *const buf_ = (const char *)buf; \
-    if (buf == NULL) { \
+    const char *const buf_ = (buf); \
+    if ((buf) == NULL) { \
       ga_concat(gap, "''"); \
     } else { \
       const size_t len_ = (len); \
@@ -305,7 +309,7 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
         if (buf_[i_] == '\'') { \
           ga_append(gap, '\''); \
         } \
-        ga_append(gap, buf_[i_]); \
+        ga_append(gap, (uint8_t)buf_[i_]); \
       } \
       ga_append(gap, '\''); \
     } \
@@ -372,7 +376,7 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
 
 #define TYPVAL_ENCODE_CONV_FUNC_START(tv, fun) \
   do { \
-    const char *const fun_ = (const char *)(fun); \
+    const char *const fun_ = (fun); \
     if (fun_ == NULL) { \
       internal_error("string(): NULL function name"); \
       ga_concat(gap, "function(NULL"); \
@@ -384,14 +388,14 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
 
 #define TYPVAL_ENCODE_CONV_FUNC_BEFORE_ARGS(tv, len) \
   do { \
-    if (len != 0) { \
+    if ((len) != 0) { \
       ga_concat(gap, ", "); \
     } \
   } while (0)
 
 #define TYPVAL_ENCODE_CONV_FUNC_BEFORE_SELF(tv, len) \
   do { \
-    if ((ptrdiff_t)len != -1) { \
+    if ((ptrdiff_t)(len) != -1) { \
       ga_concat(gap, ", "); \
     } \
   } while (0)
@@ -453,12 +457,12 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
     size_t backref = 0; \
     for (; backref < kv_size(*mpstack); backref++) { \
       const MPConvStackVal mpval = kv_A(*mpstack, backref); \
-      if (mpval.type == conv_type) { \
-        if (conv_type == kMPConvDict) { \
+      if (mpval.type == (conv_type)) { \
+        if ((conv_type) == kMPConvDict) { \
           if ((void *)mpval.data.d.dict == (void *)(val)) { \
             break; \
           } \
-        } else if (conv_type == kMPConvList) { \
+        } else if ((conv_type) == kMPConvList) { \
           if ((void *)mpval.data.l.list == (void *)(val)) { \
             break; \
           } \
@@ -488,19 +492,19 @@ int encode_read_from_list(ListReaderState *const state, char *const buf, const s
     size_t backref = 0; \
     for (; backref < kv_size(*mpstack); backref++) { \
       const MPConvStackVal mpval = kv_A(*mpstack, backref); \
-      if (mpval.type == conv_type) { \
-        if (conv_type == kMPConvDict) { \
-          if ((void *)mpval.data.d.dict == (void *)val) { \
+      if (mpval.type == (conv_type)) { \
+        if ((conv_type) == kMPConvDict) { \
+          if ((void *)mpval.data.d.dict == (void *)(val)) { \
             break; \
           } \
-        } else if (conv_type == kMPConvList) { \
-          if ((void *)mpval.data.l.list == (void *)val) { \
+        } else if ((conv_type) == kMPConvList) { \
+          if ((void *)mpval.data.l.list == (void *)(val)) { \
             break; \
           } \
         } \
       } \
     } \
-    if (conv_type == kMPConvDict) { \
+    if ((conv_type) == kMPConvDict) { \
       vim_snprintf(ebuf, ARRAY_SIZE(ebuf), "{...@%zu}", backref); \
     } else { \
       vim_snprintf(ebuf, ARRAY_SIZE(ebuf), "[...@%zu]", backref); \
@@ -610,10 +614,10 @@ static inline int convert_to_json_string(garray_T *const gap, const char *const
     // This is done to make resulting values displayable on screen also not from
     // Neovim.
 #define ENCODE_RAW(ch) \
-  (ch >= 0x20 && utf_printable(ch))
+  ((ch) >= 0x20 && utf_printable(ch))
     for (size_t i = 0; i < utf_len;) {
-      const int ch = utf_ptr2char((char_u *)utf_buf + i);
-      const size_t shift = (ch == 0 ? 1 : ((size_t)utf_ptr2len((char_u *)utf_buf + i)));
+      const int ch = utf_ptr2char(utf_buf + i);
+      const size_t shift = (ch == 0 ? 1 : ((size_t)utf_ptr2len(utf_buf + i)));
       assert(shift > 0);
       i += shift;
       switch (ch) {
@@ -652,11 +656,11 @@ static inline int convert_to_json_string(garray_T *const gap, const char *const
     ga_append(gap, '"');
     ga_grow(gap, (int)str_len);
     for (size_t i = 0; i < utf_len;) {
-      const int ch = utf_ptr2char((char_u *)utf_buf + i);
-      const size_t shift = (ch == 0? 1: utf_char2len(ch));
+      const int ch = utf_ptr2char(utf_buf + i);
+      const size_t shift = (ch == 0 ? 1 : ((size_t)utf_char2len(ch)));
       assert(shift > 0);
       // Is false on invalid unicode, but this should already be handled.
-      assert(ch == 0 || shift == ((size_t)utf_ptr2len((char_u *)utf_buf + i)));
+      assert(ch == 0 || shift == ((size_t)utf_ptr2len(utf_buf + i)));
       switch (ch) {
       case BS:
       case TAB:
@@ -708,7 +712,7 @@ static inline int convert_to_json_string(garray_T *const gap, const char *const
 #undef TYPVAL_ENCODE_CONV_STRING
 #define TYPVAL_ENCODE_CONV_STRING(tv, buf, len) \
   do { \
-    if (convert_to_json_string(gap, (const char *)(buf), (len)) != OK) { \
+    if (convert_to_json_string(gap, (buf), (len)) != OK) { \
       return FAIL; \
     } \
   } while (0)
@@ -789,7 +793,7 @@ bool encode_check_json_key(const typval_T *const tv)
 #undef TYPVAL_ENCODE_SPECIAL_DICT_KEY_CHECK
 #define TYPVAL_ENCODE_SPECIAL_DICT_KEY_CHECK(label, key) \
   do { \
-    if (!encode_check_json_key(&key)) { \
+    if (!encode_check_json_key(&(key))) { \
       emsg(_("E474: Invalid key in special dictionary")); \
       goto label; \
     } \
@@ -871,7 +875,7 @@ char *encode_tv2echo(typval_T *tv, size_t *len)
   ga_init(&ga, (int)sizeof(char), 80);
   if (tv->v_type == VAR_STRING || tv->v_type == VAR_FUNC) {
     if (tv->vval.v_string != NULL) {
-      ga_concat(&ga, (char *)tv->vval.v_string);
+      ga_concat(&ga, tv->vval.v_string);
     }
   } else {
     const int eve_ret = encode_vim_to_echo(&ga, tv, N_(":echo argument"));
@@ -912,7 +916,7 @@ char *encode_tv2json(typval_T *tv, size_t *len)
 
 #define TYPVAL_ENCODE_CONV_STRING(tv, buf, len) \
   do { \
-    if (buf == NULL) { \
+    if ((buf) == NULL) { \
       msgpack_pack_bin(packer, 0); \
     } else { \
       const size_t len_ = (len); \
@@ -923,7 +927,7 @@ char *encode_tv2json(typval_T *tv, size_t *len)
 
 #define TYPVAL_ENCODE_CONV_STR_STRING(tv, buf, len) \
   do { \
-    if (buf == NULL) { \
+    if ((buf) == NULL) { \
       msgpack_pack_str(packer, 0); \
     } else { \
       const size_t len_ = (len); \
@@ -934,11 +938,11 @@ char *encode_tv2json(typval_T *tv, size_t *len)
 
 #define TYPVAL_ENCODE_CONV_EXT_STRING(tv, buf, len, type) \
   do { \
-    if (buf == NULL) { \
-      msgpack_pack_ext(packer, 0, (int8_t)type); \
+    if ((buf) == NULL) { \
+      msgpack_pack_ext(packer, 0, (int8_t)(type)); \
     } else { \
       const size_t len_ = (len); \
-      msgpack_pack_ext(packer, len_, (int8_t)type); \
+      msgpack_pack_ext(packer, len_, (int8_t)(type)); \
       msgpack_pack_ext_body(packer, buf, len_); \
     } \
   } while (0)
