@@ -2,23 +2,24 @@
 // it. PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
 
 #include <assert.h>
-#include <stdint.h>
+#include <inttypes.h>
+#include <stdbool.h>
+#include <stdio.h>
+#include <string.h>
 #include <uv.h>
 
 #include "nvim/ascii.h"
 #include "nvim/charset.h"
 #include "nvim/event/loop.h"
-#include "nvim/event/rstream.h"
 #include "nvim/event/socket.h"
-#include "nvim/event/wstream.h"
+#include "nvim/event/stream.h"
+#include "nvim/gettext.h"
 #include "nvim/log.h"
 #include "nvim/macros.h"
 #include "nvim/main.h"
 #include "nvim/memory.h"
 #include "nvim/os/os.h"
 #include "nvim/path.h"
-#include "nvim/strings.h"
-#include "nvim/vim.h"
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "event/socket.c.generated.h"
@@ -38,7 +39,7 @@ int socket_watcher_init(Loop *loop, SocketWatcher *watcher, const char *endpoint
     char *port = host_end + 1;
     intmax_t iport;
 
-    int ok = try_getdigits(&(char_u *){ (char_u *)port }, &iport);
+    int ok = try_getdigits(&(char *){ port }, &iport);
     if (!ok || iport < 0 || iport > UINT16_MAX) {
       ELOG("Invalid port: %s", port);
       return UV_EINVAL;
@@ -106,7 +107,7 @@ int socket_watcher_start(SocketWatcher *watcher, int backlog, socket_cb cb)
                                    : (STRUCT_CAST(struct sockaddr_in6, &sas))->sin6_port);
         // v:servername uses the string from watcher->addr
         size_t len = strlen(watcher->addr);
-        snprintf(watcher->addr+len, sizeof(watcher->addr)-len, ":%" PRIu16,
+        snprintf(watcher->addr + len, sizeof(watcher->addr) - len, ":%" PRIu16,
                  ntohs(port));
         break;
       }
@@ -124,8 +125,8 @@ int socket_watcher_start(SocketWatcher *watcher, int backlog, socket_cb cb)
     if (result == UV_EACCES) {
       // Libuv converts ENOENT to EACCES for Windows compatibility, but if
       // the parent directory does not exist, ENOENT would be more accurate.
-      *path_tail((char_u *)watcher->addr) = NUL;
-      if (!os_path_exists((char_u *)watcher->addr)) {
+      *path_tail(watcher->addr) = NUL;
+      if (!os_path_exists(watcher->addr)) {
         result = UV_ENOENT;
       }
     }
@@ -225,7 +226,7 @@ bool socket_connect(Loop *loop, Stream *stream, bool is_tcp, const char *address
                                     .ai_socktype = SOCK_STREAM,
                                     .ai_flags  = AI_NUMERICSERV };
     int retval = uv_getaddrinfo(&loop->uv, &addr_req, NULL,
-                                addr, host_end+1, &hints);
+                                addr, host_end + 1, &hints);
     if (retval != 0) {
       *error = _("failed to lookup host or port");
       goto cleanup;
