@@ -2,12 +2,15 @@
 #define NVIM_EVENT_WSTREAM_H
 
 #include <stdbool.h>
+#include <stddef.h>
 #include <stdint.h>
 #include <uv.h>
 
 #include "nvim/event/loop.h"
 #include "nvim/event/stream.h"
 
+struct wbuffer;
+
 typedef struct wbuffer WBuffer;
 typedef void (*wbuffer_data_finalizer)(void *data);
 
