@@ -6,7 +6,7 @@
 
 #include "nvim/eval/typval.h"
 #include "nvim/normal.h"
-#include "nvim/pos.h"      // for linenr_T
+#include "nvim/pos.h"
 #include "nvim/regexp_defs.h"
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
@@ -37,32 +37,34 @@
 // 4. Add documentation in ../doc/xxx.txt.  Add a tag for both the short and
 //    long name of the command.
 
-#define EX_RANGE           0x001  // allow a linespecs
-#define EX_BANG            0x002  // allow a ! after the command name
-#define EX_EXTRA           0x004  // allow extra args after command name
-#define EX_XFILE           0x008  // expand wildcards in extra part
-#define EX_NOSPC           0x010  // no spaces allowed in the extra part
-#define EX_DFLALL          0x020  // default file range is 1,$
-#define EX_WHOLEFOLD       0x040  // extend range to include whole fold also
-                                  // when less than two numbers given
-#define EX_NEEDARG         0x080  // argument required
-#define EX_TRLBAR          0x100  // check for trailing vertical bar
-#define EX_REGSTR          0x200  // allow "x for register designation
-#define EX_COUNT           0x400  // allow count in argument, after command
-#define EX_NOTRLCOM        0x800  // no trailing comment allowed
-#define EX_ZEROR          0x1000  // zero line number allowed
-#define EX_CTRLV          0x2000  // do not remove CTRL-V from argument
-#define EX_CMDARG         0x4000  // allow "+command" argument
-#define EX_BUFNAME        0x8000  // accepts buffer name
-#define EX_BUFUNL        0x10000  // accepts unlisted buffer too
-#define EX_ARGOPT        0x20000  // allow "++opt=val" argument
-#define EX_SBOXOK        0x40000  // allowed in the sandbox
-#define EX_CMDWIN        0x80000  // allowed in cmdline window; when missing
-                                  // disallows editing another buffer when
-                                  // current buffer is locked
-#define EX_MODIFY       0x100000  // forbidden in non-'modifiable' buffer
-#define EX_FLAGS        0x200000  // allow flags after count in argument
-#define EX_KEEPSCRIPT  0x4000000  // keep sctx of where command was invoked
+#define EX_RANGE           0x001u  // allow a linespecs
+#define EX_BANG            0x002u  // allow a ! after the command name
+#define EX_EXTRA           0x004u  // allow extra args after command name
+#define EX_XFILE           0x008u  // expand wildcards in extra part
+#define EX_NOSPC           0x010u  // no spaces allowed in the extra part
+#define EX_DFLALL          0x020u  // default file range is 1,$
+#define EX_WHOLEFOLD       0x040u  // extend range to include whole fold also
+                                   // when less than two numbers given
+#define EX_NEEDARG         0x080u  // argument required
+#define EX_TRLBAR          0x100u  // check for trailing vertical bar
+#define EX_REGSTR          0x200u  // allow "x for register designation
+#define EX_COUNT           0x400u  // allow count in argument, after command
+#define EX_NOTRLCOM        0x800u  // no trailing comment allowed
+#define EX_ZEROR          0x1000u  // zero line number allowed
+#define EX_CTRLV          0x2000u  // do not remove CTRL-V from argument
+#define EX_CMDARG         0x4000u  // allow "+command" argument
+#define EX_BUFNAME        0x8000u  // accepts buffer name
+#define EX_BUFUNL        0x10000u  // accepts unlisted buffer too
+#define EX_ARGOPT        0x20000u  // allow "++opt=val" argument
+#define EX_SBOXOK        0x40000u  // allowed in the sandbox
+#define EX_CMDWIN        0x80000u  // allowed in cmdline window
+#define EX_MODIFY       0x100000u  // forbidden in non-'modifiable' buffer
+#define EX_FLAGS        0x200000u  // allow flags after count in argument
+#define EX_LOCK_OK     0x1000000u  // command can be executed when textlock is
+                                   // set; when missing disallows editing another
+                                   // buffer when curbuf->b_ro_locked is set
+#define EX_KEEPSCRIPT  0x4000000u  // keep sctx of where command was invoked
+#define EX_PREVIEW     0x8000000u  // allow incremental command preview
 #define EX_FILES (EX_XFILE | EX_EXTRA)  // multiple extra files allowed
 #define EX_FILE1 (EX_FILES | EX_NOSPC)  // 1 file, defaults to current file
 #define EX_WORD1 (EX_EXTRA | EX_NOSPC)  // one extra word allowed
@@ -91,6 +93,7 @@ typedef struct exarg exarg_T;
 #define BAD_DROP        (-2)    // erase it
 
 typedef void (*ex_func_T)(exarg_T *eap);
+typedef int (*ex_preview_func_T)(exarg_T *eap, long cmdpreview_ns, handle_T cmdpreview_bufnr);
 
 // NOTE: These possible could be removed and changed so that
 // Callback could take a "command" style string, and simply
@@ -114,21 +117,22 @@ typedef struct aucmd_executable_t AucmdExecutable;
 struct aucmd_executable_t {
   AucmdExecutableType type;
   union {
-    char_u *cmd;
+    char *cmd;
     Callback cb;
   } callable;
 };
 
 #define AUCMD_EXECUTABLE_INIT { .type = CALLABLE_NONE }
 
-typedef char_u *(*LineGetter)(int, void *, int, bool);
+typedef char *(*LineGetter)(int, void *, int, bool);
 
 /// Structure for command definition.
 typedef struct cmdname {
-  char_u *cmd_name;    ///< Name of the command.
-  ex_func_T cmd_func;  ///< Function with implementation of this command.
-  uint32_t cmd_argt;     ///< Relevant flags from the declared above.
-  cmd_addr_T cmd_addr_type;  ///< Flag for address type
+  char *cmd_name;                         ///< Name of the command.
+  ex_func_T cmd_func;                     ///< Function with implementation of this command.
+  ex_preview_func_T cmd_preview_func;     ///< Preview callback function of this command.
+  uint32_t cmd_argt;                      ///< Relevant flags from the declared above.
+  cmd_addr_T cmd_addr_type;               ///< Flag for address type.
 } CommandDefinition;
 
 // A list used for saving values of "emsg_silent".  Used by ex_try() to save the
@@ -174,39 +178,39 @@ enum {
 
 /// Arguments used for Ex commands.
 struct exarg {
-  char_u *arg;             ///< argument of the command
-  char_u *nextcmd;         ///< next command (NULL if none)
-  char_u *cmd;             ///< the name of the command (except for :make)
-  char_u **cmdlinep;       ///< pointer to pointer of allocated cmdline
+  char *arg;                    ///< argument of the command
+  char **args;                  ///< starting position of command arguments
+  size_t *arglens;              ///< length of command arguments
+  size_t argc;                  ///< number of command arguments
+  char *nextcmd;                ///< next command (NULL if none)
+  char *cmd;                    ///< the name of the command (except for :make)
+  char **cmdlinep;              ///< pointer to pointer of allocated cmdline
   cmdidx_T cmdidx;              ///< the index for the command
   uint32_t argt;                ///< flags for the command
   int skip;                     ///< don't execute the command, only parse it
-  int forceit;                  ///< TRUE if ! present
+  int forceit;                  ///< true if ! present
   int addr_count;               ///< the number of addresses given
   linenr_T line1;               ///< the first line number
   linenr_T line2;               ///< the second line number or count
   cmd_addr_T addr_type;         ///< type of the count/range
   int flags;                    ///< extra flags after count: EXFLAG_
-  char_u *do_ecmd_cmd;     ///< +command arg to be used in edited file
+  char *do_ecmd_cmd;            ///< +command arg to be used in edited file
   linenr_T do_ecmd_lnum;        ///< the line number in an edited file
-  int append;                   ///< TRUE with ":w >>file" command
-  int usefilter;                ///< TRUE with ":w !command" and ":r!command"
+  int append;                   ///< true with ":w >>file" command
+  int usefilter;                ///< true with ":w !command" and ":r!command"
   int amount;                   ///< number of '>' or '<' for shift command
   int regname;                  ///< register name (NUL if none)
   int force_bin;                ///< 0, FORCE_BIN or FORCE_NOBIN
   int read_edit;                ///< ++edit argument
+  int mkdir_p;                  ///< ++p argument
   int force_ff;                 ///< ++ff= argument (first char of argument)
   int force_enc;                ///< ++enc= argument (index in cmd[])
   int bad_char;                 ///< BAD_KEEP, BAD_DROP or replacement byte
   int useridx;                  ///< user command index
   char *errmsg;                 ///< returned error message
   LineGetter getline;           ///< Function used to get the next line
-  void *cookie;               ///< argument for getline()
+  void *cookie;                 ///< argument for getline()
   cstack_T *cstack;             ///< condition stack for ":if" etc.
-  long verbose_save;            ///< saved value of p_verbose
-  int save_msg_silent;          ///< saved value of msg_silent
-  int did_esilent;              ///< how many times emsg_silent was incremented
-  bool did_sandbox;             ///< when true did sandbox++
 };
 
 #define FORCE_BIN 1             // ":edit ++bin file"
@@ -219,21 +223,25 @@ struct exarg {
 
 // used for completion on the command line
 struct expand {
-  char_u *xp_pattern;           // start of item to expand
+  char *xp_pattern;             // start of item to expand
   int xp_context;               // type of expansion
   size_t xp_pattern_len;        // bytes in xp_pattern before cursor
-  char_u *xp_arg;               // completion function
+  char *xp_arg;                 // completion function
   LuaRef xp_luaref;             // Ref to Lua completion function
   sctx_T xp_script_ctx;         // SCTX for completion function
   int xp_backslash;             // one of the XP_BS_ values
 #ifndef BACKSLASH_IN_FILENAME
-  int xp_shell;                 // TRUE for a shell command, more
+  int xp_shell;                 // true for a shell command, more
                                 // characters need to be escaped
 #endif
   int xp_numfiles;              // number of files found by file name completion
   int xp_col;                   // cursor position in line
-  char_u **xp_files;            // list of files
-  char_u *xp_line;              // text being completed
+  int xp_selected;              // selected index in completion
+  char *xp_orig;                // originally expanded string
+  char **xp_files;              // list of files
+  char *xp_line;                // text being completed
+#define EXPAND_BUF_LEN 256
+  char xp_buf[EXPAND_BUF_LEN];  // buffer for returned match
 };
 
 // values for xp_backslash
@@ -241,24 +249,53 @@ struct expand {
 #define XP_BS_ONE       1       // uses one backslash before a space
 #define XP_BS_THREE     2       // uses three backslashes before a space
 
+enum {
+  CMOD_SANDBOX      = 0x0001,  ///< ":sandbox"
+  CMOD_SILENT       = 0x0002,  ///< ":silent"
+  CMOD_ERRSILENT    = 0x0004,  ///< ":silent!"
+  CMOD_UNSILENT     = 0x0008,  ///< ":unsilent"
+  CMOD_NOAUTOCMD    = 0x0010,  ///< ":noautocmd"
+  CMOD_HIDE         = 0x0020,  ///< ":hide"
+  CMOD_BROWSE       = 0x0040,  ///< ":browse" - invoke file dialog
+  CMOD_CONFIRM      = 0x0080,  ///< ":confirm" - invoke yes/no dialog
+  CMOD_KEEPALT      = 0x0100,  ///< ":keepalt"
+  CMOD_KEEPMARKS    = 0x0200,  ///< ":keepmarks"
+  CMOD_KEEPJUMPS    = 0x0400,  ///< ":keepjumps"
+  CMOD_LOCKMARKS    = 0x0800,  ///< ":lockmarks"
+  CMOD_KEEPPATTERNS = 0x1000,  ///< ":keeppatterns"
+  CMOD_NOSWAPFILE   = 0x2000,  ///< ":noswapfile"
+};
+
 /// Command modifiers ":vertical", ":browse", ":confirm", ":hide", etc. set a
 /// flag.  This needs to be saved for recursive commands, put them in a
 /// structure for easy manipulation.
 typedef struct {
-  int split;                   ///< flags for win_split()
-  int tab;                     ///< > 0 when ":tab" was used
-  bool browse;                 ///< true to invoke file dialog
-  bool confirm;                ///< true to invoke yes/no dialog
-  bool hide;                   ///< true when ":hide" was used
-  bool keepalt;                ///< true when ":keepalt" was used
-  bool keepjumps;              ///< true when ":keepjumps" was used
-  bool keepmarks;              ///< true when ":keepmarks" was used
-  bool keeppatterns;           ///< true when ":keeppatterns" was used
-  bool lockmarks;              ///< true when ":lockmarks" was used
-  bool noswapfile;             ///< true when ":noswapfile" was used
-  char_u *save_ei;             ///< saved value of 'eventignore'
-  regmatch_T filter_regmatch;  ///< set by :filter /pat/
-  bool filter_force;           ///< set for :filter!
+  int cmod_flags;  ///< CMOD_ flags
+
+  int cmod_split;  ///< flags for win_split()
+  int cmod_tab;  ///< > 0 when ":tab" was used
+  char *cmod_filter_pat;
+  regmatch_T cmod_filter_regmatch;  ///< set by :filter /pat/
+  bool cmod_filter_force;  ///< set for :filter!
+
+  int cmod_verbose;  ///< 0 if not set, > 0 to set 'verbose' to cmod_verbose - 1
+
+  // values for undo_cmdmod()
+  char *cmod_save_ei;  ///< saved value of 'eventignore'
+  int cmod_did_sandbox;  ///< set when "sandbox" was incremented
+  long cmod_verbose_save;  ///< if 'verbose' was set: value of p_verbose plus one
+  int cmod_save_msg_silent;  ///< if non-zero: saved value of msg_silent + 1
+  int cmod_save_msg_scroll;  ///< for restoring msg_scroll
+  int cmod_did_esilent;  ///< incremented when emsg_silent is
 } cmdmod_T;
 
+/// Stores command modifier info used by `nvim_parse_cmd`
+typedef struct {
+  cmdmod_T cmdmod;
+  struct {
+    bool file;
+    bool bar;
+  } magic;
+} CmdParseInfo;
+
 #endif  // NVIM_EX_CMDS_DEFS_H
