@@ -1,11 +1,18 @@
 #ifndef NVIM_EXTMARK_H
 #define NVIM_EXTMARK_H
 
+#include <stdbool.h>
+#include <stddef.h>
+#include <stdint.h>
+
+#include "klib/kvec.h"
 #include "nvim/buffer_defs.h"
 #include "nvim/decoration.h"
 #include "nvim/extmark_defs.h"
+#include "nvim/macros.h"
 #include "nvim/marktree.h"
 #include "nvim/pos.h"
+#include "nvim/types.h"
 
 EXTERN int extmark_splice_pending INIT(= 0);
 
@@ -26,7 +33,6 @@ typedef kvec_t(ExtmarkInfo) ExtmarkInfoArray;
 // TODO(bfredl): good enough name for now.
 typedef ptrdiff_t bcount_t;
 
-
 // delete the columns between mincol and endcol
 typedef struct {
   int start_row;
@@ -70,6 +76,14 @@ typedef enum {
   kExtmarkClear,
 } UndoObjectType;
 
+typedef enum {
+  kExtmarkNone = 0x1,
+  kExtmarkSign = 0x2,
+  kExtmarkVirtText = 0x4,
+  kExtmarkVirtLines = 0x8,
+  kExtmarkHighlight = 0x10,
+} ExtmarkType;
+
 // TODO(bfredl): reduce the number of undo action types
 struct undo_object {
   UndoObjectType type;
@@ -80,7 +94,6 @@ struct undo_object {
   } data;
 };
 
-
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "extmark.h.generated.h"
 #endif
