@@ -1,7 +1,7 @@
 #ifndef NVIM_EXTMARK_DEFS_H
 #define NVIM_EXTMARK_DEFS_H
 
-#include "nvim/lib/kvec.h"
+#include "klib/kvec.h"
 #include "nvim/types.h"
 
 typedef struct {
@@ -9,6 +9,8 @@ typedef struct {
   int hl_id;
 } VirtTextChunk;
 
+typedef kvec_t(VirtTextChunk) VirtText;
+
 typedef struct undo_object ExtmarkUndoObject;
 typedef kvec_t(ExtmarkUndoObject) extmark_undo_vec_t;
 
