@@ -1,9 +1,9 @@
 #ifndef NVIM_FILEIO_H
 #define NVIM_FILEIO_H
 
-#include "nvim/autocmd.h"
 #include "nvim/buffer_defs.h"
 #include "nvim/eval/typval.h"
+#include "nvim/eval/typval_defs.h"
 #include "nvim/garray.h"
 #include "nvim/os/os.h"
 
@@ -16,8 +16,7 @@
 #define READ_KEEP_UNDO  0x20    // keep undo info
 #define READ_FIFO       0x40    // read from fifo or socket
 #define READ_NOWINENTER 0x80    // do not trigger BufWinEnter
-
-#define READ_STRING(x, y) (char_u *)read_string((x), (size_t)(y))
+#define READ_NOFILE     0x100   // do not read a file, do trigger BufReadCmd
 
 typedef varnumber_T (*CheckItem)(void *expr, const char *name);
 
