@@ -15,19 +15,23 @@ typedef enum {
   HL_INVERSE         = 0x01,
   HL_BOLD            = 0x02,
   HL_ITALIC          = 0x04,
+  // The next three bits are all underline styles
+  HL_UNDERLINE_MASK  = 0x38,
   HL_UNDERLINE       = 0x08,
   HL_UNDERCURL       = 0x10,
-  HL_STANDOUT        = 0x20,
-  HL_STRIKETHROUGH   = 0x40,
-  HL_NOCOMBINE       = 0x80,
-  HL_BG_INDEXED    = 0x0100,
-  HL_FG_INDEXED    = 0x0200,
-  HL_DEFAULT       = 0x0400,
-  HL_GLOBAL        = 0x0800,
-  HL_UNDERLINELINE = 0x1000,
-  HL_UNDERDOT      = 0x2000,
-  HL_UNDERDASH     = 0x4000,
-  HL_ANY_UNDERLINE = HL_UNDERLINE | HL_UNDERLINELINE | HL_UNDERCURL | HL_UNDERDOT | HL_UNDERDASH,
+  HL_UNDERDOUBLE     = 0x18,
+  HL_UNDERDOTTED     = 0x20,
+  HL_UNDERDASHED     = 0x28,
+  // 0x30 and 0x38 spare for underline styles
+  HL_STANDOUT      = 0x0040,
+  HL_STRIKETHROUGH = 0x0080,
+  HL_ALTFONT       = 0x0100,
+  // 0x0200 spare
+  HL_NOCOMBINE     = 0x0400,
+  HL_BG_INDEXED    = 0x0800,
+  HL_FG_INDEXED    = 0x1000,
+  HL_DEFAULT       = 0x2000,
+  HL_GLOBAL        = 0x4000,
 } HlAttrFlags;
 
 /// Stores a complete highlighting entry, including colors and attributes
@@ -63,6 +67,7 @@ typedef enum {
   HLF_E,          // error messages
   HLF_I,          // incremental search
   HLF_L,          // last search string
+  HLF_LC,         // current search match
   HLF_M,          // "--More--" message
   HLF_CM,         // Mode (e.g., "-- INSERT --")
   HLF_N,          // line number for ":number" and ":#" commands
@@ -95,6 +100,10 @@ typedef enum {
   HLF_SPL,        // SpellLocal
   HLF_PNI,        // popup menu normal item
   HLF_PSI,        // popup menu selected item
+  HLF_PNK,        // popup menu normal item "kind"
+  HLF_PSK,        // popup menu selected item "kind"
+  HLF_PNX,        // popup menu normal item "menu" (extra text)
+  HLF_PSX,        // popup menu selected item "menu" (extra text)
   HLF_PSB,        // popup menu scrollbar
   HLF_PST,        // popup menu scrollbar thumb
   HLF_TP,         // tabpage line
@@ -110,7 +119,10 @@ typedef enum {
   HLF_NFLOAT,     // Floating window
   HLF_MSG,        // Message area
   HLF_BORDER,     // Floating window border
+  HLF_WBR,        // Window bars
+  HLF_WBRNC,      // Window bars of not-current windows
   HLF_CU,         // Cursor
+  HLF_BTITLE,     // Float Border Title
   HLF_COUNT,      // MUST be the last one
 } hlf_T;
 
@@ -124,6 +136,7 @@ EXTERN const char *hlf_names[] INIT(= {
   [HLF_E] = "ErrorMsg",
   [HLF_I] = "IncSearch",
   [HLF_L] = "Search",
+  [HLF_LC] = "CurSearch",
   [HLF_M] = "MoreMsg",
   [HLF_CM] = "ModeMsg",
   [HLF_N] = "LineNr",
@@ -156,6 +169,10 @@ EXTERN const char *hlf_names[] INIT(= {
   [HLF_SPL] = "SpellLocal",
   [HLF_PNI] = "Pmenu",
   [HLF_PSI] = "PmenuSel",
+  [HLF_PNK] = "PmenuKind",
+  [HLF_PSK] = "PmenuKindSel",
+  [HLF_PNX] = "PmenuExtra",
+  [HLF_PSX] = "PmenuExtraSel",
   [HLF_PSB] = "PmenuSbar",
   [HLF_PST] = "PmenuThumb",
   [HLF_TP] = "TabLine",
@@ -171,11 +188,13 @@ EXTERN const char *hlf_names[] INIT(= {
   [HLF_NFLOAT] = "NormalFloat",
   [HLF_MSG] = "MsgArea",
   [HLF_BORDER] = "FloatBorder",
+  [HLF_WBR] = "WinBar",
+  [HLF_WBRNC] = "WinBarNC",
   [HLF_CU] = "Cursor",
+  [HLF_BTITLE] = "FloatTitle",
 });
 
-
-EXTERN int highlight_attr[HLF_COUNT];       // Highl. attr for each context.
+EXTERN int highlight_attr[HLF_COUNT + 1];     // Highl. attr for each context.
 EXTERN int highlight_attr_last[HLF_COUNT];  // copy for detecting changed groups
 EXTERN int highlight_user[9];                   // User[1-9] attributes
 EXTERN int highlight_stlnc[9];                  // On top of user
@@ -185,6 +204,13 @@ EXTERN RgbValue normal_fg INIT(= -1);
 EXTERN RgbValue normal_bg INIT(= -1);
 EXTERN RgbValue normal_sp INIT(= -1);
 
+EXTERN NS ns_hl_global INIT(= 0);  // global highlight namespace
+EXTERN NS ns_hl_win INIT(= -1);    // highlight namespace for the current window
+EXTERN NS ns_hl_fast INIT(= -1);   // highlight namespace specified in a fast callback
+EXTERN NS ns_hl_active INIT(= 0);  // currently active/cached namespace
+
+EXTERN int *hl_attr_active INIT(= highlight_attr);
+
 typedef enum {
   kHlUnknown,
   kHlUI,
@@ -214,9 +240,15 @@ typedef struct {
   int link_id;
   int version;
   bool is_default;
+  bool link_global;
 } ColorItem;
-#define COLOR_ITEM_INITIALIZER { .attr_id = -1, .link_id = -1, \
-                                 .version = -1, .is_default = false }
+#define COLOR_ITEM_INITIALIZER { .attr_id = -1, .link_id = -1, .version = -1, \
+                                 .is_default = false, .link_global = false }
 
+/// highlight attributes with associated priorities
+typedef struct {
+  int hl_id;
+  int priority;
+} HlPriId;
 
 #endif  // NVIM_HIGHLIGHT_DEFS_H
