@@ -4,27 +4,31 @@
 #include <assert.h>
 #include <lauxlib.h>
 #include <lua.h>
-#include <lualib.h>
 #include <stdbool.h>
+#include <stddef.h>
 #include <stdint.h>
+#include <stdlib.h>
+#include <string.h>
 
 #include "nvim/api/private/defs.h"
 #include "nvim/api/private/helpers.h"
-#include "nvim/assert.h"
-#include "nvim/func_attr.h"
 #include "nvim/memory.h"
 // FIXME: vim.h is not actually needed, but otherwise it states MAXPATHL is
 //        redefined
+#include "klib/kvec.h"
 #include "nvim/ascii.h"
 #include "nvim/eval/decode.h"
 #include "nvim/eval/typval.h"
+#include "nvim/eval/typval_defs.h"
+#include "nvim/eval/typval_encode.h"
 #include "nvim/eval/userfunc.h"
-#include "nvim/globals.h"
-#include "nvim/lib/kvec.h"
+#include "nvim/garray.h"
+#include "nvim/gettext.h"
 #include "nvim/lua/converter.h"
 #include "nvim/lua/executor.h"
 #include "nvim/macros.h"
 #include "nvim/message.h"
+#include "nvim/types.h"
 #include "nvim/vim.h"
 
 /// Determine, which keys lua table contains
@@ -49,7 +53,6 @@ typedef struct {
 #define LUA_PUSH_STATIC_STRING(lstate, s) \
   lua_pushlstring(lstate, s, sizeof(s) - 1)
 
-
 static LuaTableProps nlua_traverse_table(lua_State *const lstate)
   FUNC_ATTR_NONNULL_ALL FUNC_ATTR_WARN_UNUSED_RESULT
 {
@@ -60,7 +63,7 @@ static LuaTableProps nlua_traverse_table(lua_State *const lstate)
   size_t other_keys_num = 0;  // Number of keys that are not string, integral
                               // or type keys.
   LuaTableProps ret;
-  memset(&ret, 0, sizeof(ret));
+  CLEAR_FIELD(ret);
   if (!lua_checkstack(lstate, lua_gettop(lstate) + 3)) {
     semsg(_("E1502: Lua failed to grow stack to %i"), lua_gettop(lstate) + 2);
     ret.type = kObjectTypeNil;
@@ -386,15 +389,12 @@ nlua_pop_typval_table_processing_end:
       break;
     }
     case LUA_TFUNCTION: {
-      LuaCFunctionState *state = xmalloc(sizeof(LuaCFunctionState));
-      state->lua_callable.func_ref = nlua_ref_global(lstate, -1);
+      LuaRef func = nlua_ref_global(lstate, -1);
 
-      char_u *name = register_cfunc(&nlua_CFunction_func_call,
-                                    &nlua_CFunction_func_free,
-                                    state);
+      char *name = register_luafunc(func);
 
       cur.tv->v_type = VAR_FUNC;
-      cur.tv->vval.v_string = vim_strsave(name);
+      cur.tv->vval.v_string = xstrdup(name);
       break;
     }
     case LUA_TUSERDATA: {
@@ -459,7 +459,7 @@ static bool typval_conv_special = false;
   TYPVAL_ENCODE_CONV_NUMBER(tv, flt)
 
 #define TYPVAL_ENCODE_CONV_STRING(tv, str, len) \
-  lua_pushlstring(lstate, (const char *)(str), (len))
+  lua_pushlstring(lstate, (str), (len))
 
 #define TYPVAL_ENCODE_CONV_STR_STRING TYPVAL_ENCODE_CONV_STRING
 
@@ -469,16 +469,14 @@ static bool typval_conv_special = false;
 #define TYPVAL_ENCODE_CONV_BLOB(tv, blob, len) \
   do { \
     const blob_T *const blob_ = (blob); \
-    lua_pushlstring(lstate, \
-                    blob_ != NULL ? (const char *)blob_->bv_ga.ga_data : "", \
-                    (size_t)(len)); \
+    lua_pushlstring(lstate, blob_ != NULL ? blob_->bv_ga.ga_data : "", (size_t)(len)); \
   } while (0)
 
 #define TYPVAL_ENCODE_CONV_FUNC_START(tv, fun) \
   do { \
     ufunc_T *fp = find_func(fun); \
-    if (fp != NULL && fp->uf_cb == nlua_CFunction_func_call) { \
-      nlua_pushref(lstate, ((LuaCFunctionState *)fp->uf_cb_state)->lua_callable.func_ref); \
+    if (fp != NULL && fp->uf_flags & FC_LUAREF) { \
+      nlua_pushref(lstate, fp->uf_luaref); \
     } else { \
       TYPVAL_ENCODE_CONV_NIL(tv); \
     } \
@@ -569,6 +567,7 @@ static bool typval_conv_special = false;
 #define TYPVAL_ENCODE_FIRST_ARG_TYPE lua_State *const
 #define TYPVAL_ENCODE_FIRST_ARG_NAME lstate
 #include "nvim/eval/typval_encode.c.h"
+
 #undef TYPVAL_ENCODE_SCOPE
 #undef TYPVAL_ENCODE_NAME
 #undef TYPVAL_ENCODE_FIRST_ARG_TYPE
@@ -674,7 +673,6 @@ static inline void nlua_create_typed_table(lua_State *lstate, const size_t narr,
   lua_rawset(lstate, -3);
 }
 
-
 /// Convert given String to lua string
 ///
 /// Leaves converted string on top of the stack.
@@ -808,7 +806,6 @@ void nlua_push_Object(lua_State *lstate, const Object obj, bool special)
   }
 }
 
-
 /// Convert lua value to string
 ///
 /// Always pops one value from the stack.
@@ -912,9 +909,8 @@ Float nlua_pop_Float(lua_State *lstate, Error *err)
   lua_pop(lstate, 1);
   if (table_props.type != kObjectTypeFloat) {
     return 0;
-  } else {
-    return (Float)table_props.val;
   }
+  return (Float)table_props.val;
 }
 
 /// Convert lua table to array without determining whether it is array
@@ -1236,7 +1232,7 @@ LuaRef nlua_pop_LuaRef(lua_State *const lstate, Error *err)
     type ret; \
     if (lua_type(lstate, -1) != LUA_TNUMBER) { \
       api_set_error(err, kErrorTypeValidation, "Expected Lua number"); \
-      ret = (type)-1; \
+      ret = (type) - 1; \
     } else { \
       ret = (type)lua_tonumber(lstate, -1); \
     } \
@@ -1297,7 +1293,6 @@ void nlua_init_types(lua_State *const lstate)
   lua_rawset(lstate, -3);
 }
 
-
 void nlua_pop_keydict(lua_State *L, void *retval, field_hash hashy, Error *err)
 {
   if (!lua_istable(L, -1)) {
