@@ -11,7 +11,6 @@
 #else
 # ifndef INIT
 #  define INIT(...) __VA_ARGS__
-#  define COMMA ,
 # endif
 #endif
 
@@ -31,7 +30,7 @@
 /// @return `s, sizeof(s) - 1`
 #define S_LEN(s) (s), (sizeof(s) - 1)
 
-/// LINEEMPTY() - return TRUE if the line is empty
+/// LINEEMPTY() - return true if the line is empty
 #define LINEEMPTY(p) (*ml_get(p) == NUL)
 
 // toupper() and tolower() that use the current locale.
@@ -54,14 +53,16 @@
 #define ASCII_ISALNUM(c) (ASCII_ISALPHA(c) || ascii_isdigit(c))
 
 // Returns empty string if it is NULL.
-#define EMPTY_IF_NULL(x) (char *)((x) ? (x) : (char_u *)"")
+#define EMPTY_IF_NULL(x) ((x) ? (x) : "")
 
-// Adjust chars in a language according to 'langmap' option.
-// NOTE that there is no noticeable overhead if 'langmap' is not set.
-// When set the overhead for characters < 256 is small.
-// Don't apply 'langmap' if the character comes from the Stuff buffer or from a
-// mapping and the langnoremap option was set.
-// The do-while is just to ignore a ';' after the macro.
+/// Adjust chars in a language according to 'langmap' option.
+/// NOTE that there is no noticeable overhead if 'langmap' is not set.
+/// When set the overhead for characters < 256 is small.
+/// Don't apply 'langmap' if the character comes from the Stuff buffer or from a
+/// mapping and the langnoremap option was set.
+/// The do-while is just to ignore a ';' after the macro.
+///
+/// -V:LANGMAP_ADJUST:560
 #define LANGMAP_ADJUST(c, condition) \
   do { \
     if (*p_langmap \
@@ -83,30 +84,29 @@
 
 // mch_open_rw(): invoke os_open() with third argument for user R/W.
 #if defined(UNIX)  // open in rw------- mode
-# define mch_open_rw(n, f)      os_open((n), (f), (mode_t)0600)
-#elif defined(WIN32)
-# define mch_open_rw(n, f)      os_open((n), (f), S_IREAD | S_IWRITE)
+# define MCH_OPEN_RW(n, f)      os_open((n), (f), (mode_t)0600)
+#elif defined(MSWIN)
+# define MCH_OPEN_RW(n, f)      os_open((n), (f), S_IREAD | S_IWRITE)
 #else
-# define mch_open_rw(n, f)      os_open((n), (f), 0)
+# define MCH_OPEN_RW(n, f)      os_open((n), (f), 0)
 #endif
 
 #define REPLACE_NORMAL(s) (((s) & REPLACE_FLAG) && !((s) & VREPLACE_FLAG))
 
-
 // MB_PTR_ADV(): advance a pointer to the next character, taking care of
 // multi-byte characters if needed. Skip over composing chars.
-#define MB_PTR_ADV(p)      (p += utfc_ptr2len((char_u *)p))
+#define MB_PTR_ADV(p)      (p += utfc_ptr2len((char *)p))
 
 // Advance multi-byte pointer, do not skip over composing chars.
-#define MB_CPTR_ADV(p)     (p += utf_ptr2len(p))
+#define MB_CPTR_ADV(p)     (p += utf_ptr2len((char *)p))
 
 // MB_PTR_BACK(): backup a pointer to the previous character, taking care of
 // multi-byte characters if needed. Only use with "p" > "s" !
 #define MB_PTR_BACK(s, p) \
-  (p -= utf_head_off((char_u *)(s), (char_u *)(p) - 1) + 1)
+  (p -= utf_head_off((char *)(s), (char *)(p) - 1) + 1)
 
 // MB_CHAR2BYTES(): convert character to bytes and advance pointer to bytes
-#define MB_CHAR2BYTES(c, b) ((b) += utf_char2bytes((c), (b)))
+#define MB_CHAR2BYTES(c, b) ((b) += utf_char2bytes((c), ((char *)b)))
 
 #define RESET_BINDING(wp) \
   do { \
@@ -177,14 +177,14 @@
 
 // Type of uv_buf_t.len is platform-dependent.
 // Related: https://github.com/libuv/libuv/pull/1236
-#if defined(WIN32)
+#if defined(MSWIN)
 # define UV_BUF_LEN(x)  (ULONG)(x)
 #else
 # define UV_BUF_LEN(x)  (x)
 #endif
 
 // Type of read()/write() `count` param is platform-dependent.
-#if defined(WIN32)
+#if defined(MSWIN)
 # define IO_COUNT(x)  (unsigned)(x)
 #else
 # define IO_COUNT(x)  (x)
