@@ -2,12 +2,17 @@
 #define NVIM_MAP_H
 
 #include <stdbool.h>
+#include <stdint.h>
+#include <stdio.h>
 
+#include "klib/khash.h"
 #include "nvim/api/private/defs.h"
-#include "nvim/api/private/dispatch.h"
 #include "nvim/extmark_defs.h"
+#include "nvim/gettext.h"
 #include "nvim/highlight_defs.h"
 #include "nvim/map_defs.h"
+#include "nvim/tui/input_defs.h"
+#include "nvim/types.h"
 #include "nvim/ui_client.h"
 
 #if defined(__NetBSD__)
@@ -35,24 +40,26 @@
 // NOTE: Keys AND values must be allocated! khash.h does not make a copy.
 //
 MAP_DECLS(int, int)
+MAP_DECLS(int, cstr_t)
 MAP_DECLS(cstr_t, ptr_t)
 MAP_DECLS(cstr_t, int)
 MAP_DECLS(ptr_t, ptr_t)
+MAP_DECLS(uint32_t, ptr_t)
 MAP_DECLS(uint64_t, ptr_t)
 MAP_DECLS(uint64_t, ssize_t)
 MAP_DECLS(uint64_t, uint64_t)
 MAP_DECLS(uint32_t, uint32_t)
 
 MAP_DECLS(handle_T, ptr_t)
-MAP_DECLS(String, MsgpackRpcRequestHandler)
 MAP_DECLS(HlEntry, int)
 MAP_DECLS(String, handle_T)
 MAP_DECLS(String, int)
 MAP_DECLS(int, String)
-MAP_DECLS(String, UIClientHandler)
 
 MAP_DECLS(ColorKey, ColorItem)
 
+MAP_DECLS(KittyKey, cstr_t)
+
 #define MAP_INIT { { 0, 0, 0, 0, NULL, NULL, NULL } }
 #define map_init(k, v, map) do { *(map) = (Map(k, v)) MAP_INIT; } while (false)
 
