@@ -6,8 +6,10 @@
 
 #include "nvim/api/private/defs.h"
 #include "nvim/channel.h"
+#include "nvim/event/multiqueue.h"
 #include "nvim/event/process.h"
 #include "nvim/event/socket.h"
+#include "nvim/macros.h"
 #include "nvim/vim.h"
 
 #define METHOD_MAXLEN 512
@@ -17,7 +19,6 @@
 ///       of os_inchar(), so they are processed "just-in-time".
 EXTERN MultiQueue *ch_before_blocking_events INIT(= NULL);
 
-
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "msgpack_rpc/channel.h.generated.h"
 #endif
