@@ -6,16 +6,19 @@
 #include <uv.h>
 
 #include "nvim/api/private/defs.h"
+#include "nvim/api/private/dispatch.h"
 #include "nvim/event/process.h"
 #include "nvim/event/socket.h"
 #include "nvim/vim.h"
 
 typedef struct Channel Channel;
+typedef struct Unpacker Unpacker;
 
 typedef struct {
   uint32_t request_id;
   bool returned, errored;
   Object result;
+  ArenaMem result_mem;
 } ChannelCallFrame;
 
 typedef struct {
@@ -24,12 +27,13 @@ typedef struct {
   MsgpackRpcRequestHandler handler;
   Array args;
   uint32_t request_id;
+  Arena used_mem;
 } RequestEvent;
 
 typedef struct {
   PMap(cstr_t) subscribed_events[1];
   bool closed;
-  msgpack_unpacker *unpacker;
+  Unpacker *unpacker;
   uint32_t next_request_id;
   kvec_t(ChannelCallFrame *) call_stack;
   Dictionary info;
