@@ -10,6 +10,7 @@
 --    secure=nil, gettext=nil, noglob=nil, normal_fname_chars=nil,
 --    pri_mkrc=nil, deny_in_modelines=nil, normal_dname_chars=nil,
 --    modelineexpr=nil,
+--    func=nil,
 --    expand=nil, nodefault=nil, no_mkrc=nil,
 --    alloced=nil,
 --    save_pv_indir=nil,
@@ -19,8 +20,8 @@
 -- types: bool, number, string
 -- lists: (nil), comma, onecomma, flags, flagscomma
 -- scopes: global, buffer, window
--- redraw options: statuslines, current_window, curent_window_only,
---                 current_buffer, all_windows, everything, curswant
+-- redraw options: statuslines, tabline, current_window, current_window_only,
+--                 current_buffer, all_windows, curswant
 -- defaults: {condition=#if condition, if_true=default, if_false=default}
 -- #if condition:
 --    string: #ifdef string
@@ -54,8 +55,6 @@ return {
       full_name='aleph', abbreviation='al',
       short_desc=N_("ASCII code of the letter Aleph (Hebrew)"),
       type='number', scope={'global'},
-      redraw={'curswant'},
-      varname='p_aleph',
       defaults={if_true=224}
     },
     {
@@ -128,7 +127,6 @@ return {
       full_name='background', abbreviation='bg',
       short_desc=N_("\"dark\" or \"light\", used for highlight colors"),
       type='string', scope={'global'},
-      redraw={'all_windows'},
       varname='p_bg',
       defaults={if_true="dark"}
     },
@@ -395,7 +393,6 @@ return {
       short_desc=N_("number of columns in the display"),
       type='number', scope={'global'},
       no_mkrc=true,
-      redraw={'everything'},
       varname='p_columns',
       defaults={if_true=macros('DFLT_COLS')}
     },
@@ -416,13 +413,12 @@ return {
       alloced=true,
       redraw={'curswant'},
       varname='p_cms',
-      defaults={if_true="/*%s*/"}
+      defaults={if_true=""}
     },
     {
       full_name='compatible', abbreviation='cp',
       short_desc=N_("No description"),
       type='bool', scope={'global'},
-      redraw={'all_windows'},
       varname='p_force_off',
       -- pri_mkrc isn't needed here, optval_default()
       -- always returns TRUE for 'compatible'
@@ -458,6 +454,7 @@ return {
       type='string', scope={'buffer'},
       secure=true,
       alloced=true,
+      func=true,
       varname='p_cfu',
       defaults={if_true=""}
     },
@@ -498,58 +495,6 @@ return {
       varname='p_cpo',
       defaults={if_true=macros('CPO_VIM')}
     },
-    {
-      full_name='cscopepathcomp', abbreviation='cspc',
-      short_desc=N_("how many components of the path to show"),
-      type='number', scope={'global'},
-      varname='p_cspc',
-      defaults={if_true=0}
-    },
-    {
-      full_name='cscopeprg', abbreviation='csprg',
-      short_desc=N_("command to execute cscope"),
-      type='string', scope={'global'},
-      secure=true,
-      expand=true,
-      varname='p_csprg',
-      defaults={if_true="cscope"}
-    },
-    {
-      full_name='cscopequickfix', abbreviation='csqf',
-      short_desc=N_("use quickfix window for cscope results"),
-      type='string', list='onecomma', scope={'global'},
-      deny_duplicates=true,
-      varname='p_csqf',
-      defaults={if_true=""}
-    },
-    {
-      full_name='cscoperelative', abbreviation='csre',
-      short_desc=N_("Use cscope.out path basename as prefix"),
-      type='bool', scope={'global'},
-      varname='p_csre',
-      defaults={if_true=0}
-    },
-    {
-      full_name='cscopetag', abbreviation='cst',
-      short_desc=N_("use cscope for tag commands"),
-      type='bool', scope={'global'},
-      varname='p_cst',
-      defaults={if_true=0}
-    },
-    {
-      full_name='cscopetagorder', abbreviation='csto',
-      short_desc=N_("determines \":cstag\" search order"),
-      type='number', scope={'global'},
-      varname='p_csto',
-      defaults={if_true=0}
-    },
-    {
-      full_name='cscopeverbose', abbreviation='csverb',
-      short_desc=N_("give messages when adding a cscope database"),
-      type='bool', scope={'global'},
-      varname='p_csverbose',
-      defaults={if_true=1}
-    },
     {
       full_name='cursorbind', abbreviation='crb',
       short_desc=N_("move cursor in window as it moves in other windows"),
@@ -663,7 +608,7 @@ return {
       deny_duplicates=true,
       redraw={'all_windows'},
       varname='p_dy',
-      defaults={if_true="lastline,msgsep"}
+      defaults={if_true="lastline"}
     },
     {
       full_name='eadirection', abbreviation='ead',
@@ -695,6 +640,15 @@ return {
       varname='p_enc',
       defaults={if_true=macros('ENC_DFLT')}
     },
+    {
+      full_name='endoffile', abbreviation='eof',
+      short_desc=N_("write CTRL-Z for last line in file"),
+      type='bool', scope={'buffer'},
+      no_mkrc=true,
+      redraw={'statuslines'},
+      varname='p_eof',
+      defaults={if_true=false}
+    },
     {
       full_name='endofline', abbreviation='eol',
       short_desc=N_("write <EOL> for last line in file"),
@@ -708,7 +662,6 @@ return {
       full_name='equalalways', abbreviation='ea',
       short_desc=N_("windows are automatically made the same size"),
       type='bool', scope={'global'},
-      redraw={'all_windows'},
       varname='p_ea',
       defaults={if_true=true}
     },
@@ -989,7 +942,7 @@ return {
       type='bool', scope={'global'},
       secure=true,
       varname='p_fs',
-      defaults={if_true=false}
+      defaults={if_true=true}
     },
     {
       full_name='gdefault', abbreviation='gd',
@@ -1014,7 +967,7 @@ return {
       expand=true,
       varname='p_gp',
       defaults={
-        condition='WIN32',
+        condition='MSWIN',
         -- Add an extra file name so that grep will always
         -- insert a file name in the match line. */
         if_true="findstr /n $* nul",
@@ -1051,7 +1004,6 @@ return {
       full_name='guioptions', abbreviation='go',
       short_desc=N_("GUI: Which components and options are used"),
       type='string', list='flags', scope={'global'},
-      redraw={'all_windows'},
       enable_if=false,
     },
     {
@@ -1117,16 +1069,16 @@ return {
     },
     {
       full_name='hkmap', abbreviation='hk',
-      short_desc=N_("Hebrew keyboard mapping"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
-      varname='p_hkmap',
+      varname='p_force_off',
       defaults={if_true=false}
     },
     {
       full_name='hkmapp', abbreviation='hkp',
-      short_desc=N_("phonetic Hebrew keyboard mapping"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
-      varname='p_hkmapp',
+      varname='p_force_off',
       defaults={if_true=false}
     },
     {
@@ -1195,7 +1147,6 @@ return {
       full_name='inccommand', abbreviation='icm',
       short_desc=N_("Live preview of substitution"),
       type='string', scope={'global'},
-      redraw={'all_windows'},
       varname='p_icm',
       defaults={if_true="nosplit"}
     },
@@ -1250,9 +1201,9 @@ return {
     },
     {
       full_name='insertmode', abbreviation='im',
-      short_desc=N_("start the edit of a file in Insert mode"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
-      varname='p_im',
+      varname='p_force_off',
       defaults={if_true=false}
     },
     {
@@ -1276,7 +1227,7 @@ return {
       deny_duplicates=true,
       varname='p_isi',
       defaults={
-        condition='WIN32',
+        condition='MSWIN',
         if_true="@,48-57,_,128-167,224-235",
         if_false="@,48-57,_,192-255"
       }
@@ -1403,7 +1354,6 @@ return {
       short_desc=N_("of lines in the display"),
       type='number', scope={'global'},
       no_mkrc=true,
-      redraw={'everything'},
       varname='p_lines',
       defaults={if_true=macros('DFLT_ROWS')}
     },
@@ -1422,6 +1372,14 @@ return {
       varname='p_lisp',
       defaults={if_true=false}
     },
+    {
+      full_name='lispoptions', abbreviation='lop',
+      short_desc=N_("options for lisp indenting"),
+      type='string', list='onecomma', scope={'buffer'},
+      deny_duplicates=true,
+      varname='p_lop', pv_name='p_lop',
+      defaults={if_true=''}
+    },
     {
       full_name='lispwords', abbreviation='lw',
       short_desc=N_("words that change how lisp indenting works"),
@@ -1597,7 +1555,7 @@ return {
       short_desc=N_("the use of mouse clicks"),
       type='string', list='flags', scope={'global'},
       varname='p_mouse',
-      defaults={if_true=""}
+      defaults={if_true="nvi"}
     },
     {
       full_name='mousefocus', abbreviation='mousef',
@@ -1619,7 +1577,23 @@ return {
       short_desc=N_("changes meaning of mouse buttons"),
       type='string', scope={'global'},
       varname='p_mousem',
-      defaults={if_true="extend"}
+      defaults={if_true="popup_setpos"}
+    },
+    {
+      full_name='mousemoveevent', abbreviation='mousemev',
+      short_desc=N_("deliver mouse move events to input queue"),
+      type='bool', scope={'global'},
+      redraw={'ui_option'},
+      varname='p_mousemev',
+      defaults={if_true=false}
+    },
+    {
+      full_name='mousescroll',
+      short_desc=N_("amount to scroll by when scrolling with a mouse"),
+      type='string', list='comma', scope={'global'},
+      vi_def=true,
+      varname='p_mousescroll',
+      defaults={if_true="ver:3,hor:6"}
     },
     {
       full_name='mouseshape', abbreviation='mouses',
@@ -1664,6 +1638,7 @@ return {
       type='string', scope={'buffer'},
       secure=true,
       alloced=true,
+      func=true,
       varname='p_ofu',
       defaults={if_true=""}
     },
@@ -1679,6 +1654,7 @@ return {
       short_desc=N_("function to be called for |g@| operator"),
       type='string', scope={'global'},
       secure=true,
+      func=true,
       varname='p_opfunc',
       defaults={if_true=""}
     },
@@ -1709,9 +1685,8 @@ return {
     },
     {
       full_name='pastetoggle', abbreviation='pt',
-      short_desc=N_("key code that causes 'paste' to toggle"),
+      short_desc=N_("No description"),
       type='string', scope={'global'},
-      varname='p_pt',
       defaults={if_true=""}
     },
     {
@@ -1761,65 +1736,6 @@ return {
       redraw={'statuslines'},
       defaults={if_true=false}
     },
-    {
-      full_name='printdevice', abbreviation='pdev',
-      short_desc=N_("name of the printer to be used for :hardcopy"),
-      type='string', scope={'global'},
-      secure=true,
-      varname='p_pdev',
-      defaults={if_true=""}
-    },
-    {
-      full_name='printencoding', abbreviation='penc',
-      short_desc=N_("encoding to be used for printing"),
-      type='string', scope={'global'},
-      varname='p_penc',
-      defaults={if_true=""}
-    },
-    {
-      full_name='printexpr', abbreviation='pexpr',
-      short_desc=N_("expression used to print PostScript for :hardcopy"),
-      type='string', scope={'global'},
-      secure=true,
-      varname='p_pexpr',
-      defaults={if_true=""}
-    },
-    {
-      full_name='printfont', abbreviation='pfn',
-      short_desc=N_("name of the font to be used for :hardcopy"),
-      type='string', scope={'global'},
-      varname='p_pfn',
-      defaults={if_true="courier"}
-    },
-    {
-      full_name='printheader', abbreviation='pheader',
-      short_desc=N_("format of the header used for :hardcopy"),
-      type='string', scope={'global'},
-      varname='p_header',
-      defaults={if_true="%<%f%h%m%=Page %N"}
-    },
-    {
-      full_name='printmbcharset', abbreviation='pmbcs',
-      short_desc=N_("CJK character set to be used for :hardcopy"),
-      type='string', scope={'global'},
-      varname='p_pmcs',
-      defaults={if_true=""}
-    },
-    {
-      full_name='printmbfont', abbreviation='pmbfn',
-      short_desc=N_("font names to be used for CJK output of :hardcopy"),
-      type='string', scope={'global'},
-      varname='p_pmfn',
-      defaults={if_true=""}
-    },
-    {
-      full_name='printoptions', abbreviation='popt',
-      short_desc=N_("controls the format of :hardcopy output"),
-      type='string', list='onecomma', scope={'global'},
-      deny_duplicates=true,
-      varname='p_popt',
-      defaults={if_true=""}
-    },
     {
       full_name='prompt',
       short_desc=N_("enable prompt in Ex mode"),
@@ -1861,6 +1777,8 @@ return {
       full_name='quickfixtextfunc', abbreviation='qftf',
       short_desc=N_("customize the quickfix window"),
       type='string', scope={'global'},
+      secure=true,
+      func=true,
       varname='p_qftf',
       defaults={if_true=""}
     },
@@ -1911,9 +1829,9 @@ return {
     },
     {
       full_name='remap',
-      short_desc=N_("mappings to work recursively"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
-      varname='p_remap',
+      varname='p_force_on',
       defaults={if_true=true}
     },
     {
@@ -2007,7 +1925,6 @@ return {
       full_name='scrolloff', abbreviation='so',
       short_desc=N_("minimum nr. of lines above and below cursor"),
       type='number', scope={'global', 'window'},
-      redraw={'all_windows'},
       varname='p_so',
       defaults={if_true=0}
     },
@@ -2028,7 +1945,7 @@ return {
     },
     {
       full_name='secure',
-      short_desc=N_("mode for reading .vimrc in current dir"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
       secure=true,
       varname='p_secure',
@@ -2055,7 +1972,7 @@ return {
       type='string', list='onecomma', scope={'global'},
       deny_duplicates=true,
       varname='p_ssop',
-      defaults={if_true="blank,buffers,curdir,folds,help,tabpages,winsize"}
+      defaults={if_true="blank,buffers,curdir,folds,help,tabpages,winsize,terminal"}
     },
     {
       full_name='shada', abbreviation='sd',
@@ -2084,7 +2001,7 @@ return {
       expand=true,
       varname='p_sh',
       defaults={
-        condition='WIN32',
+        condition='MSWIN',
         if_true="cmd.exe",
         if_false="sh"
       }
@@ -2096,7 +2013,7 @@ return {
       secure=true,
       varname='p_shcf',
       defaults={
-        condition='WIN32',
+        condition='MSWIN',
         if_true="/s /c",
         if_false="-c"
       }
@@ -2108,8 +2025,8 @@ return {
       secure=true,
       varname='p_sp',
       defaults={
-        condition='WIN32',
-        if_true=">%s 2>&1",
+        condition='MSWIN',
+        if_true="2>&1| tee",
         if_false="| tee",
       }
     },
@@ -2128,7 +2045,7 @@ return {
       secure=true,
       varname='p_srr',
       defaults={
-        condition='WIN32',
+        condition='MSWIN',
         if_true=">%s 2>&1",
         if_false=">"
       }
@@ -2155,7 +2072,7 @@ return {
       secure=true,
       varname='p_sxq',
       defaults={
-        condition='WIN32',
+        condition='MSWIN',
         if_true="\"",
         if_false="",
       }
@@ -2204,6 +2121,13 @@ return {
       varname='p_sc',
       defaults={if_true=true}
     },
+    {
+      full_name='showcmdloc', abbreviation='sloc',
+      short_desc=N_("change location of partial command"),
+      type='string', scope={'global'},
+      varname='p_sloc',
+      defaults={if_true="last"}
+    },
     {
       full_name='showfulltag', abbreviation='sft',
       short_desc=N_("show full tag pattern when completing tag"),
@@ -2244,7 +2168,6 @@ return {
       full_name='sidescrolloff', abbreviation='siso',
       short_desc=N_("min. nr. of columns to left and right of cursor"),
       type='number', scope={'global', 'window'},
-      redraw={'all_windows'},
       varname='p_siso',
       defaults={if_true=0}
     },
@@ -2339,6 +2262,7 @@ return {
       secure=true,
       expand=true,
       varname='p_spo',
+      redraw={'current_buffer'},
       defaults={if_true=""}
     },
     {
@@ -2348,6 +2272,13 @@ return {
       varname='p_sb',
       defaults={if_true=false}
     },
+    {
+      full_name='splitkeep', abbreviation='spk',
+      short_desc=N_("determines scroll behavior for split windows"),
+      type='string', scope={'global'},
+      varname='p_spk',
+      defaults={if_true='cursor'}
+    },
     {
       full_name='splitright', abbreviation='spr',
       short_desc=N_("new window is put right of the current one"),
@@ -2363,6 +2294,15 @@ return {
       varname='p_sol',
       defaults={if_true=false}
     },
+    {
+      full_name='statuscolumn', abbreviation='stc',
+      short_desc=N_("custom format for the status column"),
+      type='string', scope={'window'},
+      redraw={'current_window'},
+      secure=true,
+      alloced=true,
+      defaults={if_true=""}
+    },
     {
       full_name='statusline', abbreviation='stl',
       short_desc=N_("custom format for the status line"),
@@ -2428,6 +2368,8 @@ return {
       full_name='tagfunc', abbreviation='tfu',
       short_desc=N_("function used to perform tag searches"),
       type='string', scope={'buffer'},
+      secure=true,
+      func=true,
       varname='p_tfu',
       defaults={if_true=""}
     },
@@ -2436,7 +2378,7 @@ return {
       short_desc=N_("custom format for the console tab pages line"),
       type='string', scope={'global'},
       modelineexpr=true,
-      redraw={'all_windows'},
+      redraw={'tabline'},
       varname='p_tal',
       defaults={if_true=""}
     },
@@ -2529,9 +2471,9 @@ return {
     },
     {
       full_name='terse',
-      short_desc=N_("hides notification of search wrap"),
+      short_desc=N_("No description"),
       type='bool', scope={'global'},
-      varname='p_terse',
+      varname='p_force_off',
       defaults={if_true=false}
     },
     {
@@ -2558,6 +2500,7 @@ return {
       type='string', scope={'global', 'buffer'},
       secure=true,
       alloced=true,
+      func=true,
       varname='p_tsrfu',
       defaults={if_true=""}
     },
@@ -2701,7 +2644,7 @@ return {
       full_name='verbose', abbreviation='vbs',
       short_desc=N_("give informative messages"),
       type='number', scope={'global'},
-      varname='p_verbose',
+      varname='p_verbose', redraw={'ui_option'},
       defaults={if_true=0}
     },
     {
@@ -2831,6 +2774,16 @@ return {
       varname='p_wak',
       defaults={if_true="menu"}
     },
+    {
+      full_name='winbar', abbreviation='wbr',
+      short_desc=N_("custom format for the window bar"),
+      type='string', scope={'global', 'window'},
+      alloced=true,
+      modelineexpr=true,
+      redraw={'statuslines'},
+      varname='p_wbr',
+      defaults={if_true=""}
+    },
     {
       full_name='winblend', abbreviation='winbl',
       short_desc=N_("Controls transparency level for floating windows"),
@@ -2841,7 +2794,8 @@ return {
     {
       full_name='winhighlight', abbreviation='winhl',
       short_desc=N_("Setup window-local highlights");
-      type='string', scope={'window'},
+      type='string', list='onecomma', scope={'window'},
+      deny_duplicates=true,
       alloced=true,
       redraw={'current_window'},
       defaults={if_true=""}
