@@ -3,6 +3,9 @@
 
 // users.c -- operating system user information
 
+#include <stdbool.h>
+#include <stdio.h>
+#include <string.h>
 #include <uv.h>
 
 #include "auto/config.h"
@@ -10,11 +13,12 @@
 #include "nvim/garray.h"
 #include "nvim/memory.h"
 #include "nvim/os/os.h"
-#include "nvim/strings.h"
-#ifdef HAVE_PWD_H
+#include "nvim/types.h"
+#include "nvim/vim.h"
+#ifdef HAVE_PWD_FUNCS
 # include <pwd.h>
 #endif
-#ifdef WIN32
+#ifdef MSWIN
 # include <lm.h>
 #endif
 
@@ -30,7 +34,7 @@ static void add_user(garray_T *users, char *user, bool need_copy)
 
   if (user_copy == NULL || *user_copy == NUL) {
     if (need_copy) {
-      xfree(user);
+      xfree(user_copy);
     }
     return;
   }
@@ -46,7 +50,7 @@ int os_get_usernames(garray_T *users)
   }
   ga_init(users, sizeof(char *), 20);
 
-#if defined(HAVE_GETPWENT) && defined(HAVE_PWD_H)
+#ifdef HAVE_PWD_FUNCS
   {
     struct passwd *pw;
 
@@ -56,7 +60,7 @@ int os_get_usernames(garray_T *users)
     }
     endpwent();
   }
-#elif defined(WIN32)
+#elif defined(MSWIN)
   {
     DWORD nusers = 0, ntotal = 0, i;
     PUSER_INFO_0 uinfo;
@@ -77,7 +81,7 @@ int os_get_usernames(garray_T *users)
     }
   }
 #endif
-#if defined(HAVE_GETPWNAM)
+#ifdef HAVE_PWD_FUNCS
   {
     const char *user_env = os_getenv("USER");
 
@@ -93,7 +97,7 @@ int os_get_usernames(garray_T *users)
       for (i = 0; i < users->ga_len; i++) {
         char *local_user = ((char **)users->ga_data)[i];
 
-        if (STRCMP(local_user, user_env) == 0) {
+        if (strcmp(local_user, user_env) == 0) {
           break;
         }
       }
@@ -112,9 +116,13 @@ int os_get_usernames(garray_T *users)
   return OK;
 }
 
-// Insert user name in s[len].
-// Return OK if a name found.
-int os_get_user_name(char *s, size_t len)
+/// Gets the username that owns the current Nvim process.
+///
+/// @param s[out] Username.
+/// @param len Length of `s`.
+///
+/// @return OK if a name found.
+int os_get_username(char *s, size_t len)
 {
 #ifdef UNIX
   return os_get_uname((uv_uid_t)getuid(), s, len);
@@ -124,17 +132,21 @@ int os_get_user_name(char *s, size_t len)
 #endif
 }
 
-// Insert user name for "uid" in s[len].
-// Return OK if a name found.
-// If the name is not found, write the uid into s[len] and return FAIL.
+/// Gets the username associated with `uid`.
+///
+/// @param uid User id.
+/// @param s[out] Username, or `uid` on failure.
+/// @param len Length of `s`.
+///
+/// @return OK if a username was found, else FAIL.
 int os_get_uname(uv_uid_t uid, char *s, size_t len)
 {
-#if defined(HAVE_PWD_H) && defined(HAVE_GETPWUID)
+#ifdef HAVE_PWD_FUNCS
   struct passwd *pw;
 
   if ((pw = getpwuid(uid)) != NULL  // NOLINT(runtime/threadsafe_fn)
       && pw->pw_name != NULL && *(pw->pw_name) != NUL) {
-    STRLCPY(s, pw->pw_name, len);
+    xstrlcpy(s, pw->pw_name, len);
     return OK;
   }
 #endif
@@ -142,12 +154,12 @@ int os_get_uname(uv_uid_t uid, char *s, size_t len)
   return FAIL;  // a number is not a name
 }
 
-// Returns the user directory for the given username.
-// The caller has to free() the returned string.
-// If the username is not found, NULL is returned.
-char *os_get_user_directory(const char *name)
+/// Gets the user directory for the given username, or NULL on failure.
+///
+/// Caller must free() the returned string.
+char *os_get_userdir(const char *name)
 {
-#if defined(HAVE_GETPWNAM) && defined(HAVE_PWD_H)
+#ifdef HAVE_PWD_FUNCS
   if (name == NULL || *name == NUL) {
     return NULL;
   }
@@ -160,7 +172,6 @@ char *os_get_user_directory(const char *name)
   return NULL;
 }
 
-
 #if defined(EXITFREE)
 
 void free_users(void)
@@ -187,11 +198,11 @@ static void init_users(void)
 }
 
 /// Given to ExpandGeneric() to obtain an user names.
-char_u *get_users(expand_T *xp, int idx)
+char *get_users(expand_T *xp, int idx)
 {
   init_users();
   if (idx < ga_users.ga_len) {
-    return ((char_u **)ga_users.ga_data)[idx];
+    return ((char **)ga_users.ga_data)[idx];
   }
   return NULL;
 }
@@ -201,17 +212,17 @@ char_u *get_users(expand_T *xp, int idx)
 /// @return 0 if name does not match any user name.
 ///         1 if name partially matches the beginning of a user name.
 ///         2 is name fully matches a user name.
-int match_user(char_u *name)
+int match_user(char *name)
 {
-  int n = (int)STRLEN(name);
+  int n = (int)strlen(name);
   int result = 0;
 
   init_users();
   for (int i = 0; i < ga_users.ga_len; i++) {
-    if (STRCMP(((char_u **)ga_users.ga_data)[i], name) == 0) {
+    if (strcmp(((char **)ga_users.ga_data)[i], name) == 0) {
       return 2;       // full match
     }
-    if (STRNCMP(((char_u **)ga_users.ga_data)[i], name, n) == 0) {
+    if (strncmp(((char **)ga_users.ga_data)[i], name, (size_t)n) == 0) {
       result = 1;       // partial match
     }
   }
