@@ -1,10 +1,12 @@
 #ifndef NVIM_POS_H
 #define NVIM_POS_H
 
+#include <inttypes.h>
+
 /// Line number type
-typedef long linenr_T;
+typedef int32_t linenr_T;
 /// Format used to print values which have linenr_T type
-#define PRIdLINENR "ld"
+#define PRIdLINENR PRId32
 
 /// Column number type
 typedef int colnr_T;
