@@ -17,6 +17,8 @@
 #include <stddef.h>
 #include <stdint.h>
 
+struct rbuffer;
+
 // Macros that simplify working with the read/write pointers directly by hiding
 // ring buffer wrap logic. Some examples:
 //
@@ -36,6 +38,8 @@
 //
 // Note that the rbuffer_{produced,consumed} calls are necessary or these macros
 // create infinite loops
+//
+// -V:RBUFFER_UNTIL_EMPTY:1044
 #define RBUFFER_UNTIL_EMPTY(buf, rptr, rcnt) \
   for (size_t rcnt = 0, _r = 1; _r; _r = 0)  /* NOLINT(readability/braces) */ \
   for (char *rptr = rbuffer_read_ptr(buf, &rcnt);  /* NOLINT(readability/braces) */ \
@@ -48,7 +52,6 @@
        rbuffer_space(buf); \
        wptr = rbuffer_write_ptr(buf, &wcnt))
 
-
 // Iteration
 #define RBUFFER_EACH(buf, c, i) \
   for (size_t i = 0;  /* NOLINT(readability/braces) */ \
