@@ -1,30 +1,39 @@
 // This is an open source non-commercial project. Dear PVS-Studio, please check
 // it. PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
 
-#include <assert.h>
+#include <stdbool.h>
+#include <stddef.h>
+#include <string.h>
 
 #include "nvim/ascii.h"
 #include "nvim/autocmd.h"
-#include "nvim/edit.h"
+#include "nvim/buffer_defs.h"
+#include "nvim/drawscreen.h"
 #include "nvim/eval.h"
-#include "nvim/ex_docmd.h"
+#include "nvim/eval/typval.h"
+#include "nvim/eval/typval_defs.h"
+#include "nvim/event/defs.h"
+#include "nvim/event/loop.h"
+#include "nvim/event/multiqueue.h"
 #include "nvim/getchar.h"
-#include "nvim/lib/kvec.h"
+#include "nvim/globals.h"
+#include "nvim/insexpand.h"
+#include "nvim/keycodes.h"
 #include "nvim/log.h"
+#include "nvim/macros.h"
 #include "nvim/main.h"
 #include "nvim/option.h"
-#include "nvim/option_defs.h"
 #include "nvim/os/input.h"
-#include "nvim/screen.h"
 #include "nvim/state.h"
+#include "nvim/strings.h"
+#include "nvim/types.h"
 #include "nvim/ui.h"
 #include "nvim/vim.h"
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
-# include "state.c.generated.h"
+# include "state.c.generated.h"  // IWYU pragma: export
 #endif
 
-
 void state_enter(VimState *s)
 {
   for (;;) {
@@ -56,9 +65,9 @@ getkey:
       key = K_EVENT;
     } else {
       // Duplicate display updating logic in vgetorpeek()
-      if (((State & INSERT) != 0 || p_lz) && (State & CMDLINE) == 0
+      if (((State & MODE_INSERT) != 0 || p_lz) && (State & MODE_CMDLINE) == 0
           && must_redraw != 0 && !need_wait_return) {
-        update_screen(0);
+        update_screen();
         setcursor();  // put cursor back where it belongs
       }
       // Flush screen updates before blocking
@@ -66,7 +75,7 @@ getkey:
       // Call `os_inchar` directly to block for events or user input without
       // consuming anything from `input_buffer`(os/input.c) or calling the
       // mapping engine.
-      (void)os_inchar(NULL, 0, -1, 0, main_loop.events);
+      (void)os_inchar(NULL, 0, -1, typebuf.tb_change_cnt, main_loop.events);
       // If an event was put into the queue, we send K_EVENT directly.
       if (!multiqueue_empty(main_loop.events)) {
         key = K_EVENT;
@@ -82,8 +91,9 @@ getkey:
       may_sync_undo();
     }
 
-#if MIN_LOG_LEVEL <= DEBUG_LOG_LEVEL
-    log_key(DEBUG_LOG_LEVEL, key);
+#ifdef NVIM_LOG_DEBUG
+    char *keyname = key == K_EVENT ? "K_EVENT" : get_special_key_name(key, mod_mask);
+    DLOG("input: %s", keyname);
 #endif
 
     int execute_result = s->execute(s, key);
@@ -122,7 +132,6 @@ void state_handle_k_event(void)
   }
 }
 
-
 /// Return true if in the current mode we need to use virtual.
 bool virtual_active(void)
 {
@@ -136,21 +145,22 @@ bool virtual_active(void)
   }
   return cur_ve_flags == VE_ALL
          || ((cur_ve_flags & VE_BLOCK) && VIsual_active && VIsual_mode == Ctrl_V)
-         || ((cur_ve_flags & VE_INSERT) && (State & INSERT));
+         || ((cur_ve_flags & VE_INSERT) && (State & MODE_INSERT));
 }
 
-/// VISUAL, SELECTMODE and OP_PENDING State are never set, they are equal to
-/// NORMAL State with a condition.  This function returns the real State.
+/// MODE_VISUAL, MODE_SELECT and MODE_OP_PENDING State are never set, they are
+/// equal to MODE_NORMAL State with a condition.  This function returns the real
+/// State.
 int get_real_state(void)
 {
-  if (State & NORMAL) {
+  if (State & MODE_NORMAL) {
     if (VIsual_active) {
       if (VIsual_select) {
-        return SELECTMODE;
+        return MODE_SELECT;
       }
-      return VISUAL;
+      return MODE_VISUAL;
     } else if (finish_op) {
-      return OP_PENDING;
+      return MODE_OP_PENDING;
     }
   }
   return State;
@@ -173,43 +183,38 @@ void get_mode(char *buf)
         buf[i++] = 's';
       }
     }
-  } else if (State == HITRETURN || State == ASKMORE || State == SETWSIZE
-             || State == CONFIRM) {
+  } else if (State == MODE_HITRETURN || State == MODE_ASKMORE || State == MODE_SETWSIZE
+             || State == MODE_CONFIRM) {
     buf[i++] = 'r';
-    if (State == ASKMORE) {
+    if (State == MODE_ASKMORE) {
       buf[i++] = 'm';
-    } else if (State == CONFIRM) {
+    } else if (State == MODE_CONFIRM) {
       buf[i++] = '?';
     }
-  } else if (State == EXTERNCMD) {
+  } else if (State == MODE_EXTERNCMD) {
     buf[i++] = '!';
-  } else if (State & INSERT) {
+  } else if (State & MODE_INSERT) {
     if (State & VREPLACE_FLAG) {
       buf[i++] = 'R';
       buf[i++] = 'v';
-      if (ins_compl_active()) {
-        buf[i++] = 'c';
-      } else if (ctrl_x_mode_not_defined_yet()) {
-        buf[i++] = 'x';
-      }
     } else {
       if (State & REPLACE_FLAG) {
         buf[i++] = 'R';
       } else {
         buf[i++] = 'i';
       }
-      if (ins_compl_active()) {
-        buf[i++] = 'c';
-      } else if (ctrl_x_mode_not_defined_yet()) {
-        buf[i++] = 'x';
-      }
     }
-  } else if ((State & CMDLINE) || exmode_active) {
+    if (ins_compl_active()) {
+      buf[i++] = 'c';
+    } else if (ctrl_x_mode_not_defined_yet()) {
+      buf[i++] = 'x';
+    }
+  } else if ((State & MODE_CMDLINE) || exmode_active) {
     buf[i++] = 'c';
     if (exmode_active) {
       buf[i++] = 'v';
     }
-  } else if (State & TERM_FOCUS) {
+  } else if (State & MODE_TERMINAL) {
     buf[i++] = 't';
   } else {
     buf[i++] = 'n';
@@ -217,12 +222,15 @@ void get_mode(char *buf)
       buf[i++] = 'o';
       // to be able to detect force-linewise/blockwise/charwise operations
       buf[i++] = (char)motion_force;
+    } else if (curbuf->terminal) {
+      buf[i++] = 't';
+      if (restart_edit == 'I') {
+        buf[i++] = 'T';
+      }
     } else if (restart_edit == 'I' || restart_edit == 'R'
                || restart_edit == 'V') {
       buf[i++] = 'i';
       buf[i++] = (char)restart_edit;
-    } else if (curbuf->terminal) {
-      buf[i++] = 't';
     }
   }
 
@@ -232,15 +240,17 @@ void get_mode(char *buf)
 /// Fires a ModeChanged autocmd if appropriate.
 void may_trigger_modechanged(void)
 {
-  if (!has_event(EVENT_MODECHANGED)) {
+  // Skip this when got_int is set, the autocommand will not be executed.
+  // Better trigger it next time.
+  if (!has_event(EVENT_MODECHANGED) || got_int) {
     return;
   }
 
   char curr_mode[MODE_MAX_LENGTH];
-  char_u pattern_buf[2 * MODE_MAX_LENGTH];
+  char pattern_buf[2 * MODE_MAX_LENGTH];
 
   get_mode(curr_mode);
-  if (STRCMP(curr_mode, last_mode) == 0) {
+  if (strcmp(curr_mode, last_mode) == 0) {
     return;
   }
 
@@ -251,7 +261,7 @@ void may_trigger_modechanged(void)
   tv_dict_set_keys_readonly(v_event);
 
   // concatenate modes in format "old_mode:new_mode"
-  vim_snprintf((char *)pattern_buf, sizeof(pattern_buf), "%s:%s", last_mode, curr_mode);
+  vim_snprintf(pattern_buf, sizeof(pattern_buf), "%s:%s", last_mode, curr_mode);
 
   apply_autocmds(EVENT_MODECHANGED, pattern_buf, NULL, false, curbuf);
   STRCPY(last_mode, curr_mode);
