@@ -6,106 +6,72 @@
 #include <math.h>
 #include <stdarg.h>
 #include <stdbool.h>
+#include <stddef.h>
+#include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 
+#include "auto/config.h"
 #include "nvim/ascii.h"
 #include "nvim/assert.h"
-#include "nvim/buffer.h"
 #include "nvim/charset.h"
-#include "nvim/diff.h"
-#include "nvim/edit.h"
-#include "nvim/eval.h"
 #include "nvim/eval/encode.h"
-#include "nvim/ex_cmds.h"
+#include "nvim/eval/typval.h"
+#include "nvim/eval/typval_defs.h"
 #include "nvim/ex_docmd.h"
-#include "nvim/ex_getln.h"
-#include "nvim/file_search.h"
-#include "nvim/fileio.h"
-#include "nvim/fold.h"
-#include "nvim/func_attr.h"
-#include "nvim/getchar.h"
-#include "nvim/mark.h"
+#include "nvim/gettext.h"
+#include "nvim/macros.h"
 #include "nvim/math.h"
 #include "nvim/mbyte.h"
-#include "nvim/memfile.h"
-#include "nvim/memline.h"
 #include "nvim/memory.h"
 #include "nvim/message.h"
-#include "nvim/move.h"
-#include "nvim/ops.h"
 #include "nvim/option.h"
-#include "nvim/os/os.h"
-#include "nvim/os/shell.h"
-#include "nvim/os_unix.h"
-#include "nvim/path.h"
-#include "nvim/quickfix.h"
-#include "nvim/regexp.h"
-#include "nvim/screen.h"
-#include "nvim/search.h"
-#include "nvim/spell.h"
 #include "nvim/strings.h"
-#include "nvim/syntax.h"
-#include "nvim/tag.h"
+#include "nvim/types.h"
 #include "nvim/vim.h"
-#include "nvim/window.h"
-
-/// Copy "string" into newly allocated memory.
-char_u *vim_strsave(const char_u *string)
-  FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
-{
-  return (char_u *)xstrdup((char *)string);
-}
 
 /// Copy up to `len` bytes of `string` into newly allocated memory and
 /// terminate with a NUL. The allocated memory always has size `len + 1`, even
 /// when `string` is shorter.
-char_u *vim_strnsave(const char_u *string, size_t len)
+char *xstrnsave(const char *string, size_t len)
   FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
 {
-  // strncpy is intentional: some parts of Vim use `string` shorter than `len`
-  // and expect the remainder to be zeroed out.
-  return (char_u *)strncpy(xmallocz(len), (char *)string, len);
+  return strncpy(xmallocz(len), string, len);  // NOLINT(runtime/printf)
 }
 
-/*
- * Same as vim_strsave(), but any characters found in esc_chars are preceded
- * by a backslash.
- */
-char_u *vim_strsave_escaped(const char_u *string, const char_u *esc_chars)
+// Same as vim_strsave(), but any characters found in esc_chars are preceded
+// by a backslash.
+char *vim_strsave_escaped(const char *string, const char *esc_chars)
   FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
 {
   return vim_strsave_escaped_ext(string, esc_chars, '\\', false);
 }
 
-/*
- * Same as vim_strsave_escaped(), but when "bsl" is true also escape
- * characters where rem_backslash() would remove the backslash.
- * Escape the characters with "cc".
- */
-char_u *vim_strsave_escaped_ext(const char_u *string, const char_u *esc_chars, char_u cc, bool bsl)
+// Same as vim_strsave_escaped(), but when "bsl" is true also escape
+// characters where rem_backslash() would remove the backslash.
+// Escape the characters with "cc".
+char *vim_strsave_escaped_ext(const char *string, const char *esc_chars, char cc, bool bsl)
   FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
 {
-  /*
-   * First count the number of backslashes required.
-   * Then allocate the memory and insert them.
-   */
+  // First count the number of backslashes required.
+  // Then allocate the memory and insert them.
   size_t length = 1;                    // count the trailing NUL
-  for (const char_u *p = string; *p; p++) {
+  for (const char *p = string; *p; p++) {
     const size_t l = (size_t)(utfc_ptr2len(p));
     if (l > 1) {
       length += l;                      // count a multibyte char
       p += l - 1;
       continue;
     }
-    if (vim_strchr(esc_chars, *p) != NULL || (bsl && rem_backslash(p))) {
-      ++length;                         // count a backslash
+    if (vim_strchr(esc_chars, (uint8_t)(*p)) != NULL || (bsl && rem_backslash(p))) {
+      length++;                         // count a backslash
     }
-    ++length;                           // count an ordinary char
+    length++;                           // count an ordinary char
   }
 
-  char_u *escaped_string = xmalloc(length);
-  char_u *p2 = escaped_string;
-  for (const char_u *p = string; *p; p++) {
+  char *escaped_string = xmalloc(length);
+  char *p2 = escaped_string;
+  for (const char *p = string; *p; p++) {
     const size_t l = (size_t)(utfc_ptr2len(p));
     if (l > 1) {
       memcpy(p2, p, l);
@@ -113,7 +79,7 @@ char_u *vim_strsave_escaped_ext(const char_u *string, const char_u *esc_chars, c
       p += l - 1;                     // skip multibyte char
       continue;
     }
-    if (vim_strchr(esc_chars, *p) != NULL || (bsl && rem_backslash(p))) {
+    if (vim_strchr(esc_chars, (uint8_t)(*p)) != NULL || (bsl && rem_backslash(p))) {
       *p2++ = cc;
     }
     *p2++ = *p;
@@ -139,7 +105,7 @@ char *vim_strnsave_unquoted(const char *const string, const size_t length)
   FUNC_ATTR_NONNULL_RET
 {
 #define ESCAPE_COND(p, inquote, string_end) \
-  (*p == '\\' && inquote && p + 1 < string_end && (p[1] == '\\' || p[1] == '"'))
+  (*(p) == '\\' && (inquote) && (p) + 1 < (string_end) && ((p)[1] == '\\' || (p)[1] == '"'))
   size_t ret_length = 0;
   bool inquote = false;
   const char *const string_end = string + length;
@@ -171,29 +137,28 @@ char *vim_strnsave_unquoted(const char *const string, const size_t length)
   return ret;
 }
 
-/*
- * Escape "string" for use as a shell argument with system().
- * This uses single quotes, except when we know we need to use double quotes
- * (MS-Windows without 'shellslash' set).
- * Escape a newline, depending on the 'shell' option.
- * When "do_special" is true also replace "!", "%", "#" and things starting
- * with "<" like "<cfile>".
- * When "do_newline" is false do not escape newline unless it is csh shell.
- * Returns the result in allocated memory.
- */
-char_u *vim_strsave_shellescape(const char_u *string, bool do_special, bool do_newline)
+/// Escape "string" for use as a shell argument with system().
+/// This uses single quotes, except when we know we need to use double quotes
+/// (MS-Windows without 'shellslash' set).
+/// Escape a newline, depending on the 'shell' option.
+/// When "do_special" is true also replace "!", "%", "#" and things starting
+/// with "<" like "<cfile>".
+/// When "do_newline" is false do not escape newline unless it is csh shell.
+///
+/// @return  the result in allocated memory.
+char *vim_strsave_shellescape(const char *string, bool do_special, bool do_newline)
   FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
 {
-  char_u *d;
-  char_u *escaped_string;
+  char *d;
+  char *escaped_string;
   size_t l;
   int csh_like;
   bool fish_like;
 
-  /* Only csh and similar shells expand '!' within single quotes.  For sh and
-   * the like we must not put a backslash before it, it will be taken
-   * literally.  If do_special is set the '!' will be escaped twice.
-   * Csh also needs to have "\n" escaped twice when do_special is set. */
+  // Only csh and similar shells expand '!' within single quotes.  For sh and
+  // the like we must not put a backslash before it, it will be taken
+  // literally.  If do_special is set the '!' will be escaped twice.
+  // Csh also needs to have "\n" escaped twice when do_special is set.
   csh_like = csh_like_shell();
 
   // Fish shell uses '\' as an escape character within single quotes, so '\'
@@ -201,9 +166,9 @@ char_u *vim_strsave_shellescape(const char_u *string, bool do_special, bool do_n
   fish_like = fish_like_shell();
 
   // First count the number of extra bytes required.
-  size_t length = STRLEN(string) + 3;       // two quotes and a trailing NUL
-  for (const char_u *p = string; *p != NUL; MB_PTR_ADV(p)) {
-#ifdef WIN32
+  size_t length = strlen(string) + 3;       // two quotes and a trailing NUL
+  for (const char *p = string; *p != NUL; MB_PTR_ADV(p)) {
+#ifdef MSWIN
     if (!p_ssl) {
       if (*p == '"') {
         length++;                       // " -> ""
@@ -215,13 +180,13 @@ char_u *vim_strsave_shellescape(const char_u *string, bool do_special, bool do_n
     }
     if ((*p == '\n' && (csh_like || do_newline))
         || (*p == '!' && (csh_like || do_special))) {
-      ++length;                         // insert backslash
+      length++;                         // insert backslash
       if (csh_like && do_special) {
-        ++length;                       // insert backslash
+        length++;                       // insert backslash
       }
     }
     if (do_special && find_cmdline_var(p, &l) >= 0) {
-      ++length;                         // insert backslash
+      length++;                         // insert backslash
       p += l - 1;
     }
     if (*p == '\\' && fish_like) {
@@ -234,15 +199,15 @@ char_u *vim_strsave_shellescape(const char_u *string, bool do_special, bool do_n
   d = escaped_string;
 
   // add opening quote
-#ifdef WIN32
+#ifdef MSWIN
   if (!p_ssl) {
     *d++ = '"';
   } else
 #endif
   *d++ = '\'';
 
-  for (const char_u *p = string; *p != NUL;) {
-#ifdef WIN32
+  for (const char *p = string; *p != NUL;) {
+#ifdef MSWIN
     if (!p_ssl) {
       if (*p == '"') {
         *d++ = '"';
@@ -257,7 +222,7 @@ char_u *vim_strsave_shellescape(const char_u *string, bool do_special, bool do_n
       *d++ = '\\';
       *d++ = '\'';
       *d++ = '\'';
-      ++p;
+      p++;
       continue;
     }
     if ((*p == '\n' && (csh_like || do_newline))
@@ -286,7 +251,7 @@ char_u *vim_strsave_shellescape(const char_u *string, bool do_special, bool do_n
   }
 
   // add terminating quote and finish with a NUL
-#ifdef WIN32
+#ifdef MSWIN
   if (!p_ssl) {
     *d++ = '"';
   } else
@@ -297,41 +262,35 @@ char_u *vim_strsave_shellescape(const char_u *string, bool do_special, bool do_n
   return escaped_string;
 }
 
-/*
- * Like vim_strsave(), but make all characters uppercase.
- * This uses ASCII lower-to-upper case translation, language independent.
- */
-char_u *vim_strsave_up(const char_u *string)
+// Like vim_strsave(), but make all characters uppercase.
+// This uses ASCII lower-to-upper case translation, language independent.
+char *vim_strsave_up(const char *string)
   FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
 {
-  char_u *p1;
+  char *p1;
 
-  p1 = vim_strsave(string);
+  p1 = xstrdup(string);
   vim_strup(p1);
   return p1;
 }
 
-/*
- * Like vim_strnsave(), but make all characters uppercase.
- * This uses ASCII lower-to-upper case translation, language independent.
- */
-char_u *vim_strnsave_up(const char_u *string, size_t len)
+/// Like xstrnsave(), but make all characters uppercase.
+/// This uses ASCII lower-to-upper case translation, language independent.
+char *vim_strnsave_up(const char *string, size_t len)
   FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
 {
-  char_u *p1 = vim_strnsave(string, len);
+  char *p1 = xstrnsave(string, len);
   vim_strup(p1);
   return p1;
 }
 
-/*
- * ASCII lower-to-upper case translation, language independent.
- */
-void vim_strup(char_u *p)
+// ASCII lower-to-upper case translation, language independent.
+void vim_strup(char *p)
   FUNC_ATTR_NONNULL_ALL
 {
-  char_u c;
-  while ((c = *p) != NUL) {
-    *p++ = (char_u)(c < 'a' || c > 'z' ? c : c - 0x20);
+  uint8_t c;
+  while ((c = (uint8_t)(*p)) != NUL) {
+    *p++ = (char)(uint8_t)(c < 'a' || c > 'z' ? c : c - 0x20);
   }
 }
 
@@ -350,11 +309,11 @@ char *strcase_save(const char *const orig, bool upper)
 
   char *p = res;
   while (*p != NUL) {
-    int c = utf_ptr2char((const char_u *)p);
-    int l = utf_ptr2len((const char_u *)p);
+    int c = utf_ptr2char(p);
+    int l = utf_ptr2len(p);
     if (c == 0) {
       // overlong sequence, use only the first byte
-      c = (char_u)(*p);
+      c = (uint8_t)(*p);
       l = 1;
     }
     int uc = upper ? mb_toupper(c) : mb_tolower(c);
@@ -364,7 +323,7 @@ char *strcase_save(const char *const orig, bool upper)
     int newl = utf_char2len(uc);
     if (newl != l) {
       // TODO(philix): use xrealloc() in strcase_save()
-      char *s = xmalloc(STRLEN(res) + (size_t)(1 + newl - l));
+      char *s = xmalloc(strlen(res) + (size_t)(1 + newl - l));
       memcpy(s, res, (size_t)(p - res));
       STRCPY(s + (p - res) + newl, p + l);
       p = s + (p - res);
@@ -372,22 +331,20 @@ char *strcase_save(const char *const orig, bool upper)
       res = s;
     }
 
-    utf_char2bytes(uc, (char_u *)p);
+    utf_char2bytes(uc, p);
     p += newl;
   }
 
   return res;
 }
 
-/*
- * delete spaces at the end of a string
- */
-void del_trailing_spaces(char_u *ptr)
+// delete spaces at the end of a string
+void del_trailing_spaces(char *ptr)
   FUNC_ATTR_NONNULL_ALL
 {
-  char_u *q;
+  char *q;
 
-  q = ptr + STRLEN(ptr);
+  q = ptr + strlen(ptr);
   while (--q > ptr && ascii_iswhite(q[0]) && q[-1] != '\\' && q[-1] != Ctrl_V) {
     *q = NUL;
   }
@@ -401,58 +358,54 @@ size_t xstrnlen(const char *s, size_t n)
   if (end == NULL) {
     return n;
   }
-  return end - s;
+  return (size_t)(end - s);
 }
 #endif
 
 #if (!defined(HAVE_STRCASECMP) && !defined(HAVE_STRICMP))
-/*
- * Compare two strings, ignoring case, using current locale.
- * Doesn't work for multi-byte characters.
- * return 0 for match, < 0 for smaller, > 0 for bigger
- */
+// Compare two strings, ignoring case, using current locale.
+// Doesn't work for multi-byte characters.
+// return 0 for match, < 0 for smaller, > 0 for bigger
 int vim_stricmp(const char *s1, const char *s2)
   FUNC_ATTR_NONNULL_ALL FUNC_ATTR_PURE
 {
   int i;
 
   for (;;) {
-    i = (int)TOLOWER_LOC(*s1) - (int)TOLOWER_LOC(*s2);
+    i = (int)TOLOWER_LOC((uint8_t)(*s1)) - (int)TOLOWER_LOC((uint8_t)(*s2));
     if (i != 0) {
       return i;                             // this character different
     }
     if (*s1 == NUL) {
       break;                                // strings match until NUL
     }
-    ++s1;
-    ++s2;
+    s1++;
+    s2++;
   }
   return 0;                                 // strings match
 }
 #endif
 
 #if (!defined(HAVE_STRNCASECMP) && !defined(HAVE_STRNICMP))
-/*
- * Compare two strings, for length "len", ignoring case, using current locale.
- * Doesn't work for multi-byte characters.
- * return 0 for match, < 0 for smaller, > 0 for bigger
- */
+// Compare two strings, for length "len", ignoring case, using current locale.
+// Doesn't work for multi-byte characters.
+// return 0 for match, < 0 for smaller, > 0 for bigger
 int vim_strnicmp(const char *s1, const char *s2, size_t len)
   FUNC_ATTR_NONNULL_ALL FUNC_ATTR_PURE
 {
   int i;
 
   while (len > 0) {
-    i = (int)TOLOWER_LOC(*s1) - (int)TOLOWER_LOC(*s2);
+    i = (int)TOLOWER_LOC((uint8_t)(*s1)) - (int)TOLOWER_LOC((uint8_t)(*s2));
     if (i != 0) {
       return i;                             // this character different
     }
     if (*s1 == NUL) {
       break;                                // strings match until NUL
     }
-    ++s1;
-    ++s2;
-    --len;
+    s1++;
+    s2++;
+    len--;
   }
   return 0;                                 // strings match
 }
@@ -466,24 +419,22 @@ int vim_strnicmp(const char *s1, const char *s2, size_t len)
 /// @return Pointer to the first byte of the found character in string or NULL
 ///         if it was not found or character is invalid. NUL character is never
 ///         found, use `strlen()` instead.
-char_u *vim_strchr(const char_u *const string, const int c)
+char *vim_strchr(const char *const string, const int c)
   FUNC_ATTR_NONNULL_ALL FUNC_ATTR_PURE FUNC_ATTR_WARN_UNUSED_RESULT
 {
   if (c <= 0) {
     return NULL;
   } else if (c < 0x80) {
-    return (char_u *)strchr((const char *)string, c);
+    return strchr(string, c);
   } else {
     char u8char[MB_MAXBYTES + 1];
-    const int len = utf_char2bytes(c, (char_u *)u8char);
+    const int len = utf_char2bytes(c, u8char);
     u8char[len] = NUL;
-    return (char_u *)strstr((const char *)string, u8char);
+    return strstr(string, u8char);
   }
 }
 
-/*
- * Sort an array of strings.
- */
+// Sort an array of strings.
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "strings.c.generated.h"
@@ -491,26 +442,24 @@ char_u *vim_strchr(const char_u *const string, const int c)
 static int sort_compare(const void *s1, const void *s2)
   FUNC_ATTR_NONNULL_ALL
 {
-  return STRCMP(*(char **)s1, *(char **)s2);
+  return strcmp(*(char **)s1, *(char **)s2);
 }
 
-void sort_strings(char_u **files, int count)
+void sort_strings(char **files, int count)
 {
-  qsort((void *)files, (size_t)count, sizeof(char_u *), sort_compare);
+  qsort((void *)files, (size_t)count, sizeof(char *), sort_compare);
 }
 
-/*
- * Return true if string "s" contains a non-ASCII character (128 or higher).
- * When "s" is NULL false is returned.
- */
-bool has_non_ascii(const char_u *s)
+// Return true if string "s" contains a non-ASCII character (128 or higher).
+// When "s" is NULL false is returned.
+bool has_non_ascii(const char *s)
   FUNC_ATTR_PURE
 {
-  const char_u *p;
+  const char *p;
 
   if (s != NULL) {
-    for (p = s; *p != NUL; ++p) {
-      if (*p >= 128) {
+    for (p = s; *p != NUL; p++) {
+      if ((uint8_t)(*p) >= 128) {
         return true;
       }
     }
@@ -533,20 +482,17 @@ bool has_non_ascii_len(const char *const s, const size_t len)
   return false;
 }
 
-/*
- * Concatenate two strings and return the result in allocated memory.
- */
-char_u *concat_str(const char_u *restrict str1, const char_u *restrict str2)
+/// Concatenate two strings and return the result in allocated memory.
+char *concat_str(const char *restrict str1, const char *restrict str2)
   FUNC_ATTR_NONNULL_RET FUNC_ATTR_MALLOC FUNC_ATTR_NONNULL_ALL
 {
-  size_t l = STRLEN(str1);
-  char_u *dest = xmalloc(l + STRLEN(str2) + 1);
+  size_t l = strlen(str1);
+  char *dest = xmalloc(l + strlen(str2) + 1);
   STRCPY(dest, str1);
   STRCPY(dest + l, str2);
   return dest;
 }
 
-
 static const char *const e_printf =
   N_("E766: Insufficient arguments for printf()");
 
@@ -627,7 +573,7 @@ static const void *tv_ptr(const typval_T *const tvs, int *const idxp)
   FUNC_ATTR_NONNULL_ALL FUNC_ATTR_WARN_UNUSED_RESULT
 {
 #define OFF(attr) offsetof(union typval_vval_union, attr)
-  STATIC_ASSERT(OFF(v_string) == OFF(v_list)
+  STATIC_ASSERT(OFF(v_string) == OFF(v_list)  // -V568
                 && OFF(v_string) == OFF(v_dict)
                 && OFF(v_string) == OFF(v_partial)
                 && sizeof(tvs[0].vval.v_string) == sizeof(tvs[0].vval.v_list)
@@ -640,10 +586,9 @@ static const void *tv_ptr(const typval_T *const tvs, int *const idxp)
   if (tvs[idx].v_type == VAR_UNKNOWN) {
     emsg(_(e_printf));
     return NULL;
-  } else {
-    (*idxp)++;
-    return tvs[idx].vval.v_string;
   }
+  (*idxp)++;
+  return tvs[idx].vval.v_string;
 }
 
 /// Get float argument from idxp entry in tvs
@@ -1001,10 +946,10 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
                                  - str_arg);
           }
           if (fmt_spec == 'S') {
-            char_u *p1;
+            const char *p1;
             size_t i;
 
-            for (i = 0, p1 = (char_u *)str_arg; *p1; p1 += utfc_ptr2len(p1)) {
+            for (i = 0, p1 = str_arg; *p1; p1 += utfc_ptr2len(p1)) {
               size_t cell = (size_t)utf_ptr2cells(p1);
               if (precision_specified && i + cell > precision) {
                 break;
@@ -1012,7 +957,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
               i += cell;
             }
 
-            str_arg_l = (size_t)(p1 - (char_u *)str_arg);
+            str_arg_l = (size_t)(p1 - str_arg);
             if (min_field_width != 0) {
               min_field_width += str_arg_l - i;
             }
@@ -1071,13 +1016,11 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
                     : va_arg(ap, long long));  // NOLINT (runtime/int)
             break;
           case 'z':
-            arg = (tvs
-                       ? (ptrdiff_t)tv_nr(tvs, &arg_idx)
-                       : va_arg(ap, ptrdiff_t));
+            arg = (tvs ? (ptrdiff_t)tv_nr(tvs, &arg_idx) : va_arg(ap, ptrdiff_t));
             break;
           }
           if (arg > 0) {
-            arg_sign =  1;
+            arg_sign = 1;
           } else if (arg < 0) {
             arg_sign = -1;
           }
@@ -1085,19 +1028,13 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
           // unsigned
           switch (length_modifier) {
           case '\0':
-            uarg = (unsigned int)(tvs
-                                      ? tv_nr(tvs, &arg_idx)
-                                      : va_arg(ap, unsigned int));
+            uarg = (unsigned int)(tvs ? tv_nr(tvs, &arg_idx) : va_arg(ap, unsigned int));
             break;
           case 'h':
-            uarg = (uint16_t)(tvs
-                                  ? tv_nr(tvs, &arg_idx)
-                                  : va_arg(ap, unsigned int));
+            uarg = (uint16_t)(tvs ? tv_nr(tvs, &arg_idx) : va_arg(ap, unsigned int));
             break;
           case 'l':
-            uarg = (tvs
-                        ? (unsigned long)tv_nr(tvs, &arg_idx)
-                        : va_arg(ap, unsigned long));
+            uarg = (tvs ? (unsigned long)tv_nr(tvs, &arg_idx) : va_arg(ap, unsigned long));
             break;
           case '2':
             uarg = (uintmax_t)(unsigned long long)(  // NOLINT (runtime/int)
@@ -1107,9 +1044,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
                     : va_arg(ap, unsigned long long));  // NOLINT (runtime/int)
             break;
           case 'z':
-            uarg = (tvs
-                        ? (size_t)tv_nr(tvs, &arg_idx)
-                        : va_arg(ap, size_t));
+            uarg = (tvs ? (size_t)tv_nr(tvs, &arg_idx) : va_arg(ap, size_t));
             break;
           }
           arg_sign = (uarg != 0);
@@ -1306,8 +1241,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
             if (fmt_spec == 'f' || fmt_spec == 'F') {
               tp = tmp + str_arg_l - 1;
             } else {
-              tp = (char *)vim_strchr((char_u *)tmp,
-                                      fmt_spec == 'e' ? 'e' : 'E');
+              tp = vim_strchr(tmp, fmt_spec == 'e' ? 'e' : 'E');
               if (tp) {
                 // remove superfluous '+' and leading zeroes from exponent
                 if (tp[1] == '+') {
@@ -1336,8 +1270,7 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
           } else {
             // Be consistent: some printf("%e") use 1.0e+12 and some
             // 1.0e+012; remove one zero in the last case.
-            char *tp = (char *)vim_strchr((char_u *)tmp,
-                                          fmt_spec == 'e' ? 'e' : 'E');
+            char *tp = vim_strchr(tmp, fmt_spec == 'e' ? 'e' : 'E');
             if (tp && (tp[1] == '+' || tp[1] == '-') && tp[2] == '0'
                 && ascii_isdigit(tp[3]) && ascii_isdigit(tp[4])) {
               STRMOVE(tp + 2, tp + 3);
@@ -1473,3 +1406,96 @@ int vim_vsnprintf_typval(char *str, size_t str_m, const char *fmt, va_list ap, t
   // written to the buffer if it were large enough.
   return (int)str_l;
 }
+
+int kv_do_printf(StringBuilder *str, const char *fmt, ...)
+  FUNC_ATTR_PRINTF(2, 3)
+{
+  size_t remaining = str->capacity - str->size;
+
+  va_list ap;
+  va_start(ap, fmt);
+  int printed = vsnprintf(str->items ? str->items + str->size : NULL, remaining, fmt, ap);
+  va_end(ap);
+
+  if (printed < 0) {
+    return -1;
+  }
+
+  // printed string didn't fit, resize and try again
+  if ((size_t)printed >= remaining) {
+    kv_ensure_space(*str, (size_t)printed + 1);  // include space for NUL terminator at the end
+    assert(str->items != NULL);
+    va_start(ap, fmt);
+    printed = vsnprintf(str->items + str->size, str->capacity - str->size, fmt, ap);
+    va_end(ap);
+    if (printed < 0) {
+      return -1;
+    }
+  }
+
+  str->size += (size_t)printed;
+  return printed;
+}
+
+/// Reverse text into allocated memory.
+///
+/// @return  the allocated string.
+char *reverse_text(char *s)
+  FUNC_ATTR_NONNULL_RET
+{
+  // Reverse the pattern.
+  size_t len = strlen(s);
+  char *rev = xmalloc(len + 1);
+  size_t rev_i = len;
+  for (size_t s_i = 0; s_i < len; s_i++) {
+    const int mb_len = utfc_ptr2len(s + s_i);
+    rev_i -= (size_t)mb_len;
+    memmove(rev + rev_i, s + s_i, (size_t)mb_len);
+    s_i += (size_t)mb_len - 1;
+  }
+  rev[len] = NUL;
+
+  return rev;
+}
+
+/// Replace all occurrences of "what" with "rep" in "src". If no replacement happens then NULL is
+/// returned otherwise return a newly allocated string.
+///
+/// @param[in] src  Source text
+/// @param[in] what Substring to replace
+/// @param[in] rep  Substring to replace with
+///
+/// @return [allocated] Copy of the string.
+char *strrep(const char *src, const char *what, const char *rep)
+{
+  const char *pos = src;
+  size_t whatlen = strlen(what);
+
+  // Count occurrences
+  size_t count = 0;
+  while ((pos = strstr(pos, what)) != NULL) {
+    count++;
+    pos += whatlen;
+  }
+
+  if (count == 0) {
+    return NULL;
+  }
+
+  size_t replen = strlen(rep);
+  char *ret = xmalloc(strlen(src) + count * (replen - whatlen) + 1);
+  char *ptr = ret;
+  while ((pos = strstr(src, what)) != NULL) {
+    size_t idx = (size_t)(pos - src);
+    memcpy(ptr, src, idx);
+    ptr += idx;
+    STRCPY(ptr, rep);
+    ptr += replen;
+    src = pos + whatlen;
+  }
+
+  // Copy remaining
+  STRCPY(ptr, src);
+
+  return ret;
+}
