@@ -5,6 +5,7 @@
 #include <stdbool.h>
 #include <string.h>
 
+#include "klib/kvec.h"
 #include "nvim/eval/typval.h"
 #include "nvim/types.h"
 
@@ -25,6 +26,8 @@ static inline char *strappend(char *const dst, const char *const src)
   return (char *)memmove(dst, src, src_len) + src_len;
 }
 
+typedef kvec_t(char) StringBuilder;
+
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "strings.h.generated.h"
 #endif
