@@ -37,44 +37,56 @@
 // Some code from pangoterm http://www.leonerd.org.uk/code/pangoterm
 
 #include <assert.h>
+#include <limits.h>
 #include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <vterm.h>
+#include <vterm_keycodes.h>
 
+#include "nvim/api/private/defs.h"
 #include "nvim/api/private/helpers.h"
 #include "nvim/ascii.h"
+#include "nvim/autocmd.h"
 #include "nvim/buffer.h"
+#include "nvim/buffer_defs.h"
 #include "nvim/change.h"
+#include "nvim/channel.h"
 #include "nvim/cursor.h"
-#include "nvim/edit.h"
+#include "nvim/drawline.h"
+#include "nvim/drawscreen.h"
+#include "nvim/eval.h"
+#include "nvim/eval/typval.h"
+#include "nvim/eval/typval_defs.h"
 #include "nvim/event/loop.h"
+#include "nvim/event/multiqueue.h"
 #include "nvim/event/time.h"
-#include "nvim/ex_cmds.h"
 #include "nvim/ex_docmd.h"
-#include "nvim/fileio.h"
 #include "nvim/getchar.h"
+#include "nvim/globals.h"
 #include "nvim/highlight.h"
 #include "nvim/highlight_group.h"
-#include "nvim/keymap.h"
-#include "nvim/log.h"
+#include "nvim/keycodes.h"
 #include "nvim/macros.h"
 #include "nvim/main.h"
 #include "nvim/map.h"
 #include "nvim/mbyte.h"
 #include "nvim/memline.h"
 #include "nvim/memory.h"
-#include "nvim/message.h"
 #include "nvim/mouse.h"
 #include "nvim/move.h"
+#include "nvim/msgpack_rpc/channel_defs.h"
+#include "nvim/normal.h"
 #include "nvim/option.h"
-#include "nvim/os/input.h"
-#include "nvim/screen.h"
+#include "nvim/optionstr.h"
+#include "nvim/pos.h"
 #include "nvim/state.h"
 #include "nvim/terminal.h"
+#include "nvim/types.h"
 #include "nvim/ui.h"
 #include "nvim/vim.h"
-#include "nvim/window.h"
 
 typedef struct terminal_state {
   VimState state;
@@ -82,6 +94,7 @@ typedef struct terminal_state {
   int save_rd;              // saved value of RedrawingDisabled
   bool close;
   bool got_bsl;             // if the last input was <C-\>
+  bool got_bsl_o;           // if left terminal mode with <c-\><c-o>
 } TerminalState;
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
@@ -109,15 +122,19 @@ struct terminal {
   //  - receive data from libvterm as a result of key presses.
   char textbuf[0x1fff];
 
-  ScrollbackLine **sb_buffer;       // Scrollback buffer storage for libvterm
-  size_t sb_current;                // number of rows pushed to sb_buffer
-  size_t sb_size;                   // sb_buffer size
+  ScrollbackLine **sb_buffer;       // Scrollback storage.
+  size_t sb_current;                // Lines stored in sb_buffer.
+  size_t sb_size;                   // Capacity of sb_buffer.
   // "virtual index" that points to the first sb_buffer row that we need to
   // push to the terminal buffer when refreshing the scrollback. When negative,
   // it actually points to entries that are no longer in sb_buffer (because the
   // window height has increased) and must be deleted from the terminal buffer
   int sb_pending;
 
+  char *title;     // VTermStringFragment buffer
+  size_t title_len;    // number of rows pushed to sb_buffer
+  size_t title_size;   // sb_buffer size
+
   // buf_T instance that acts as a "drawing surface" for libvterm
   // we can't store a direct reference to the buffer because the
   // refresh_timer_cb may be called after the buffer was freed, and there's
@@ -148,7 +165,7 @@ static VTermScreenCallbacks vterm_screen_callbacks = {
   .movecursor  = term_movecursor,
   .settermprop = term_settermprop,
   .bell        = term_bell,
-  .sb_pushline = term_sb_push,
+  .sb_pushline = term_sb_push,  // Called before a line goes offscreen.
   .sb_popline  = term_sb_pop,
 };
 
@@ -186,10 +203,11 @@ static void term_output_callback(const char *s, size_t len, void *user_data)
 ///
 /// @param buf Buffer used for presentation of the terminal.
 /// @param opts PTY process channel, various terminal properties and callbacks.
-Terminal *terminal_open(buf_T *buf, TerminalOptions opts)
+void terminal_open(Terminal **termpp, buf_T *buf, TerminalOptions opts)
+  FUNC_ATTR_NONNULL_ALL
 {
   // Create a new terminal instance and configure it
-  Terminal *rv = xcalloc(1, sizeof(Terminal));
+  Terminal *rv = *termpp = xcalloc(1, sizeof(Terminal));
   rv->opts = opts;
   rv->cursor.visible = true;
   // Associate the terminal instance with the new buffer
@@ -228,21 +246,31 @@ Terminal *terminal_open(buf_T *buf, TerminalOptions opts)
   set_option_value("wrap", false, NULL, OPT_LOCAL);
   set_option_value("list", false, NULL, OPT_LOCAL);
   if (buf->b_ffname != NULL) {
-    buf_set_term_title(buf, (char *)buf->b_ffname);
+    buf_set_term_title(buf, buf->b_ffname, strlen(buf->b_ffname));
   }
   RESET_BINDING(curwin);
   // Reset cursor in current window.
   curwin->w_cursor = (pos_T){ .lnum = 1, .col = 0, .coladd = 0 };
+  // Initialize to check if the scrollback buffer has been allocated in a TermOpen autocmd.
+  rv->sb_buffer = NULL;
   // Apply TermOpen autocmds _before_ configuring the scrollback buffer.
   apply_autocmds(EVENT_TERMOPEN, NULL, NULL, false, buf);
-  // Local 'scrollback' _after_ autocmds.
-  buf->b_p_scbk = (buf->b_p_scbk < 1) ? SB_MAX : buf->b_p_scbk;
 
   aucmd_restbuf(&aco);
 
-  // Configure the scrollback buffer.
-  rv->sb_size = (size_t)buf->b_p_scbk;
-  rv->sb_buffer = xmalloc(sizeof(ScrollbackLine *) * rv->sb_size);
+  if (*termpp == NULL) {
+    return;  // Terminal has already been destroyed.
+  }
+
+  if (rv->sb_buffer == NULL) {
+    // Local 'scrollback' _after_ autocmds.
+    if (buf->b_p_scbk < 1) {
+      buf->b_p_scbk = SB_MAX;
+    }
+    // Configure the scrollback buffer.
+    rv->sb_size = (size_t)buf->b_p_scbk;
+    rv->sb_buffer = xmalloc(sizeof(ScrollbackLine *) * rv->sb_size);
+  }
 
   // Configure the color palette. Try to get the color from:
   //
@@ -255,7 +283,8 @@ Terminal *terminal_open(buf_T *buf, TerminalOptions opts)
     snprintf(var, sizeof(var), "terminal_color_%d", i);
     char *name = get_config_string(var);
     if (name) {
-      color_val = name_to_color(name);
+      int dummy;
+      color_val = name_to_color(name, &dummy);
       xfree(name);
 
       if (color_val != -1) {
@@ -269,12 +298,15 @@ Terminal *terminal_open(buf_T *buf, TerminalOptions opts)
       }
     }
   }
-
-  return rv;
 }
 
-void terminal_close(Terminal *term, int status)
+/// Closes the Terminal buffer.
+///
+/// May call terminal_destroy, which sets caller storage to NULL.
+void terminal_close(Terminal **termpp, int status)
+  FUNC_ATTR_NONNULL_ALL
 {
+  Terminal *term = *termpp;
   if (term->destroy) {
     return;
   }
@@ -283,7 +315,7 @@ void terminal_close(Terminal *term, int status)
   if (entered_free_all_mem) {
     // If called from close_buffer() inside free_all_mem(), the main loop has
     // already been freed, so it is not safe to call the close callback here.
-    terminal_destroy(term);
+    terminal_destroy(termpp);
     return;
   }
 #endif
@@ -359,8 +391,12 @@ void terminal_check_size(Terminal *term)
   vterm_get_size(term->vt, &curheight, &curwidth);
   uint16_t width = 0, height = 0;
 
-
+  // Check if there is a window that displays the terminal and find the maximum width and height.
+  // Skip the autocommand window which isn't actually displayed.
   FOR_ALL_TAB_WINDOWS(tp, wp) {
+    if (is_aucmd_win(wp)) {
+      continue;
+    }
     if (wp->w_buffer && wp->w_buffer->terminal == term) {
       const uint16_t win_width =
         (uint16_t)(MAX(0, wp->w_width_inner - win_col_off(wp)));
@@ -381,13 +417,12 @@ void terminal_check_size(Terminal *term)
   invalidate_terminal(term, -1, -1);
 }
 
-/// Implements TERM_FOCUS mode. :help Terminal-mode
-void terminal_enter(void)
+/// Implements MODE_TERMINAL state. :help Terminal-mode
+bool terminal_enter(void)
 {
   buf_T *buf = curbuf;
   assert(buf->terminal);  // Should only be called when curbuf has a terminal.
-  TerminalState state, *s = &state;
-  memset(s, 0, sizeof(TerminalState));
+  TerminalState s[1] = { 0 };
   s->term = buf->terminal;
   stop_insert_mode = false;
 
@@ -398,23 +433,23 @@ void terminal_enter(void)
 
   int save_state = State;
   s->save_rd = RedrawingDisabled;
-  State = TERM_FOCUS;
-  mapped_ctrl_c |= TERM_FOCUS;  // Always map CTRL-C to avoid interrupt.
+  State = MODE_TERMINAL;
+  mapped_ctrl_c |= MODE_TERMINAL;  // Always map CTRL-C to avoid interrupt.
   RedrawingDisabled = false;
 
   // Disable these options in terminal-mode. They are nonsense because cursor is
   // placed at end of buffer to "follow" output. #11072
-  win_T *save_curwin = curwin;
+  handle_T save_curwin = curwin->handle;
   bool save_w_p_cul = curwin->w_p_cul;
-  char_u *save_w_p_culopt = NULL;
-  char_u save_w_p_culopt_flags = curwin->w_p_culopt_flags;
+  char *save_w_p_culopt = NULL;
+  uint8_t save_w_p_culopt_flags = curwin->w_p_culopt_flags;
   int save_w_p_cuc = curwin->w_p_cuc;
   long save_w_p_so = curwin->w_p_so;
   long save_w_p_siso = curwin->w_p_siso;
   if (curwin->w_p_cul && curwin->w_p_culopt_flags & CULOPT_NBR) {
-    if (STRCMP(curwin->w_p_culopt, "number")) {
+    if (strcmp(curwin->w_p_culopt, "number") != 0) {
       save_w_p_culopt = curwin->w_p_culopt;
-      curwin->w_p_culopt = (char_u *)xstrdup("number");
+      curwin->w_p_culopt = xstrdup("number");
     }
     curwin->w_p_culopt_flags = CULOPT_NBR;
   } else {
@@ -437,15 +472,17 @@ void terminal_enter(void)
   s->state.check = terminal_check;
   state_enter(&s->state);
 
-  restart_edit = 0;
+  if (!s->got_bsl_o) {
+    restart_edit = 0;
+  }
   State = save_state;
   RedrawingDisabled = s->save_rd;
   apply_autocmds(EVENT_TERMLEAVE, NULL, NULL, false, curbuf);
 
-  if (save_curwin == curwin) {  // save_curwin may be invalid (window closed)!
+  if (save_curwin == curwin->handle) {  // Else: window was closed.
     curwin->w_p_cul = save_w_p_cul;
     if (save_w_p_culopt) {
-      xfree(curwin->w_p_culopt);
+      free_string_option(curwin->w_p_culopt);
       curwin->w_p_culopt = save_w_p_culopt;
     }
     curwin->w_p_culopt_flags = save_w_p_culopt_flags;
@@ -453,7 +490,7 @@ void terminal_enter(void)
     curwin->w_p_so = save_w_p_so;
     curwin->w_p_siso = save_w_p_siso;
   } else if (save_w_p_culopt) {
-    xfree(save_w_p_culopt);
+    free_string_option(save_w_p_culopt);
   }
 
   // draw the unfocused cursor
@@ -461,7 +498,11 @@ void terminal_enter(void)
   if (curbuf->terminal == s->term && !s->close) {
     terminal_check_cursor();
   }
-  unshowmode(true);
+  if (restart_edit) {
+    showmode();
+  } else {
+    unshowmode(true);
+  }
   ui_busy_stop();
   if (s->close) {
     bool wipe = s->term->buf_handle != 0;
@@ -471,6 +512,8 @@ void terminal_enter(void)
       do_cmdline_cmd("bwipeout!");
     }
   }
+
+  return s->got_bsl_o;
 }
 
 static void terminal_check_cursor(void)
@@ -496,9 +539,22 @@ static int terminal_check(VimState *state)
   }
 
   terminal_check_cursor();
+  validate_cursor();
 
   if (must_redraw) {
-    update_screen(0);
+    update_screen();
+
+    // Make sure an invoked autocmd doesn't delete the buffer (and the
+    // terminal) under our fingers.
+    curbuf->b_locked++;
+
+    // save and restore curwin and curbuf, in case the autocmd changes them
+    aco_save_T aco;
+    aucmd_prepbuf(&aco, curbuf);
+    apply_autocmds(EVENT_TEXTCHANGEDT, NULL, NULL, false, curbuf);
+    aucmd_restbuf(&aco);
+
+    curbuf->b_locked--;
   }
 
   if (need_maketitle) {  // Update title in terminal-mode. #7248
@@ -528,6 +584,8 @@ static int terminal_execute(VimState *state, int key)
   case K_RIGHTRELEASE:
   case K_MOUSEDOWN:
   case K_MOUSEUP:
+  case K_MOUSELEFT:
+  case K_MOUSERIGHT:
     if (send_mouse_event(s->term, key)) {
       return 0;
     }
@@ -549,7 +607,7 @@ static int terminal_execute(VimState *state, int key)
     break;
 
   case K_LUA:
-    map_execute_lua();
+    map_execute_lua(false);
     break;
 
   case Ctrl_N:
@@ -558,6 +616,14 @@ static int terminal_execute(VimState *state, int key)
     }
     FALLTHROUGH;
 
+  case Ctrl_O:
+    if (s->got_bsl) {
+      s->got_bsl_o = true;
+      restart_edit = 'I';
+      return 0;
+    }
+    FALLTHROUGH;
+
   default:
     if (key == Ctrl_BSL && !s->got_bsl) {
       s->got_bsl = true;
@@ -585,8 +651,12 @@ static int terminal_execute(VimState *state, int key)
   return 1;
 }
 
-void terminal_destroy(Terminal *term)
+/// Frees the given Terminal structure and sets the caller storage to NULL (in the spirit of
+/// XFREE_CLEAR).
+void terminal_destroy(Terminal **termpp)
+  FUNC_ATTR_NONNULL_ALL
 {
+  Terminal *term = *termpp;
   buf_T *buf = handle_get_buffer(term->buf_handle);
   if (buf) {
     term->buf_handle = 0;
@@ -605,8 +675,10 @@ void terminal_destroy(Terminal *term)
       xfree(term->sb_buffer[i]);
     }
     xfree(term->sb_buffer);
+    xfree(term->title);
     vterm_free(term->vt);
     xfree(term);
+    *termpp = NULL;  // coverity[dead-store]
   }
 }
 
@@ -650,28 +722,32 @@ static bool is_filter_char(int c)
   return !!(tpf_flags & flag);
 }
 
-void terminal_paste(long count, char_u **y_array, size_t y_size)
+void terminal_paste(long count, char **y_array, size_t y_size)
 {
   if (y_size == 0) {
     return;
   }
   vterm_keyboard_start_paste(curbuf->terminal->vt);
-  size_t buff_len = STRLEN(y_array[0]);
-  char_u *buff = xmalloc(buff_len);
+  size_t buff_len = strlen(y_array[0]);
+  char *buff = xmalloc(buff_len);
   for (int i = 0; i < count; i++) {  // -V756
     // feed the lines to the terminal
     for (size_t j = 0; j < y_size; j++) {
       if (j) {
         // terminate the previous line
+#ifdef MSWIN
+        terminal_send(curbuf->terminal, "\r\n", 2);
+#else
         terminal_send(curbuf->terminal, "\n", 1);
+#endif
       }
-      size_t len = STRLEN(y_array[j]);
+      size_t len = strlen(y_array[j]);
       if (len > buff_len) {
         buff = xrealloc(buff, len);
         buff_len = len;
       }
-      char_u *dst = buff;
-      char_u *src = y_array[j];
+      char *dst = buff;
+      char *src = y_array[j];
       while (*src != '\0') {
         len = (size_t)utf_ptr2len(src);
         int c = utf_ptr2char(src);
@@ -681,7 +757,7 @@ void terminal_paste(long count, char_u **y_array, size_t y_size)
         }
         src += len;
       }
-      terminal_send(curbuf->terminal, (char *)buff, (size_t)(dst - buff));
+      terminal_send(curbuf->terminal, buff, (size_t)(dst - buff));
     }
   }
   xfree(buff);
@@ -701,7 +777,7 @@ void terminal_send_key(Terminal *term, int c)
 
   if (key) {
     vterm_keyboard_key(term->vt, key, mod);
-  } else {
+  } else if (!IS_SPECIAL(c)) {
     vterm_keyboard_unichar(term->vt, (uint32_t)c, mod);
   }
 }
@@ -722,6 +798,21 @@ static int get_rgb(VTermState *state, VTermColor color)
   return RGB_(color.rgb.red, color.rgb.green, color.rgb.blue);
 }
 
+static int get_underline_hl_flag(VTermScreenCellAttrs attrs)
+{
+  switch (attrs.underline) {
+  case VTERM_UNDERLINE_OFF:
+    return 0;
+  case VTERM_UNDERLINE_SINGLE:
+    return HL_UNDERLINE;
+  case VTERM_UNDERLINE_DOUBLE:
+    return HL_UNDERDOUBLE;
+  case VTERM_UNDERLINE_CURLY:
+    return HL_UNDERCURL;
+  default:
+    return HL_UNDERLINE;
+  }
+}
 
 void terminal_get_line_attributes(Terminal *term, win_T *wp, int linenr, int *term_attrs)
 {
@@ -753,13 +844,13 @@ void terminal_get_line_attributes(Terminal *term, win_T *wp, int linenr, int *te
     int vt_fg_idx = ((!fg_default && fg_indexed) ? cell.fg.indexed.idx + 1 : 0);
     int vt_bg_idx = ((!bg_default && bg_indexed) ? cell.bg.indexed.idx + 1 : 0);
 
-    bool fg_set = vt_fg_idx && vt_fg_idx <= 16 && term->color_set[vt_fg_idx-1];
-    bool bg_set = vt_bg_idx && vt_bg_idx <= 16 && term->color_set[vt_bg_idx-1];
+    bool fg_set = vt_fg_idx && vt_fg_idx <= 16 && term->color_set[vt_fg_idx - 1];
+    bool bg_set = vt_bg_idx && vt_bg_idx <= 16 && term->color_set[vt_bg_idx - 1];
 
     int hl_attrs = (cell.attrs.bold ? HL_BOLD : 0)
                    | (cell.attrs.italic ? HL_ITALIC : 0)
                    | (cell.attrs.reverse ? HL_INVERSE : 0)
-                   | (cell.attrs.underline ? HL_UNDERLINE : 0)
+                   | get_underline_hl_flag(cell.attrs)
                    | (cell.attrs.strike ? HL_STRIKETHROUGH: 0)
                    | ((fg_indexed && !fg_set) ? HL_FG_INDEXED : 0)
                    | ((bg_indexed && !bg_set) ? HL_BG_INDEXED : 0);
@@ -817,23 +908,23 @@ static int term_moverect(VTermRect dest, VTermRect src, void *data)
   return 1;
 }
 
-static int term_movecursor(VTermPos new, VTermPos old, int visible, void *data)
+static int term_movecursor(VTermPos new_pos, VTermPos old_pos, int visible, void *data)
 {
   Terminal *term = data;
-  term->cursor.row = new.row;
-  term->cursor.col = new.col;
-  invalidate_terminal(term, old.row, old.row + 1);
-  invalidate_terminal(term, new.row, new.row + 1);
+  term->cursor.row = new_pos.row;
+  term->cursor.col = new_pos.col;
+  invalidate_terminal(term, old_pos.row, old_pos.row + 1);
+  invalidate_terminal(term, new_pos.row, new_pos.row + 1);
   return 1;
 }
 
-static void buf_set_term_title(buf_T *buf, char *title)
+static void buf_set_term_title(buf_T *buf, const char *title, size_t len)
   FUNC_ATTR_NONNULL_ALL
 {
   Error err = ERROR_INIT;
   dict_set_var(buf->b_vars,
                STATIC_CSTR_AS_STRING("term_title"),
-               STRING_OBJ(cstr_as_string(title)),
+               STRING_OBJ(((String){ .data = (char *)title, .size = len })),
                false,
                false,
                &err);
@@ -856,7 +947,30 @@ static int term_settermprop(VTermProp prop, VTermValue *val, void *data)
 
   case VTERM_PROP_TITLE: {
     buf_T *buf = handle_get_buffer(term->buf_handle);
-    buf_set_term_title(buf, val->string);
+    VTermStringFragment frag = val->string;
+
+    if (frag.initial && frag.final) {
+      buf_set_term_title(buf, frag.str, frag.len);
+      break;
+    }
+
+    if (frag.initial) {
+      term->title_len = 0;
+      term->title_size = MAX(frag.len, 1024);
+      term->title = xmalloc(sizeof(char *) * term->title_size);
+    } else if (term->title_len + frag.len > term->title_size) {
+      term->title_size *= 2;
+      term->title = xrealloc(term->title, sizeof(char *) * term->title_size);
+    }
+
+    memcpy(term->title + term->title_len, frag.str, frag.len);
+    term->title_len += frag.len;
+
+    if (frag.final) {
+      buf_set_term_title(buf, term->title, term->title_len);
+      xfree(term->title);
+      term->title = NULL;
+    }
     break;
   }
 
@@ -877,7 +991,10 @@ static int term_bell(void *data)
   return 1;
 }
 
-// Scrollback push handler (from pangoterm).
+/// Scrollback push handler: called just before a line goes offscreen (and libvterm will forget it),
+/// giving us a chance to store it.
+///
+/// Code adapted from pangoterm.
 static int term_sb_push(int cols, const VTermScreenCell *cells, void *data)
 {
   Terminal *term = data;
@@ -989,6 +1106,8 @@ static void convert_modifiers(int key, VTermModifier *statep)
   case K_S_DOWN:
   case K_S_LEFT:
   case K_S_RIGHT:
+  case K_S_HOME:
+  case K_S_END:
   case K_S_F1:
   case K_S_F2:
   case K_S_F3:
@@ -1006,6 +1125,8 @@ static void convert_modifiers(int key, VTermModifier *statep)
 
   case K_C_LEFT:
   case K_C_RIGHT:
+  case K_C_HOME:
+  case K_C_END:
     *statep |= VTERM_MOD_CTRL;
     break;
   }
@@ -1052,8 +1173,16 @@ static VTermKey convert_key(int key, VTermModifier *statep)
     return VTERM_KEY_INS;
   case K_DEL:
     return VTERM_KEY_DEL;
+  case K_S_HOME:
+    FALLTHROUGH;
+  case K_C_HOME:
+    FALLTHROUGH;
   case K_HOME:
     return VTERM_KEY_HOME;
+  case K_S_END:
+    FALLTHROUGH;
+  case K_C_END:
+    FALLTHROUGH;
   case K_END:
     return VTERM_KEY_END;
   case K_PAGEUP:
@@ -1287,13 +1416,14 @@ static void mouse_action(Terminal *term, int button, int row, int col, bool pres
 static bool send_mouse_event(Terminal *term, int c)
 {
   int row = mouse_row, col = mouse_col, grid = mouse_grid;
-  int offset;
   win_T *mouse_win = mouse_find_win(&grid, &row, &col);
-  if (mouse_win == NULL || (offset = win_col_off(mouse_win)) > col) {
+  if (mouse_win == NULL) {
     goto end;
   }
 
-  if (term->forward_mouse && mouse_win->w_buffer->terminal == term) {
+  int offset;
+  if (term->forward_mouse && mouse_win->w_buffer->terminal == term
+      && col >= (offset = win_col_off(mouse_win))) {
     // event in the terminal window and mouse events was enabled by the
     // program. translate and forward the event
     int button;
@@ -1321,11 +1451,17 @@ static bool send_mouse_event(Terminal *term, int c)
       pressed = true; button = 4; break;
     case K_MOUSEUP:
       pressed = true; button = 5; break;
+    case K_MOUSELEFT:
+      pressed = true; button = 7; break;
+    case K_MOUSERIGHT:
+      pressed = true; button = 6; break;
     default:
       return false;
     }
 
-    mouse_action(term, button, row, col - offset, pressed, 0);
+    VTermModifier mod = VTERM_MOD_NONE;
+    convert_modifiers(c, &mod);
+    mouse_action(term, button, row, col - offset, pressed, mod);
     return false;
   }
 
@@ -1337,34 +1473,34 @@ static bool send_mouse_event(Terminal *term, int c)
     int direction = c == K_MOUSEDOWN ? MSCR_DOWN : MSCR_UP;
     if (mod_mask & (MOD_MASK_SHIFT | MOD_MASK_CTRL)) {
       scroll_redraw(direction, curwin->w_botline - curwin->w_topline);
-    } else {
-      scroll_redraw(direction, 3L);
+    } else if (p_mousescroll_vert > 0) {
+      scroll_redraw(direction, p_mousescroll_vert);
     }
 
     curwin->w_redr_status = true;
     curwin = save_curwin;
     curbuf = curwin->w_buffer;
-    redraw_later(mouse_win, NOT_VALID);
+    redraw_later(mouse_win, UPD_NOT_VALID);
     invalidate_terminal(term, -1, -1);
     // Only need to exit focus if the scrolled window is the terminal window
     return mouse_win == curwin;
   }
 
-  // ignore left release action if it was not processed above
-  // to prevent leaving Terminal mode after entering to it using a mouse
-  if (c == K_LEFTRELEASE && mouse_win->w_buffer->terminal == term) {
+end:
+  // Ignore left release action if it was not forwarded to prevent
+  // leaving Terminal mode after entering to it using a mouse.
+  if ((c == K_LEFTRELEASE && mouse_win != NULL && mouse_win->w_buffer->terminal == term)
+      || c == K_MOUSEMOVE) {
     return false;
   }
 
-end:
-  ins_char_typebuf(c, mod_mask);
+  ins_char_typebuf(vgetc_char, vgetc_mod_mask);
   return true;
 }
 
 // }}}
 // terminal buffer refresh & misc {{{
 
-
 static void fetch_row(Terminal *term, int row, int end_col)
 {
   int col = 0;
@@ -1374,27 +1510,21 @@ static void fetch_row(Terminal *term, int row, int end_col)
   while (col < end_col) {
     VTermScreenCell cell;
     fetch_cell(term, row, col, &cell);
-    int cell_len = 0;
     if (cell.chars[0]) {
-      for (int i = 0; cell.chars[i]; i++) {
-        cell_len += utf_char2bytes((int)cell.chars[i],
-                                   (uint8_t *)ptr + cell_len);
+      int cell_len = 0;
+      for (int i = 0; i < VTERM_MAX_CHARS_PER_CELL && cell.chars[i]; i++) {
+        cell_len += utf_char2bytes((int)cell.chars[i], ptr + cell_len);
       }
-    } else {
-      *ptr = ' ';
-      cell_len = 1;
-    }
-    char c = *ptr;
-    ptr += cell_len;
-    if (c != ' ') {
-      // only increase the line length if the last character is not whitespace
+      ptr += cell_len;
       line_len = (size_t)(ptr - term->textbuf);
+    } else {
+      *ptr++ = ' ';
     }
     col += cell.width;
   }
 
-  // trim trailing whitespace
-  term->textbuf[line_len] = 0;
+  // end of line
+  term->textbuf[line_len] = NUL;
 }
 
 static bool fetch_cell(Terminal *term, int row, int col, VTermScreenCell *cell)
@@ -1490,8 +1620,16 @@ static void refresh_size(Terminal *term, buf_T *buf)
   term->opts.resize_cb((uint16_t)width, (uint16_t)height, term->opts.data);
 }
 
-/// Adjusts scrollback storage after 'scrollback' option changed.
-static void on_scrollback_option_changed(Terminal *term, buf_T *buf)
+void on_scrollback_option_changed(Terminal *term)
+{
+  // Scrollback buffer may not exist yet, e.g. if 'scrollback' is set in a TermOpen autocmd.
+  if (term->sb_buffer != NULL) {
+    refresh_terminal(term);
+  }
+}
+
+/// Adjusts scrollback storage and the terminal buffer scrollback lines
+static void adjust_scrollback(Terminal *term, buf_T *buf)
 {
   if (buf->b_p_scbk < 1) {  // Local 'scrollback' was set to -1.
     buf->b_p_scbk = SB_MAX;
@@ -1510,7 +1648,7 @@ static void on_scrollback_option_changed(Terminal *term, buf_T *buf)
       term->sb_current--;
       xfree(term->sb_buffer[term->sb_current]);
     }
-    deleted_lines(1, (long)diff);
+    deleted_lines(1, (linenr_T)diff);
   }
 
   // Resize the scrollback storage.
@@ -1533,7 +1671,7 @@ static void refresh_scrollback(Terminal *term, buf_T *buf)
   int row_offset = term->sb_pending;
   while (term->sb_pending > 0 && buf->b_ml.ml_line_count < height) {
     fetch_row(term, term->sb_pending - row_offset - 1, width);
-    ml_append(0, (uint8_t *)term->textbuf, 0, false);
+    ml_append(0, term->textbuf, 0, false);
     appended_lines(0, 1);
     term->sb_pending--;
   }
@@ -1551,7 +1689,7 @@ static void refresh_scrollback(Terminal *term, buf_T *buf)
     }
     fetch_row(term, -term->sb_pending - row_offset, width);
     int buf_index = (int)buf->b_ml.ml_line_count - height;
-    ml_append(buf_index, (uint8_t *)term->textbuf, 0, false);
+    ml_append(buf_index, term->textbuf, 0, false);
     appended_lines(buf_index, 1);
     term->sb_pending--;
   }
@@ -1563,7 +1701,7 @@ static void refresh_scrollback(Terminal *term, buf_T *buf)
     deleted_lines(buf->b_ml.ml_line_count, 1);
   }
 
-  on_scrollback_option_changed(term, buf);
+  adjust_scrollback(term, buf);
 }
 
 // Refresh the screen (visible part of the buffer when the terminal is
@@ -1591,10 +1729,10 @@ static void refresh_screen(Terminal *term, buf_T *buf)
     fetch_row(term, r, width);
 
     if (linenr <= buf->b_ml.ml_line_count) {
-      ml_replace(linenr, (uint8_t *)term->textbuf, true);
+      ml_replace(linenr, term->textbuf, true);
       changed++;
     } else {
-      ml_append(linenr - 1, (uint8_t *)term->textbuf, 0, false);
+      ml_append(linenr - 1, term->textbuf, 0, false);
       added++;
     }
   }
@@ -1638,7 +1776,7 @@ static int linenr_to_row(Terminal *term, int linenr)
 
 static bool is_focused(Terminal *term)
 {
-  return State & TERM_FOCUS && curbuf->terminal == term;
+  return State & MODE_TERMINAL && curbuf->terminal == term;
 }
 
 static char *get_config_string(char *key)
