@@ -7,10 +7,6 @@
 // dummy to pass an ACL to a function
 typedef void *vim_acl_T;
 
-// Shorthand for unsigned variables. Many systems, but not all, have u_char
-// already defined, so we use char_u to avoid trouble.
-typedef unsigned char char_u;
-
 // Can hold one decoded UTF-8 character.
 typedef uint32_t u8char_T;
 
@@ -22,16 +18,32 @@ typedef int handle_T;
 // absent callback etc.
 typedef int LuaRef;
 
+/// Type used for VimL VAR_FLOAT values
+typedef double float_T;
+
+typedef struct MsgpackRpcRequestHandler MsgpackRpcRequestHandler;
+
+typedef union {
+  float_T (*float_func)(float_T);
+  const MsgpackRpcRequestHandler *api_handler;
+  void *null;
+} EvalFuncData;
+
 typedef handle_T NS;
 
 typedef struct expand expand_T;
 
+typedef uint64_t proftime_T;
+
 typedef enum {
   kNone  = -1,
   kFalse = 0,
   kTrue  = 1,
 } TriState;
 
+#define TRISTATE_TO_BOOL(val, \
+                         default) ((val) == kTrue ? true : ((val) == kFalse ? false : (default)))
+
 typedef struct Decoration Decoration;
 
 #endif  // NVIM_TYPES_H
