@@ -2,8 +2,12 @@
 #define NVIM_UGRID_H
 
 #include "nvim/globals.h"
+#include "nvim/grid_defs.h"
 #include "nvim/ui.h"
 
+struct ucell;
+struct ugrid;
+
 typedef struct ucell UCell;
 typedef struct ugrid UGrid;
 
