@@ -7,24 +7,28 @@
 // Layer-based compositing: https://en.wikipedia.org/wiki/Digital_compositing
 
 #include <assert.h>
+#include <inttypes.h>
 #include <limits.h>
 #include <stdbool.h>
 #include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 
-#include "nvim/api/private/helpers.h"
+#include "klib/kvec.h"
+#include "nvim/api/private/defs.h"
 #include "nvim/ascii.h"
+#include "nvim/buffer_defs.h"
+#include "nvim/globals.h"
+#include "nvim/grid.h"
 #include "nvim/highlight.h"
 #include "nvim/highlight_group.h"
-#include "nvim/lib/kvec.h"
 #include "nvim/log.h"
-#include "nvim/lua/executor.h"
-#include "nvim/main.h"
+#include "nvim/macros.h"
 #include "nvim/memory.h"
 #include "nvim/message.h"
-#include "nvim/os/os.h"
-#include "nvim/popupmnu.h"
-#include "nvim/screen.h"
-#include "nvim/ugrid.h"
+#include "nvim/option_defs.h"
+#include "nvim/os/time.h"
+#include "nvim/types.h"
 #include "nvim/ui.h"
 #include "nvim/ui_compositor.h"
 #include "nvim/vim.h"
@@ -33,7 +37,6 @@
 # include "ui_compositor.c.generated.h"
 #endif
 
-static UI *compositor = NULL;
 static int composed_uis = 0;
 kvec_t(ScreenGrid *) layers = KV_INITIAL_VALUE;
 
@@ -58,33 +61,8 @@ static int dbghl_normal, dbghl_clear, dbghl_composed, dbghl_recompose;
 
 void ui_comp_init(void)
 {
-  if (compositor != NULL) {
-    return;
-  }
-  compositor = xcalloc(1, sizeof(UI));
-
-  compositor->rgb = true;
-  compositor->grid_resize = ui_comp_grid_resize;
-  compositor->grid_scroll = ui_comp_grid_scroll;
-  compositor->grid_cursor_goto = ui_comp_grid_cursor_goto;
-  compositor->raw_line = ui_comp_raw_line;
-  compositor->msg_set_pos = ui_comp_msg_set_pos;
-
-  // Be unopinionated: will be attached together with a "real" ui anyway
-  compositor->width = INT_MAX;
-  compositor->height = INT_MAX;
-  for (UIExtension i = 0; (int)i < kUIExtCount; i++) {
-    compositor->ui_ext[i] = true;
-  }
-
-  // TODO(bfredl): this will be more complicated if we implement
-  // hlstate per UI (i e reduce hl ids for non-hlstate UIs)
-  compositor->ui_ext[kUIHlState] = false;
-
   kv_push(layers, &default_grid);
   curgrid = &default_grid;
-
-  ui_attach_impl(compositor, 0);
 }
 
 void ui_comp_syn_init(void)
@@ -122,7 +100,7 @@ bool ui_comp_should_draw(void)
 ///
 /// TODO(bfredl): later on the compositor should just use win_float_pos events,
 /// though that will require slight event order adjustment: emit the win_pos
-/// events in the beginning of  update_screen(0), rather than in ui_flush()
+/// events in the beginning of update_screen(), rather than in ui_flush()
 bool ui_comp_put_grid(ScreenGrid *grid, int row, int col, int height, int width, bool valid,
                       bool on_top)
 {
@@ -138,19 +116,19 @@ bool ui_comp_put_grid(ScreenGrid *grid, int row, int col, int height, int width,
       // use it.
       grid->comp_disabled = true;
       compose_area(grid->comp_row, row,
-                   grid->comp_col, grid->comp_col + grid->Columns);
+                   grid->comp_col, grid->comp_col + grid->cols);
       if (grid->comp_col < col) {
         compose_area(MAX(row, grid->comp_row),
-                     MIN(row+height, grid->comp_row+grid->Rows),
+                     MIN(row + height, grid->comp_row + grid->rows),
                      grid->comp_col, col);
       }
-      if (col+width < grid->comp_col+grid->Columns) {
+      if (col + width < grid->comp_col + grid->cols) {
         compose_area(MAX(row, grid->comp_row),
-                     MIN(row+height, grid->comp_row+grid->Rows),
-                     col+width, grid->comp_col+grid->Columns);
+                     MIN(row + height, grid->comp_row + grid->rows),
+                     col + width, grid->comp_col + grid->cols);
       }
-      compose_area(row+height, grid->comp_row+grid->Rows,
-                   grid->comp_col, grid->comp_col + grid->Columns);
+      compose_area(row + height, grid->comp_row + grid->rows,
+                   grid->comp_col, grid->comp_col + grid->cols);
       grid->comp_disabled = false;
     }
     grid->comp_row = row;
@@ -166,19 +144,19 @@ bool ui_comp_put_grid(ScreenGrid *grid, int row, int col, int height, int width,
 #endif
 
     size_t insert_at = kv_size(layers);
-    while (insert_at > 0 && kv_A(layers, insert_at-1)->zindex > grid->zindex) {
+    while (insert_at > 0 && kv_A(layers, insert_at - 1)->zindex > grid->zindex) {
       insert_at--;
     }
 
-    if (curwin && kv_A(layers, insert_at-1) == &curwin->w_grid_alloc
-        && kv_A(layers, insert_at-1)->zindex == grid->zindex
+    if (curwin && kv_A(layers, insert_at - 1) == &curwin->w_grid_alloc
+        && kv_A(layers, insert_at - 1)->zindex == grid->zindex
         && !on_top) {
       insert_at--;
     }
     // not found: new grid
     kv_pushp(layers);
-    for (size_t i = kv_size(layers)-1; i > insert_at; i--) {
-      kv_A(layers, i) = kv_A(layers, i-1);
+    for (size_t i = kv_size(layers) - 1; i > insert_at; i--) {
+      kv_A(layers, i) = kv_A(layers, i - 1);
       kv_A(layers, i)->comp_index = i;
     }
     kv_A(layers, insert_at) = grid;
@@ -188,8 +166,8 @@ bool ui_comp_put_grid(ScreenGrid *grid, int row, int col, int height, int width,
     grid->comp_index = insert_at;
   }
   if (moved && valid && ui_comp_should_draw()) {
-    compose_area(grid->comp_row, grid->comp_row+grid->Rows,
-                 grid->comp_col, grid->comp_col+grid->Columns);
+    compose_area(grid->comp_row, grid->comp_row + grid->rows,
+                 grid->comp_col, grid->comp_col + grid->cols);
   }
   return moved;
 }
@@ -206,8 +184,8 @@ void ui_comp_remove_grid(ScreenGrid *grid)
     curgrid = &default_grid;
   }
 
-  for (size_t i = grid->comp_index; i < kv_size(layers)-1; i++) {
-    kv_A(layers, i) = kv_A(layers, i+1);
+  for (size_t i = grid->comp_index; i < kv_size(layers) - 1; i++) {
+    kv_A(layers, i) = kv_A(layers, i + 1);
     kv_A(layers, i)->comp_index = i;
   }
   (void)kv_pop(layers);
@@ -237,11 +215,11 @@ bool ui_comp_set_grid(handle_T handle)
   return false;
 }
 
-static void ui_comp_raise_grid(ScreenGrid *grid, size_t new_index)
+void ui_comp_raise_grid(ScreenGrid *grid, size_t new_index)
 {
   size_t old_index = grid->comp_index;
   for (size_t i = old_index; i < new_index; i++) {
-    kv_A(layers, i) = kv_A(layers, i+1);
+    kv_A(layers, i) = kv_A(layers, i + 1);
     kv_A(layers, i)->comp_index = i;
   }
   kv_A(layers, new_index) = grid;
@@ -249,26 +227,26 @@ static void ui_comp_raise_grid(ScreenGrid *grid, size_t new_index)
   for (size_t i = old_index; i < new_index; i++) {
     ScreenGrid *grid2 = kv_A(layers, i);
     int startcol = MAX(grid->comp_col, grid2->comp_col);
-    int endcol = MIN(grid->comp_col+grid->Columns,
-                     grid2->comp_col+grid2->Columns);
+    int endcol = MIN(grid->comp_col + grid->cols,
+                     grid2->comp_col + grid2->cols);
     compose_area(MAX(grid->comp_row, grid2->comp_row),
-                 MIN(grid->comp_row+grid->Rows, grid2->comp_row+grid2->Rows),
+                 MIN(grid->comp_row + grid->rows, grid2->comp_row + grid2->rows),
                  startcol, endcol);
   }
 }
 
-static void ui_comp_grid_cursor_goto(UI *ui, Integer grid_handle, Integer r, Integer c)
+void ui_comp_grid_cursor_goto(Integer grid_handle, Integer r, Integer c)
 {
   if (!ui_comp_should_draw() || !ui_comp_set_grid((int)grid_handle)) {
     return;
   }
-  int cursor_row = curgrid->comp_row+(int)r;
-  int cursor_col = curgrid->comp_col+(int)c;
+  int cursor_row = curgrid->comp_row + (int)r;
+  int cursor_col = curgrid->comp_col + (int)c;
 
   // TODO(bfredl): maybe not the best time to do this, for efficiency we
   // should configure all grids before entering win_update()
   if (curgrid != &default_grid) {
-    size_t new_index = kv_size(layers)-1;
+    size_t new_index = kv_size(layers) - 1;
 
     while (new_index > 1 && kv_A(layers, new_index)->zindex > curgrid->zindex) {
       new_index--;
@@ -279,7 +257,7 @@ static void ui_comp_grid_cursor_goto(UI *ui, Integer grid_handle, Integer r, Int
     }
   }
 
-  if (cursor_col >= default_grid.Columns || cursor_row >= default_grid.Rows) {
+  if (cursor_col >= default_grid.cols || cursor_row >= default_grid.rows) {
     // TODO(bfredl): this happens with 'writedelay', refactor?
     // abort();
     return;
@@ -289,17 +267,30 @@ static void ui_comp_grid_cursor_goto(UI *ui, Integer grid_handle, Integer r, Int
 
 ScreenGrid *ui_comp_mouse_focus(int row, int col)
 {
-  for (ssize_t i = (ssize_t)kv_size(layers)-1; i > 0; i--) {
+  for (ssize_t i = (ssize_t)kv_size(layers) - 1; i > 0; i--) {
     ScreenGrid *grid = kv_A(layers, i);
     if (grid->focusable
-        && row >= grid->comp_row && row < grid->comp_row+grid->Rows
-        && col >= grid->comp_col && col < grid->comp_col+grid->Columns) {
+        && row >= grid->comp_row && row < grid->comp_row + grid->rows
+        && col >= grid->comp_col && col < grid->comp_col + grid->cols) {
       return grid;
     }
   }
   return NULL;
 }
 
+/// Compute which grid is on top at supplied screen coordinates
+ScreenGrid *ui_comp_get_grid_at_coord(int row, int col)
+{
+  for (ssize_t i = (ssize_t)kv_size(layers) - 1; i > 0; i--) {
+    ScreenGrid *grid = kv_A(layers, i);
+    if (row >= grid->comp_row && row < grid->comp_row + grid->rows
+        && col >= grid->comp_col && col < grid->comp_col + grid->cols) {
+      return grid;
+    }
+  }
+  return &default_grid;
+}
+
 /// Baseline implementation. This is always correct, but we can sometimes
 /// do something more efficient (where efficiency means smaller deltas to
 /// the downstream UI.)
@@ -315,7 +306,7 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
     startcol--;
     skipstart = 1;
   }
-  if (endcol < default_grid.Columns && (flags & kLineFlagInvalid)) {
+  if (endcol < default_grid.cols && (flags & kLineFlagInvalid)) {
     endcol++;
     skipend = 1;
   }
@@ -323,9 +314,9 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
   int col = (int)startcol;
   ScreenGrid *grid = NULL;
   schar_T *bg_line = &default_grid.chars[default_grid.line_offset[row]
-                                         +(size_t)startcol];
+                                         + (size_t)startcol];
   sattr_T *bg_attrs = &default_grid.attrs[default_grid.line_offset[row]
-                                          +(size_t)startcol];
+                                          + (size_t)startcol];
 
   int grid_width, grid_height;
   while (col < endcol) {
@@ -337,8 +328,8 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
       // first check to see if any grids have pending updates to width/height,
       // to ensure that we don't accidentally put any characters into `linebuf`
       // that have been invalidated.
-      grid_width = MIN(g->Columns, g->comp_width);
-      grid_height = MIN(g->Rows, g->comp_height);
+      grid_width = MIN(g->cols, g->comp_width);
+      grid_height = MIN(g->rows, g->comp_height);
       if (g->comp_row > row || row >= g->comp_row + grid_height
           || g->comp_disabled) {
         continue;
@@ -354,8 +345,8 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
 
     assert(grid != NULL);
     assert(until > col);
-    assert(until <= default_grid.Columns);
-    size_t n = (size_t)(until-col);
+    assert(until <= default_grid.cols);
+    size_t n = (size_t)(until - col);
 
     if (row == msg_sep_row && grid->comp_index <= msg_grid.comp_index) {
       // TODO(bfredl): when we implement borders around floating windows, then
@@ -363,18 +354,18 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
       grid = &msg_grid;
       sattr_T msg_sep_attr = (sattr_T)HL_ATTR(HLF_MSGSEP);
       for (int i = col; i < until; i++) {
-        memcpy(linebuf[i-startcol], msg_sep_char, sizeof(*linebuf));
-        attrbuf[i-startcol] = msg_sep_attr;
+        memcpy(linebuf[i - startcol], msg_sep_char, sizeof(*linebuf));
+        attrbuf[i - startcol] = msg_sep_attr;
       }
     } else {
-      size_t off = grid->line_offset[row-grid->comp_row]
-                   + (size_t)(col-grid->comp_col);
-      memcpy(linebuf+(col-startcol), grid->chars+off, n * sizeof(*linebuf));
-      memcpy(attrbuf+(col-startcol), grid->attrs+off, n * sizeof(*attrbuf));
-      if (grid->comp_col+grid->Columns > until
-          && grid->chars[off+n][0] == NUL) {
-        linebuf[until-1-startcol][0] = ' ';
-        linebuf[until-1-startcol][1] = '\0';
+      size_t off = grid->line_offset[row - grid->comp_row]
+                   + (size_t)(col - grid->comp_col);
+      memcpy(linebuf + (col - startcol), grid->chars + off, n * sizeof(*linebuf));
+      memcpy(attrbuf + (col - startcol), grid->attrs + off, n * sizeof(*attrbuf));
+      if (grid->comp_col + grid->cols > until
+          && grid->chars[off + n][0] == NUL) {
+        linebuf[until - 1 - startcol][0] = ' ';
+        linebuf[until - 1 - startcol][1] = '\0';
         if (col == startcol && n == 1) {
           skipstart = 0;
         }
@@ -384,18 +375,18 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
     // 'pumblend' and 'winblend'
     if (grid->blending) {
       int width;
-      for (int i = col-(int)startcol; i < until-startcol; i += width) {
+      for (int i = col - (int)startcol; i < until - startcol; i += width) {
         width = 1;
         // negative space
         bool thru = strequal((char *)linebuf[i], " ") && bg_line[i][0] != NUL;
-        if (i+1 < endcol-startcol && bg_line[i+1][0] == NUL) {
+        if (i + 1 < endcol - startcol && bg_line[i + 1][0] == NUL) {
           width = 2;
-          thru &= strequal((char *)linebuf[i+1], " ");
+          thru &= strequal((char *)linebuf[i + 1], " ");
         }
         attrbuf[i] = (sattr_T)hl_blend_attrs(bg_attrs[i], attrbuf[i], &thru);
         if (width == 2) {
-          attrbuf[i+1] = (sattr_T)hl_blend_attrs(bg_attrs[i+1],
-                                                 attrbuf[i+1], &thru);
+          attrbuf[i + 1] = (sattr_T)hl_blend_attrs(bg_attrs[i + 1],
+                                                   attrbuf[i + 1], &thru);
         }
         if (thru) {
           memcpy(linebuf + i, bg_line + i, (size_t)width * sizeof(linebuf[i]));
@@ -405,19 +396,19 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
 
     // Tricky: if overlap caused a doublewidth char to get cut-off, must
     // replace the visible half with a space.
-    if (linebuf[col-startcol][0] == NUL) {
-      linebuf[col-startcol][0] = ' ';
-      linebuf[col-startcol][1] = NUL;
-      if (col == endcol-1) {
+    if (linebuf[col - startcol][0] == NUL) {
+      linebuf[col - startcol][0] = ' ';
+      linebuf[col - startcol][1] = NUL;
+      if (col == endcol - 1) {
         skipend = 0;
       }
-    } else if (n > 1 && linebuf[col-startcol+1][0] == NUL) {
+    } else if (col == startcol && n > 1 && linebuf[1][0] == NUL) {
       skipstart = 0;
     }
 
     col = until;
   }
-  if (linebuf[endcol-startcol-1][0] == NUL) {
+  if (linebuf[endcol - startcol - 1][0] == NUL) {
     skipend = 0;
   }
 
@@ -430,7 +421,7 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
     flags = flags & ~kLineFlagWrap;
   }
 
-  for (int i = skipstart; i < (endcol-skipend)-startcol; i++) {
+  for (int i = skipstart; i < (endcol - skipend) - startcol; i++) {
     if (attrbuf[i] < 0) {
       if (rdb_flags & RDB_INVALID) {
         abort();
@@ -439,32 +430,35 @@ static void compose_line(Integer row, Integer startcol, Integer endcol, LineFlag
       }
     }
   }
-  ui_composed_call_raw_line(1, row, startcol+skipstart,
-                            endcol-skipend, endcol-skipend, 0, flags,
-                            (const schar_T *)linebuf+skipstart,
-                            (const sattr_T *)attrbuf+skipstart);
+  ui_composed_call_raw_line(1, row, startcol + skipstart,
+                            endcol - skipend, endcol - skipend, 0, flags,
+                            (const schar_T *)linebuf + skipstart,
+                            (const sattr_T *)attrbuf + skipstart);
 }
 
 static void compose_debug(Integer startrow, Integer endrow, Integer startcol, Integer endcol,
                           int syn_id, bool delay)
 {
-  if (!(rdb_flags & RDB_COMPOSITOR)) {
+  if (!(rdb_flags & RDB_COMPOSITOR) || startcol >= endcol) {
     return;
   }
 
-  endrow = MIN(endrow, default_grid.Rows);
-  endcol = MIN(endcol, default_grid.Columns);
+  endrow = MIN(endrow, default_grid.rows);
+  endcol = MIN(endcol, default_grid.cols);
   int attr = syn_id2attr(syn_id);
 
+  if (delay) {
+    debug_delay(endrow - startrow);
+  }
+
   for (int row = (int)startrow; row < endrow; row++) {
     ui_composed_call_raw_line(1, row, startcol, startcol, endcol, attr, false,
                               (const schar_T *)linebuf,
                               (const sattr_T *)attrbuf);
   }
 
-
   if (delay) {
-    debug_delay(endrow-startrow);
+    debug_delay(endrow - startrow);
   }
 }
 
@@ -473,15 +467,14 @@ static void debug_delay(Integer lines)
   ui_call_flush();
   uint64_t wd = (uint64_t)labs(p_wd);
   uint64_t factor = (uint64_t)MAX(MIN(lines, 5), 1);
-  os_microdelay(factor * wd * 1000u, true);
+  os_microdelay(factor * wd * 1000U, true);
 }
 
-
 static void compose_area(Integer startrow, Integer endrow, Integer startcol, Integer endcol)
 {
   compose_debug(startrow, endrow, startcol, endcol, dbghl_recompose, true);
-  endrow = MIN(endrow, default_grid.Rows);
-  endcol = MIN(endcol, default_grid.Columns);
+  endrow = MIN(endrow, default_grid.rows);
+  endcol = MIN(endcol, default_grid.cols);
   if (endcol <= startcol) {
     return;
   }
@@ -497,14 +490,14 @@ static void compose_area(Integer startrow, Integer endrow, Integer startcol, Int
 void ui_comp_compose_grid(ScreenGrid *grid)
 {
   if (ui_comp_should_draw()) {
-    compose_area(grid->comp_row, grid->comp_row+grid->Rows,
-                 grid->comp_col, grid->comp_col+grid->Columns);
+    compose_area(grid->comp_row, grid->comp_row + grid->rows,
+                 grid->comp_col, grid->comp_col + grid->cols);
   }
 }
 
-static void ui_comp_raw_line(UI *ui, Integer grid, Integer row, Integer startcol, Integer endcol,
-                             Integer clearcol, Integer clearattr, LineFlags flags,
-                             const schar_T *chunk, const sattr_T *attrs)
+void ui_comp_raw_line(Integer grid, Integer row, Integer startcol, Integer endcol, Integer clearcol,
+                      Integer clearattr, LineFlags flags, const schar_T *chunk,
+                      const sattr_T *attrs)
 {
   if (!ui_comp_should_draw() || !ui_comp_set_grid((int)grid)) {
     return;
@@ -523,17 +516,17 @@ static void ui_comp_raw_line(UI *ui, Integer grid, Integer row, Integer startcol
   // TODO(bfredl): this should not really be necessary. But on some condition
   // when resizing nvim, a window will be attempted to be drawn on the older
   // and possibly larger global screen size.
-  if (row >= default_grid.Rows) {
+  if (row >= default_grid.rows) {
     DLOG("compositor: invalid row %" PRId64 " on grid %" PRId64, row, grid);
     return;
   }
-  if (clearcol > default_grid.Columns) {
+  if (clearcol > default_grid.cols) {
     DLOG("compositor: invalid last column %" PRId64 " on grid %" PRId64,
          clearcol, grid);
-    if (startcol >= default_grid.Columns) {
+    if (startcol >= default_grid.cols) {
       return;
     }
-    clearcol = default_grid.Columns;
+    clearcol = default_grid.cols;
     endcol = MIN(endcol, clearcol);
   }
 
@@ -541,13 +534,13 @@ static void ui_comp_raw_line(UI *ui, Integer grid, Integer row, Integer startcol
   // TODO(bfredl): eventually should just fix compose_line to respect clearing
   // and optimize it for uncovered lines.
   if (flags & kLineFlagInvalid || covered || curgrid->blending) {
-    compose_debug(row, row+1, startcol, clearcol, dbghl_composed, true);
+    compose_debug(row, row + 1, startcol, clearcol, dbghl_composed, true);
     compose_line(row, startcol, clearcol, flags);
   } else {
-    compose_debug(row, row+1, startcol, endcol, dbghl_normal, false);
-    compose_debug(row, row+1, endcol, clearcol, dbghl_clear, true);
+    compose_debug(row, row + 1, startcol, endcol, dbghl_normal, false);
+    compose_debug(row, row + 1, endcol, clearcol, dbghl_clear, true);
 #ifndef NDEBUG
-    for (int i = 0; i < endcol-startcol; i++) {
+    for (int i = 0; i < endcol - startcol; i++) {
       assert(attrs[i] >= 0);
     }
 #endif
@@ -559,41 +552,42 @@ static void ui_comp_raw_line(UI *ui, Integer grid, Integer row, Integer startcol
 /// The screen is invalid and will soon be cleared
 ///
 /// Don't redraw floats until screen is cleared
-void ui_comp_set_screen_valid(bool valid)
+bool ui_comp_set_screen_valid(bool valid)
 {
+  bool old_val = valid_screen;
   valid_screen = valid;
   if (!valid) {
     msg_sep_row = -1;
   }
+  return old_val;
 }
 
-static void ui_comp_msg_set_pos(UI *ui, Integer grid, Integer row, Boolean scrolled,
-                                String sep_char)
+void ui_comp_msg_set_pos(Integer grid, Integer row, Boolean scrolled, String sep_char)
 {
   msg_grid.comp_row = (int)row;
   if (scrolled && row > 0) {
-    msg_sep_row = (int)row-1;
+    msg_sep_row = (int)row - 1;
     if (sep_char.data) {
-      STRLCPY(msg_sep_char, sep_char.data, sizeof(msg_sep_char));
+      xstrlcpy(msg_sep_char, sep_char.data, sizeof(msg_sep_char));
     }
   } else {
     msg_sep_row = -1;
   }
 
   if (row > msg_current_row && ui_comp_should_draw()) {
-    compose_area(MAX(msg_current_row-1, 0), row, 0, default_grid.Columns);
+    compose_area(MAX(msg_current_row - 1, 0), row, 0, default_grid.cols);
   } else if (row < msg_current_row && ui_comp_should_draw()
-             && msg_current_row < Rows) {
+             && (msg_current_row < Rows || (scrolled && !msg_was_scrolled))) {
     int delta = msg_current_row - (int)row;
     if (msg_grid.blending) {
-      int first_row = MAX((int)row-(scrolled?1:0), 0);
-      compose_area(first_row, Rows-delta, 0, Columns);
+      int first_row = MAX((int)row - (scrolled?1:0), 0);
+      compose_area(first_row, Rows - delta, 0, Columns);
     } else {
       // scroll separator together with message text
-      int first_row = MAX((int)row-(msg_was_scrolled?1:0), 0);
+      int first_row = MAX((int)row - (msg_was_scrolled?1:0), 0);
       ui_composed_call_grid_scroll(1, first_row, Rows, 0, Columns, delta, 0);
       if (scrolled && !msg_was_scrolled && row > 0) {
-        compose_area(row-1, row, 0, Columns);
+        compose_area(row - 1, row, 0, Columns);
       }
     }
   }
@@ -607,13 +601,13 @@ static void ui_comp_msg_set_pos(UI *ui, Integer grid, Integer row, Boolean scrol
 /// TODO(bfredl): currently this only handles message row
 static bool curgrid_covered_above(int row)
 {
-  bool above_msg = (kv_A(layers, kv_size(layers)-1) == &msg_grid
-                    && row < msg_current_row-(msg_was_scrolled?1:0));
-  return kv_size(layers)-(above_msg?1:0) > curgrid->comp_index+1;
+  bool above_msg = (kv_A(layers, kv_size(layers) - 1) == &msg_grid
+                    && row < msg_current_row - (msg_was_scrolled?1:0));
+  return kv_size(layers) - (above_msg?1:0) > curgrid->comp_index + 1;
 }
 
-static void ui_comp_grid_scroll(UI *ui, Integer grid, Integer top, Integer bot, Integer left,
-                                Integer right, Integer rows, Integer cols)
+void ui_comp_grid_scroll(Integer grid, Integer top, Integer bot, Integer left, Integer right,
+                         Integer rows, Integer cols)
 {
   if (!ui_comp_should_draw() || !ui_comp_set_grid((int)grid)) {
     return;
@@ -634,8 +628,8 @@ static void ui_comp_grid_scroll(UI *ui, Integer grid, Integer top, Integer bot,
       // row, where the latter might scroll invalid space created by the first.
       // ideally win_update() should keep track of this itself and not scroll
       // the invalid space.
-      if (curgrid->attrs[curgrid->line_offset[r-curgrid->comp_row]
-                         +left-curgrid->comp_col] >= 0) {
+      if (curgrid->attrs[curgrid->line_offset[r - curgrid->comp_row]
+                         + (size_t)left - (size_t)curgrid->comp_col] >= 0) {
         compose_line(r, left, right, 0);
       }
     }
@@ -647,7 +641,7 @@ static void ui_comp_grid_scroll(UI *ui, Integer grid, Integer top, Integer bot,
   }
 }
 
-static void ui_comp_grid_resize(UI *ui, Integer grid, Integer width, Integer height)
+void ui_comp_grid_resize(Integer grid, Integer width, Integer height)
 {
   if (grid == 1) {
     ui_composed_call_grid_resize(1, width, height);
@@ -665,4 +659,3 @@ static void ui_comp_grid_resize(UI *ui, Integer grid, Integer width, Integer hei
     }
   }
 }
-
