@@ -4,10 +4,9 @@
 #include "nvim/viml/parser/parser.h"
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
-# include "viml/parser/parser.c.generated.h"
+# include "viml/parser/parser.c.generated.h"  // IWYU pragma: export
 #endif
 
-
 void parser_simple_get_line(void *cookie, ParserLine *ret_pline)
 {
   ParserLine **plines_p = (ParserLine **)cookie;
