@@ -1,4 +1,5 @@
 local pretty = require 'pl.pretty'
+local global_helpers = require('test.helpers')
 
 -- Colors are disabled by default. #15610
 local colors = setmetatable({}, {__index = function() return function(s) return s == nil and '' or tostring(s) end end})
@@ -25,35 +26,35 @@ return function(options)
 
   local repeatSuiteString = '\nRepeating all tests (run %d of %d) . . .\n\n'
   local randomizeString  = c.note('Note: Randomizing test order with a seed of %d.\n')
-  local globalSetup      = c.sect('[----------]') .. ' Global test environment setup.\n'
-  local fileStartString  = c.sect('[----------]') .. ' Running tests from ' .. c.file('%s') .. '\n'
-  local runString        = c.sect('[ RUN      ]') .. ' ' .. c.test('%s') .. ': '
+  local globalSetup      = c.sect('--------') .. ' Global test environment setup.\n'
+  local fileStartString  = c.sect('--------') .. ' Running tests from ' .. c.file('%s') .. '\n'
+  local runString        = c.sect('RUN     ') .. ' ' .. c.test('%s') .. ': '
   local successString    = c.succ('OK')   .. '\n'
   local skippedString    = c.skip('SKIP') .. '\n'
   local failureString    = c.fail('FAIL') .. '\n'
   local errorString      = c.errr('ERR')  .. '\n'
-  local fileEndString    = c.sect('[----------]') .. ' '.. c.nmbr('%d') .. ' %s from ' .. c.file('%s') .. ' ' .. c.time('(%.2f ms total)') .. '\n\n'
-  local globalTeardown   = c.sect('[----------]') .. ' Global test environment teardown.\n'
-  local suiteEndString   = c.sect('[==========]') .. ' ' .. c.nmbr('%d') .. ' %s from ' .. c.nmbr('%d') .. ' test %s ran. ' .. c.time('(%.2f ms total)') .. '\n'
-  local successStatus    = c.succ('[  PASSED  ]') .. ' ' .. c.nmbr('%d') .. ' %s.\n'
+  local fileEndString    = c.sect('--------') .. ' '.. c.nmbr('%d') .. ' %s from ' .. c.file('%s') .. ' ' .. c.time('(%.2f ms total)') .. '\n\n'
+  local globalTeardown   = c.sect('--------') .. ' Global test environment teardown.\n'
+  local suiteEndString   = c.sect('========') .. ' ' .. c.nmbr('%d') .. ' %s from ' .. c.nmbr('%d') .. ' test %s ran. ' .. c.time('(%.2f ms total)') .. '\n'
+  local successStatus    = c.succ('PASSED  ') .. ' ' .. c.nmbr('%d') .. ' %s.\n'
   local timeString       = c.time('%.2f ms')
 
   local summaryStrings = {
     skipped = {
-      header = c.skip('[ SKIPPED  ]') .. ' ' .. c.nmbr('%d') .. ' %s, listed below:\n',
-      test   = c.skip('[ SKIPPED  ]') .. ' %s\n',
+      header = c.skip('SKIPPED ') .. ' ' .. c.nmbr('%d') .. ' %s, listed below:\n',
+      test   = c.skip('SKIPPED ') .. ' %s\n',
       footer = ' ' .. c.nmbr('%d') .. ' SKIPPED %s\n',
     },
 
     failure = {
-      header = c.fail('[  FAILED  ]') .. ' ' .. c.nmbr('%d') .. ' %s, listed below:\n',
-      test   = c.fail('[  FAILED  ]') .. ' %s\n',
+      header = c.fail('FAILED  ') .. ' ' .. c.nmbr('%d') .. ' %s, listed below:\n',
+      test   = c.fail('FAILED  ') .. ' %s\n',
       footer = ' ' .. c.nmbr('%d') .. ' FAILED %s\n',
     },
 
     error = {
-      header = c.errr('[  ERROR   ]') .. ' ' .. c.nmbr('%d') .. ' %s, listed below:\n',
-      test   = c.errr('[  ERROR   ]') .. ' %s\n',
+      header = c.errr('ERROR   ') .. ' ' .. c.nmbr('%d') .. ' %s, listed below:\n',
+      test   = c.errr('ERROR   ') .. ' %s\n',
       footer = ' ' .. c.nmbr('%d') .. ' %s\n',
     },
   }
@@ -193,6 +194,9 @@ return function(options)
     io.write(globalTeardown)
     io.write(suiteEndString:format(testCount, tests, fileCount, files, elapsedTime_ms))
     io.write(getSummaryString())
+    if failureCount > 0 or errorCount > 0 then
+      io.write(global_helpers.read_nvim_log(nil, true))
+    end
     io.flush()
 
     return nil, true
@@ -200,7 +204,7 @@ return function(options)
 
   handler.fileStart = function(file)
     fileTestCount = 0
-    io.write(fileStartString:format(file.name))
+    io.write(fileStartString:format(vim.fs.normalize(file.name)))
     io.flush()
     return nil, true
   end
@@ -209,13 +213,15 @@ return function(options)
     local elapsedTime_ms = getElapsedTime(file)
     local tests = (fileTestCount == 1 and 'test' or 'tests')
     fileCount = fileCount + 1
-    io.write(fileEndString:format(fileTestCount, tests, file.name, elapsedTime_ms))
+    io.write(fileEndString:format(fileTestCount, tests, vim.fs.normalize(file.name), elapsedTime_ms))
     io.flush()
     return nil, true
   end
 
   handler.testStart = function(element, _parent)
-    io.write(runString:format(handler.getFullName(element)))
+    local testid = _G._nvim_test_id or ''
+    local desc = ('%s %s'):format(testid, handler.getFullName(element))
+    io.write(runString:format(desc))
     io.flush()
 
     return nil, true
