@@ -14,14 +14,40 @@ before_each(clear)
 
 describe('autocmd api', function()
   describe('nvim_create_autocmd', function()
-    it('does not allow "command" and "callback" in the same autocmd', function()
-      local ok, _ = pcall(meths.create_autocmd, "BufReadPost", {
-        pattern = "*.py,*.pyi",
+    it('validation', function()
+      eq("Cannot use both 'callback' and 'command'", pcall_err(meths.create_autocmd, 'BufReadPost', {
+        pattern = '*.py,*.pyi',
         command = "echo 'Should Have Errored",
-        callback = "not allowed",
-      })
+        callback = 'NotAllowed',
+      }))
+      eq("Cannot use both 'pattern' and 'buffer' for the same autocmd", pcall_err(meths.create_autocmd, 'FileType', {
+        command = 'let g:called = g:called + 1',
+        buffer = 0,
+        pattern = '*.py',
+      }))
+      eq("Required: 'event'", pcall_err(meths.create_autocmd, {}, {
+        command = 'ls',
+      }))
+      eq("Required: 'command' or 'callback'", pcall_err(meths.create_autocmd, 'FileType', {
+      }))
+      eq("Invalid 'desc': expected String, got Integer", pcall_err(meths.create_autocmd, 'FileType', {
+        command = 'ls',
+        desc = 42,
+      }))
+      eq("Invalid 'callback': expected Lua function or Vim function name, got Integer", pcall_err(meths.create_autocmd, 'FileType', {
+        callback = 0,
+      }))
+      eq("Invalid 'event' item: expected String, got Array", pcall_err(meths.create_autocmd,
+        {'FileType', {}}, {}))
+      eq("Invalid 'group': 0", pcall_err(meths.create_autocmd, 'FileType', {
+        group = 0,
+        command = 'ls',
+      }))
 
-      eq(false, ok)
+      eq("Invalid 'event': 'foo'", pcall_err(meths.create_autocmd, 'foo', { command = '' }))
+      eq("Invalid 'event': 'VimEnter '", pcall_err(meths.create_autocmd, 'VimEnter ', { command = '' }))
+      eq("Invalid 'event': 'VimEnter foo'", pcall_err(meths.create_autocmd, 'VimEnter foo', { command = '' }))
+      eq("Invalid 'event': 'BufAdd,BufDelete'", pcall_err(meths.create_autocmd, 'BufAdd,BufDelete', { command = '' }))
     end)
 
     it('doesnt leak when you use ++once', function()
@@ -59,18 +85,8 @@ describe('autocmd api', function()
       eq(1, meths.get_var('called'))
     end)
 
-    it('does not allow passing buffer and patterns', function()
-      local ok = pcall(meths.create_autocmd, "Filetype", {
-        command = "let g:called = g:called + 1",
-        buffer = 0,
-        pattern = "*.py",
-      })
-
-      eq(false, ok)
-    end)
-
     it('does not allow passing invalid buffers', function()
-      local ok, msg = pcall(meths.create_autocmd, "Filetype", {
+      local ok, msg = pcall(meths.create_autocmd, 'FileType', {
         command = "let g:called = g:called + 1",
         buffer = -1,
       })
@@ -135,17 +151,22 @@ describe('autocmd api', function()
         local desc = 'Can show description'
         meths.set_var('desc', desc)
 
-        exec_lua([[
+        local result = exec_lua([[
           local callback = function() print 'Should Not Have Errored' end
           vim.api.nvim_create_autocmd("BufReadPost", {
             pattern = "*.py",
             callback = callback,
             desc = vim.g.desc,
           })
+          local aus = vim.api.nvim_get_autocmds({ event = 'BufReadPost' })
+          local first = aus[1]
+          return {
+            desc = first.desc,
+            cbtype = type(first.callback)
+          }
         ]])
 
-        eq(desc, meths.get_autocmds({ event = 'BufReadPost' })[1].desc)
-        matches('<lua: %d+>', meths.get_autocmds({ event = 'BufReadPost' })[1].command)
+        eq({ desc = desc, cbtype = 'function' }, result)
       end)
 
       it('will not add a description unless it was provided', function()
@@ -204,7 +225,7 @@ describe('autocmd api', function()
 
       local aus = meths.get_autocmds({ event = 'User', pattern = 'Test' })
       local first = aus[1]
-      eq(first.id, 1)
+      eq(true, first.id > 0)
 
       meths.set_var("some_condition", true)
       meths.exec_autocmds("User", {pattern = "Test"})
@@ -212,23 +233,28 @@ describe('autocmd api', function()
     end)
 
     it('receives an args table', function()
-      local res = exec_lua [[
-        local group_id = vim.api.nvim_create_augroup("TestGroup", {})
-        local autocmd_id = vim.api.nvim_create_autocmd("User", {
+      local group_id = meths.create_augroup("TestGroup", {})
+      -- Having an existing autocmd calling expand("<afile>") shouldn't change args #18964
+      meths.create_autocmd('User', {
+        group = 'TestGroup',
+        pattern = 'Te*',
+        command = 'call expand("<afile>")',
+      })
+
+      local autocmd_id = exec_lua [[
+        return vim.api.nvim_create_autocmd("User", {
           group = "TestGroup",
           pattern = "Te*",
           callback = function(args)
             vim.g.autocmd_args = args
           end,
         })
-
-        return {group_id, autocmd_id}
       ]]
 
       meths.exec_autocmds("User", {pattern = "Test pattern"})
       eq({
-        id = res[2],
-        group = res[1],
+        id = autocmd_id,
+        group = group_id,
         event = "User",
         match = "Test pattern",
         file = "Test pattern",
@@ -236,33 +262,82 @@ describe('autocmd api', function()
       }, meths.get_var("autocmd_args"))
 
       -- Test without a group
-      res = exec_lua [[
-        local autocmd_id = vim.api.nvim_create_autocmd("User", {
+      autocmd_id = exec_lua [[
+        return vim.api.nvim_create_autocmd("User", {
           pattern = "*",
           callback = function(args)
             vim.g.autocmd_args = args
           end,
         })
-
-        return {autocmd_id}
       ]]
 
       meths.exec_autocmds("User", {pattern = "some_pat"})
       eq({
-        id = res[1],
+        id = autocmd_id,
         group = nil,
         event = "User",
         match = "some_pat",
         file = "some_pat",
         buf = 1,
       }, meths.get_var("autocmd_args"))
+    end)
 
+    it('can receive arbitrary data', function()
+      local function test(data)
+        eq(data, exec_lua([[
+          local input = ...
+          local output
+          vim.api.nvim_create_autocmd("User", {
+            pattern = "Test",
+            callback = function(args)
+              output = args.data
+            end,
+          })
+
+          vim.api.nvim_exec_autocmds("User", {
+            pattern = "Test",
+            data = input,
+          })
+
+          return output
+        ]], data))
+      end
+
+      test("Hello")
+      test(42)
+      test(true)
+      test({ "list" })
+      test({ foo = "bar" })
     end)
   end)
 
   describe('nvim_get_autocmds', function()
+    it('validation', function()
+      eq("Invalid 'group': 9997999", pcall_err(meths.get_autocmds, {
+        group = 9997999,
+      }))
+      eq("Invalid 'group': 'bogus'", pcall_err(meths.get_autocmds, {
+        group = 'bogus',
+      }))
+      eq("Invalid 'group': 0", pcall_err(meths.get_autocmds, {
+        group = 0,
+      }))
+      eq("Invalid 'group': expected String or Integer, got Array", pcall_err(meths.get_autocmds, {
+        group = {},
+      }))
+      eq("Invalid 'buffer': expected Integer or Array, got Boolean", pcall_err(meths.get_autocmds, {
+        buffer = true,
+      }))
+      eq("Invalid 'event': expected String or Array", pcall_err(meths.get_autocmds, {
+        event = true,
+      }))
+      eq("Invalid 'pattern': expected String or Array, got Boolean", pcall_err(meths.get_autocmds, {
+        pattern = true,
+      }))
+    end)
+
     describe('events', function()
-      it('should return one autocmd when there is only one for an event', function()
+      it('returns one autocmd when there is only one for an event', function()
         command [[au! InsertEnter]]
         command [[au InsertEnter * :echo "1"]]
 
@@ -270,7 +345,7 @@ describe('autocmd api', function()
         eq(1, #aus)
       end)
 
-      it('should return two autocmds when there are two for an event', function()
+      it('returns two autocmds when there are two for an event', function()
         command [[au! InsertEnter]]
         command [[au InsertEnter * :echo "1"]]
         command [[au InsertEnter * :echo "2"]]
@@ -279,7 +354,7 @@ describe('autocmd api', function()
         eq(2, #aus)
       end)
 
-      it('should return the same thing if you use string or list', function()
+      it('returns the same thing if you use string or list', function()
         command [[au! InsertEnter]]
         command [[au InsertEnter * :echo "1"]]
         command [[au InsertEnter * :echo "2"]]
@@ -289,7 +364,7 @@ describe('autocmd api', function()
         eq(string_aus, array_aus)
       end)
 
-      it('should return two autocmds when there are two for an event', function()
+      it('returns two autocmds when there are two for an event', function()
         command [[au! InsertEnter]]
         command [[au! InsertLeave]]
         command [[au InsertEnter * :echo "1"]]
@@ -299,7 +374,7 @@ describe('autocmd api', function()
         eq(2, #aus)
       end)
 
-      it('should return different IDs for different autocmds', function()
+      it('returns different IDs for different autocmds', function()
         command [[au! InsertEnter]]
         command [[au! InsertLeave]]
         command [[au InsertEnter * :echo "1"]]
@@ -323,7 +398,7 @@ describe('autocmd api', function()
         eq(first, new_aus[1])
       end)
 
-      it('should return event name', function()
+      it('returns event name', function()
         command [[au! InsertEnter]]
         command [[au InsertEnter * :echo "1"]]
 
@@ -331,7 +406,7 @@ describe('autocmd api', function()
         eq({ { buflocal = false, command = ':echo "1"', event = "InsertEnter", once = false, pattern = "*" } }, aus)
       end)
 
-      it('should work with buffer numbers', function()
+      it('works with buffer numbers', function()
         command [[new]]
         command [[au! InsertEnter]]
         command [[au InsertEnter <buffer=1> :echo "1"]]
@@ -374,8 +449,8 @@ describe('autocmd api', function()
           pattern = "<buffer=2>",
         }}, aus)
 
-        eq("Invalid value for 'buffer': must be an integer or array of integers", pcall_err(meths.get_autocmds, { event = "InsertEnter", buffer = "foo" }))
-        eq("Invalid value for 'buffer': must be an integer", pcall_err(meths.get_autocmds, { event = "InsertEnter", buffer = { "foo", 42 } }))
+        eq("Invalid 'buffer': expected Integer or Array, got String", pcall_err(meths.get_autocmds, { event = "InsertEnter", buffer = "foo" }))
+        eq("Invalid 'buffer': expected Integer, got String", pcall_err(meths.get_autocmds, { event = "InsertEnter", buffer = { "foo", 42 } }))
         eq("Invalid buffer id: 42", pcall_err(meths.get_autocmds, { event = "InsertEnter", buffer = { 42 } }))
 
         local bufs = {}
@@ -383,10 +458,10 @@ describe('autocmd api', function()
           table.insert(bufs, meths.create_buf(true, false))
         end
 
-        eq("Too many buffers. Please limit yourself to 256 or fewer", pcall_err(meths.get_autocmds, { event = "InsertEnter", buffer = bufs }))
+        eq("Too many buffers (maximum of 256)", pcall_err(meths.get_autocmds, { event = "InsertEnter", buffer = bufs }))
       end)
 
-      it('should return autocmds when group is specified by id', function()
+      it('returns autocmds when group is specified by id', function()
         local auid = meths.create_augroup("nvim_test_augroup", { clear = true })
         meths.create_autocmd("FileType", { group = auid, command = 'echo "1"' })
         meths.create_autocmd("FileType", { group = auid, command = 'echo "2"' })
@@ -398,7 +473,7 @@ describe('autocmd api', function()
         eq(0, #aus2)
       end)
 
-      it('should return autocmds when group is specified by name', function()
+      it('returns autocmds when group is specified by name', function()
         local auname = "nvim_test_augroup"
         meths.create_augroup(auname, { clear = true })
         meths.create_autocmd("FileType", { group = auname, command = 'echo "1"' })
@@ -437,6 +512,49 @@ describe('autocmd api', function()
         -- 3-7 for the 5 we make in the autocmd
         eq({1, 2, 3, 4, 5, 6, 7}, bufs)
       end)
+
+      it('can retrieve a callback from an autocmd', function()
+        local content = 'I Am A Callback'
+        meths.set_var('content', content)
+
+        local result = exec_lua([[
+          local cb = function() return vim.g.content end
+          vim.api.nvim_create_autocmd("User", {
+            pattern = "TestTrigger",
+            desc = "A test autocommand with a callback",
+            callback = cb,
+          })
+          local aus = vim.api.nvim_get_autocmds({ event = 'User', pattern = 'TestTrigger'})
+          local first = aus[1]
+          return {
+            cb = {
+              type = type(first.callback),
+              can_retrieve = first.callback() == vim.g.content
+            }
+          }
+        ]])
+
+        eq("function", result.cb.type)
+        eq(true, result.cb.can_retrieve)
+      end)
+
+      it('will return an empty string as the command for an autocmd that uses a callback', function()
+        local result = exec_lua([[
+          local callback = function() print 'I Am A Callback' end
+          vim.api.nvim_create_autocmd("BufWritePost", {
+            pattern = "*.py",
+            callback = callback,
+          })
+          local aus = vim.api.nvim_get_autocmds({ event = 'BufWritePost' })
+          local first = aus[1]
+          return {
+            command = first.command,
+            cbtype = type(first.callback)
+          }
+        ]])
+
+        eq({ command = "", cbtype = 'function' }, result)
+      end)
     end)
 
     describe('groups', function()
@@ -455,7 +573,7 @@ describe('autocmd api', function()
         command [[augroup END]]
       end)
 
-      it('should return all groups if no group is specified', function()
+      it('returns all groups if no group is specified', function()
         local aus = meths.get_autocmds { event = "InsertEnter" }
         if #aus ~= 4 then
           eq({}, aus)
@@ -464,7 +582,7 @@ describe('autocmd api', function()
         eq(4, #aus)
       end)
 
-      it('should return only the group specified', function()
+      it('returns only the group specified', function()
         local aus = meths.get_autocmds {
           event = "InsertEnter",
           group = "GroupOne",
@@ -475,7 +593,7 @@ describe('autocmd api', function()
         eq("GroupOne", aus[1].group_name)
       end)
 
-      it('should return only the group specified, multiple values', function()
+      it('returns only the group specified, multiple values', function()
         local aus = meths.get_autocmds {
           event = "InsertEnter",
           group = "GroupTwo",
@@ -502,7 +620,7 @@ describe('autocmd api', function()
         ]], {}))
 
         eq(false, success)
-        matches('invalid augroup: NotDefined', code)
+        matches("Invalid 'group': 'NotDefined'", code)
       end)
 
       it('raises error for undefined augroup id', function()
@@ -520,7 +638,7 @@ describe('autocmd api', function()
         ]], {}))
 
         eq(false, success)
-        matches('invalid augroup: 1', code)
+        matches("Invalid 'group': 1", code)
       end)
 
       it('raises error for invalid group type', function()
@@ -535,7 +653,21 @@ describe('autocmd api', function()
         ]], {}))
 
         eq(false, success)
-        matches("'group' must be a string or an integer", code)
+        matches("Invalid 'group': expected String or Integer, got Boolean", code)
+      end)
+
+      it('raises error for invalid pattern array', function()
+        local success, code = unpack(meths.exec_lua([[
+          return {pcall(function()
+            vim.api.nvim_create_autocmd("FileType", {
+              pattern = {{}},
+              command = "echo 'hello'",
+            })
+          end)}
+        ]], {}))
+
+        eq(false, success)
+        matches("Invalid 'pattern' item: expected String, got Array", code)
       end)
     end)
 
@@ -550,7 +682,7 @@ describe('autocmd api', function()
         command [[au InsertEnter <buffer> :echo "Buffer"]]
       end)
 
-      it('should should return for literal match', function()
+      it('returns for literal match', function()
         local aus = meths.get_autocmds {
           event = "InsertEnter",
           pattern = "*"
@@ -560,7 +692,7 @@ describe('autocmd api', function()
         eq([[:echo "No Group"]], aus[1].command)
       end)
 
-      it('should return for multiple matches', function()
+      it('returns for multiple matches', function()
         -- vim.api.nvim_get_autocmds
         local aus = meths.get_autocmds {
           event = "InsertEnter",
@@ -597,6 +729,26 @@ describe('autocmd api', function()
   end)
 
   describe('nvim_exec_autocmds', function()
+    it('validation', function()
+      eq("Invalid 'group': 9997999", pcall_err(meths.exec_autocmds, 'FileType', {
+        group = 9997999,
+      }))
+      eq("Invalid 'group': 'bogus'", pcall_err(meths.exec_autocmds, 'FileType', {
+        group = 'bogus',
+      }))
+      eq("Invalid 'group': expected String or Integer, got Array", pcall_err(meths.exec_autocmds, 'FileType', {
+        group = {},
+      }))
+      eq("Invalid 'group': 0", pcall_err(meths.exec_autocmds, 'FileType', {
+        group = 0,
+      }))
+      eq("Invalid 'buffer': expected Integer, got Array", pcall_err(meths.exec_autocmds, 'FileType', {
+        buffer = {},
+      }))
+      eq("Invalid 'event' item: expected String, got Array", pcall_err(meths.exec_autocmds,
+        {'FileType', {}}, {}))
+    end)
+
     it("can trigger builtin autocmds", function()
       meths.set_var("autocmd_executed", false)
 
@@ -610,6 +762,26 @@ describe('autocmd api', function()
       eq(true, meths.get_var("autocmd_executed"))
     end)
 
+    it("can trigger multiple patterns", function()
+      meths.set_var("autocmd_executed", 0)
+
+      meths.create_autocmd("BufReadPost", {
+        pattern = "*",
+        command = "let g:autocmd_executed += 1",
+      })
+
+      meths.exec_autocmds("BufReadPost", { pattern = { "*.lua", "*.vim" } })
+      eq(2, meths.get_var("autocmd_executed"))
+
+      meths.create_autocmd("BufReadPre", {
+        pattern = { "bar", "foo" },
+        command = "let g:autocmd_executed += 10",
+      })
+
+      meths.exec_autocmds("BufReadPre", { pattern = { "foo", "bar", "baz", "frederick" }})
+      eq(22, meths.get_var("autocmd_executed"))
+    end)
+
     it("can pass the buffer", function()
       meths.set_var("buffer_executed", -1)
       eq(-1, meths.get_var("buffer_executed"))
@@ -666,7 +838,7 @@ describe('autocmd api', function()
       meths.exec_autocmds("CursorHoldI", { buffer = 1 })
       eq('none', meths.get_var("filename_executed"))
 
-      meths.exec_autocmds("CursorHoldI", { buffer = tonumber(meths.get_current_buf()) })
+      meths.exec_autocmds("CursorHoldI", { buffer = meths.get_current_buf() })
       eq('__init__.py', meths.get_var("filename_executed"))
 
       -- Reset filename
@@ -758,7 +930,7 @@ describe('autocmd api', function()
       set_ft("txt")
       set_ft("python")
 
-      eq(get_executed_count(), 2)
+      eq(2, get_executed_count())
     end)
 
     it('works getting called multiple times', function()
@@ -767,7 +939,7 @@ describe('autocmd api', function()
       set_ft()
       set_ft()
 
-      eq(get_executed_count(), 3)
+      eq(3, get_executed_count())
     end)
 
     it('handles ++once', function()
@@ -777,7 +949,7 @@ describe('autocmd api', function()
       set_ft('txt')
       set_ft('help')
 
-      eq(get_executed_count(), 1)
+      eq(1, get_executed_count())
     end)
 
     it('errors on unexpected keys', function()
@@ -894,6 +1066,12 @@ describe('autocmd api', function()
 
       eq(false, exec_lua[[return pcall(vim.api.nvim_del_augroup_by_id, -12342)]])
       eq('Vim:E367: No such group: "--Deleted--"', pcall_err(meths.del_augroup_by_id, -12312))
+
+      eq(false, exec_lua[[return pcall(vim.api.nvim_del_augroup_by_id, 0)]])
+      eq('Vim:E367: No such group: "[NULL]"', pcall_err(meths.del_augroup_by_id, 0))
+
+      eq(false, exec_lua[[return pcall(vim.api.nvim_del_augroup_by_id, 12342)]])
+      eq('Vim:E367: No such group: "[NULL]"', pcall_err(meths.del_augroup_by_id, 12312))
     end)
 
     it('groups work with once', function()
@@ -905,7 +1083,7 @@ describe('autocmd api', function()
       set_ft("txt")
       set_ft("python")
 
-      eq(get_executed_count(), 1)
+      eq(1, get_executed_count())
     end)
 
     it('autocmds can be registered multiple times.', function()
@@ -919,14 +1097,14 @@ describe('autocmd api', function()
       set_ft("txt")
       set_ft("python")
 
-      eq(get_executed_count(), 3 * 2)
+      eq(3 * 2, get_executed_count())
     end)
 
     it('can be deleted', function()
       local augroup = "WillBeDeleted"
 
       meths.create_augroup(augroup, { clear = true })
-      meths.create_autocmd({"Filetype"}, {
+      meths.create_autocmd({"FileType"}, {
         pattern = "*",
         command = "echo 'does not matter'",
       })
@@ -945,7 +1123,7 @@ describe('autocmd api', function()
       meths.set_var("value_set", false)
 
       meths.create_augroup(augroup, { clear = true })
-      meths.create_autocmd("Filetype", {
+      meths.create_autocmd("FileType", {
         pattern = "<buffer>",
         command = "let g:value_set = v:true",
       })
@@ -1061,6 +1239,17 @@ describe('autocmd api', function()
   end)
 
   describe('nvim_clear_autocmds', function()
+    it('validation', function()
+      eq("Cannot use both 'pattern' and 'buffer'", pcall_err(meths.clear_autocmds, {
+        pattern = '*',
+        buffer = 42,
+      }))
+      eq("Invalid 'event' item: expected String, got Array", pcall_err(meths.clear_autocmds, {
+        event = {'FileType', {}}
+      }))
+      eq("Invalid 'group': 0", pcall_err(meths.clear_autocmds, {group = 0}))
+    end)
+
     it('should clear based on event + pattern', function()
       command('autocmd InsertEnter *.py  :echo "Python can be cool sometimes"')
       command('autocmd InsertEnter *.txt :echo "Text Files Are Cool"')
