@@ -3,10 +3,12 @@ local Screen = require('test.functional.ui.screen')
 local clear, nvim, buffer = helpers.clear, helpers.nvim, helpers.buffer
 local curbuf, curwin, eq = helpers.curbuf, helpers.curwin, helpers.eq
 local curbufmeths, ok = helpers.curbufmeths, helpers.ok
+local describe_lua_and_rpc = helpers.describe_lua_and_rpc(describe)
 local meths = helpers.meths
 local funcs = helpers.funcs
 local request = helpers.request
 local exc_exec = helpers.exc_exec
+local exec_lua = helpers.exec_lua
 local feed_command = helpers.feed_command
 local insert = helpers.insert
 local NIL = helpers.NIL
@@ -184,20 +186,21 @@ describe('api/buf', function()
     end)
   end)
 
-  describe('nvim_buf_get_lines, nvim_buf_set_lines', function()
-    local get_lines, set_lines = curbufmeths.get_lines, curbufmeths.set_lines
-    local line_count = curbufmeths.line_count
+  describe_lua_and_rpc('nvim_buf_get_lines, nvim_buf_set_lines', function(api)
+    local get_lines = api.curbufmeths.get_lines
+    local set_lines = api.curbufmeths.set_lines
+    local line_count = api.curbufmeths.line_count
 
     it('fails correctly when input is not valid', function()
-      eq(1, curbufmeths.get_number())
-      eq([[String cannot contain newlines]],
+      eq(1, api.curbufmeths.get_number())
+      eq([['replacement string' item contains newlines]],
         pcall_err(bufmeths.set_lines, 1, 1, 2, false, {'b\na'}))
     end)
 
     it("fails if 'nomodifiable'", function()
       command('set nomodifiable')
       eq([[Buffer is not 'modifiable']],
-        pcall_err(bufmeths.set_lines, 1, 1, 2, false, {'a','b'}))
+        pcall_err(api.bufmeths.set_lines, 1, 1, 2, false, {'a','b'}))
     end)
 
     it('has correct line_count when inserting and deleting', function()
@@ -353,7 +356,7 @@ describe('api/buf', function()
         Who would win?
         A real window
         with proper text]])
-      local buf = meths.create_buf(false,true)
+      local buf = api.meths.create_buf(false,true)
       screen:expect([[
         Who would win?      |
         A real window       |
@@ -362,7 +365,7 @@ describe('api/buf', function()
                             |
       ]])
 
-      meths.buf_set_lines(buf, 0, -1, true, {'or some', 'scratchy text'})
+      api.meths.buf_set_lines(buf, 0, -1, true, {'or some', 'scratchy text'})
       feed('i') -- provoke redraw
       screen:expect([[
         Who would win?      |
@@ -378,15 +381,15 @@ describe('api/buf', function()
         visible buffer line 1
         line 2
       ]])
-      local hiddenbuf = meths.create_buf(false,true)
+      local hiddenbuf = api.meths.create_buf(false,true)
       command('vsplit')
       command('vsplit')
       feed('<c-w>l<c-w>l<c-w>l')
       eq(3, funcs.winnr())
       feed('<c-w>h')
       eq(2, funcs.winnr())
-      meths.buf_set_lines(hiddenbuf, 0, -1, true,
-                          {'hidden buffer line 1', 'line 2'})
+      api.meths.buf_set_lines(hiddenbuf, 0, -1, true,
+                              {'hidden buffer line 1', 'line 2'})
       feed('<c-w>p')
       eq(3, funcs.winnr())
     end)
@@ -550,30 +553,41 @@ describe('api/buf', function()
       insert([[
       hello foo!
       text]])
-      eq('start_row out of bounds', pcall_err(set_text, 2, 0, 3, 0, {}))
-      eq('start_row out of bounds', pcall_err(set_text, -3, 0, 0, 0, {}))
-      eq('end_row out of bounds', pcall_err(set_text, 0, 0, 2, 0, {}))
-      eq('end_row out of bounds', pcall_err(set_text, 0, 0, -3, 0, {}))
-      eq('start_col out of bounds', pcall_err(set_text, 1, 5, 1, 5, {}))
-      eq('end_col out of bounds', pcall_err(set_text, 1, 0, 1, 5, {}))
+      eq("Invalid 'start_row': out of range", pcall_err(set_text, 2, 0, 3, 0, {}))
+      eq("Invalid 'start_row': out of range", pcall_err(set_text, -3, 0, 0, 0, {}))
+      eq("Invalid 'end_row': out of range", pcall_err(set_text, 0, 0, 2, 0, {}))
+      eq("Invalid 'end_row': out of range", pcall_err(set_text, 0, 0, -3, 0, {}))
+      eq("Invalid 'start_col': out of range", pcall_err(set_text, 1, 5, 1, 5, {}))
+      eq("Invalid 'end_col': out of range", pcall_err(set_text, 1, 0, 1, 5, {}))
     end)
 
     it('errors when start is greater than end', function()
       insert([[
       hello foo!
       text]])
-      eq('start is higher than end', pcall_err(set_text, 1, 0, 0, 0, {}))
-      eq('start is higher than end', pcall_err(set_text, 0, 1, 0, 0, {}))
+      eq("'start' is higher than 'end'", pcall_err(set_text, 1, 0, 0, 0, {}))
+      eq("'start' is higher than 'end'", pcall_err(set_text, 0, 1, 0, 0, {}))
+    end)
+
+    it('no heap-use-after-free when called consecutively #19643', function()
+      set_text(0, 0, 0, 0, {'one', '', '', 'two'})
+      eq({'one', '', '', 'two'}, get_lines(0, 4, true))
+      meths.win_set_cursor(0, {1, 0})
+      exec_lua([[
+        vim.api.nvim_buf_set_text(0, 0, 3, 1, 0, {''})
+        vim.api.nvim_buf_set_text(0, 0, 3, 1, 0, {''})
+      ]])
+      eq({'one', 'two'}, get_lines(0, 2, true))
     end)
   end)
 
-  describe('nvim_buf_get_text', function()
-    local get_text = curbufmeths.get_text
-
+  describe_lua_and_rpc('nvim_buf_get_text', function(api)
+    local get_text = api.curbufmeths.get_text
     before_each(function()
       insert([[
       hello foo!
-      text]])
+      text
+      more]])
     end)
 
     it('works', function()
@@ -581,22 +595,23 @@ describe('api/buf', function()
       eq({'hello foo!'}, get_text(0, 0, 0, 42, {}))
       eq({'foo!'}, get_text(0, 6, 0, 10, {}))
       eq({'foo!', 'tex'}, get_text(0, 6, 1, 3, {}))
-      eq({'foo!', 'tex'}, get_text(-2, 6, -1, 3, {}))
+      eq({'foo!', 'tex'}, get_text(-3, 6, -2, 3, {}))
       eq({''}, get_text(0, 18, 0, 20, {}))
-      eq({'ext'}, get_text(-1, 1, -1, 4, {}))
+      eq({'ext'}, get_text(-2, 1, -2, 4, {}))
+      eq({'hello foo!', 'text', 'm'}, get_text(0, 0, 2, 1, {}))
     end)
 
     it('errors on out-of-range', function()
-      eq('Index out of bounds', pcall_err(get_text, 2, 0, 3, 0, {}))
-      eq('Index out of bounds', pcall_err(get_text, -3, 0, 0, 0, {}))
-      eq('Index out of bounds', pcall_err(get_text, 0, 0, 2, 0, {}))
-      eq('Index out of bounds', pcall_err(get_text, 0, 0, -3, 0, {}))
+      eq('Index out of bounds', pcall_err(get_text, 2, 0, 4, 0, {}))
+      eq('Index out of bounds', pcall_err(get_text, -4, 0, 0, 0, {}))
+      eq('Index out of bounds', pcall_err(get_text, 0, 0, 3, 0, {}))
+      eq('Index out of bounds', pcall_err(get_text, 0, 0, -4, 0, {}))
       -- no ml_get errors should happen #19017
       eq('', meths.get_vvar('errmsg'))
     end)
 
     it('errors when start is greater than end', function()
-      eq('start is higher than end', pcall_err(get_text, 1, 0, 0, 0, {}))
+      eq("'start' is higher than 'end'", pcall_err(get_text, 1, 0, 0, 0, {}))
       eq('start_col must be less than end_col', pcall_err(get_text, 0, 1, 0, 0, {}))
     end)
   end)
@@ -784,11 +799,13 @@ describe('api/buf', function()
       curbufmeths.set_lines(-1, -1, true, {'a', 'bit of', 'text'})
       eq(true, curbufmeths.set_mark('z', 1, 1, {}))
       eq({1, 1}, curbufmeths.get_mark('z'))
+      eq({0, 1, 2, 0}, funcs.getpos("'z"))
     end)
     it('works with file/uppercase marks', function()
       curbufmeths.set_lines(-1, -1, true, {'a', 'bit of', 'text'})
-      eq(true, curbufmeths.set_mark('Z', 3, 1, {}))
-      eq({3, 1}, curbufmeths.get_mark('Z'))
+      eq(true, curbufmeths.set_mark('Z', 3, 2, {}))
+      eq({3, 2}, curbufmeths.get_mark('Z'))
+      eq({curbuf().id, 3, 3, 0}, funcs.getpos("'Z"))
     end)
     it('fails when invalid marks names are used', function()
       eq(false, pcall(curbufmeths.set_mark, '!', 1, 0, {}))
