@@ -6,6 +6,7 @@ local command = helpers.command
 local curbufmeths = helpers.curbufmeths
 local eq, neq = helpers.eq, helpers.neq
 local exec_lua = helpers.exec_lua
+local exec = helpers.exec
 local feed = helpers.feed
 local funcs = helpers.funcs
 local meths = helpers.meths
@@ -25,6 +26,7 @@ describe('nvim_get_keymap', function()
   local foo_bar_string = 'nnoremap foo bar'
   local foo_bar_map_table = {
     lhs='foo',
+    lhsraw='foo',
     script=0,
     silent=0,
     rhs='bar',
@@ -56,6 +58,7 @@ describe('nvim_get_keymap', function()
     command('nnoremap foo_longer bar_longer')
     local foolong_bar_map_table = shallowcopy(foo_bar_map_table)
     foolong_bar_map_table['lhs'] = 'foo_longer'
+    foolong_bar_map_table['lhsraw'] = 'foo_longer'
     foolong_bar_map_table['rhs'] = 'bar_longer'
 
     eq({foolong_bar_map_table, foo_bar_map_table},
@@ -87,6 +90,7 @@ describe('nvim_get_keymap', function()
     command('nnoremap foo_longer bar_longer')
     local foolong_bar_map_table = shallowcopy(foo_bar_map_table)
     foolong_bar_map_table['lhs'] = 'foo_longer'
+    foolong_bar_map_table['lhsraw'] = 'foo_longer'
     foolong_bar_map_table['rhs'] = 'bar_longer'
 
     local buffer_table = shallowcopy(foo_bar_map_table)
@@ -283,6 +287,16 @@ describe('nvim_get_keymap', function()
     command('onoremap \\<C-a><C-a><LT>C-a>\\  \\<C-b><C-b><LT>C-b>\\')
     command('onoremap <special> \\<C-c><C-c><LT>C-c>\\  \\<C-d><C-d><LT>C-d>\\')
 
+    -- wrapper around get_keymap() that drops "lhsraw" and "lhsrawalt" which are hard to check
+    local function get_keymap_noraw(...)
+      local ret = meths.get_keymap(...)
+      for _, item in ipairs(ret) do
+        item.lhsraw = nil
+        item.lhsrawalt = nil
+      end
+      return ret
+    end
+
     for _, cmd in ipairs({
       'set cpo-=B',
       'set cpo+=B',
@@ -290,22 +304,23 @@ describe('nvim_get_keymap', function()
       command(cmd)
       eq({cpomap('\\<C-C><C-C><lt>C-c>\\', '\\<C-D><C-D><lt>C-d>\\', 'n'),
           cpomap('\\<C-A><C-A><lt>C-a>\\', '\\<C-B><C-B><lt>C-b>\\', 'n')},
-         meths.get_keymap('n'))
+         get_keymap_noraw('n'))
       eq({cpomap('\\<C-C><C-C><lt>C-c>\\', '\\<C-D><C-D><lt>C-d>\\', 'x'),
           cpomap('\\<C-A><C-A><lt>C-a>\\', '\\<C-B><C-B><lt>C-b>\\', 'x')},
-         meths.get_keymap('x'))
+         get_keymap_noraw('x'))
       eq({cpomap('<lt>C-c><C-C><lt>C-c> ', '<lt>C-d><C-D><lt>C-d>', 's'),
           cpomap('<lt>C-a><C-A><lt>C-a> ', '<lt>C-b><C-B><lt>C-b>', 's')},
-         meths.get_keymap('s'))
+         get_keymap_noraw('s'))
       eq({cpomap('<lt>C-c><C-C><lt>C-c> ', '<lt>C-d><C-D><lt>C-d>', 'o'),
           cpomap('<lt>C-a><C-A><lt>C-a> ', '<lt>C-b><C-B><lt>C-b>', 'o')},
-         meths.get_keymap('o'))
+         get_keymap_noraw('o'))
     end
   end)
 
   it('always uses space for space and bar for bar', function()
     local space_table = {
       lhs='|   |',
+      lhsraw='|   |',
       rhs='|    |',
       mode='n',
       script=0,
@@ -321,25 +336,31 @@ describe('nvim_get_keymap', function()
     eq({space_table}, meths.get_keymap('n'))
   end)
 
-  it('can handle lua keymaps', function()
-    eq(0, exec_lua [[
+  it('can handle lua mappings', function()
+    eq(0, exec_lua([[
       GlobalCount = 0
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
       return GlobalCount
-    ]])
+    ]]))
 
     feed('asdf\n')
-    eq(1, exec_lua[[return GlobalCount]])
+    eq(1, exec_lua([[return GlobalCount]]))
 
-    eq(2, exec_lua[[
+    eq(2, exec_lua([[
       vim.api.nvim_get_keymap('n')[1].callback()
       return GlobalCount
+    ]]))
+
+    exec([[
+      call nvim_get_keymap('n')[0].callback()
     ]])
+    eq(3, exec_lua([[return GlobalCount]]))
+
     local mapargs = meths.get_keymap('n')
-    assert.Truthy(type(mapargs[1].callback) == 'number', 'callback is not luaref number')
     mapargs[1].callback = nil
     eq({
       lhs='asdf',
+      lhsraw='asdf',
       script=0,
       silent=0,
       expr=0,
@@ -356,6 +377,7 @@ describe('nvim_get_keymap', function()
     meths.set_keymap('n', 'lhs', 'rhs', {desc="map description"})
     eq({
       lhs='lhs',
+      lhsraw='lhs',
       rhs='rhs',
       script=0,
       silent=0,
@@ -413,7 +435,11 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
 
   -- Gets a maparg() dict from Nvim, if one exists.
   local function get_mapargs(mode, lhs)
-    return funcs.maparg(lhs, normalize_mapmode(mode), false, true)
+    local mapargs = funcs.maparg(lhs, normalize_mapmode(mode), false, true)
+    -- drop "lhsraw" and "lhsrawalt" which are hard to check
+    mapargs.lhsraw = nil
+    mapargs.lhsrawalt = nil
+    return mapargs
   end
 
   it('error on empty LHS', function()
@@ -503,6 +529,11 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
       pcall_err(meths.set_keymap, 'n', 'lhs', 'rhs', {buffer = true}))
   end)
 
+  it('error when "replace_keycodes" is used without "expr"', function()
+    eq('"replace_keycodes" requires "expr"',
+      pcall_err(meths.set_keymap, 'n', 'lhs', 'rhs', {replace_keycodes = true}))
+  end)
+
   local optnames = {'nowait', 'silent', 'script', 'expr', 'unique'}
   for _, opt in ipairs(optnames) do
     -- note: need '%' to escape hyphens, which have special meaning in lua
@@ -582,7 +613,7 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
   it('can set mappings containing literal keycodes', function()
     meths.set_keymap('n', '\n\r\n', 'rhs', {})
     local expected = generate_mapargs('n', '<NL><CR><NL>', 'rhs')
-    eq(expected, get_mapargs('n', '<C-j><CR><C-j>'))
+    eq(expected, get_mapargs('n', '<NL><CR><NL>'))
   end)
 
   it('can set mappings whose RHS is a <Nop>', function()
@@ -606,6 +637,13 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
     eq({''}, curbufmeths.get_lines(0, -1, 0))
     eq(generate_mapargs('i', 'lhs', '<NOP>', {}),
        get_mapargs('i', 'lhs'))
+
+    -- a single ^V in RHS is also <Nop> (see :h map-empty-rhs)
+    meths.set_keymap('i', 'lhs', '\022', {})
+    command('normal ilhs')
+    eq({''}, curbufmeths.get_lines(0, -1, 0))
+    eq(generate_mapargs('i', 'lhs', '\022', {}),
+       get_mapargs('i', 'lhs'))
   end)
 
   it('treats an empty RHS in a mapping like a <Nop>', function()
@@ -643,13 +681,13 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
   end)
 
   it('can set <expr> mappings whose RHS change dynamically', function()
-    meths.exec([[
+    exec([[
         function! FlipFlop() abort
           if !exists('g:flip') | let g:flip = 0 | endif
           let g:flip = !g:flip
           return g:flip
         endfunction
-        ]], true)
+        ]])
     eq(1, meths.call_function('FlipFlop', {}))
     eq(0, meths.call_function('FlipFlop', {}))
     eq(1, meths.call_function('FlipFlop', {}))
@@ -775,68 +813,117 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
   it('can make lua mappings', function()
     eq(0, exec_lua [[
       GlobalCount = 0
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
       return GlobalCount
     ]])
 
     feed('asdf\n')
 
     eq(1, exec_lua[[return GlobalCount]])
-
   end)
 
-  it (':map command shows lua keymap correctly', function()
+  it (':map command shows lua mapping correctly', function()
     exec_lua [[
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() print('jkl;') end })
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() print('jkl;') end })
     ]]
-    assert.truthy(string.match(exec_lua[[return vim.api.nvim_exec(':nmap asdf', true)]],
-                  "^\nn  asdf          <Lua function %d+>"))
+    assert.truthy(
+      string.match(
+        exec_lua[[return vim.api.nvim_exec2(':nmap asdf', { output = true }).output]],
+        "^\nn  asdf          <Lua %d+>"
+      )
+    )
   end)
 
-  it ('mapcheck() returns lua keymap correctly', function()
+  it ('mapcheck() returns lua mapping correctly', function()
     exec_lua [[
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() print('jkl;') end })
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() print('jkl;') end })
     ]]
     assert.truthy(string.match(funcs.mapcheck('asdf', 'n'),
-                  "^<Lua function %d+>"))
+                  "^<Lua %d+>"))
   end)
 
-  it ('maparg() returns lua keymap correctly', function()
-    exec_lua [[
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() print('jkl;') end })
-    ]]
-    assert.truthy(string.match(funcs.maparg('asdf', 'n'),
-                  "^<Lua function %d+>"))
+  it ('maparg() returns lua mapping correctly', function()
+    eq(0, exec_lua([[
+      GlobalCount = 0
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      return GlobalCount
+    ]]))
+
+    assert.truthy(string.match(funcs.maparg('asdf', 'n'), "^<Lua %d+>"))
+
     local mapargs = funcs.maparg('asdf', 'n', false, true)
-    assert.Truthy(type(mapargs.callback) == 'number', 'callback is not luaref number')
     mapargs.callback = nil
+    mapargs.lhsraw = nil
+    mapargs.lhsrawalt = nil
     eq(generate_mapargs('n', 'asdf', nil, {sid=sid_lua}), mapargs)
+
+    eq(1, exec_lua([[
+      vim.fn.maparg('asdf', 'n', false, true).callback()
+      return GlobalCount
+    ]]))
+
+    exec([[
+      call maparg('asdf', 'n', v:false, v:true).callback()
+    ]])
+    eq(2, exec_lua([[return GlobalCount]]))
   end)
 
-  it('can make lua expr mappings', function()
+  it('can make lua expr mappings replacing keycodes', function()
     exec_lua [[
-      vim.api.nvim_set_keymap ('n', 'aa', '', {callback = function() return vim.api.nvim_replace_termcodes(':lua SomeValue = 99<cr>', true, false, true) end, expr = true })
+      vim.api.nvim_set_keymap('n', 'aa', '', {callback = function() return '<Insert>π<C-V><M-π>foo<lt><Esc>' end, expr = true, replace_keycodes = true })
     ]]
 
     feed('aa')
 
-    eq(99, exec_lua[[return SomeValue]])
+    eq({'π<M-π>foo<'}, meths.buf_get_lines(0, 0, -1, false))
+  end)
+
+  it('can make lua expr mappings without replacing keycodes', function()
+    exec_lua [[
+      vim.api.nvim_set_keymap('i', 'aa', '', {callback = function() return '<space>' end, expr = true })
+    ]]
+
+    feed('iaa<esc>')
+
+    eq({'<space>'}, meths.buf_get_lines(0, 0, -1, false))
+  end)
+
+  it('lua expr mapping returning nil is equivalent to returning an empty string', function()
+    exec_lua [[
+      vim.api.nvim_set_keymap('i', 'aa', '', {callback = function() return nil end, expr = true })
+    ]]
+
+    feed('iaa<esc>')
+
+    eq({''}, meths.buf_get_lines(0, 0, -1, false))
   end)
 
   it('does not reset pum in lua mapping', function()
     eq(0, exec_lua [[
       VisibleCount = 0
-      vim.api.nvim_set_keymap ('i', '<F2>', '', {callback = function() VisibleCount = VisibleCount + vim.fn.pumvisible() end})
+      vim.api.nvim_set_keymap('i', '<F2>', '', {callback = function() VisibleCount = VisibleCount + vim.fn.pumvisible() end})
       return VisibleCount
     ]])
     feed('i<C-X><C-V><F2><F2><esc>')
     eq(2, exec_lua[[return VisibleCount]])
   end)
 
+  it('redo of lua mappings in op-pending mode work', function()
+    eq(0, exec_lua [[
+      OpCount = 0
+      vim.api.nvim_set_keymap('o', '<F2>', '', {callback = function() OpCount = OpCount + 1 end})
+      return OpCount
+    ]])
+    feed('d<F2>')
+    eq(1, exec_lua[[return OpCount]])
+    feed('.')
+    eq(2, exec_lua[[return OpCount]])
+  end)
+
   it('can overwrite lua mappings', function()
     eq(0, exec_lua [[
       GlobalCount = 0
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
       return GlobalCount
     ]])
 
@@ -845,7 +932,7 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
     eq(1, exec_lua[[return GlobalCount]])
 
     exec_lua [[
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount - 1 end })
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount - 1 end })
     ]]
 
     feed('asdf\n')
@@ -856,7 +943,7 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
   it('can unmap lua mappings', function()
     eq(0, exec_lua [[
       GlobalCount = 0
-      vim.api.nvim_set_keymap ('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      vim.api.nvim_set_keymap('n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
       return GlobalCount
     ]])
 
@@ -874,7 +961,28 @@ describe('nvim_set_keymap, nvim_del_keymap', function()
     eq('\nNo mapping found', helpers.exec_capture('nmap asdf'))
   end)
 
-  it('can set descriptions on keymaps', function()
+  it('no double-free when unmapping simplifiable lua mappings', function()
+    eq(0, exec_lua [[
+      GlobalCount = 0
+      vim.api.nvim_set_keymap('n', '<C-I>', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      return GlobalCount
+    ]])
+
+    feed('<C-I>\n')
+
+    eq(1, exec_lua[[return GlobalCount]])
+
+    exec_lua [[
+      vim.api.nvim_del_keymap('n', '<C-I>')
+    ]]
+
+    feed('<C-I>\n')
+
+    eq(1, exec_lua[[return GlobalCount]])
+    eq('\nNo mapping found', helpers.exec_capture('nmap <C-I>'))
+  end)
+
+  it('can set descriptions on mappings', function()
     meths.set_keymap('n', 'lhs', 'rhs', {desc="map description"})
     eq(generate_mapargs('n', 'lhs', 'rhs', {desc="map description"}), get_mapargs('n', 'lhs'))
     eq("\nn  lhs           rhs\n                 map description",
@@ -973,7 +1081,7 @@ describe('nvim_buf_set_keymap, nvim_buf_del_keymap', function()
     eq({'rhs'}, bufmeths.get_lines(0, 0, 1, 1))
   end)
 
-  it("does not crash when setting keymap in a non-existing buffer #13541", function()
+  it("does not crash when setting mapping in a non-existing buffer #13541", function()
     pcall_err(bufmeths.set_keymap, 100, '', 'lsh', 'irhs<Esc>', {})
     helpers.assert_alive()
   end)
@@ -981,7 +1089,7 @@ describe('nvim_buf_set_keymap, nvim_buf_del_keymap', function()
   it('can make lua mappings', function()
     eq(0, exec_lua [[
       GlobalCount = 0
-      vim.api.nvim_buf_set_keymap (0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      vim.api.nvim_buf_set_keymap(0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
       return GlobalCount
     ]])
 
@@ -990,20 +1098,31 @@ describe('nvim_buf_set_keymap, nvim_buf_del_keymap', function()
     eq(1, exec_lua[[return GlobalCount]])
   end)
 
-  it('can make lua expr mappings', function()
+  it('can make lua expr mappings replacing keycodes', function()
     exec_lua [[
-      vim.api.nvim_buf_set_keymap (0, 'n', 'aa', '', {callback = function() return vim.api.nvim_replace_termcodes(':lua SomeValue = 99<cr>', true, false, true) end, expr = true })
+      vim.api.nvim_buf_set_keymap(0, 'n', 'aa', '', {callback = function() return '<Insert>π<C-V><M-π>foo<lt><Esc>' end, expr = true, replace_keycodes = true })
     ]]
 
     feed('aa')
 
-    eq(99, exec_lua[[return SomeValue ]])
+    eq({'π<M-π>foo<'}, meths.buf_get_lines(0, 0, -1, false))
   end)
 
+  it('can make lua expr mappings without replacing keycodes', function()
+    exec_lua [[
+      vim.api.nvim_buf_set_keymap(0, 'i', 'aa', '', {callback = function() return '<space>' end, expr = true })
+    ]]
+
+    feed('iaa<esc>')
+
+    eq({'<space>'}, meths.buf_get_lines(0, 0, -1, false))
+  end)
+
+
   it('can overwrite lua mappings', function()
     eq(0, exec_lua [[
       GlobalCount = 0
-      vim.api.nvim_buf_set_keymap (0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      vim.api.nvim_buf_set_keymap(0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
       return GlobalCount
     ]])
 
@@ -1012,7 +1131,7 @@ describe('nvim_buf_set_keymap, nvim_buf_del_keymap', function()
     eq(1, exec_lua[[return GlobalCount]])
 
     exec_lua [[
-      vim.api.nvim_buf_set_keymap (0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount - 1 end })
+      vim.api.nvim_buf_set_keymap(0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount - 1 end })
     ]]
 
     feed('asdf\n')
@@ -1023,7 +1142,7 @@ describe('nvim_buf_set_keymap, nvim_buf_del_keymap', function()
   it('can unmap lua mappings', function()
     eq(0, exec_lua [[
       GlobalCount = 0
-      vim.api.nvim_buf_set_keymap (0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      vim.api.nvim_buf_set_keymap(0, 'n', 'asdf', '', {callback = function() GlobalCount = GlobalCount + 1 end })
       return GlobalCount
     ]])
 
@@ -1040,4 +1159,25 @@ describe('nvim_buf_set_keymap, nvim_buf_del_keymap', function()
     eq(1, exec_lua[[return GlobalCount]])
     eq('\nNo mapping found', helpers.exec_capture('nmap asdf'))
   end)
+
+  it('no double-free when unmapping simplifiable lua mappings', function()
+    eq(0, exec_lua [[
+      GlobalCount = 0
+      vim.api.nvim_buf_set_keymap(0, 'n', '<C-I>', '', {callback = function() GlobalCount = GlobalCount + 1 end })
+      return GlobalCount
+    ]])
+
+    feed('<C-I>\n')
+
+    eq(1, exec_lua[[return GlobalCount]])
+
+    exec_lua [[
+      vim.api.nvim_buf_del_keymap(0, 'n', '<C-I>')
+    ]]
+
+    feed('<C-I>\n')
+
+    eq(1, exec_lua[[return GlobalCount]])
+    eq('\nNo mapping found', helpers.exec_capture('nmap <C-I>'))
+  end)
 end)
