@@ -4,9 +4,8 @@
 package.path = arg[1]
 package.cpath = arg[2]
 
-local mpack = require('mpack')
-local StdioStream = require('nvim.stdio_stream')
-local Session = require('nvim.session')
+local StdioStream = require'test.client.uv_stream'.StdioStream
+local Session = require'test.client.session'
 
 local stdio_stream = StdioStream.open()
 local session = Session.new(stdio_stream)
@@ -19,7 +18,7 @@ local function on_request(method, args)
     return "done!"
   elseif method == "exit" then
     session:stop()
-    return mpack.NIL
+    return vim.NIL
   end
 end
 
