@@ -7,6 +7,7 @@ local clear, nvim, curbuf, curbuf_contents, window, curwin, eq, neq,
   helpers.tabpage
 local poke_eventloop = helpers.poke_eventloop
 local curwinmeths = helpers.curwinmeths
+local exec = helpers.exec
 local funcs = helpers.funcs
 local request = helpers.request
 local NIL = helpers.NIL
@@ -15,25 +16,6 @@ local command = helpers.command
 local pcall_err = helpers.pcall_err
 local assert_alive = helpers.assert_alive
 
--- check if str is visible at the beginning of some line
-local function is_visible(str)
-    local slen = string.len(str)
-    local nlines = eval("&lines")
-    for i = 1,nlines do
-        local iseq = true
-        for j = 1,slen do
-            if string.byte(str,j) ~= eval("screenchar("..i..","..j..")") then
-                iseq = false
-                break
-            end
-        end
-        if iseq then
-            return true
-        end
-    end
-    return false
-end
-
 describe('API/win', function()
   before_each(clear)
 
@@ -79,27 +61,61 @@ describe('API/win', function()
     end)
 
     it('updates the screen, and also when the window is unfocused', function()
+      local screen = Screen.new(30, 9)
+      screen:set_default_attr_ids({
+        [1] = {bold = true, foreground = Screen.colors.Blue},
+        [2] = {bold = true, reverse = true};
+        [3] = {reverse = true};
+      })
+      screen:attach()
+
       insert("prologue")
       feed('100o<esc>')
       insert("epilogue")
       local win = curwin()
       feed('gg')
-      poke_eventloop() -- let nvim process the 'gg' command
 
+      screen:expect{grid=[[
+        ^prologue                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+      ]]}
       -- cursor position is at beginning
       eq({1, 0}, window('get_cursor', win))
-      eq(true, is_visible("prologue"))
-      eq(false, is_visible("epilogue"))
 
       -- move cursor to end
       window('set_cursor', win, {101, 0})
-      eq(false, is_visible("prologue"))
-      eq(true, is_visible("epilogue"))
+      screen:expect{grid=[[
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+        ^epilogue                      |
+                                      |
+      ]]}
 
       -- move cursor to the beginning again
       window('set_cursor', win, {1, 0})
-      eq(true, is_visible("prologue"))
-      eq(false, is_visible("epilogue"))
+      screen:expect{grid=[[
+        ^prologue                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+                                      |
+      ]]}
 
       -- move focus to new window
       nvim('command',"new")
@@ -107,18 +123,45 @@ describe('API/win', function()
 
       -- sanity check, cursor position is kept
       eq({1, 0}, window('get_cursor', win))
-      eq(true, is_visible("prologue"))
-      eq(false, is_visible("epilogue"))
+      screen:expect{grid=[[
+        ^                              |
+        {1:~                             }|
+        {1:~                             }|
+        {2:[No Name]                     }|
+        prologue                      |
+                                      |
+                                      |
+        {3:[No Name] [+]                 }|
+                                      |
+      ]]}
 
       -- move cursor to end
       window('set_cursor', win, {101, 0})
-      eq(false, is_visible("prologue"))
-      eq(true, is_visible("epilogue"))
+      screen:expect{grid=[[
+        ^                              |
+        {1:~                             }|
+        {1:~                             }|
+        {2:[No Name]                     }|
+                                      |
+                                      |
+        epilogue                      |
+        {3:[No Name] [+]                 }|
+                                      |
+      ]]}
 
       -- move cursor to the beginning again
       window('set_cursor', win, {1, 0})
-      eq(true, is_visible("prologue"))
-      eq(false, is_visible("epilogue"))
+      screen:expect{grid=[[
+        ^                              |
+        {1:~                             }|
+        {1:~                             }|
+        {2:[No Name]                     }|
+        prologue                      |
+                                      |
+                                      |
+        {3:[No Name] [+]                 }|
+                                      |
+      ]]}
 
       -- curwin didn't change back
       neq(win, curwin())
@@ -153,7 +196,7 @@ describe('API/win', function()
         [1] = {bold = true, foreground = Screen.colors.Blue},  -- NonText
         [2] = {background = Screen.colors.Grey90},  -- CursorLine
         [3] = {bold = true, reverse = true},  -- StatusLine
-        [4] = {reverse = true},  -- VertSplit, StatusLineNC
+        [4] = {reverse = true},  -- StatusLineNC
       })
       screen:attach()
       command('set ruler')
@@ -166,27 +209,67 @@ describe('API/win', function()
       local oldwin = curwin()
       command('vsplit')
       screen:expect([[
-        aaa                           {4:│}aaa                          |
-        bbb                           {4:│}bbb                          |
-        ccc                           {4:│}ccc                          |
-        {2:dd^d                           }{4:│}{2:ddd                          }|
-        {1:~                             }{4:│}{1:~                            }|
-        {1:~                             }{4:│}{1:~                            }|
+        aaa                           │aaa                          |
+        bbb                           │bbb                          |
+        ccc                           │ccc                          |
+        {2:dd^d                           }│{2:ddd                          }|
+        {1:~                             }│{1:~                            }|
+        {1:~                             }│{1:~                            }|
         {3:[No Name] [+]  4,3         All }{4:[No Name] [+]  4,3        All}|
                                                                     |
       ]])
       window('set_cursor', oldwin, {1, 0})
       screen:expect([[
-        aaa                           {4:│}{2:aaa                          }|
-        bbb                           {4:│}bbb                          |
-        ccc                           {4:│}ccc                          |
-        {2:dd^d                           }{4:│}ddd                          |
-        {1:~                             }{4:│}{1:~                            }|
-        {1:~                             }{4:│}{1:~                            }|
+        aaa                           │{2:aaa                          }|
+        bbb                           │bbb                          |
+        ccc                           │ccc                          |
+        {2:dd^d                           }│ddd                          |
+        {1:~                             }│{1:~                            }|
+        {1:~                             }│{1:~                            }|
         {3:[No Name] [+]  4,3         All }{4:[No Name] [+]  1,1        All}|
                                                                     |
       ]])
     end)
+
+    it('updates cursorcolumn in non-current window', function()
+      local screen = Screen.new(60, 8)
+      screen:set_default_attr_ids({
+        [1] = {bold = true, foreground = Screen.colors.Blue},  -- NonText
+        [2] = {background = Screen.colors.Grey90},  -- CursorColumn
+        [3] = {bold = true, reverse = true},  -- StatusLine
+        [4] = {reverse = true},  -- StatusLineNC
+      })
+      screen:attach()
+      command('set cursorcolumn')
+      insert([[
+        aaa
+        bbb
+        ccc
+        ddd]])
+      local oldwin = curwin()
+      command('vsplit')
+      screen:expect([[
+        aa{2:a}                           │aa{2:a}                          |
+        bb{2:b}                           │bb{2:b}                          |
+        cc{2:c}                           │cc{2:c}                          |
+        dd^d                           │ddd                          |
+        {1:~                             }│{1:~                            }|
+        {1:~                             }│{1:~                            }|
+        {3:[No Name] [+]                  }{4:[No Name] [+]                }|
+                                                                    |
+      ]])
+      window('set_cursor', oldwin, {2, 0})
+      screen:expect([[
+        aa{2:a}                           │{2:a}aa                          |
+        bb{2:b}                           │bbb                          |
+        cc{2:c}                           │{2:c}cc                          |
+        dd^d                           │{2:d}dd                          |
+        {1:~                             }│{1:~                            }|
+        {1:~                             }│{1:~                            }|
+        {3:[No Name] [+]                  }{4:[No Name] [+]                }|
+                                                                    |
+      ]])
+    end)
   end)
 
   describe('{get,set}_height', function()
@@ -201,6 +284,22 @@ describe('API/win', function()
       window('set_height', nvim('list_wins')[2], 2)
       eq(2, window('get_height', nvim('list_wins')[2]))
     end)
+
+    it('do not cause ml_get errors with foldmethod=expr #19989', function()
+      insert([[
+        aaaaa
+        bbbbb
+        ccccc]])
+      command('set foldmethod=expr')
+      exec([[
+        new
+        let w = nvim_get_current_win()
+        wincmd w
+        call nvim_win_set_height(w, 5)
+      ]])
+      feed('l')
+      eq('', meths.get_vvar('errmsg'))
+    end)
   end)
 
   describe('{get,set}_width', function()
@@ -215,6 +314,22 @@ describe('API/win', function()
       window('set_width', nvim('list_wins')[2], 2)
       eq(2, window('get_width', nvim('list_wins')[2]))
     end)
+
+    it('do not cause ml_get errors with foldmethod=expr #19989', function()
+      insert([[
+        aaaaa
+        bbbbb
+        ccccc]])
+      command('set foldmethod=expr')
+      exec([[
+        vnew
+        let w = nvim_get_current_win()
+        wincmd w
+        call nvim_win_set_width(w, 5)
+      ]])
+      feed('l')
+      eq('', meths.get_vvar('errmsg'))
+    end)
   end)
 
   describe('{get,set,del}_var', function()
@@ -408,6 +523,8 @@ describe('API/win', function()
 
     it('closing current (float) window of another tabpage #15313', function()
       command('tabedit')
+      command('botright split')
+      local prevwin = curwin().id
       eq(2, eval('tabpagenr()'))
       local win = meths.open_win(0, true, {
         relative='editor', row=10, col=10, width=50, height=10
@@ -417,7 +534,7 @@ describe('API/win', function()
       eq(1, eval('tabpagenr()'))
       meths.win_close(win, false)
 
-      eq(1001, meths.tabpage_get_win(tab).id)
+      eq(prevwin, meths.tabpage_get_win(tab).id)
       assert_alive()
     end)
   end)
