@@ -21,9 +21,9 @@ describe('TabNewEntered', function()
     end)
     describe('with FILE as <afile>', function()
       it('matches when opening a new tab for FILE', function()
+        clear()
         nvim('command', 'au! TabNewEntered Xtest-tabnewentered echom "tabnewentered:match"')
-        eq('tabnewentered:4:4\ntabnewentered:match',
-          nvim('exec', 'tabnew Xtest-tabnewentered', true))
+        eq('tabnewentered:match', nvim('exec', 'tabnew Xtest-tabnewentered', true))
      end)
     end)
     describe('with CTRL-W T', function()
