@@ -1,13 +1,19 @@
 local luv = require('luv')
 local helpers = require('test.functional.helpers')(after_each)
+local thelpers = require('test.functional.terminal.helpers')
 
 local clear, command, nvim, testprg =
   helpers.clear, helpers.command, helpers.nvim, helpers.testprg
 local eval, eq, neq, retry =
   helpers.eval, helpers.eq, helpers.neq, helpers.retry
+local matches = helpers.matches
 local ok = helpers.ok
 local feed = helpers.feed
-local iswin = helpers.iswin
+local meths = helpers.meths
+local pcall_err = helpers.pcall_err
+local assert_alive = helpers.assert_alive
+local skip = helpers.skip
+local is_os = helpers.is_os
 
 describe('autocmd TermClose', function()
   before_each(function()
@@ -16,6 +22,26 @@ describe('autocmd TermClose', function()
     command('set shellcmdflag=EXE shellredir= shellpipe= shellquote= shellxquote=')
   end)
 
+  local function test_termclose_delete_own_buf()
+    -- The terminal process needs to keep running so that TermClose isn't triggered immediately.
+    nvim('set_option', 'shell', string.format('"%s" INTERACT', testprg('shell-test')))
+    command('autocmd TermClose * bdelete!')
+    command('terminal')
+    matches('^TermClose Autocommands for "%*": Vim%(bdelete%):E937: Attempt to delete a buffer that is in use: term://',
+            pcall_err(command, 'bdelete!'))
+    assert_alive()
+  end
+
+  -- TODO: fixed after merging patches for `can_unload_buffer`?
+  pending('TermClose deleting its own buffer, altbuf = buffer 1 #10386', function()
+    test_termclose_delete_own_buf()
+  end)
+
+  it('TermClose deleting its own buffer, altbuf NOT buffer 1 #10386', function()
+    command('edit foo1')
+    test_termclose_delete_own_buf()
+  end)
+
   it('triggers when fast-exiting terminal job stops', function()
     command('autocmd TermClose * let g:test_termclose = 23')
     command('terminal')
@@ -25,7 +51,8 @@ describe('autocmd TermClose', function()
   end)
 
   it('triggers when long-running terminal job gets stopped', function()
-    nvim('set_option', 'shell', iswin() and 'cmd.exe' or 'sh')
+    skip(is_os('win'))
+    nvim('set_option', 'shell', is_os('win') and 'cmd.exe' or 'sh')
     command('autocmd TermClose * let g:test_termclose = 23')
     command('terminal')
     command('call jobstop(b:terminal_job_id)')
@@ -33,7 +60,7 @@ describe('autocmd TermClose', function()
   end)
 
   it('kills job trapping SIGTERM', function()
-    if iswin() then return end
+    skip(is_os('win'))
     nvim('set_option', 'shell', 'sh')
     nvim('set_option', 'shellcmdflag', '-c')
     command([[ let g:test_job = jobstart('trap "" TERM && echo 1 && sleep 60', { ]]
@@ -53,7 +80,7 @@ describe('autocmd TermClose', function()
   end)
 
   it('kills PTY job trapping SIGHUP and SIGTERM', function()
-    if iswin() then return end
+    skip(is_os('win'))
     nvim('set_option', 'shell', 'sh')
     nvim('set_option', 'shellcmdflag', '-c')
     command([[ let g:test_job = jobstart('trap "" HUP TERM && echo 1 && sleep 60', { ]]
@@ -120,19 +147,37 @@ it('autocmd TermEnter, TermLeave', function()
 
   -- TermLeave is also triggered by :quit.
   command('split foo')
+  feed('<Ignore>')  -- Add input to separate two RPC requests
   command('wincmd w')
   feed('i')
   command('q!')
-  eq(
-    {
-     {'TermOpen',  'n'},
-     {'TermEnter', 't'},
-     {'TermLeave', 'n'},
-     {'TermEnter', 't'},
-     {'TermLeave', 'n'},
-     {'TermEnter', 't'},
-     {'TermClose', 't'},
-     {'TermLeave', 'n'},
-    },
-    eval('g:evs'))
+  feed('<Ignore>')  -- Add input to separate two RPC requests
+  eq({
+    {'TermOpen',  'n'},
+    {'TermEnter', 't'},
+    {'TermLeave', 'n'},
+    {'TermEnter', 't'},
+    {'TermLeave', 'n'},
+    {'TermEnter', 't'},
+    {'TermClose', 't'},
+    {'TermLeave', 'n'},
+  }, eval('g:evs'))
+end)
+
+describe('autocmd TextChangedT', function()
+  clear()
+  local screen = thelpers.screen_setup()
+
+  it('works', function()
+    command('autocmd TextChangedT * ++once let g:called = 1')
+    thelpers.feed_data('a')
+    retry(nil, nil, function() eq(1, meths.get_var('called')) end)
+  end)
+
+  it('cannot delete terminal buffer', function()
+    command([[autocmd TextChangedT * call nvim_input('<CR>') | bwipe!]])
+    thelpers.feed_data('a')
+    screen:expect({any = 'E937: '})
+    matches('^E937: Attempt to delete a buffer that is in use: term://', meths.get_vvar('errmsg'))
+  end)
 end)
