@@ -13,7 +13,6 @@ local retry = helpers.retry
 local meths = helpers.meths
 local NIL = helpers.NIL
 local poke_eventloop = helpers.poke_eventloop
-local iswin = helpers.iswin
 local get_pathsep = helpers.get_pathsep
 local pathroot = helpers.pathroot
 local exec_lua = helpers.exec_lua
@@ -21,7 +20,10 @@ local nvim_set = helpers.nvim_set
 local expect_twostreams = helpers.expect_twostreams
 local expect_msg_seq = helpers.expect_msg_seq
 local pcall_err = helpers.pcall_err
+local matches = helpers.matches
 local Screen = require('test.functional.ui.screen')
+local skip = helpers.skip
+local is_os = helpers.is_os
 
 describe('jobs', function()
   local channel
@@ -54,7 +56,7 @@ describe('jobs', function()
   it('must specify env option as a dict', function()
     command("let g:job_opts.env = v:true")
     local _, err = pcall(function()
-      if iswin() then
+      if is_os('win') then
         nvim('command', "let j = jobstart('set', g:job_opts)")
       else
         nvim('command', "let j = jobstart('env', g:job_opts)")
@@ -67,31 +69,44 @@ describe('jobs', function()
     nvim('command', "let $VAR = 'abc'")
     nvim('command', "let $TOTO = 'goodbye world'")
     nvim('command', "let g:job_opts.env = {'TOTO': 'hello world'}")
-    if iswin() then
+    if is_os('win') then
       nvim('command', [[call jobstart('echo %TOTO% %VAR%', g:job_opts)]])
     else
       nvim('command', [[call jobstart('echo $TOTO $VAR', g:job_opts)]])
     end
 
-    expect_msg_seq({
-      {'notification', 'stdout', {0, {'hello world abc', ''}}},
-    })
+    expect_msg_seq(
+      {
+        {'notification', 'stdout', {0, {'hello world abc'}}},
+        {'notification', 'stdout', {0, {'', ''}}},
+      },
+      {
+        {'notification', 'stdout', {0, {'hello world abc', ''}}},
+        {'notification', 'stdout', {0, {''}}}
+      }
+    )
   end)
 
   it('append environment with pty #env', function()
-    if helpers.pending_win32(pending) then return end
     nvim('command', "let $VAR = 'abc'")
     nvim('command', "let $TOTO = 'goodbye world'")
     nvim('command', "let g:job_opts.pty = v:true")
     nvim('command', "let g:job_opts.env = {'TOTO': 'hello world'}")
-    if iswin() then
+    if is_os('win') then
       nvim('command', [[call jobstart('echo %TOTO% %VAR%', g:job_opts)]])
     else
       nvim('command', [[call jobstart('echo $TOTO $VAR', g:job_opts)]])
     end
-    expect_msg_seq({
-      {'notification', 'stdout', {0, {'hello world abc', ''}}},
-    })
+    expect_msg_seq(
+      {
+        {'notification', 'stdout', {0, {'hello world abc'}}},
+        {'notification', 'stdout', {0, {'', ''}}},
+      },
+      {
+        {'notification', 'stdout', {0, {'hello world abc', ''}}},
+        {'notification', 'stdout', {0, {''}}}
+      }
+    )
   end)
 
   it('replace environment #env', function()
@@ -107,7 +122,7 @@ describe('jobs', function()
     --
     -- Rather than expecting a completely empty environment, ensure that $VAR
     -- is *not* in the environment but $TOTO is.
-    if iswin() then
+    if is_os('win') then
       nvim('command', [[call jobstart('echo %TOTO% %VAR%', g:job_opts)]])
       expect_msg_seq({
         {'notification', 'stdout', {0, {'hello world %VAR%', ''}}}
@@ -126,7 +141,7 @@ describe('jobs', function()
     -- Since $Toto is being set in the job, it should take precedence over the
     -- global $TOTO on Windows
     nvim('command', "let g:job_opts = {'env': {'Toto': 'def'}, 'stdout_buffered': v:true}")
-    if iswin() then
+    if is_os('win') then
       nvim('command', [[let j = jobstart('set | find /I "toto="', g:job_opts)]])
     else
       nvim('command', [[let j = jobstart('env | grep -i toto=', g:job_opts)]])
@@ -135,7 +150,7 @@ describe('jobs', function()
     nvim('command', "let g:output = Normalize(g:job_opts.stdout)")
     local actual = eval('g:output')
     local expected
-    if iswin() then
+    if is_os('win') then
       -- Toto is normalized to TOTO so we can detect duplicates, and because
       -- Windows doesn't care about case
       expected = {'TOTO=def', ''}
@@ -149,7 +164,7 @@ describe('jobs', function()
 
   it('uses &shell and &shellcmdflag if passed a string', function()
     nvim('command', "let $VAR = 'abc'")
-    if iswin() then
+    if is_os('win') then
       nvim('command', "let j = jobstart('echo %VAR%', g:job_opts)")
     else
       nvim('command', "let j = jobstart('echo $VAR', g:job_opts)")
@@ -161,7 +176,7 @@ describe('jobs', function()
 
   it('changes to given / directory', function()
     nvim('command', "let g:job_opts.cwd = '/'")
-    if iswin() then
+    if is_os('win') then
       nvim('command', "let j = jobstart('cd', g:job_opts)")
     else
       nvim('command', "let j = jobstart('pwd', g:job_opts)")
@@ -176,7 +191,7 @@ describe('jobs', function()
     local dir = eval("resolve(tempname())"):gsub("/", get_pathsep())
     mkdir(dir)
     nvim('command', "let g:job_opts.cwd = '" .. dir .. "'")
-    if iswin() then
+    if is_os('win') then
       nvim('command', "let j = jobstart('cd', g:job_opts)")
     else
       nvim('command', "let j = jobstart('pwd', g:job_opts)")
@@ -200,7 +215,7 @@ describe('jobs', function()
     local dir = eval('resolve(tempname())."-bogus"')
     local _, err = pcall(function()
       nvim('command', "let g:job_opts.cwd = '" .. dir .. "'")
-      if iswin() then
+      if is_os('win') then
         nvim('command', "let j = jobstart('cd', g:job_opts)")
       else
         nvim('command', "let j = jobstart('pwd', g:job_opts)")
@@ -210,13 +225,13 @@ describe('jobs', function()
   end)
 
   it('error on non-executable `cwd`', function()
-    if iswin() then return end  -- N/A for Windows
+    skip(is_os('win'), 'Not applicable for Windows')
 
     local dir = 'Xtest_not_executable_dir'
     mkdir(dir)
     funcs.setfperm(dir, 'rw-------')
-    eq('Vim(call):E475: Invalid argument: expected valid directory',
-      pcall_err(nvim, 'command', "call jobstart('pwd', {'cwd': '"..dir.."'})"))
+    matches('^Vim%(call%):E903: Process failed to start: permission denied: .*',
+            pcall_err(nvim, 'command', "call jobstart(['pwd'], {'cwd': '"..dir.."'})"))
     rmdir(dir)
   end)
 
@@ -233,7 +248,7 @@ describe('jobs', function()
     end
     local executable_jobid = new_job()
 
-    local exe = iswin() and './test/functional/fixtures' or './test/functional/fixtures/non_executable.txt'
+    local exe = is_os('win') and './test/functional/fixtures' or './test/functional/fixtures/non_executable.txt'
     eq("Vim:E475: Invalid value for argument cmd: '"..exe.."' is not executable",
       pcall_err(eval, "jobstart(['"..exe.."'])"))
     eq("", eval("v:errmsg"))
@@ -676,17 +691,18 @@ describe('jobs', function()
     -- jobstart() shares its v:servername with the child via $NVIM.
     eq('NVIM='..addr, get_env_in_child_job('NVIM'))
     -- $NVIM_LISTEN_ADDRESS is unset by server_init in the child.
-    eq('NVIM_LISTEN_ADDRESS=null', get_env_in_child_job('NVIM_LISTEN_ADDRESS'))
-    eq('NVIM_LISTEN_ADDRESS=null', get_env_in_child_job('NVIM_LISTEN_ADDRESS',
+    eq('NVIM_LISTEN_ADDRESS=v:null', get_env_in_child_job('NVIM_LISTEN_ADDRESS'))
+    eq('NVIM_LISTEN_ADDRESS=v:null', get_env_in_child_job('NVIM_LISTEN_ADDRESS',
       { NVIM_LISTEN_ADDRESS='Xtest_jobstart_env' }))
     -- User can explicitly set $NVIM_LOG_FILE, $VIM, $VIMRUNTIME.
     eq('NVIM_LOG_FILE=Xtest_jobstart_env',
       get_env_in_child_job('NVIM_LOG_FILE', { NVIM_LOG_FILE='Xtest_jobstart_env' }))
+    os.remove('Xtest_jobstart_env')
   end)
 
   describe('jobwait', function()
     before_each(function()
-      if iswin() then
+      if is_os('win') then
         helpers.set_shell_powershell()
       end
     end)
@@ -770,7 +786,7 @@ describe('jobs', function()
       feed_command('call rpcnotify(g:channel, "ready") | '..
               'call rpcnotify(g:channel, "wait", '..
               'jobwait([jobstart("'..
-              (iswin() and 'Start-Sleep 10' or 'sleep 10')..
+              (is_os('win') and 'Start-Sleep 10' or 'sleep 10')..
               '; exit 55")]))')
       eq({'notification', 'ready', {}}, next_msg())
       feed('<c-c>')
@@ -781,7 +797,7 @@ describe('jobs', function()
       feed_command('call rpcnotify(g:channel, "ready") | '..
               'call rpcnotify(g:channel, "wait", '..
               'jobwait([jobstart("'..
-              (iswin() and 'Start-Sleep 10' or 'sleep 10')..
+              (is_os('win') and 'Start-Sleep 10' or 'sleep 10')..
               '; exit 55")], 10000))')
       eq({'notification', 'ready', {}}, next_msg())
       feed('<c-c>')
@@ -789,7 +805,6 @@ describe('jobs', function()
     end)
 
     it('can be called recursively', function()
-      if helpers.pending_win32(pending) then return end  -- TODO: Need `cat`.
       source([[
       let g:opts = {}
       let g:counter = 0
@@ -914,7 +929,7 @@ describe('jobs', function()
     --                ..c.."', '-c', '"..c.."'])")
 
     -- Create child with several descendants.
-    if iswin() then
+    if is_os('win') then
       source([[
       function! s:formatprocs(pid, prefix)
         let result = ''
@@ -963,13 +978,13 @@ describe('jobs', function()
       endfunction
       ]])
     end
-    local sleep_cmd = (iswin()
+    local sleep_cmd = (is_os('win')
       and 'ping -n 31 127.0.0.1'
       or  'sleep 30')
     local j = eval("jobstart('"..sleep_cmd..' | '..sleep_cmd..' | '..sleep_cmd.."')")
     local ppid = funcs.jobpid(j)
     local children
-    if iswin() then
+    if is_os('win') then
       local status, result = pcall(retry, nil, nil, function()
         children = meths.get_proc_children(ppid)
         -- On Windows conhost.exe may exist, and
@@ -990,7 +1005,7 @@ describe('jobs', function()
     -- Assert that nvim_get_proc() sees the children.
     for _, child_pid in ipairs(children) do
       local info = meths.get_proc(child_pid)
-      -- eq((iswin() and 'nvim.exe' or 'nvim'), info.name)
+      -- eq((is_os('win') and 'nvim.exe' or 'nvim'), info.name)
       eq(ppid, info.ppid)
     end
     -- Kill the root of the tree.
@@ -1011,7 +1026,7 @@ describe('jobs', function()
   end)
 
   describe('running tty-test program', function()
-    if helpers.pending_win32(pending) then return end
+    if skip(is_os('win')) then return end
     local function next_chunk()
       local rv
       while true do
@@ -1108,7 +1123,7 @@ describe("pty process teardown", function()
   end)
 
   it("does not prevent/delay exit. #4798 #4900", function()
-    if helpers.pending_win32(pending) then return end
+    skip(is_os('win'))
     -- Use a nested nvim (in :term) to test without --headless.
     feed_command(":terminal '"..helpers.nvim_prog
       .."' -u NONE -i NONE --cmd '"..nvim_set.."' "
