@@ -2,15 +2,19 @@ local helpers = require('test.functional.helpers')(after_each)
 local Screen = require('test.functional.ui.screen')
 
 local assert_alive = helpers.assert_alive
+local assert_log = helpers.assert_log
 local clear = helpers.clear
 local command = helpers.command
 local ok = helpers.ok
 local eq = helpers.eq
 local matches = helpers.matches
 local eval = helpers.eval
+local exec = helpers.exec
+local exec_capture = helpers.exec_capture
 local exec_lua = helpers.exec_lua
 local feed = helpers.feed
 local funcs = helpers.funcs
+local pesc = helpers.pesc
 local mkdir = helpers.mkdir
 local mkdir_p = helpers.mkdir_p
 local nvim_prog = helpers.nvim_prog
@@ -19,18 +23,216 @@ local read_file = helpers.read_file
 local retry = helpers.retry
 local rmdir = helpers.rmdir
 local sleep = helpers.sleep
-local iswin = helpers.iswin
 local startswith = helpers.startswith
 local write_file = helpers.write_file
 local meths = helpers.meths
+local alter_slashes = helpers.alter_slashes
+local is_os = helpers.is_os
+local dedent = helpers.dedent
 
 describe('startup', function()
-  before_each(function()
+  it('--clean', function()
     clear()
-    os.remove('Xtest_startup_ttyout')
+    ok(string.find(alter_slashes(meths.get_option('runtimepath')), funcs.stdpath('config'), 1, true) ~= nil)
+    clear('--clean')
+    ok(string.find(alter_slashes(meths.get_option('runtimepath')), funcs.stdpath('config'), 1, true) == nil)
   end)
-  after_each(function()
-    os.remove('Xtest_startup_ttyout')
+
+  it('--startuptime', function()
+    local testfile = 'Xtest_startuptime'
+    finally(function()
+      os.remove(testfile)
+    end)
+    clear({ args = {'--startuptime', testfile}})
+    assert_log('sourcing', testfile, 100)
+    assert_log("require%('vim%._editor'%)", testfile, 100)
+  end)
+
+  it('-D does not hang #12647', function()
+    clear()
+    local screen
+    screen = Screen.new(60, 7)
+    screen:attach()
+    command([[let g:id = termopen('"]]..nvim_prog..
+    [[" -u NONE -i NONE --cmd "set noruler" -D')]])
+    screen:expect([[
+      ^                                                            |
+                                                                  |
+      Entering Debug mode.  Type "cont" to continue.              |
+      nvim_exec2()                                                |
+      cmd: aunmenu *                                              |
+      >                                                           |
+                                                                  |
+    ]])
+    command([[call chansend(g:id, "cont\n")]])
+    screen:expect([[
+      ^                                                            |
+      ~                                                           |
+      ~                                                           |
+      ~                                                           |
+      [No Name]                                                   |
+                                                                  |
+                                                                  |
+    ]])
+  end)
+end)
+
+describe('startup', function()
+  before_each(clear)
+
+  describe('-l Lua', function()
+    local function assert_l_out(expected, nvim_args, lua_args, script, input)
+      local args = { nvim_prog }
+      vim.list_extend(args, nvim_args or {})
+      vim.list_extend(args, { '-l', (script or 'test/functional/fixtures/startup.lua') })
+      vim.list_extend(args, lua_args or {})
+      local out = funcs.system(args, input):gsub('\r\n', '\n')
+      return eq(dedent(expected), out)
+    end
+
+    it('failure modes', function()
+      -- nvim -l <empty>
+      matches('nvim%.?e?x?e?: Argument missing after: "%-l"', funcs.system({ nvim_prog, '-l' }))
+      eq(1, eval('v:shell_error'))
+    end)
+
+    it('os.exit() sets Nvim exitcode', function()
+      -- tricky: LeakSanitizer triggers on os.exit() and disrupts the return value, disable it
+      exec_lua [[
+        local asan_options = os.getenv('ASAN_OPTIONS') or ''
+        if asan_options ~= '' then
+          asan_options = asan_options .. ':'
+        end
+        vim.loop.os_setenv('ASAN_OPTIONS', asan_options .. ':detect_leaks=0')
+      ]]
+      -- nvim -l foo.lua -arg1 -- a b c
+      assert_l_out([[
+          bufs:
+          nvim args: 7
+          lua args: { "-arg1", "--exitcode", "73", "--arg2",
+            [0] = "test/functional/fixtures/startup.lua"
+          }]],
+        {},
+        { '-arg1', "--exitcode", "73", '--arg2' }
+      )
+      eq(73, eval('v:shell_error'))
+    end)
+
+    it('Lua-error sets Nvim exitcode', function()
+      eq(0, eval('v:shell_error'))
+      matches('E5113: .* my pearls!!',
+        funcs.system({ nvim_prog, '-l', 'test/functional/fixtures/startup-fail.lua' }))
+      eq(1, eval('v:shell_error'))
+      matches('E5113: .* %[string "error%("whoa"%)"%]:1: whoa',
+        funcs.system({ nvim_prog, '-l', '-' }, 'error("whoa")'))
+      eq(1, eval('v:shell_error'))
+    end)
+
+    it('executes stdin "-"', function()
+      assert_l_out('arg0=- args=2 whoa',
+        nil,
+        { 'arg1', 'arg 2' },
+        '-',
+        "print(('arg0=%s args=%d %s'):format(_G.arg[0], #_G.arg, 'whoa'))")
+      assert_l_out('biiig input: 1000042',
+        nil,
+        nil,
+        '-',
+        ('print("biiig input: "..("%s"):len())'):format(string.rep('x', (1000 * 1000) + 42)))
+      eq(0, eval('v:shell_error'))
+    end)
+
+    it('does not truncate long print() message', function()
+      assert_l_out(('k'):rep(1234),
+        nil,
+        nil,
+        '-',
+        "print(('k'):rep(1234))")
+    end)
+
+    it('sets _G.arg', function()
+      -- nvim -l foo.lua
+      assert_l_out([[
+          bufs:
+          nvim args: 3
+          lua args: {
+            [0] = "test/functional/fixtures/startup.lua"
+          }]],
+        {},
+        {}
+      )
+      eq(0, eval('v:shell_error'))
+
+      -- nvim -l foo.lua [args]
+      assert_l_out([[
+          bufs:
+          nvim args: 7
+          lua args: { "-arg1", "--arg2", "--", "arg3",
+            [0] = "test/functional/fixtures/startup.lua"
+          }]],
+        {},
+        { '-arg1', '--arg2', '--', 'arg3' }
+      )
+      eq(0, eval('v:shell_error'))
+
+      -- nvim file1 file2 -l foo.lua -arg1 -- file3 file4
+      assert_l_out([[
+          bufs: file1 file2
+          nvim args: 10
+          lua args: { "-arg1", "arg 2", "--", "file3", "file4",
+            [0] = "test/functional/fixtures/startup.lua"
+          }]],
+        { 'file1', 'file2', },
+        { '-arg1', 'arg 2', '--', 'file3', 'file4' }
+      )
+      eq(0, eval('v:shell_error'))
+
+      -- nvim -l foo.lua <vim args>
+      assert_l_out([[
+          bufs:
+          nvim args: 5
+          lua args: { "-c", "set wrap?",
+            [0] = "test/functional/fixtures/startup.lua"
+          }]],
+        {},
+        { '-c', 'set wrap?' }
+      )
+      eq(0, eval('v:shell_error'))
+
+      -- nvim <vim args> -l foo.lua <vim args>
+      assert_l_out(
+        -- luacheck: ignore 611 (Line contains only whitespaces)
+        [[
+            wrap
+          
+          bufs:
+          nvim args: 7
+          lua args: { "-c", "set wrap?",
+            [0] = "test/functional/fixtures/startup.lua"
+          }]],
+        { '-c', 'set wrap?' },
+        { '-c', 'set wrap?' }
+      )
+      eq(0, eval('v:shell_error'))
+    end)
+
+    it('disables swapfile/shada/config/plugins', function()
+      assert_l_out('updatecount=0 shadafile=NONE loadplugins=false scriptnames=1',
+        nil,
+        nil,
+        '-',
+        [[print(('updatecount=%d shadafile=%s loadplugins=%s scriptnames=%d'):format(
+          vim.o.updatecount, vim.o.shadafile, tostring(vim.o.loadplugins), math.max(1, #vim.fn.split(vim.fn.execute('scriptnames'),'\n'))))]])
+    end)
+  end)
+
+  it('--cmd/-c/+ do not truncate long Lua print() message with --headless', function()
+    local out = funcs.system({ nvim_prog, '-u', 'NONE', '-i', 'NONE', '--headless',
+                               '--cmd', 'lua print(("A"):rep(1234))',
+                               '-c', 'lua print(("B"):rep(1234))',
+                               '+lua print(("C"):rep(1234))',
+                               '+q' })
+    eq(('A'):rep(1234) .. '\r\n' .. ('B'):rep(1234) .. '\r\n' .. ('C'):rep(1234), out)
   end)
 
   it('pipe at both ends: has("ttyin")==0 has("ttyout")==0', function()
@@ -41,6 +243,7 @@ describe('startup', function()
                                '+q' })
     eq('0 0', out)
   end)
+
   it('with --embed: has("ttyin")==0 has("ttyout")==0', function()
     local screen = Screen.new(25, 3)
     -- Remote UI connected by --embed.
@@ -52,11 +255,11 @@ describe('startup', function()
       0 0                      |
     ]])
   end)
+
   it('in a TTY: has("ttyin")==1 has("ttyout")==1', function()
-    if helpers.pending_win32(pending) then return end
     local screen = Screen.new(25, 4)
     screen:attach()
-    if iswin() then
+    if is_os('win') then
       command([[set shellcmdflag=/s\ /c shellxquote=\"]])
     end
     -- Running in :terminal
@@ -71,10 +274,15 @@ describe('startup', function()
                                |
     ]])
   end)
+
   it('output to pipe: has("ttyin")==1 has("ttyout")==0', function()
-    if iswin() then
+    if is_os('win') then
       command([[set shellcmdflag=/s\ /c shellxquote=\"]])
     end
+    os.remove('Xtest_startup_ttyout')
+    finally(function()
+      os.remove('Xtest_startup_ttyout')
+    end)
     -- Running in :terminal
     command([[exe printf("terminal %s -u NONE -i NONE --cmd \"]]
             ..nvim_set..[[\"]]
@@ -87,10 +295,15 @@ describe('startup', function()
          read_file('Xtest_startup_ttyout'))
     end)
   end)
+
   it('input from pipe: has("ttyin")==0 has("ttyout")==1', function()
-    if iswin() then
+    if is_os('win') then
       command([[set shellcmdflag=/s\ /c shellxquote=\"]])
     end
+    os.remove('Xtest_startup_ttyout')
+    finally(function()
+      os.remove('Xtest_startup_ttyout')
+    end)
     -- Running in :terminal
     command([[exe printf("terminal echo foo | ]]  -- Input from a pipe.
             ..[[%s -u NONE -i NONE --cmd \"]]
@@ -104,11 +317,11 @@ describe('startup', function()
          read_file('Xtest_startup_ttyout'))
     end)
   end)
+
   it('input from pipe (implicit) #7679', function()
-    if helpers.pending_win32(pending) then return end
     local screen = Screen.new(25, 4)
     screen:attach()
-    if iswin() then
+    if is_os('win') then
       command([[set shellcmdflag=/s\ /c shellxquote=\"]])
     end
     -- Running in :terminal
@@ -124,6 +337,7 @@ describe('startup', function()
                                |
     ]])
   end)
+
   it('input from pipe + file args #7679', function()
     eq('ohyeah\r\n0 0 bufs=3',
        funcs.system({nvim_prog, '-n', '-u', 'NONE', '-i', 'NONE', '--headless',
@@ -149,28 +363,6 @@ describe('startup', function()
                     { '' }))
   end)
 
-  it('-e/-E interactive #7679', function()
-    clear('-e')
-    local screen = Screen.new(25, 3)
-    screen:attach()
-    feed("put ='from -e'<CR>")
-    screen:expect([[
-      :put ='from -e'          |
-      from -e                  |
-      :^                        |
-    ]])
-
-    clear('-E')
-    screen = Screen.new(25, 3)
-    screen:attach()
-    feed("put ='from -E'<CR>")
-    screen:expect([[
-      :put ='from -E'          |
-      from -E                  |
-      :^                        |
-    ]])
-  end)
-
   it('stdin with -es/-Es #7679', function()
     local input = { 'append', 'line1', 'line2', '.', '%print', '' }
     local inputstr = table.concat(input, '\n')
@@ -215,53 +407,27 @@ describe('startup', function()
   it('-es/-Es disables swapfile, user config #8540', function()
     for _,arg in ipairs({'-es', '-Es'}) do
       local out = funcs.system({nvim_prog, arg,
-                                '+set swapfile? updatecount? shada?',
+                                '+set swapfile? updatecount? shadafile?',
                                 "+put =execute('scriptnames')", '+%print'})
       local line1 = string.match(out, '^.-\n')
       -- updatecount=0 means swapfile was disabled.
-      eq("  swapfile  updatecount=0  shada=!,'100,<50,s10,h\n", line1)
+      eq("  swapfile  updatecount=0  shadafile=\n", line1)
       -- Standard plugins were loaded, but not user config.
       eq('health.vim', string.match(out, 'health.vim'))
       eq(nil, string.match(out, 'init.vim'))
     end
   end)
 
-  it('-e sets ex mode', function()
-    local screen = Screen.new(25, 3)
-    clear('-e')
-    screen:attach()
-    -- Verify we set the proper mode both before and after :vi.
-    feed("put =mode(1)<CR>vi<CR>:put =mode(1)<CR>")
-    screen:expect([[
-      cv                       |
-      ^n                        |
-      :put =mode(1)            |
-    ]])
-
-    eq('cv\n',
-       funcs.system({nvim_prog, '-n', '-es' },
-                    { 'put =mode(1)', 'print', '' }))
-  end)
-
-  it('fails on --embed with -es/-Es', function()
-    matches('nvim[.exe]*: %-%-embed conflicts with %-es/%-Es',
+  it('fails on --embed with -es/-Es/-l', function()
+    matches('nvim[.exe]*: %-%-embed conflicts with %-es/%-Es/%-l',
       funcs.system({nvim_prog, '--embed', '-es' }))
-    matches('nvim[.exe]*: %-%-embed conflicts with %-es/%-Es',
+    matches('nvim[.exe]*: %-%-embed conflicts with %-es/%-Es/%-l',
       funcs.system({nvim_prog, '--embed', '-Es' }))
-  end)
-
-  it('does not crash if --embed is given twice', function()
-    clear{args={'--embed'}}
-    assert_alive()
-  end)
-
-  it('does not crash when expanding cdpath during early_init', function()
-    clear{env={CDPATH='~doesnotexist'}}
-    eq(',~doesnotexist', eval('&cdpath'))
+    matches('nvim[.exe]*: %-%-embed conflicts with %-es/%-Es/%-l',
+      funcs.system({nvim_prog, '--embed', '-l', 'foo.lua' }))
   end)
 
   it('ENTER dismisses early message #7967', function()
-    if helpers.pending_win32(pending) then return end
     local screen
     screen = Screen.new(60, 6)
     screen:attach()
@@ -286,22 +452,24 @@ describe('startup', function()
     ]])
   end)
 
-  it("sets 'shortmess' when loading other tabs", function()
-    clear({args={'-p', 'a', 'b', 'c'}})
-    local screen = Screen.new(25, 4)
-    screen:attach()
-    screen:expect({grid=[[
-        {1: a }{2: b  c }{3:               }{2:X}|
-        ^                         |
-        {4:~                        }|
-                                 |
-          ]],
-      attr_ids={
-        [1] = {bold = true},
-        [2] = {background = Screen.colors.LightGrey, underline = true},
-        [3] = {reverse = true},
-        [4] = {bold = true, foreground = Screen.colors.Blue1},
-    }})
+  it('-r works without --headless in PTY #23294', function()
+    exec([[
+      func Normalize(data) abort
+        " Windows: remove ^M and term escape sequences
+        return map(a:data, 'substitute(substitute(v:val, "\r", "", "g"), "\x1b\\%(\\]\\d\\+;.\\{-}\x07\\|\\[.\\{-}[\x40-\x7E]\\)", "", "g")')
+      endfunc
+      func OnOutput(id, data, event) dict
+        let g:stdout = Normalize(a:data)
+      endfunc
+      call jobstart([v:progpath, '-u', 'NONE', '-i', 'NONE', '-r'], {
+      \ 'pty': v:true,
+      \ 'stdout_buffered': v:true,
+      \ 'on_stdout': function('OnOutput'),
+      \ })
+    ]])
+    retry(nil, nil, function()
+      eq('Swap files found:', eval('g:stdout[0]'))
+    end)
   end)
 
   it('fixed hang issue with --headless (#11386)', function()
@@ -329,13 +497,109 @@ describe('startup', function()
                                '+q' })
     eq('[\'+q\'] 1', out)
   end)
+end)
 
+describe('startup', function()
+  it('-e/-E interactive #7679', function()
+    clear('-e')
+    local screen = Screen.new(25, 3)
+    screen:attach()
+    feed("put ='from -e'<CR>")
+    screen:expect([[
+      :put ='from -e'          |
+      from -e                  |
+      :^                        |
+    ]])
+
+    clear('-E')
+    screen = Screen.new(25, 3)
+    screen:attach()
+    feed("put ='from -E'<CR>")
+    screen:expect([[
+      :put ='from -E'          |
+      from -E                  |
+      :^                        |
+    ]])
+  end)
+
+  it('-e sets ex mode', function()
+    local screen = Screen.new(25, 3)
+    clear('-e')
+    screen:attach()
+    -- Verify we set the proper mode both before and after :vi.
+    feed("put =mode(1)<CR>vi<CR>:put =mode(1)<CR>")
+    screen:expect([[
+      cv                       |
+      ^n                        |
+      :put =mode(1)            |
+    ]])
+
+    eq('cv\n',
+       funcs.system({nvim_prog, '-n', '-es' },
+                    { 'put =mode(1)', 'print', '' }))
+  end)
+
+  it('-d does not diff non-arglist windows #13720 #21289', function()
+    write_file('Xdiff.vim', [[
+      let bufnr = nvim_create_buf(0, 1)
+      let config = {
+            \   'relative': 'editor',
+            \   'focusable': v:false,
+            \   'width': 1,
+            \   'height': 1,
+            \   'row': 3,
+            \   'col': 3
+            \ }
+      autocmd WinEnter * call nvim_open_win(bufnr, v:false, config)]])
+    finally(function()
+      os.remove('Xdiff.vim')
+    end)
+    clear{args={'-u', 'Xdiff.vim', '-d', 'Xdiff.vim', 'Xdiff.vim'}}
+    eq(true, meths.get_option_value('diff', {win = funcs.win_getid(1)}))
+    eq(true, meths.get_option_value('diff', {win = funcs.win_getid(2)}))
+    local float_win = funcs.win_getid(3)
+    eq('editor', meths.win_get_config(float_win).relative)
+    eq(false, meths.get_option_value('diff', {win = float_win}))
+  end)
+
+  it('does not crash if --embed is given twice', function()
+    clear{args={'--embed'}}
+    assert_alive()
+  end)
+
+  it('does not crash when expanding cdpath during early_init', function()
+    clear{env={CDPATH='~doesnotexist'}}
+    assert_alive()
+    eq(',~doesnotexist', eval('&cdpath'))
+  end)
+
+  it("sets 'shortmess' when loading other tabs", function()
+    clear({args={'-p', 'a', 'b', 'c'}})
+    local screen = Screen.new(25, 4)
+    screen:attach()
+    screen:expect({grid=[[
+        {1: a }{2: b  c }{3:               }{2:X}|
+        ^                         |
+        {4:~                        }|
+                                 |
+          ]],
+      attr_ids={
+        [1] = {bold = true},
+        [2] = {background = Screen.colors.LightGrey, underline = true},
+        [3] = {reverse = true},
+        [4] = {bold = true, foreground = Screen.colors.Blue1},
+    }})
+  end)
+end)
+
+describe('startup', function()
   local function pack_clear(cmd)
     -- add packages after config dir in rtp but before config/after
-    clear{args={'--cmd', 'set packpath=test/functional/fixtures', '--cmd', 'let paths=split(&rtp, ",")', '--cmd', 'let &rtp = paths[0]..",test/functional/fixtures,test/functional/fixtures/middle,"..join(paths[1:],",")', '--cmd', cmd}, env={XDG_CONFIG_HOME='test/functional/fixtures/'}}
+    clear{args={'--cmd', 'set packpath=test/functional/fixtures', '--cmd', 'let paths=split(&rtp, ",")', '--cmd', 'let &rtp = paths[0]..",test/functional/fixtures,test/functional/fixtures/middle,"..join(paths[1:],",")', '--cmd', cmd}, env={XDG_CONFIG_HOME='test/functional/fixtures/'},
+          args_rm={'runtimepath'},
+    }
   end
 
-
   it("handles &packpath during startup", function()
     pack_clear [[
       let g:x = bar#test()
@@ -401,7 +665,8 @@ describe('startup', function()
     eq({'ordinary', 'FANCY', 'mittel', 'FANCY after', 'ordinary after'}, exec_lua [[ return _G.test_loadorder ]])
 
     local rtp = meths.get_option'rtp'
-    ok(startswith(rtp, 'test/functional/fixtures/nvim,test/functional/fixtures/pack/*/start/*,test/functional/fixtures/start/*,test/functional/fixtures,test/functional/fixtures/middle,'), 'rtp='..rtp)
+    ok(startswith(rtp, 'test/functional/fixtures/nvim,test/functional/fixtures/pack/*/start/*,test/functional/fixtures/start/*,test/functional/fixtures,test/functional/fixtures/middle,'),
+      'startswith(…)', 'rtp='..rtp)
   end)
 
   it("handles the correct order with opt packages and after/", function()
@@ -439,6 +704,19 @@ describe('startup', function()
     clear{args={'--cmd', 'set packpath^=test/functional/fixtures',  '--cmd', [[ lua _G.test_loadorder = {} vim.cmd "runtime! filen.lua" ]]}, env={XDG_CONFIG_HOME='test/functional/fixtures/'}}
     eq({'ordinary', 'FANCY', 'FANCY after', 'ordinary after'}, exec_lua [[ return _G.test_loadorder ]])
   end)
+
+  it('window widths are correct when modelines set &columns with tabpages', function()
+    write_file('Xtab1.noft', 'vim: columns=81')
+    write_file('Xtab2.noft', 'vim: columns=81')
+    finally(function()
+      os.remove('Xtab1.noft')
+      os.remove('Xtab2.noft')
+    end)
+    clear({args = {'-p', 'Xtab1.noft', 'Xtab2.noft'}})
+    eq(81, meths.win_get_width(0))
+    command('tabnext')
+    eq(81, meths.win_get_width(0))
+  end)
 end)
 
 describe('sysinit', function()
@@ -492,39 +770,6 @@ describe('sysinit', function()
     eq('loaded 1 xdg 0 vim 1',
        eval('printf("loaded %d xdg %d vim %d", g:loaded, get(g:, "xdg", 0), get(g:, "vim", 0))'))
   end)
-
-  it('fixed hang issue with -D (#12647)', function()
-    if helpers.pending_win32(pending) then return end
-    local screen
-    screen = Screen.new(60, 6)
-    screen:attach()
-    command([[let g:id = termopen('"]]..nvim_prog..
-    [[" -u NONE -i NONE --cmd "set noruler" -D')]])
-    screen:expect([[
-      ^                                                            |
-      Entering Debug mode.  Type "cont" to continue.              |
-      cmd: augroup nvim_terminal                                  |
-      >                                                           |
-      <" -u NONE -i NONE --cmd "set noruler" -D 1,0-1          All|
-                                                                  |
-    ]])
-    command([[call chansend(g:id, "cont\n")]])
-    screen:expect([[
-      ^                                                            |
-      ~                                                           |
-      [No Name]                                                   |
-                                                                  |
-      <" -u NONE -i NONE --cmd "set noruler" -D 1,0-1          All|
-                                                                  |
-    ]])
-  end)
-end)
-
-describe('clean', function()
-  clear()
-  ok(string.find(meths.get_option('runtimepath'), funcs.stdpath('config'), 1, true) ~= nil)
-  clear('--clean')
-  ok(string.find(meths.get_option('runtimepath'), funcs.stdpath('config'), 1, true) == nil)
 end)
 
 describe('user config init', function()
@@ -551,13 +796,88 @@ describe('user config init', function()
   end)
 
   it('loads init.lua from XDG config home by default', function()
-    clear{ args_rm={'-u' }, env=xenv }
+    clear{ args_rm={'-u'}, env=xenv }
 
     eq(1, eval('g:lua_rc'))
     eq(funcs.fnamemodify(init_lua_path, ':p'), eval('$MYVIMRC'))
   end)
 
-  describe 'with explicitly provided config'(function()
+  describe('loads existing', function()
+    local exrc_path = '.exrc'
+    local xstate = 'Xstate'
+    local xstateenv = { XDG_CONFIG_HOME=xconfig, XDG_DATA_HOME=xdata, XDG_STATE_HOME=xstate }
+
+    local function setup_exrc_file(filename)
+      exrc_path = filename
+
+      if string.find(exrc_path, "%.lua$") then
+        write_file(exrc_path, string.format([[
+          vim.g.exrc_file = "%s"
+        ]], exrc_path))
+      else
+        write_file(exrc_path, string.format([[
+          let g:exrc_file = "%s"
+        ]], exrc_path))
+      end
+    end
+
+    before_each(function()
+      write_file(init_lua_path, [[
+        vim.o.exrc = true
+        vim.g.exrc_file = '---'
+      ]])
+      mkdir_p(xstate .. pathsep .. (is_os('win') and 'nvim-data' or 'nvim'))
+    end)
+
+    after_each(function()
+      os.remove(exrc_path)
+      rmdir(xstate)
+    end)
+
+    for _, filename in ipairs({ '.exrc', '.nvimrc', '.nvim.lua' }) do
+      it(filename .. ' in cwd', function()
+        setup_exrc_file(filename)
+
+        clear{ args_rm={'-u'}, env=xstateenv }
+        -- The 'exrc' file is not trusted, and the prompt is skipped because there is no UI.
+        eq('---', eval('g:exrc_file'))
+
+        local screen = Screen.new(50, 8)
+        screen:attach()
+        funcs.termopen({nvim_prog})
+        screen:expect({ any = pesc('[i]gnore, (v)iew, (d)eny, (a)llow:') })
+        -- `i` to enter Terminal mode, `a` to allow
+        feed('ia')
+        screen:expect([[
+                                                            |
+          ~                                                 |
+          ~                                                 |
+          ~                                                 |
+          ~                                                 |
+          [No Name]                       0,0-1          All|
+                                                            |
+          -- TERMINAL --                                    |
+        ]])
+        feed(':echo g:exrc_file<CR>')
+        screen:expect(string.format([[
+                                                            |
+          ~                                                 |
+          ~                                                 |
+          ~                                                 |
+          ~                                                 |
+          [No Name]                       0,0-1          All|
+          %s%s|
+          -- TERMINAL --                                    |
+        ]], filename, string.rep(' ', 50 - #filename)))
+
+        clear{ args_rm={'-u'}, env=xstateenv }
+        -- The 'exrc' file is now trusted.
+        eq(filename, eval('g:exrc_file'))
+      end)
+    end
+  end)
+
+  describe('with explicitly provided config', function()
     local custom_lua_path = table.concat({xhome, 'custom.lua'}, pathsep)
     before_each(function()
       write_file(custom_lua_path, [[
@@ -572,7 +892,7 @@ describe('user config init', function()
     end)
   end)
 
-  describe 'VIMRC also exists'(function()
+  describe('VIMRC also exists', function()
     before_each(function()
       write_file(table.concat({xconfig, 'nvim', 'init.vim'}, pathsep), [[
       let g:vim_rc = 1
@@ -581,9 +901,9 @@ describe('user config init', function()
 
     it('loads default lua config, but shows an error', function()
       clear{ args_rm={'-u'}, env=xenv }
-      feed('<cr>') -- confirm "Conflicting config ..." message
+      feed('<cr><c-c>')  -- Dismiss "Conflicting config …" message.
       eq(1, eval('g:lua_rc'))
-      matches('^E5422: Conflicting configs', meths.exec('messages', true))
+      matches('^E5422: Conflicting configs', exec_capture('messages'))
     end)
   end)
 end)
@@ -618,7 +938,7 @@ describe('runtime:', function()
   end)
 
   it('loads plugin/*.lua from start packages', function()
-    local plugin_path = table.concat({xconfig, 'nvim', 'pack', 'catagory',
+    local plugin_path = table.concat({xconfig, 'nvim', 'pack', 'category',
     'start', 'test_plugin'}, pathsep)
     local plugin_folder_path = table.concat({plugin_path, 'plugin'}, pathsep)
     local plugin_file_path = table.concat({plugin_folder_path, 'plugin.lua'},
@@ -632,21 +952,21 @@ describe('runtime:', function()
 
     eq(2, eval('g:lua_plugin'))
     -- Check if plugin_file_path is listed in :scriptname
-    local scripts = meths.exec(':scriptnames', true)
-    assert.Truthy(scripts:find(plugin_file_path))
+    local scripts = exec_capture('scriptnames')
+    assert(scripts:find(plugin_file_path))
 
     -- Check if plugin_file_path is listed in startup profile
     local profile_reader = io.open(profiler_file, 'r')
     local profile_log = profile_reader:read('*a')
     profile_reader:close()
-    assert.Truthy(profile_log :find(plugin_file_path))
+    assert(profile_log:find(plugin_file_path))
 
     os.remove(profiler_file)
     rmdir(plugin_path)
   end)
 
   it('loads plugin/*.lua from site packages', function()
-    local nvimdata = iswin() and "nvim-data" or "nvim"
+    local nvimdata = is_os('win') and "nvim-data" or "nvim"
     local plugin_path = table.concat({xdata, nvimdata, 'site', 'pack', 'xa', 'start', 'yb'}, pathsep)
     local plugin_folder_path = table.concat({plugin_path, 'plugin'}, pathsep)
     local plugin_after_path = table.concat({plugin_path, 'after', 'plugin'}, pathsep)
