@@ -1,19 +1,26 @@
 -- Insert-mode tests.
 
 local helpers = require('test.functional.helpers')(after_each)
+local Screen = require('test.functional.ui.screen')
 local clear, feed, insert = helpers.clear, helpers.feed, helpers.insert
 local expect = helpers.expect
 local command = helpers.command
 local eq = helpers.eq
 local eval = helpers.eval
-local meths = helpers.meths
-local poke_eventloop = helpers.poke_eventloop
+local curbuf_contents = helpers.curbuf_contents
 
 describe('insert-mode', function()
   before_each(function()
     clear()
   end)
 
+  it('indents only once after "!" keys #12894', function()
+    command('let counter = []')
+    command('set indentexpr=len(add(counter,0))')
+    feed('i<C-F>x')
+    eq(' x', curbuf_contents())
+  end)
+
   it('CTRL-@', function()
     -- Inserts last-inserted text, leaves insert-mode.
     insert('hello')
@@ -42,6 +49,56 @@ describe('insert-mode', function()
       feed('i<C-r>"')
       expect('påskägg')
     end)
+
+    it('double quote is removed after hit-enter prompt #22609', function()
+      local screen = Screen.new(60, 6)
+      screen:set_default_attr_ids({
+        [0] = {bold = true, foreground = Screen.colors.Blue},  -- NonText
+        [1] = {foreground = Screen.colors.Blue},  -- SpecialKey
+        [2] = {foreground = Screen.colors.SlateBlue},
+        [3] = {bold = true},  -- ModeMsg
+        [4] = {reverse = true, bold = true},  -- MsgSeparator
+        [5] = {background = Screen.colors.Red, foreground = Screen.colors.White},  -- ErrorMsg
+        [6] = {foreground = Screen.colors.SeaGreen, bold = true},  -- MoreMsg
+      })
+      screen:attach()
+      feed('i<C-R>')
+      screen:expect([[
+        {1:^"}                                                           |
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {3:-- INSERT --}                                                |
+      ]])
+      feed('={}')
+      screen:expect([[
+        {1:"}                                                           |
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {0:~                                                           }|
+        ={2:{}}^                                                         |
+      ]])
+      feed('<CR>')
+      screen:expect([[
+        {1:"}                                                           |
+        {0:~                                                           }|
+        {4:                                                            }|
+        ={2:{}}                                                         |
+        {5:E731: using Dictionary as a String}                          |
+        {6:Press ENTER or type command to continue}^                     |
+      ]])
+      feed('<CR>')
+      screen:expect([[
+        ^                                                            |
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {0:~                                                           }|
+        {3:-- INSERT --}                                                |
+      ]])
+    end)
   end)
 
   describe('Ctrl-O', function()
@@ -131,25 +188,8 @@ describe('insert-mode', function()
     end)
   end)
 
-  describe([[With 'insertmode', Insert mode is not re-entered immediately after <C-L>]], function()
-    before_each(function()
-      command('set insertmode')
-      poke_eventloop()
-      eq({mode = 'i', blocking = false}, meths.get_mode())
-    end)
-
-    it('after calling :edit from <Cmd> mapping', function()
-      command('inoremap <C-B> <Cmd>edit Xfoo<CR>')
-      feed('<C-B><C-L>')
-      poke_eventloop()
-      eq({mode = 'n', blocking = false}, meths.get_mode())
-    end)
-
-    it('after calling :edit from RPC #16823', function()
-      command('edit Xfoo')
-      feed('<C-L>')
-      poke_eventloop()
-      eq({mode = 'n', blocking = false}, meths.get_mode())
-    end)
+  it('Ctrl-Shift-V supports entering unsimplified key notations', function()
+    feed('i<C-S-V><C-J><C-S-V><C-@><C-S-V><C-[><C-S-V><C-S-M><C-S-V><M-C-I><C-S-V><C-D-J><Esc>')
+    expect('<C-J><C-@><C-[><C-S-M><M-C-I><C-D-J>')
   end)
 end)
