@@ -93,19 +93,6 @@ describe('mappings with <Cmd>', function()
       {2:E5521: <Cmd> mapping must end with <CR> before second <Cmd>}      |
     ]])
 
-    command('noremap <F3> <Cmd><F3>let x = 2<cr>')
-    feed('<F3>')
-    screen:expect([[
-      ^some short lines                                                 |
-      of test text                                                     |
-      {1:~                                                                }|
-      {1:~                                                                }|
-      {1:~                                                                }|
-      {1:~                                                                }|
-      {1:~                                                                }|
-      {2:E5522: <Cmd> mapping must not include <F3> key}                   |
-    ]])
-
     command('noremap <F3> <Cmd>let x = 3')
     feed('<F3>')
     screen:expect([[
@@ -121,6 +108,47 @@ describe('mappings with <Cmd>', function()
     eq(0, eval('x'))
   end)
 
+  it('allows special keys and modifiers', function()
+    command('noremap <F3> <Cmd>normal! <Down><CR>')
+    feed('<F3>')
+    screen:expect([[
+      some short lines                                                 |
+      ^of test text                                                     |
+      {1:~                                                                }|
+      {1:~                                                                }|
+      {1:~                                                                }|
+      {1:~                                                                }|
+      {1:~                                                                }|
+                                                                       |
+    ]])
+
+    command('noremap <F3> <Cmd>normal! <C-Right><CR>')
+    feed('<F3>')
+    screen:expect([[
+      some short lines                                                 |
+      of ^test text                                                     |
+      {1:~                                                                }|
+      {1:~                                                                }|
+      {1:~                                                                }|
+      {1:~                                                                }|
+      {1:~                                                                }|
+                                                                       |
+    ]])
+  end)
+
+  it('handles string containing K_SPECIAL (0x80) bytes correctly', function()
+    command([[noremap <F3> <Cmd>let g:str = 'foo…bar'<CR>]])
+    feed('<F3>')
+    eq('foo…bar', eval('g:str'))
+    local str = eval([["foo\<D-…>bar"]])
+    command([[noremap <F3> <Cmd>let g:str = ']]..str..[['<CR>]])
+    feed('<F3>')
+    eq(str, eval('g:str'))
+    command([[noremap <F3> <Cmd>let g:str = 'foo<D-…>bar'<CR>]])
+    feed('<F3>')
+    eq(str, eval('g:str'))
+  end)
+
   it('works in various modes and sees correct `mode()` value', function()
     -- normal mode
     feed('<F3>')
