@@ -16,7 +16,7 @@ describe('Ex cmds', function()
     command(':echo expand("#<9999999999999999999999999999999999999999")')
     command(':lockvar 9999999999999999999999999999999999999999')
     command(':winsize 9999999999999999999999999999999999999999 9999999999999999999999999999999999999999')
-    eq('Vim(tabnext):E474: Invalid argument',
+    eq('Vim(tabnext):E475: Invalid argument: 9999999999999999999999999999999999999999',
       pcall_err(command, ':tabnext 9999999999999999999999999999999999999999'))
     eq('Vim(Next):E939: Positive count required',
       pcall_err(command, ':N 9999999999999999999999999999999999999999'))
@@ -26,5 +26,9 @@ describe('Ex cmds', function()
       pcall_err(command, ':bdelete 9999999999999999999999999999999999999999'))
     assert_alive()
   end)
-end)
 
+  it('listing long user command does not crash', function()
+    command('execute "command" repeat("T", 255) ":"')
+    command('command')
+  end)
+end)
