@@ -28,7 +28,10 @@ local function assert_eq(a, b, ...)
   if not vim.deep_equal(a, b) then
     error(message_parts(": ",
       ..., "assert_eq failed",
-      string.format("left == %q, right == %q", vim.inspect(a), vim.inspect(b))
+      string.format("left == %q, right == %q",
+        table.concat(vim.split(vim.inspect(a), "\n"), ""),
+        table.concat(vim.split(vim.inspect(b), "\n"), "")
+      )
     ))
   end
 end
@@ -64,10 +67,12 @@ local function expect_notification(method, params, ...)
   local message = read_message()
   assert_eq(method, message.method,
       ..., "expect_notification", "method")
-  assert_eq(params, message.params,
-      ..., "expect_notification", method, "params")
-  assert_eq({jsonrpc = "2.0"; method=method, params=params}, message,
-      ..., "expect_notification", "message")
+  if params then
+    assert_eq(params, message.params,
+        ..., "expect_notification", method, "params")
+    assert_eq({jsonrpc = "2.0"; method=method, params=params}, message,
+        ..., "expect_notification", "message")
+  end
 end
 
 local function expect_request(method, handler, ...)
@@ -100,8 +105,12 @@ local tests = {}
 
 function tests.basic_init()
   skeleton {
-    on_init = function(_params)
-      return { capabilities = {} }
+    on_init = function(_)
+      return {
+        capabilities = {
+          textDocumentSync = protocol.TextDocumentSyncKind.None;
+        }
+      }
     end;
     body = function()
       notify('test')
@@ -109,6 +118,19 @@ function tests.basic_init()
   }
 end
 
+function tests.basic_init_did_change_configuration()
+  skeleton({
+    on_init = function(_)
+      return {
+        capabilities = {},
+      }
+    end,
+    body = function()
+      expect_notification('workspace/didChangeConfiguration', { settings = { dummy = 1 } })
+    end,
+  })
+end
+
 function tests.check_workspace_configuration()
   skeleton {
     on_init = function(_params)
@@ -132,8 +154,11 @@ function tests.prepare_rename_nil()
   skeleton {
     on_init = function()
       return { capabilities = {
-        renameProvider = true,
-      } }
+        renameProvider = {
+            prepareProvider = true
+          }
+        }
+      }
     end;
     body = function()
       notify('start')
@@ -149,8 +174,11 @@ function tests.prepare_rename_placeholder()
   skeleton {
     on_init = function()
       return { capabilities = {
-        renameProvider = true,
-      } }
+        renameProvider = {
+            prepareProvider = true
+          }
+        }
+      }
     end;
     body = function()
       notify('start')
@@ -170,8 +198,11 @@ function tests.prepare_rename_range()
   skeleton {
     on_init = function()
       return { capabilities = {
-        renameProvider = true,
-      } }
+        renameProvider = {
+            prepareProvider = true
+          }
+        }
+      }
     end;
     body = function()
       notify('start')
@@ -193,19 +224,19 @@ end
 function tests.prepare_rename_error()
   skeleton {
     on_init = function()
-      return { capabilities = {
-        renameProvider = true,
-      } }
+      return {
+        capabilities = {
+          renameProvider = {
+            prepareProvider = true
+          },
+        }
+      }
     end;
     body = function()
       notify('start')
       expect_request('textDocument/prepareRename', function()
         return {}, nil
       end)
-      expect_request('textDocument/rename', function(params)
-        assert_eq(params.newName, 'renameto')
-        return nil, nil
-      end)
       notify('shutdown')
     end;
   }
@@ -219,6 +250,7 @@ function tests.basic_check_capabilities()
       return {
         capabilities = {
           textDocumentSync = protocol.TextDocumentSyncKind.Full;
+          codeLensProvider = false
         }
       }
     end;
@@ -227,6 +259,72 @@ function tests.basic_check_capabilities()
   }
 end
 
+function tests.text_document_save_did_open()
+  skeleton {
+    on_init = function()
+      return {
+        capabilities = {
+          textDocumentSync = {
+            save = true
+          }
+        }
+      }
+    end;
+    body = function()
+      notify('start')
+      expect_notification('textDocument/didClose')
+      expect_notification('textDocument/didOpen')
+      expect_notification('textDocument/didSave')
+      notify('shutdown')
+    end;
+  }
+end
+
+function tests.text_document_sync_save_bool()
+  skeleton {
+    on_init = function()
+      return {
+        capabilities = {
+          textDocumentSync = {
+            save = true
+          }
+        }
+      }
+    end;
+    body = function()
+      notify('start')
+      expect_notification('textDocument/didSave', {textDocument = { uri = "file://" }})
+      notify('shutdown')
+    end;
+  }
+end
+
+function tests.text_document_sync_save_includeText()
+  skeleton {
+    on_init = function()
+      return {
+        capabilities = {
+          textDocumentSync = {
+            save = {
+              includeText = true
+            }
+          }
+        }
+      }
+    end;
+    body = function()
+      notify('start')
+      expect_notification('textDocument/didSave', {
+        textDocument = {
+          uri = "file://"
+        },
+        text = "help me\n"
+      })
+      notify('shutdown')
+    end;
+  }
+end
+
 function tests.capabilities_for_client_supports_method()
   skeleton {
     on_init = function(params)
@@ -237,6 +335,7 @@ function tests.capabilities_for_client_supports_method()
           textDocumentSync = protocol.TextDocumentSyncKind.Full;
           completionProvider = true;
           hoverProvider = true;
+          renameProvider = false;
           definitionProvider = false;
           referencesProvider = false;
           codeLensProvider = { resolveProvider = true; };
@@ -544,7 +643,15 @@ function tests.basic_check_buffer_open_and_change_incremental()
       assert_eq(params.capabilities, expected_capabilities)
       return {
         capabilities = {
-          textDocumentSync = protocol.TextDocumentSyncKind.Incremental;
+          textDocumentSync = {
+            openClose = true,
+            change = protocol.TextDocumentSyncKind.Incremental,
+            willSave = true,
+            willSaveWaitUntil = true,
+            save = {
+              includeText = true,
+            }
+          }
         }
       }
     end;
@@ -701,6 +808,50 @@ function tests.code_action_server_side_command()
   })
 end
 
+
+function tests.code_action_filter()
+  skeleton {
+    on_init = function()
+      return {
+        capabilities = {
+          codeActionProvider = {
+            resolveProvider = false
+          }
+        }
+      }
+    end;
+    body = function()
+      notify('start')
+      local action = {
+        title = 'Action 1',
+        command = 'command'
+      }
+      local preferred_action = {
+        title = 'Action 2',
+        isPreferred = true,
+        command = 'preferred_command',
+      }
+      local quickfix_action = {
+        title = 'Action 3',
+        kind = 'quickfix',
+        command = 'quickfix_command',
+      }
+      local quickfix_foo_action = {
+        title = 'Action 4',
+        kind = 'quickfix.foo',
+        command = 'quickfix_foo_command',
+      }
+      expect_request('textDocument/codeAction', function()
+        return nil, { action, preferred_action, quickfix_action, quickfix_foo_action, }
+      end)
+      expect_request('textDocument/codeAction', function()
+        return nil, { action, preferred_action, quickfix_action, quickfix_foo_action, }
+      end)
+      notify('shutdown')
+    end;
+  }
+end
+
 function tests.clientside_commands()
   skeleton {
     on_init = function()
@@ -758,10 +909,49 @@ function tests.codelens_refresh_lock()
   }
 end
 
--- Tests will be indexed by TEST_NAME
+function tests.basic_formatting()
+  skeleton {
+    on_init = function()
+      return {
+        capabilities = {
+          documentFormattingProvider = true,
+        }
+      }
+    end;
+    body = function()
+      notify('start')
+      expect_request('textDocument/formatting', function()
+        return nil, {}
+      end)
+      notify('shutdown')
+    end;
+  }
+end
+
+function tests.set_defaults_all_capabilities()
+  skeleton {
+    on_init = function(_)
+      return {
+        capabilities = {
+          definitionProvider = true,
+          completionProvider = true,
+          documentRangeFormattingProvider = true,
+        }
+      }
+    end;
+    body = function()
+      notify('test')
+    end;
+  }
+end
+
+-- Tests will be indexed by test_name
+local test_name = arg[1]
+local timeout = arg[2]
+assert(type(test_name) == 'string', 'test_name must be specified as first arg.')
 
 local kill_timer = vim.loop.new_timer()
-kill_timer:start(_G.TIMEOUT or 1e3, 0, function()
+kill_timer:start(timeout or 1e3, 0, function()
   kill_timer:stop()
   kill_timer:close()
   log('ERROR', 'LSP', 'TIMEOUT')
@@ -769,14 +959,11 @@ kill_timer:start(_G.TIMEOUT or 1e3, 0, function()
   os.exit(100)
 end)
 
-local test_name = _G.TEST_NAME -- lualint workaround
-assert(type(test_name) == 'string', 'TEST_NAME must be specified.')
 local status, err = pcall(assert(tests[test_name], "Test not found"))
 kill_timer:stop()
 kill_timer:close()
 if not status then
   log('ERROR', 'LSP', tostring(err))
   io.stderr:write(err)
-  os.exit(101)
+  vim.cmd [[101cquit]]
 end
-os.exit(0)
