@@ -126,27 +126,57 @@ describe('mapping', function()
       ]])
   end)
 
+  it('dragging starts Select mode even if coming from mapping vim-patch:8.2.4806', function()
+    command('set mouse=a')
+    command('set selectmode=mouse')
+
+    command('nnoremap <LeftDrag> <LeftDrag><Cmd><CR>')
+    poke_eventloop()
+    meths.input_mouse('left', 'press', '', 0, 0, 0)
+    poke_eventloop()
+    meths.input_mouse('left', 'drag', '', 0, 0, 1)
+    poke_eventloop()
+    eq('s', eval('mode()'))
+  end)
+
   it('<LeftDrag> mapping in Insert mode works correctly vim-patch:8.2.4692', function()
     command('set mouse=a')
 
-    command([[inoremap <LeftDrag> <LeftDrag><Cmd>let g:dragged = 1<CR>]])
+    command('inoremap <LeftDrag> <LeftDrag><Cmd>let g:dragged = 1<CR>')
     feed('i')
-    sleep(10)
+    poke_eventloop()
     meths.input_mouse('left', 'press', '', 0, 0, 0)
-    sleep(10)
+    poke_eventloop()
     meths.input_mouse('left', 'drag', '', 0, 0, 1)
-    sleep(10)
+    poke_eventloop()
     eq(1, eval('g:dragged'))
     eq('v', eval('mode()'))
     feed([[<C-\><C-N>]])
 
     command([[inoremap <LeftDrag> <LeftDrag><C-\><C-N>]])
     feed('i')
-    sleep(10)
+    poke_eventloop()
     meths.input_mouse('left', 'press', '', 0, 0, 0)
-    sleep(10)
+    poke_eventloop()
     meths.input_mouse('left', 'drag', '', 0, 0, 1)
-    sleep(10)
+    poke_eventloop()
     eq('n', eval('mode()'))
   end)
+
+  it('timeout works after an <Nop> mapping is triggered on timeout vim-patch:8.1.0052', function()
+    command('set timeout timeoutlen=400')
+    command('inoremap ab TEST')
+    command('inoremap a <Nop>')
+    -- Enter Insert mode
+    feed('i')
+    -- Wait for the "a" mapping to time out
+    feed('a')
+    sleep(500)
+    -- Send "a" and wait for a period shorter than 'timeoutlen'
+    feed('a')
+    sleep(100)
+    -- Send "b", should trigger the "ab" mapping
+    feed('b')
+    expect('TEST')
+  end)
 end)
