@@ -7,13 +7,14 @@ local eval = helpers.eval
 local feed = helpers.feed
 local funcs = helpers.funcs
 local poke_eventloop = helpers.poke_eventloop
+local exec = helpers.exec
 
 describe('search cmdline', function()
   local screen
 
   before_each(function()
     clear()
-    command('set nohlsearch')
+    command('set nohlsearch inccommand=')
     screen = Screen.new(20, 3)
     screen:attach()
     screen:set_default_attr_ids({
@@ -289,8 +290,8 @@ describe('search cmdline', function()
     -- First match
     feed('/thei')
     screen:expect([[
+        3 the             |
         4 {inc:thei}r           |
-        5 there           |
       /thei^               |
     ]])
     -- Match from initial cursor position when modifying search
@@ -471,8 +472,8 @@ describe('search cmdline', function()
        funcs.winsaveview())
   end)
 
+  -- oldtest: Test_search_cmdline4().
   it("CTRL-G with 'incsearch' and ? goes in the right direction", function()
-    -- oldtest: Test_search_cmdline4().
     screen:try_resize(40, 4)
     command('enew!')
     funcs.setline(1, {'  1 the first', '  2 the second', '  3 the third'})
@@ -572,8 +573,8 @@ describe('search cmdline', function()
     ]])
   end)
 
+  -- oldtest: Test_incsearch_sort_dump().
   it('incsearch works with :sort', function()
-    -- oldtest: Test_incsearch_sort_dump().
     screen:try_resize(20, 4)
     command('set incsearch hlsearch scrolloff=0')
     funcs.setline(1, {'another one 2', 'that one 3', 'the one 1'})
@@ -588,8 +589,8 @@ describe('search cmdline', function()
     feed('<esc>')
   end)
 
+  -- oldtest: Test_incsearch_vimgrep_dump().
   it('incsearch works with :vimgrep family', function()
-    -- oldtest: Test_incsearch_vimgrep_dump().
     screen:try_resize(30, 4)
     command('set incsearch hlsearch scrolloff=0')
     funcs.setline(1, {'another one 2', 'that one 3', 'the one 1'})
@@ -639,4 +640,98 @@ describe('search cmdline', function()
     ]])
     feed('<esc>')
   end)
+
+  -- oldtest: Test_incsearch_substitute_dump2()
+  it('detects empty pattern properly vim-patch:8.2.2295', function()
+    screen:try_resize(70, 6)
+    exec([[
+      set incsearch hlsearch scrolloff=0
+      for n in range(1, 4)
+        call setline(n, "foo " . n)
+      endfor
+      call setline(5, "abc|def")
+      3
+    ]])
+
+    feed([[:%s/\vabc|]])
+    screen:expect([[
+      foo 1                                                                 |
+      foo 2                                                                 |
+      foo 3                                                                 |
+      foo 4                                                                 |
+      abc|def                                                               |
+      :%s/\vabc|^                                                            |
+    ]])
+    feed('<Esc>')
+
+    -- The following should not be highlighted
+    feed([[:1,5s/\v|]])
+    screen:expect([[
+      foo 1                                                                 |
+      foo 2                                                                 |
+      foo 3                                                                 |
+      foo 4                                                                 |
+      abc|def                                                               |
+      :1,5s/\v|^                                                             |
+    ]])
+  end)
+end)
+
+describe('Search highlight', function()
+  before_each(clear)
+
+  -- oldtest: Test_hlsearch_dump()
+  it('beyond line end vim-patch:8.2.2542', function()
+    local screen = Screen.new(50, 6)
+    screen:set_default_attr_ids({
+      [1] = {bold = true, foreground = Screen.colors.Blue},  -- NonText
+      [2] = {background = Screen.colors.Yellow},  -- Search
+      [3] = {background = Screen.colors.Grey90},  -- CursorLine
+    })
+    screen:attach()
+    exec([[
+      set hlsearch noincsearch cursorline
+      call setline(1, ["xxx", "xxx", "xxx"])
+      /.*
+      2
+    ]])
+    feed([[/\_.*<CR>]])
+    screen:expect([[
+      {2:xxx }                                              |
+      {2:xxx }                                              |
+      {2:^xxx }{3:                                              }|
+      {1:~                                                 }|
+      {1:~                                                 }|
+      /\_.*                                             |
+    ]])
+  end)
+
+  -- oldtest: Test_hlsearch_and_visual()
+  it('is combined with Visual highlight vim-patch:8.2.2797', function()
+    local screen = Screen.new(40, 6)
+    screen:set_default_attr_ids({
+      [1] = {bold = true, foreground = Screen.colors.Blue},  -- NonText
+      [2] = {bold = true}, -- ModeMsg, Search
+      [3] = {background = Screen.colors.LightGrey},  -- Visual
+      [4] = {background = Screen.colors.Yellow, bold = true},  -- Search
+      [5] = {background = Screen.colors.LightGrey, bold = true},  -- Visual + Search
+    })
+    screen:attach()
+    exec([[
+      set hlsearch noincsearch
+      call setline(1, repeat(["xxx yyy zzz"], 3))
+      hi Search gui=bold
+      /yyy
+      call cursor(1, 6)
+    ]])
+    feed('vjj')
+    screen:expect([[
+      xxx {4:y}{5:yy}{3: zzz}                             |
+      {3:xxx }{5:yyy}{3: zzz}                             |
+      {3:xxx }{5:y}{4:^yy} zzz                             |
+      {1:~                                       }|
+      {1:~                                       }|
+      {2:-- VISUAL --}                            |
+    ]])
+  end)
 end)
