@@ -467,7 +467,6 @@ describe('v:lua', function()
           end
         end
       end
-      vim.api.nvim_buf_set_option(0, 'omnifunc', 'v:lua.mymod.omni')
     ]])
   end)
 
@@ -515,6 +514,7 @@ describe('v:lua', function()
       [5] = {bold = true, foreground = Screen.colors.SeaGreen4},
     })
     screen:attach()
+    meths.buf_set_option(0, 'omnifunc', 'v:lua.mymod.omni')
     feed('isome st<c-x><c-o>')
     screen:expect{grid=[[
       some stuff^                                                  |
@@ -526,12 +526,16 @@ describe('v:lua', function()
       {1:~                                                           }|
       {4:-- Omni completion (^O^N^P) }{5:match 1 of 3}                    |
     ]]}
+    meths.set_option('operatorfunc', 'v:lua.mymod.noisy')
+    feed('<Esc>g@g@')
+    eq("hey line", meths.get_current_line())
   end)
 
   it('supports packages', function()
     command('set pp+=test/functional/fixtures')
     eq('\tbadval', eval("v:lua.require'leftpad'('badval')"))
     eq(9003, eval("v:lua.require'bar'.doit()"))
+    eq(9004, eval("v:lua.require'baz-quux'.doit()"))
   end)
 
   it('throw errors for invalid use', function()
