@@ -8,14 +8,12 @@ local feed = helpers.feed
 local clear = helpers.clear
 local funcs = helpers.funcs
 local meths = helpers.meths
-local iswin = helpers.iswin
 local command = helpers.command
 local write_file = helpers.write_file
 local exec_capture = helpers.exec_capture
 local exec_lua = helpers.exec_lua
 local pcall_err = helpers.pcall_err
-
-local screen
+local is_os = helpers.is_os
 
 local fname = 'Xtest-functional-lua-overrides-luafile'
 
@@ -135,22 +133,58 @@ describe('print', function()
         print("very slow")
         vim.api.nvim_command("sleep 1m") -- force deferred event processing
       end
-    ]], (iswin() and "timeout 1") or "sleep 0.1")
+    ]], (is_os('win') and "timeout 1") or "sleep 0.1")
     eq('very slow\nvery fast', exec_capture('lua test()'))
   end)
+
+  it('blank line in message works', function()
+    local screen = Screen.new(40, 8)
+    screen:attach()
+    screen:set_default_attr_ids({
+      [0] = {bold = true, foreground=Screen.colors.Blue},
+      [1] = {bold = true, foreground = Screen.colors.SeaGreen},
+      [2] = {bold = true, reverse = true},
+    })
+    feed([[:lua print('\na')<CR>]])
+    screen:expect{grid=[[
+                                              |
+      {0:~                                       }|
+      {0:~                                       }|
+      {0:~                                       }|
+      {2:                                        }|
+                                              |
+      a                                       |
+      {1:Press ENTER or type command to continue}^ |
+    ]]}
+    feed('<CR>')
+    feed([[:lua print('b\n\nc')<CR>]])
+    screen:expect{grid=[[
+                                              |
+      {0:~                                       }|
+      {0:~                                       }|
+      {2:                                        }|
+      b                                       |
+                                              |
+      c                                       |
+      {1:Press ENTER or type command to continue}^ |
+    ]]}
+  end)
 end)
 
 describe('debug.debug', function()
+  local screen
+
   before_each(function()
     screen = Screen.new()
     screen:attach()
-    screen:set_default_attr_ids({
-      [0] = {bold=true, foreground=255},
-      E = {foreground = Screen.colors.Grey100, background = Screen.colors.Red},
-      cr = {bold = true, foreground = Screen.colors.SeaGreen4},
-    })
-    command("set display-=msgsep")
+    screen:set_default_attr_ids {
+      [0] = {bold=true, foreground=255};
+      [1] = {bold = true, reverse = true};
+      E = {foreground = Screen.colors.Grey100, background = Screen.colors.Red};
+      cr = {bold = true, foreground = Screen.colors.SeaGreen4};
+    }
   end)
+
   it('works', function()
     command([[lua
       function Test(a)
@@ -160,9 +194,8 @@ describe('debug.debug', function()
       end
     ]])
     feed(':lua Test()\n')
-    screen:expect([[
-      {0:~                                                    }|
-      {0:~                                                    }|
+    screen:expect{grid=[[
+                                                           |
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
@@ -173,11 +206,13 @@ describe('debug.debug', function()
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
+      {1:                                                     }|
       nil                                                  |
       lua_debug> ^                                          |
-    ]])
+    ]]}
     feed('print("TEST")\n')
     screen:expect([[
+                                                           |
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
@@ -186,8 +221,7 @@ describe('debug.debug', function()
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
+      {1:                                                     }|
       nil                                                  |
       lua_debug> print("TEST")                             |
       TEST                                                 |
@@ -195,10 +229,10 @@ describe('debug.debug', function()
     ]])
     feed('<C-c>')
     screen:expect{grid=[[
+                                                           |
       {0:~                                                    }|
       {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
+      {1:                                                     }|
       nil                                                  |
       lua_debug> print("TEST")                             |
       TEST                                                 |
@@ -212,6 +246,7 @@ describe('debug.debug', function()
     ]]}
     feed('<C-l>:lua Test()\n')
     screen:expect([[
+                                                           |
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
@@ -222,19 +257,18 @@ describe('debug.debug', function()
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
+      {1:                                                     }|
       nil                                                  |
       lua_debug> ^                                          |
     ]])
     feed('\n')
     screen:expect{grid=[[
+                                                           |
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
+      {1:                                                     }|
       nil                                                  |
       lua_debug>                                           |
       {E:E5108: Error executing lua [string ":lua"]:5: attempt}|
@@ -268,6 +302,7 @@ describe('debug.debug', function()
 
     feed("conttt<cr>") -- misspelled cont; invalid syntax
     screen:expect{grid=[[
+                                                           |
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
@@ -276,8 +311,7 @@ describe('debug.debug', function()
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
+      {1:                                                     }|
       lua_debug> conttt                                    |
       {E:E5115: Error while loading debug string: (debug comma}|
       {E:nd):1: '=' expected near '<eof>'}                     |
@@ -286,14 +320,14 @@ describe('debug.debug', function()
 
     feed("cont<cr>") -- exactly "cont", exit now
     screen:expect{grid=[[
+                                                           |
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
       {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
+      {1:                                                     }|
       lua_debug> conttt                                    |
       {E:E5115: Error while loading debug string: (debug comma}|
       {E:nd):1: '=' expected near '<eof>'}                     |
@@ -337,3 +371,11 @@ describe('os.getenv', function()
     eq(value, funcs.luaeval('os.getenv("XTEST_1")'))
   end)
 end)
+
+-- "bit" module is always available, regardless if nvim is built with
+-- luajit or PUC lua 5.1.
+describe('bit module', function()
+  it('works', function()
+    eq (9, exec_lua [[ return require'bit'.band(11,13) ]])
+  end)
+end)
