@@ -3,20 +3,25 @@ local helpers = require('test.functional.helpers')(after_each)
 local Screen = require('test.functional.ui.screen')
 
 local assert_alive = helpers.assert_alive
+local assert_log = helpers.assert_log
 local meths = helpers.meths
 local command = helpers.command
 local clear = helpers.clear
 local exc_exec = helpers.exc_exec
 local eval = helpers.eval
 local eq = helpers.eq
+local ok = helpers.ok
 local funcs = helpers.funcs
 local insert = helpers.insert
-local iswin = helpers.iswin
 local neq = helpers.neq
 local mkdir = helpers.mkdir
 local rmdir = helpers.rmdir
 local alter_slashes = helpers.alter_slashes
 local tbl_contains = helpers.tbl_contains
+local expect_exit = helpers.expect_exit
+local is_os = helpers.is_os
+
+local testlog = 'Xtest-defaults-log'
 
 describe('startup defaults', function()
   describe(':filetype', function()
@@ -159,11 +164,41 @@ describe('startup defaults', function()
         ~                        |~                       |
                                                           |
       ]])
+
+      -- change "vert" character to single-cell
+      funcs.setcellwidths({{0x2502, 0x2502, 1}})
+      screen:expect([[
+        1                        │1                       |
+        ^+--  2 lines: 2----------│+--  2 lines: 2---------|
+        4                        │4                       |
+        ~                        │~                       |
+                                                          |
+      ]])
+
+      -- change "vert" character to double-cell
+      funcs.setcellwidths({{0x2502, 0x2502, 2}})
+      screen:expect([[
+        1                        |1                       |
+        ^+--  2 lines: 2----------|+--  2 lines: 2---------|
+        4                        |4                       |
+        ~                        |~                       |
+                                                          |
+      ]])
+
+      -- "vert" character should still default to single-byte fillchars because of setcellwidths().
+      command('set ambiwidth=single')
+      screen:expect([[
+        1                        |1                       |
+        ^+--  2 lines: 2··········|+--  2 lines: 2·········|
+        4                        |4                       |
+        ~                        |~                       |
+                                                          |
+      ]])
     end)
   end)
 
   it("'shadafile' ('viminfofile')", function()
-    local env = {XDG_DATA_HOME='Xtest-userdata', XDG_CONFIG_HOME='Xtest-userconfig'}
+    local env = {XDG_DATA_HOME='Xtest-userdata', XDG_STATE_HOME='Xtest-userstate', XDG_CONFIG_HOME='Xtest-userconfig'}
     clear{args={}, args_rm={'-i'}, env=env}
     -- Default 'shadafile' is empty.
     -- This means use the default location. :help shada-file-name
@@ -174,11 +209,11 @@ describe('startup defaults', function()
     command('write')
     local f = eval('fnamemodify(@%,":p")')
     assert(string.len(f) > 3)
-    command('qall')
+    expect_exit(command, 'qall')
     clear{args={}, args_rm={'-i'}, env=env}
     eq({ f }, eval('v:oldfiles'))
     os.remove('Xtest-foo')
-    rmdir('Xtest-userdata')
+    rmdir('Xtest-userstate')
 
     -- Handles viminfo/viminfofile as alias for shada/shadafile.
     eq('\n  shadafile=', eval('execute("set shadafile?")'))
@@ -188,7 +223,9 @@ describe('startup defaults', function()
   end)
 
   it("'packpath'", function()
-    clear()
+    clear{
+      args_rm={'runtimepath'},
+    }
     -- Defaults to &runtimepath.
     eq(meths.get_option('runtimepath'), meths.get_option('packpath'))
 
@@ -206,7 +243,7 @@ describe('startup defaults', function()
 
   describe('$NVIM_LOG_FILE', function()
     local xdgdir = 'Xtest-startup-xdg-logpath'
-    local xdgcachedir = xdgdir..'/nvim'
+    local xdgstatedir = is_os('win') and xdgdir..'/nvim-data' or xdgdir..'/nvim'
     after_each(function()
       os.remove('Xtest-logpath')
       rmdir(xdgdir)
@@ -218,34 +255,38 @@ describe('startup defaults', function()
       }})
       eq('Xtest-logpath', eval('$NVIM_LOG_FILE'))
     end)
-    it('defaults to stdpath("cache")/log if empty', function()
-      eq(true, mkdir(xdgdir) and mkdir(xdgcachedir))
+    it('defaults to stdpath("log")/log if empty', function()
+      eq(true, mkdir(xdgdir) and mkdir(xdgstatedir))
       clear({env={
-        XDG_CACHE_HOME=xdgdir,
+        XDG_STATE_HOME=xdgdir,
         NVIM_LOG_FILE='',  -- Empty is invalid.
       }})
-      eq(xdgcachedir..'/log', string.gsub(eval('$NVIM_LOG_FILE'), '\\', '/'))
+      eq(xdgstatedir..'/log', string.gsub(eval('$NVIM_LOG_FILE'), '\\', '/'))
     end)
-    it('defaults to stdpath("cache")/log if invalid', function()
-      eq(true, mkdir(xdgdir) and mkdir(xdgcachedir))
+    it('defaults to stdpath("log")/log if invalid', function()
+      eq(true, mkdir(xdgdir) and mkdir(xdgstatedir))
       clear({env={
-        XDG_CACHE_HOME=xdgdir,
+        XDG_STATE_HOME=xdgdir,
         NVIM_LOG_FILE='.',  -- Any directory is invalid.
       }})
-      eq(xdgcachedir..'/log', string.gsub(eval('$NVIM_LOG_FILE'), '\\', '/'))
+      eq(xdgstatedir..'/log', string.gsub(eval('$NVIM_LOG_FILE'), '\\', '/'))
     end)
   end)
 end)
 
-describe('XDG-based defaults', function()
+describe('XDG defaults', function()
   -- Need separate describe() blocks to not run clear() twice.
   -- Do not put before_each() here for the same reasons.
 
+  after_each(function()
+    os.remove(testlog)
+  end)
+
   it("&runtimepath data-dir matches stdpath('data') #9910", function()
     clear()
     local rtp = eval('split(&runtimepath, ",")')
     local rv = {}
-    local expected = (iswin()
+    local expected = (is_os('win')
                       and { [[\nvim-data\site]], [[\nvim-data\site\after]], }
                       or { '/nvim/site', '/nvim/site/after', })
 
@@ -264,6 +305,7 @@ describe('XDG-based defaults', function()
         XDG_CONFIG_HOME=nil,
         XDG_DATA_HOME=nil,
         XDG_CACHE_HOME=nil,
+        XDG_STATE_HOME=nil,
         XDG_RUNTIME_DIR=nil,
         XDG_CONFIG_DIRS=nil,
         XDG_DATA_DIRS=nil,
@@ -280,6 +322,7 @@ describe('XDG-based defaults', function()
       eq('.', meths.get_option('viewdir'))
       eq('.', meths.get_option('directory'))
       eq('.', meths.get_option('undodir'))
+      ok((funcs.tempname()):len() > 4)
     end)
   end)
 
@@ -291,25 +334,35 @@ describe('XDG-based defaults', function()
     return vimruntime, libdir
   end
 
-  local env_sep = iswin() and ';' or ':'
-  local data_dir = iswin() and 'nvim-data' or 'nvim'
-  local root_path = iswin() and 'C:' or ''
+  local env_sep = is_os('win') and ';' or ':'
+  local data_dir = is_os('win') and 'nvim-data' or 'nvim'
+  local state_dir = is_os('win') and 'nvim-data' or 'nvim'
+  local root_path = is_os('win') and 'C:' or ''
 
   describe('with too long XDG variables', function()
     before_each(function()
-      clear({env={
-        XDG_CONFIG_HOME=(root_path .. ('/x'):rep(4096)),
-        XDG_CONFIG_DIRS=(root_path .. ('/a'):rep(2048)
-                         .. env_sep.. root_path .. ('/b'):rep(2048)
-                         .. (env_sep .. root_path .. '/c'):rep(512)),
-        XDG_DATA_HOME=(root_path .. ('/X'):rep(4096)),
-        XDG_DATA_DIRS=(root_path .. ('/A'):rep(2048)
-                       .. env_sep .. root_path .. ('/B'):rep(2048)
-                       .. (env_sep .. root_path .. '/C'):rep(512)),
+      clear({
+        args_rm={'runtimepath'},
+        env={
+          NVIM_LOG_FILE=testlog,
+          XDG_CONFIG_HOME=(root_path .. ('/x'):rep(4096)),
+          XDG_CONFIG_DIRS=(root_path .. ('/a'):rep(2048)
+                           .. env_sep.. root_path .. ('/b'):rep(2048)
+                           .. (env_sep .. root_path .. '/c'):rep(512)),
+          XDG_DATA_HOME=(root_path .. ('/X'):rep(4096)),
+          XDG_RUNTIME_DIR=(root_path .. ('/X'):rep(4096)),
+          XDG_STATE_HOME=(root_path .. ('/X'):rep(4096)),
+          XDG_DATA_DIRS=(root_path .. ('/A'):rep(2048)
+                         .. env_sep .. root_path .. ('/B'):rep(2048)
+                         .. (env_sep .. root_path .. '/C'):rep(512)),
       }})
     end)
 
     it('are correctly set', function()
+      if not is_os('win') then
+        assert_log('Failed to start server: no such file or directory: /X/X/X', testlog, 10)
+      end
+
       local vimruntime, libdir = vimruntime_and_libdir()
 
       eq(((root_path .. ('/x'):rep(4096) .. '/nvim'
@@ -355,28 +408,37 @@ describe('XDG-based defaults', function()
           .. ',' .. root_path .. ('/a'):rep(2048) .. '/nvim/after'
           .. ',' .. root_path .. ('/x'):rep(4096) .. '/nvim/after'
       ):gsub('\\', '/')), (meths.get_option('runtimepath')):gsub('\\', '/'))
-      eq('.,' .. root_path .. ('/X'):rep(4096).. '/' .. data_dir .. '/backup//',
+      eq('.,' .. root_path .. ('/X'):rep(4096).. '/' .. state_dir .. '/backup//',
          (meths.get_option('backupdir'):gsub('\\', '/')))
-      eq(root_path .. ('/X'):rep(4096) .. '/' .. data_dir .. '/swap//',
+      eq(root_path .. ('/X'):rep(4096) .. '/' .. state_dir .. '/swap//',
          (meths.get_option('directory')):gsub('\\', '/'))
-      eq(root_path .. ('/X'):rep(4096) .. '/' .. data_dir .. '/undo//',
+      eq(root_path .. ('/X'):rep(4096) .. '/' .. state_dir .. '/undo//',
          (meths.get_option('undodir')):gsub('\\', '/'))
-      eq(root_path .. ('/X'):rep(4096) .. '/'  ..  data_dir .. '/view//',
+      eq(root_path .. ('/X'):rep(4096) .. '/'  ..  state_dir .. '/view//',
          (meths.get_option('viewdir')):gsub('\\', '/'))
     end)
   end)
 
   describe('with XDG variables that can be expanded', function()
     before_each(function()
-      clear({env={
-        XDG_CONFIG_HOME='$XDG_DATA_HOME',
-        XDG_CONFIG_DIRS='$XDG_DATA_DIRS',
-        XDG_DATA_HOME='$XDG_CONFIG_HOME',
-        XDG_DATA_DIRS='$XDG_CONFIG_DIRS',
+      clear({
+        args_rm={'runtimepath'},
+        env={
+          NVIM_LOG_FILE=testlog,
+          XDG_CONFIG_HOME='$XDG_DATA_HOME',
+          XDG_CONFIG_DIRS='$XDG_DATA_DIRS',
+          XDG_DATA_HOME='$XDG_CONFIG_HOME',
+          XDG_RUNTIME_DIR='$XDG_RUNTIME_DIR',
+          XDG_STATE_HOME='$XDG_CONFIG_HOME',
+          XDG_DATA_DIRS='$XDG_CONFIG_DIRS',
       }})
     end)
 
     it('are not expanded', function()
+      if not is_os('win') then
+        assert_log('Failed to start server: no such file or directory: %$XDG_RUNTIME_DIR%/', testlog, 10)
+      end
+
       local vimruntime, libdir = vimruntime_and_libdir()
       eq((('$XDG_DATA_HOME/nvim'
           .. ',$XDG_DATA_DIRS/nvim'
@@ -405,13 +467,13 @@ describe('XDG-based defaults', function()
           .. ',$XDG_DATA_DIRS/nvim/after'
           .. ',$XDG_DATA_HOME/nvim/after'
       ):gsub('\\', '/')), (meths.get_option('runtimepath')):gsub('\\', '/'))
-      eq(('.,$XDG_CONFIG_HOME/' .. data_dir .. '/backup//'),
+      eq(('.,$XDG_CONFIG_HOME/' .. state_dir .. '/backup//'),
           meths.get_option('backupdir'):gsub('\\', '/'))
-      eq(('$XDG_CONFIG_HOME/' .. data_dir .. '/swap//'),
+      eq(('$XDG_CONFIG_HOME/' .. state_dir .. '/swap//'),
           meths.get_option('directory'):gsub('\\', '/'))
-      eq(('$XDG_CONFIG_HOME/' .. data_dir .. '/undo//'),
+      eq(('$XDG_CONFIG_HOME/' .. state_dir .. '/undo//'),
           meths.get_option('undodir'):gsub('\\', '/'))
-      eq(('$XDG_CONFIG_HOME/' .. data_dir .. '/view//'),
+      eq(('$XDG_CONFIG_HOME/' .. state_dir .. '/view//'),
           meths.get_option('viewdir'):gsub('\\', '/'))
       meths.command('set all&')
       eq(('$XDG_DATA_HOME/nvim'
@@ -425,30 +487,34 @@ describe('XDG-based defaults', function()
           .. ',$XDG_DATA_DIRS/nvim/after'
           .. ',$XDG_DATA_HOME/nvim/after'
       ):gsub('\\', '/'), (meths.get_option('runtimepath')):gsub('\\', '/'))
-      eq(('.,$XDG_CONFIG_HOME/' .. data_dir .. '/backup//'),
+      eq(('.,$XDG_CONFIG_HOME/' .. state_dir .. '/backup//'),
           meths.get_option('backupdir'):gsub('\\', '/'))
-      eq(('$XDG_CONFIG_HOME/' .. data_dir .. '/swap//'),
+      eq(('$XDG_CONFIG_HOME/' .. state_dir .. '/swap//'),
           meths.get_option('directory'):gsub('\\', '/'))
-      eq(('$XDG_CONFIG_HOME/' .. data_dir .. '/undo//'),
+      eq(('$XDG_CONFIG_HOME/' .. state_dir .. '/undo//'),
           meths.get_option('undodir'):gsub('\\', '/'))
-      eq(('$XDG_CONFIG_HOME/' .. data_dir .. '/view//'),
+      eq(('$XDG_CONFIG_HOME/' .. state_dir .. '/view//'),
           meths.get_option('viewdir'):gsub('\\', '/'))
+      eq(nil, (funcs.tempname()):match('XDG_RUNTIME_DIR'))
     end)
   end)
 
   describe('with commas', function()
     before_each(function()
-      clear({env={
-        XDG_CONFIG_HOME=', , ,',
-        XDG_CONFIG_DIRS=',-,-,' .. env_sep .. '-,-,-',
-        XDG_DATA_HOME=',=,=,',
-        XDG_DATA_DIRS=',≡,≡,' .. env_sep .. '≡,≡,≡',
+      clear({
+        args_rm={'runtimepath'},
+        env={
+          XDG_CONFIG_HOME=', , ,',
+          XDG_CONFIG_DIRS=',-,-,' .. env_sep .. '-,-,-',
+          XDG_DATA_HOME=',=,=,',
+          XDG_STATE_HOME=',=,=,',
+          XDG_DATA_DIRS=',≡,≡,' .. env_sep .. '≡,≡,≡',
       }})
     end)
 
     it('are escaped properly', function()
       local vimruntime, libdir = vimruntime_and_libdir()
-      local path_sep = iswin() and '\\' or '/'
+      local path_sep = is_os('win') and '\\' or '/'
       eq(('\\, \\, \\,' .. path_sep .. 'nvim'
           .. ',\\,-\\,-\\,' .. path_sep .. 'nvim'
           .. ',-\\,-\\,-' .. path_sep .. 'nvim'
@@ -484,13 +550,13 @@ describe('XDG-based defaults', function()
           .. ',\\,-\\,-\\,' .. path_sep ..'nvim' .. path_sep ..'after'
           .. ',\\, \\, \\,' .. path_sep ..'nvim' .. path_sep ..'after'
       ), meths.get_option('runtimepath'))
-      eq('.,\\,=\\,=\\,' .. path_sep .. data_dir .. '' .. path_sep ..'backup' .. (path_sep):rep(2),
+      eq('.,\\,=\\,=\\,' .. path_sep .. state_dir .. '' .. path_sep ..'backup' .. (path_sep):rep(2),
           meths.get_option('backupdir'))
-      eq('\\,=\\,=\\,' .. path_sep ..'' .. data_dir .. '' .. path_sep ..'swap' .. (path_sep):rep(2),
+      eq('\\,=\\,=\\,' .. path_sep ..'' .. state_dir .. '' .. path_sep ..'swap' .. (path_sep):rep(2),
           meths.get_option('directory'))
-      eq('\\,=\\,=\\,' .. path_sep ..'' .. data_dir .. '' .. path_sep ..'undo' .. (path_sep):rep(2),
+      eq('\\,=\\,=\\,' .. path_sep ..'' .. state_dir .. '' .. path_sep ..'undo' .. (path_sep):rep(2),
           meths.get_option('undodir'))
-      eq('\\,=\\,=\\,' .. path_sep ..'' .. data_dir .. '' .. path_sep ..'view' .. (path_sep):rep(2),
+      eq('\\,=\\,=\\,' .. path_sep ..'' .. state_dir .. '' .. path_sep ..'view' .. (path_sep):rep(2),
           meths.get_option('viewdir'))
     end)
   end)
@@ -499,9 +565,14 @@ end)
 
 describe('stdpath()', function()
   -- Windows appends 'nvim-data' instead of just 'nvim' to prevent collisions
-  -- due to XDG_CONFIG_HOME and XDG_DATA_HOME being the same.
-  local datadir = iswin() and 'nvim-data' or 'nvim'
-  local env_sep = iswin() and ';' or ':'
+  -- due to XDG_CONFIG_HOME, XDG_DATA_HOME and XDG_STATE_HOME being the same.
+  local function maybe_data(name)
+    return is_os('win') and name .. '-data' or name
+  end
+
+  local datadir = maybe_data('nvim')
+  local statedir = maybe_data('nvim')
+  local env_sep = is_os('win') and ';' or ':'
 
   it('acceptance', function()
     clear()  -- Do not explicitly set any env vars.
@@ -509,8 +580,28 @@ describe('stdpath()', function()
     eq('nvim', funcs.fnamemodify(funcs.stdpath('cache'), ':t'))
     eq('nvim', funcs.fnamemodify(funcs.stdpath('config'), ':t'))
     eq(datadir, funcs.fnamemodify(funcs.stdpath('data'), ':t'))
+    eq(statedir, funcs.fnamemodify(funcs.stdpath('state'), ':t'))
     eq('table', type(funcs.stdpath('config_dirs')))
     eq('table', type(funcs.stdpath('data_dirs')))
+    eq('string', type(funcs.stdpath('run')))
+    assert_alive()  -- Check for crash. #8393
+  end)
+
+  it('reacts to #NVIM_APPNAME', function()
+    local appname = "NVIM_APPNAME_TEST____________________________________" ..
+      "______________________________________________________________________"
+    clear({env={ NVIM_APPNAME=appname }})
+    eq(appname, funcs.fnamemodify(funcs.stdpath('config'), ':t'))
+    eq(appname, funcs.fnamemodify(funcs.stdpath('cache'), ':t'))
+    eq(maybe_data(appname), funcs.fnamemodify(funcs.stdpath('log'), ':t'))
+    eq(maybe_data(appname), funcs.fnamemodify(funcs.stdpath('data'), ':t'))
+    eq(maybe_data(appname), funcs.fnamemodify(funcs.stdpath('state'), ':t'))
+    -- config_dirs and data_dirs are empty on windows, so don't check them on
+    -- that platform
+    if not is_os('win') then
+      eq(appname, funcs.fnamemodify(funcs.stdpath('config_dirs')[1], ':t'))
+      eq(appname, funcs.fnamemodify(funcs.stdpath('data_dirs')[1], ':t'))
+    end
     assert_alive()  -- Check for crash. #8393
   end)
 
@@ -582,6 +673,39 @@ describe('stdpath()', function()
       end)
     end)
 
+    describe('with "state"' , function ()
+      it('knows XDG_STATE_HOME', function()
+        clear({env={
+          XDG_STATE_HOME=alter_slashes('/home/docwhat/.local'),
+        }})
+        eq(alter_slashes('/home/docwhat/.local/'..statedir), funcs.stdpath('state'))
+      end)
+
+      it('handles changes during runtime', function()
+        clear({env={
+          XDG_STATE_HOME=alter_slashes('/home/original'),
+        }})
+        eq(alter_slashes('/home/original/'..statedir), funcs.stdpath('state'))
+        command("let $XDG_STATE_HOME='"..alter_slashes('/home/new').."'")
+        eq(alter_slashes('/home/new/'..statedir), funcs.stdpath('state'))
+      end)
+
+      it("doesn't expand $VARIABLES", function()
+        clear({env={
+          XDG_STATE_HOME='$VARIABLES',
+          VARIABLES='this-should-not-happen',
+        }})
+        eq(alter_slashes('$VARIABLES/'..statedir), funcs.stdpath('state'))
+      end)
+
+      it("doesn't expand ~/", function()
+        clear({env={
+          XDG_STATE_HOME=alter_slashes('~/frobnitz'),
+        }})
+        eq(alter_slashes('~/frobnitz/'..statedir), funcs.stdpath('state'))
+      end)
+    end)
+
     describe('with "cache"' , function ()
       it('knows XDG_CACHE_HOME', function()
         clear({env={
@@ -619,7 +743,7 @@ describe('stdpath()', function()
   context('returns a List', function()
     -- Some OS specific variables the system would have set.
     local function base_env()
-      if iswin() then
+      if is_os('win') then
         return {
           HOME='C:\\Users\\docwhat', -- technically, is not a usual PATH
           HOMEDRIVE='C:',
