@@ -4,21 +4,27 @@ local assert_alive = helpers.assert_alive
 local feed, clear, nvim = helpers.feed, helpers.clear, helpers.nvim
 local poke_eventloop = helpers.poke_eventloop
 local eval, feed_command, source = helpers.eval, helpers.feed_command, helpers.source
+local pcall_err = helpers.pcall_err
 local eq, neq = helpers.eq, helpers.neq
+local meths = helpers.meths
+local retry = helpers.retry
 local write_file = helpers.write_file
 local command = helpers.command
 local exc_exec = helpers.exc_exec
 local matches = helpers.matches
 local exec_lua = helpers.exec_lua
 local sleep = helpers.sleep
+local funcs = helpers.funcs
+local is_os = helpers.is_os
+local skip = helpers.skip
+local nvim_prog = helpers.nvim_prog
 
 describe(':terminal buffer', function()
   local screen
 
   before_each(function()
     clear()
-    feed_command('set modifiable swapfile undolevels=20')
-    poke_eventloop()
+    command('set modifiable swapfile undolevels=20')
     screen = thelpers.screen_setup()
   end)
 
@@ -191,12 +197,11 @@ describe(':terminal buffer', function()
 
   it('handles loss of focus gracefully', function()
     -- Change the statusline to avoid printing the file name, which varies.
-    nvim('set_option', 'statusline', '==========')
-    feed_command('set laststatus=0')
+    nvim('set_option_value', 'statusline', '==========', {})
 
     -- Save the buffer number of the terminal for later testing.
     local tbuf = eval('bufnr("%")')
-    local exitcmd = helpers.iswin()
+    local exitcmd = is_os('win')
       and "['cmd', '/c', 'exit']"
       or "['sh', '-c', 'exit']"
     source([[
@@ -225,8 +230,6 @@ describe(':terminal buffer', function()
     neq(tbuf, eval('bufnr("%")'))
     feed_command('quit!')  -- Should exit the new window, not the terminal.
     eq(tbuf, eval('bufnr("%")'))
-
-    feed_command('set laststatus=1')  -- Restore laststatus to the default.
   end)
 
   it('term_close() use-after-free #4393', function()
@@ -258,7 +261,7 @@ describe(':terminal buffer', function()
   end)
 
   it('it works with set rightleft #11438', function()
-    if helpers.pending_win32(pending) then return end
+    skip(is_os('win'))
     local columns = eval('&columns')
     feed(string.rep('a', columns))
     command('set rightleft')
@@ -298,6 +301,49 @@ describe(':terminal buffer', function()
     feed_command('put a')  -- register a is empty
     helpers.assert_alive()
   end)
+
+  it([[can use temporary normal mode <c-\><c-o>]], function()
+    eq('t', funcs.mode(1))
+    feed [[<c-\><c-o>]]
+    screen:expect{grid=[[
+      tty ready                                         |
+      {2:^ }                                                 |
+                                                        |
+                                                        |
+                                                        |
+                                                        |
+      {3:-- (terminal) --}                                  |
+    ]]}
+    eq('ntT', funcs.mode(1))
+
+    feed [[:let g:x = 17]]
+    screen:expect{grid=[[
+      tty ready                                         |
+      {2: }                                                 |
+                                                        |
+                                                        |
+                                                        |
+                                                        |
+      :let g:x = 17^                                     |
+    ]]}
+
+    feed [[<cr>]]
+    screen:expect{grid=[[
+      tty ready                                         |
+      {1: }                                                 |
+                                                        |
+                                                        |
+                                                        |
+                                                        |
+      {3:-- TERMINAL --}                                    |
+    ]]}
+    eq('t', funcs.mode(1))
+  end)
+
+  it('writing to an existing file with :w fails #13549', function()
+    eq('Vim(write):E13: File exists (add ! to override)',
+       pcall_err(command, 'write test/functional/fixtures/tty-test.c'))
+  end)
 end)
 
 describe('No heap-buffer-overflow when using', function()
@@ -364,3 +410,135 @@ describe('on_lines does not emit out-of-bounds line indexes when', function()
     eq('', exec_lua([[return _G.cb_error]]))
   end)
 end)
+
+it('terminal truncates number of composing characters to 5', function()
+  clear()
+  local chan = meths.open_term(0, {})
+  local composing = ('a̳'):sub(2)
+  meths.chan_send(chan, 'a' .. composing:rep(8))
+  retry(nil, nil, function() eq('a' .. composing:rep(5), meths.get_current_line()) end)
+end)
+
+describe('terminal input', function()
+  before_each(function()
+    clear()
+    exec_lua([[
+      _G.input_data = ''
+      vim.api.nvim_open_term(0, { on_input = function(_, _, _, data)
+        _G.input_data = _G.input_data .. data
+      end })
+    ]])
+    feed('i')
+    poke_eventloop()
+  end)
+
+  it('<C-Space> is sent as NUL byte', function()
+    feed('aaa<C-Space>bbb')
+    eq('aaa\0bbb', exec_lua([[return _G.input_data]]))
+  end)
+
+  it('unknown special keys are not sent', function()
+    feed('aaa<Help>bbb')
+    eq('aaabbb', exec_lua([[return _G.input_data]]))
+  end)
+end)
+
+describe('terminal input', function()
+  it('sends various special keys with modifiers', function()
+    clear()
+    local screen = thelpers.screen_setup(0,
+      string.format([=[["%s", "-u", "NONE", "-i", "NONE", "--cmd", "startinsert"]]=], nvim_prog))
+    screen:expect{grid=[[
+      {1: }                                                 |
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {4:~                                                 }|
+      {5:[No Name]                       0,1            All}|
+      {3:-- INSERT --}                                      |
+      {3:-- TERMINAL --}                                    |
+    ]]}
+    for _, key in ipairs({
+      '<M-Tab>', '<M-CR>', '<M-Esc>',
+      '<BS>', '<S-Tab>', '<Insert>', '<Del>', '<PageUp>', '<PageDown>',
+      '<S-Up>', '<C-Up>', '<Up>', '<S-Down>', '<C-Down>', '<Down>',
+      '<S-Left>', '<C-Left>', '<Left>', '<S-Right>', '<C-Right>', '<Right>',
+      '<S-Home>', '<C-Home>', '<Home>', '<S-End>', '<C-End>', '<End>',
+      '<C-LeftMouse>', '<C-LeftRelease>', '<2-LeftMouse>', '<2-LeftRelease>',
+      '<S-RightMouse>', '<S-RightRelease>', '<2-RightMouse>', '<2-RightRelease>',
+      '<M-MiddleMouse>', '<M-MiddleRelease>', '<2-MiddleMouse>', '<2-MiddleRelease>',
+      '<S-ScrollWheelUp>', '<S-ScrollWheelDown>', '<ScrollWheelUp>', '<ScrollWheelDown>',
+      '<S-ScrollWheelLeft>', '<S-ScrollWheelRight>', '<ScrollWheelLeft>', '<ScrollWheelRight>',
+    }) do
+      feed('<CR><C-V>' .. key)
+      retry(nil, nil, function() eq(key, meths.get_current_line()) end)
+    end
+  end)
+end)
+
+if is_os('win') then
+  describe(':terminal in Windows', function()
+    local screen
+
+    before_each(function()
+      clear()
+      feed_command('set modifiable swapfile undolevels=20')
+      poke_eventloop()
+      local cmd = '["cmd.exe","/K","PROMPT=$g$s"]'
+      screen = thelpers.screen_setup(nil, cmd)
+    end)
+
+    it('"put" operator sends data normally', function()
+      feed('<c-\\><c-n>G')
+      feed_command('let @a = ":: tty ready"')
+      feed_command('let @a = @a . "\\n:: appended " . @a . "\\n\\n"')
+      feed('"ap"ap')
+      screen:expect([[
+                                                        |
+      > :: tty ready                                    |
+      > :: appended :: tty ready                        |
+      > :: tty ready                                    |
+      > :: appended :: tty ready                        |
+      ^> {2: }                                               |
+      :let @a = @a . "\n:: appended " . @a . "\n\n"     |
+      ]])
+      -- operator count is also taken into consideration
+      feed('3"ap')
+      screen:expect([[
+      > :: appended :: tty ready                        |
+      > :: tty ready                                    |
+      > :: appended :: tty ready                        |
+      > :: tty ready                                    |
+      > :: appended :: tty ready                        |
+      ^> {2: }                                               |
+      :let @a = @a . "\n:: appended " . @a . "\n\n"     |
+      ]])
+    end)
+
+    it('":put" command sends data normally', function()
+      feed('<c-\\><c-n>G')
+      feed_command('let @a = ":: tty ready"')
+      feed_command('let @a = @a . "\\n:: appended " . @a . "\\n\\n"')
+      feed_command('put a')
+      screen:expect([[
+                                                        |
+      > :: tty ready                                    |
+      > :: appended :: tty ready                        |
+      > {2: }                                               |
+                                                        |
+      ^                                                  |
+      :put a                                            |
+      ]])
+      -- line argument is only used to move the cursor
+      feed_command('6put a')
+      screen:expect([[
+                                                        |
+      > :: tty ready                                    |
+      > :: appended :: tty ready                        |
+      > :: tty ready                                    |
+      > :: appended :: tty ready                        |
+      ^> {2: }                                               |
+      :6put a                                           |
+      ]])
+    end)
+  end)
+end
