@@ -5,9 +5,12 @@ local feed, clear, nvim = helpers.feed, helpers.clear, helpers.nvim
 local testprg, command = helpers.testprg, helpers.command
 local nvim_prog = helpers.nvim_prog
 local eq, eval = helpers.eq, helpers.eval
-local feed_command = helpers.feed_command
+local matches = helpers.matches
+local poke_eventloop = helpers.poke_eventloop
 local hide_cursor = thelpers.hide_cursor
 local show_cursor = thelpers.show_cursor
+local is_os = helpers.is_os
+local skip = helpers.skip
 
 describe(':terminal cursor', function()
   local screen
@@ -87,7 +90,7 @@ describe(':terminal cursor', function()
 
   describe('when invisible', function()
     it('is not highlighted and is detached from screen cursor', function()
-      if helpers.pending_win32(pending) then return end
+      skip(is_os('win'))
       hide_cursor()
       screen:expect([[
         tty ready                                         |
@@ -150,7 +153,8 @@ describe('cursor with customized highlighting', function()
     })
     screen:attach({rgb=false})
     command('call termopen(["'..testprg('tty-test')..'"])')
-    feed_command('startinsert')
+    feed('i')
+    poke_eventloop()
   end)
 
   it('overrides the default highlighting', function()
@@ -177,7 +181,6 @@ describe('cursor with customized highlighting', function()
 end)
 
 describe('buffer cursor position is correct in terminal without number column', function()
-  if helpers.pending_win32(pending) then return end
   local screen
 
   local function setup_ex_register(str)
@@ -185,13 +188,16 @@ describe('buffer cursor position is correct in terminal without number column',
       ..[[", "-u", "NONE", "-i", "NONE", "-E", "--cmd", "let @r = ']]..str..[['", ]]
       -- <Left> and <Right> don't always work
       ..[["--cmd", "cnoremap <C-X> <Left>", "--cmd", "cnoremap <C-O> <Right>"]]..']', 70)
+    -- Also check for real cursor position, as it is used for stuff like input methods
+    screen._handle_busy_start = function() end
+    screen._handle_busy_stop = function() end
     screen:expect([[
                                                                             |
                                                                             |
                                                                             |
                                                                             |
       Entering Ex mode.  Type "visual" to go to Normal mode.                |
-      :{1: }                                                                    |
+      :{1:^ }                                                                    |
       {3:-- TERMINAL --}                                                        |
     ]])
   end
@@ -211,7 +217,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :aaaaaaaa{1: }                                                            |
+        :aaaaaaaa{1:^ }                                                            |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 9}, eval('nvim_win_get_cursor(0)'))
@@ -236,7 +242,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :aaaaaa{1:a}a                                                             |
+        :aaaaaa{1:^a}a                                                             |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 7}, eval('nvim_win_get_cursor(0)'))
@@ -261,7 +267,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :a{1:a}aaaaaa                                                             |
+        :a{1:^a}aaaaaa                                                             |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 2}, eval('nvim_win_get_cursor(0)'))
@@ -292,7 +298,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :µµµµµµµµ{1: }                                                            |
+        :µµµµµµµµ{1:^ }                                                            |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 17}, eval('nvim_win_get_cursor(0)'))
@@ -317,7 +323,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :µµµµµµ{1:µ}µ                                                             |
+        :µµµµµµ{1:^µ}µ                                                             |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 13}, eval('nvim_win_get_cursor(0)'))
@@ -342,7 +348,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :µ{1:µ}µµµµµµ                                                             |
+        :µ{1:^µ}µµµµµµ                                                             |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 3}, eval('nvim_win_get_cursor(0)'))
@@ -361,7 +367,7 @@ describe('buffer cursor position is correct in terminal without number column',
   end)
 
   describe('in a line with single-cell composed multibyte characters and no trailing spaces,', function()
-    if helpers.pending_win32(pending) then return end  -- These tests fail on Windows. Encoding problem?
+    if skip(is_os('win'), "Encoding problem?") then return end
 
     before_each(function()
       setup_ex_register('µ̳µ̳µ̳µ̳µ̳µ̳µ̳µ̳')
@@ -375,7 +381,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :µ̳µ̳µ̳µ̳µ̳µ̳µ̳µ̳{1: }                                                            |
+        :µ̳µ̳µ̳µ̳µ̳µ̳µ̳µ̳{1:^ }                                                            |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 33}, eval('nvim_win_get_cursor(0)'))
@@ -400,7 +406,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :µ̳µ̳µ̳µ̳µ̳µ̳{1:µ̳}µ̳                                                             |
+        :µ̳µ̳µ̳µ̳µ̳µ̳{1:^µ̳}µ̳                                                             |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 25}, eval('nvim_win_get_cursor(0)'))
@@ -425,7 +431,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :µ̳{1:µ̳}µ̳µ̳µ̳µ̳µ̳µ̳                                                             |
+        :µ̳{1:^µ̳}µ̳µ̳µ̳µ̳µ̳µ̳                                                             |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 5}, eval('nvim_win_get_cursor(0)'))
@@ -444,7 +450,7 @@ describe('buffer cursor position is correct in terminal without number column',
   end)
 
   describe('in a line with double-cell multibyte characters and no trailing spaces,', function()
-    if helpers.pending_win32(pending) then return end  -- These tests fail on Windows. Encoding problem?
+    if skip(is_os('win'), "Encoding problem?") then return end
 
     before_each(function()
       setup_ex_register('哦哦哦哦哦哦哦哦')
@@ -458,7 +464,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :哦哦哦哦哦哦哦哦{1: }                                                    |
+        :哦哦哦哦哦哦哦哦{1:^ }                                                    |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 25}, eval('nvim_win_get_cursor(0)'))
@@ -483,7 +489,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :哦哦哦哦哦哦{1:哦}哦                                                     |
+        :哦哦哦哦哦哦{1:^哦}哦                                                     |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 19}, eval('nvim_win_get_cursor(0)'))
@@ -508,7 +514,7 @@ describe('buffer cursor position is correct in terminal without number column',
                                                                               |
                                                                               |
         Entering Ex mode.  Type "visual" to go to Normal mode.                |
-        :哦{1:哦}哦哦哦哦哦哦                                                     |
+        :哦{1:^哦}哦哦哦哦哦哦                                                     |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 4}, eval('nvim_win_get_cursor(0)'))
@@ -525,10 +531,36 @@ describe('buffer cursor position is correct in terminal without number column',
       eq({6, 1}, eval('nvim_win_get_cursor(0)'))
     end)
   end)
+
+  it('at the end of a line with trailing spaces #16234', function()
+    setup_ex_register('aaaaaaaa    ')
+    feed('<C-R>r')
+    screen:expect([[
+                                                                            |
+                                                                            |
+                                                                            |
+                                                                            |
+      Entering Ex mode.  Type "visual" to go to Normal mode.                |
+      :aaaaaaaa    {1:^ }                                                        |
+      {3:-- TERMINAL --}                                                        |
+    ]])
+    matches('^:aaaaaaaa    [ ]*$', eval('nvim_get_current_line()'))
+    eq({6, 13}, eval('nvim_win_get_cursor(0)'))
+    feed([[<C-\><C-N>]])
+    screen:expect([[
+                                                                            |
+                                                                            |
+                                                                            |
+                                                                            |
+      Entering Ex mode.  Type "visual" to go to Normal mode.                |
+      :aaaaaaaa   ^ {2: }                                                        |
+                                                                            |
+    ]])
+    eq({6, 12}, eval('nvim_win_get_cursor(0)'))
+  end)
 end)
 
 describe('buffer cursor position is correct in terminal with number column', function()
-  if helpers.pending_win32(pending) then return end
   local screen
 
   local function setup_ex_register(str)
@@ -536,13 +568,16 @@ describe('buffer cursor position is correct in terminal with number column', fun
       ..[[", "-u", "NONE", "-i", "NONE", "-E", "--cmd", "let @r = ']]..str..[['", ]]
       -- <Left> and <Right> don't always work
       ..[["--cmd", "cnoremap <C-X> <Left>", "--cmd", "cnoremap <C-O> <Right>"]]..']', 70)
+    -- Also check for real cursor position, as it is used for stuff like input methods
+    screen._handle_busy_start = function() end
+    screen._handle_busy_stop = function() end
     screen:expect([[
       {7:  1 }                                                                  |
       {7:  2 }                                                                  |
       {7:  3 }                                                                  |
       {7:  4 }                                                                  |
       {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-      {7:  6 }:{1: }                                                                |
+      {7:  6 }:{1:^ }                                                                |
       {3:-- TERMINAL --}                                                        |
     ]])
   end
@@ -565,7 +600,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:aaaaaaaa{1: }                                                        |
+        {7:  6 }:aaaaaaaa{1:^ }                                                        |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 9}, eval('nvim_win_get_cursor(0)'))
@@ -590,7 +625,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:aaaaaa{1:a}a                                                         |
+        {7:  6 }:aaaaaa{1:^a}a                                                         |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 7}, eval('nvim_win_get_cursor(0)'))
@@ -615,7 +650,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:a{1:a}aaaaaa                                                         |
+        {7:  6 }:a{1:^a}aaaaaa                                                         |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 2}, eval('nvim_win_get_cursor(0)'))
@@ -646,7 +681,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:µµµµµµµµ{1: }                                                        |
+        {7:  6 }:µµµµµµµµ{1:^ }                                                        |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 17}, eval('nvim_win_get_cursor(0)'))
@@ -671,7 +706,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:µµµµµµ{1:µ}µ                                                         |
+        {7:  6 }:µµµµµµ{1:^µ}µ                                                         |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 13}, eval('nvim_win_get_cursor(0)'))
@@ -696,7 +731,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:µ{1:µ}µµµµµµ                                                         |
+        {7:  6 }:µ{1:^µ}µµµµµµ                                                         |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 3}, eval('nvim_win_get_cursor(0)'))
@@ -715,7 +750,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
   end)
 
   describe('in a line with single-cell composed multibyte characters and no trailing spaces,', function()
-    if helpers.pending_win32(pending) then return end  -- These tests fail on Windows. Encoding problem?
+    if skip(is_os('win'), "Encoding problem?") then return end
 
     before_each(function()
       setup_ex_register('µ̳µ̳µ̳µ̳µ̳µ̳µ̳µ̳')
@@ -729,7 +764,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:µ̳µ̳µ̳µ̳µ̳µ̳µ̳µ̳{1: }                                                        |
+        {7:  6 }:µ̳µ̳µ̳µ̳µ̳µ̳µ̳µ̳{1:^ }                                                        |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 33}, eval('nvim_win_get_cursor(0)'))
@@ -754,7 +789,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:µ̳µ̳µ̳µ̳µ̳µ̳{1:µ̳}µ̳                                                         |
+        {7:  6 }:µ̳µ̳µ̳µ̳µ̳µ̳{1:^µ̳}µ̳                                                         |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 25}, eval('nvim_win_get_cursor(0)'))
@@ -779,7 +814,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:µ̳{1:µ̳}µ̳µ̳µ̳µ̳µ̳µ̳                                                         |
+        {7:  6 }:µ̳{1:^µ̳}µ̳µ̳µ̳µ̳µ̳µ̳                                                         |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 5}, eval('nvim_win_get_cursor(0)'))
@@ -798,7 +833,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
   end)
 
   describe('in a line with double-cell multibyte characters and no trailing spaces,', function()
-    if helpers.pending_win32(pending) then return end  -- These tests fail on Windows. Encoding problem?
+    if skip(is_os('win'), "Encoding problem?") then return end
 
     before_each(function()
       setup_ex_register('哦哦哦哦哦哦哦哦')
@@ -812,7 +847,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:哦哦哦哦哦哦哦哦{1: }                                                |
+        {7:  6 }:哦哦哦哦哦哦哦哦{1:^ }                                                |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 25}, eval('nvim_win_get_cursor(0)'))
@@ -837,7 +872,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:哦哦哦哦哦哦{1:哦}哦                                                 |
+        {7:  6 }:哦哦哦哦哦哦{1:^哦}哦                                                 |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 19}, eval('nvim_win_get_cursor(0)'))
@@ -862,7 +897,7 @@ describe('buffer cursor position is correct in terminal with number column', fun
         {7:  3 }                                                                  |
         {7:  4 }                                                                  |
         {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
-        {7:  6 }:哦{1:哦}哦哦哦哦哦哦                                                 |
+        {7:  6 }:哦{1:^哦}哦哦哦哦哦哦                                                 |
         {3:-- TERMINAL --}                                                        |
       ]])
       eq({6, 4}, eval('nvim_win_get_cursor(0)'))
@@ -879,4 +914,31 @@ describe('buffer cursor position is correct in terminal with number column', fun
       eq({6, 1}, eval('nvim_win_get_cursor(0)'))
     end)
   end)
+
+  it('at the end of a line with trailing spaces #16234', function()
+    setup_ex_register('aaaaaaaa    ')
+    feed('<C-R>r')
+    screen:expect([[
+      {7:  1 }                                                                  |
+      {7:  2 }                                                                  |
+      {7:  3 }                                                                  |
+      {7:  4 }                                                                  |
+      {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
+      {7:  6 }:aaaaaaaa    {1:^ }                                                    |
+      {3:-- TERMINAL --}                                                        |
+    ]])
+    matches('^:aaaaaaaa    [ ]*$', eval('nvim_get_current_line()'))
+    eq({6, 13}, eval('nvim_win_get_cursor(0)'))
+    feed([[<C-\><C-N>]])
+    screen:expect([[
+      {7:  1 }                                                                  |
+      {7:  2 }                                                                  |
+      {7:  3 }                                                                  |
+      {7:  4 }                                                                  |
+      {7:  5 }Entering Ex mode.  Type "visual" to go to Normal mode.            |
+      {7:  6 }:aaaaaaaa   ^ {2: }                                                    |
+                                                                            |
+    ]])
+    eq({6, 12}, eval('nvim_win_get_cursor(0)'))
+  end)
 end)
