@@ -8,8 +8,10 @@ local feed_command, eval = helpers.feed_command, helpers.eval
 local funcs = helpers.funcs
 local retry = helpers.retry
 local ok = helpers.ok
-local iswin = helpers.iswin
 local command = helpers.command
+local skip = helpers.skip
+local is_os = helpers.is_os
+local is_ci = helpers.is_ci
 
 describe(':terminal', function()
   local screen
@@ -45,8 +47,8 @@ describe(':terminal', function()
   end)
 
   it("reads output buffer on terminal reporting #4151", function()
-    if helpers.pending_win32(pending) then return end
-    if iswin() then
+    skip(is_ci('cirrus') or is_os('win'))
+    if is_os('win') then
       feed_command([[terminal powershell -NoProfile -NoLogo -Command Write-Host -NoNewline "\"$([char]27)[6n\""; Start-Sleep -Milliseconds 500 ]])
     else
       feed_command([[terminal printf '\e[6n'; sleep 0.5 ]])
@@ -55,7 +57,7 @@ describe(':terminal', function()
   end)
 
   it("in normal-mode :split does not move cursor", function()
-    if iswin() then
+    if is_os('win') then
       feed_command([[terminal for /L \\%I in (1,0,2) do ( echo foo & ping -w 100 -n 1 127.0.0.1 > nul )]])
     else
       feed_command([[terminal while true; do echo foo; sleep .1; done]])
@@ -97,7 +99,7 @@ describe(':terminal', function()
   end)
 
   it('nvim_get_mode() in :terminal', function()
-    command(':terminal')
+    command('terminal')
     eq({ blocking=false, mode='nt' }, nvim('get_mode'))
     feed('i')
     eq({ blocking=false, mode='t' }, nvim('get_mode'))
@@ -106,24 +108,46 @@ describe(':terminal', function()
   end)
 
   it(':stopinsert RPC request exits terminal-mode #7807', function()
-    command(':terminal')
+    command('terminal')
     feed('i[tui] insert-mode')
     eq({ blocking=false, mode='t' }, nvim('get_mode'))
     command('stopinsert')
+    feed('<Ignore>')  -- Add input to separate two RPC requests
     eq({ blocking=false, mode='nt' }, nvim('get_mode'))
   end)
 
   it(':stopinsert in normal mode doesn\'t break insert mode #9889', function()
-    command(':terminal')
+    command('terminal')
     eq({ blocking=false, mode='nt' }, nvim('get_mode'))
-    command(':stopinsert')
+    command('stopinsert')
+    feed('<Ignore>')  -- Add input to separate two RPC requests
     eq({ blocking=false, mode='nt' }, nvim('get_mode'))
     feed('a')
     eq({ blocking=false, mode='t' }, nvim('get_mode'))
   end)
+
+  it('switching to terminal buffer in Insert mode goes to Terminal mode #7164', function()
+    command('terminal')
+    command('vnew')
+    feed('i')
+    command('let g:events = []')
+    command('autocmd InsertLeave * let g:events += ["InsertLeave"]')
+    command('autocmd TermEnter * let g:events += ["TermEnter"]')
+    command('inoremap <F2> <Cmd>wincmd p<CR>')
+    eq({ blocking=false, mode='i' }, nvim('get_mode'))
+    feed('<F2>')
+    eq({ blocking=false, mode='t' }, nvim('get_mode'))
+    eq({'InsertLeave', 'TermEnter'}, eval('g:events'))
+  end)
 end)
 
-describe(':terminal (with fake shell)', function()
+local function test_terminal_with_fake_shell(backslash)
+  -- shell-test.c is a fake shell that prints its arguments and exits.
+  local shell_path = testprg('shell-test')
+  if backslash then
+    shell_path = shell_path:gsub('/', [[\]])
+  end
+
   local screen
 
   before_each(function()
@@ -131,7 +155,7 @@ describe(':terminal (with fake shell)', function()
     screen = Screen.new(50, 4)
     screen:attach({rgb=false})
     -- shell-test.c is a fake shell that prints its arguments and exits.
-    nvim('set_option', 'shell', testprg('shell-test'))
+    nvim('set_option', 'shell', shell_path)
     nvim('set_option', 'shellcmdflag', 'EXE')
   end)
 
@@ -142,7 +166,6 @@ describe(':terminal (with fake shell)', function()
   end
 
   it('with no argument, acts like termopen()', function()
-    if helpers.pending_win32(pending) then return end
     terminal_with_fake_shell()
     retry(nil, 4 * screen.timeout, function()
     screen:expect([[
@@ -166,8 +189,7 @@ describe(':terminal (with fake shell)', function()
   end)
 
   it("with no argument, but 'shell' has arguments, acts like termopen()", function()
-    if helpers.pending_win32(pending) then return end
-    nvim('set_option', 'shell', testprg('shell-test')..' -t jeff')
+    nvim('set_option', 'shell', shell_path ..' -t jeff')
     terminal_with_fake_shell()
     screen:expect([[
       ^jeff $                                            |
@@ -178,7 +200,6 @@ describe(':terminal (with fake shell)', function()
   end)
 
   it('executes a given command through the shell', function()
-    if helpers.pending_win32(pending) then return end
     command('set shellxquote=')   -- win: avoid extra quotes
     terminal_with_fake_shell('echo hi')
     screen:expect([[
@@ -190,8 +211,7 @@ describe(':terminal (with fake shell)', function()
   end)
 
   it("executes a given command through the shell, when 'shell' has arguments", function()
-    if helpers.pending_win32(pending) then return end
-    nvim('set_option', 'shell', testprg('shell-test')..' -t jeff')
+    nvim('set_option', 'shell', shell_path ..' -t jeff')
     command('set shellxquote=')   -- win: avoid extra quotes
     terminal_with_fake_shell('echo hi')
     screen:expect([[
@@ -203,7 +223,6 @@ describe(':terminal (with fake shell)', function()
   end)
 
   it('allows quotes and slashes', function()
-    if helpers.pending_win32(pending) then return end
     command('set shellxquote=')   -- win: avoid extra quotes
     terminal_with_fake_shell([[echo 'hello' \ "world"]])
     screen:expect([[
@@ -240,7 +259,6 @@ describe(':terminal (with fake shell)', function()
   end)
 
   it('works with :find', function()
-    if helpers.pending_win32(pending) then return end
     terminal_with_fake_shell()
     screen:expect([[
       ^ready $                                           |
@@ -251,7 +269,7 @@ describe(':terminal (with fake shell)', function()
     eq('term://', string.match(eval('bufname("%")'), "^term://"))
     feed([[<C-\><C-N>]])
     feed_command([[find */shadacat.py]])
-    if iswin() then
+    if is_os('win') then
       eq('scripts\\shadacat.py', eval('bufname("%")'))
     else
       eq('scripts/shadacat.py', eval('bufname("%")'))
@@ -259,7 +277,6 @@ describe(':terminal (with fake shell)', function()
   end)
 
   it('works with gf', function()
-    if helpers.pending_win32(pending) then return end
     command('set shellxquote=')   -- win: avoid extra quotes
     terminal_with_fake_shell([[echo "scripts/shadacat.py"]])
     retry(nil, 4 * screen.timeout, function()
@@ -286,4 +303,13 @@ describe(':terminal (with fake shell)', function()
       terminal]])
     end
   end)
+end
+
+describe(':terminal (with fake shell)', function()
+  test_terminal_with_fake_shell(false)
+  if is_os('win') then
+    describe("when 'shell' uses backslashes", function()
+      test_terminal_with_fake_shell(true)
+    end)
+  end
 end)
