@@ -3,6 +3,8 @@ local thelpers = require('test.functional.terminal.helpers')
 local clear, eq, eval = helpers.clear, helpers.eq, helpers.eval
 local feed, nvim, command = helpers.feed, helpers.nvim, helpers.command
 local feed_data = thelpers.feed_data
+local is_os = helpers.is_os
+local skip = helpers.skip
 
 describe(':terminal mouse', function()
   local screen
@@ -65,9 +67,23 @@ describe(':terminal mouse', function()
       eq('nt', eval('mode(1)'))
     end)
 
-    it('does not leave terminal mode on left-release', function()
-      if helpers.pending_win32(pending) then return end
-      feed('<LeftRelease>')
+    it('will not exit focus on left-release', function()
+      eq('t', eval('mode(1)'))
+      feed('<LeftRelease><0,0>')
+      eq('t', eval('mode(1)'))
+      command('setlocal number')
+      eq('t', eval('mode(1)'))
+      feed('<LeftRelease><0,0>')
+      eq('t', eval('mode(1)'))
+    end)
+
+    it('will not exit focus on mouse movement', function()
+      eq('t', eval('mode(1)'))
+      feed('<MouseMove><0,0>')
+      eq('t', eval('mode(1)'))
+      command('setlocal number')
+      eq('t', eval('mode(1)'))
+      feed('<MouseMove><0,0>')
       eq('t', eval('mode(1)'))
     end)
 
@@ -87,7 +103,7 @@ describe(':terminal mouse', function()
       end)
 
       it('will forward mouse press, drag and release to the program', function()
-        if helpers.pending_win32(pending) then return end
+        skip(is_os('win'))
         feed('<LeftMouse><1,2>')
         screen:expect([[
           line27                                            |
@@ -131,7 +147,7 @@ describe(':terminal mouse', function()
       end)
 
       it('will forward mouse scroll to the program', function()
-        if helpers.pending_win32(pending) then return end
+        skip(is_os('win'))
         feed('<ScrollWheelUp><0,0>')
         screen:expect([[
           line27                                            |
@@ -145,7 +161,7 @@ describe(':terminal mouse', function()
       end)
 
       it('dragging and scrolling do not interfere with each other', function()
-        if helpers.pending_win32(pending) then return end
+        skip(is_os('win'))
         feed('<LeftMouse><1,2>')
         screen:expect([[
           line27                                            |
@@ -199,7 +215,7 @@ describe(':terminal mouse', function()
       end)
 
       it('will forward mouse clicks to the program with the correct even if set nu', function()
-        if helpers.pending_win32(pending) then return end
+        skip(is_os('win'))
         command('set number')
         -- When the display area such as a number is clicked, it returns to the
         -- normal mode.
@@ -230,7 +246,6 @@ describe(':terminal mouse', function()
     end)
 
     describe('with a split window and other buffer', function()
-      if helpers.pending_win32(pending) then return end
       before_each(function()
         feed('<c-\\><c-n>:vsp<cr>')
         screen:expect([[
@@ -287,7 +302,7 @@ describe(':terminal mouse', function()
         ]])
       end)
 
-      it('wont lose focus if another window is scrolled', function()
+      it("won't lose focus if another window is scrolled", function()
         feed('<ScrollWheelUp><4,0><ScrollWheelUp><4,0>')
         screen:expect([[
           {7: 21 }line                 │line30                  |
@@ -310,6 +325,34 @@ describe(':terminal mouse', function()
         ]])
       end)
 
+      it("scrolling another window respects 'mousescroll'", function()
+        command('set mousescroll=ver:1')
+        feed('<ScrollWheelUp><4,0>')
+        screen:expect([[
+          {7: 26 }line                 │line30                  |
+          {7: 27 }line                 │rows: 5, cols: 25       |
+          {7: 28 }line                 │rows: 5, cols: 24       |
+          {7: 29 }line                 │mouse enabled           |
+          {7: 30 }line                 │{1: }                       |
+          ==========                ==========              |
+          {3:-- TERMINAL --}                                    |
+        ]])
+        command('set mousescroll=ver:10')
+        feed('<ScrollWheelUp><0,0>')
+        screen:expect([[
+          {7: 16 }line                 │line30                  |
+          {7: 17 }line                 │rows: 5, cols: 25       |
+          {7: 18 }line                 │rows: 5, cols: 24       |
+          {7: 19 }line                 │mouse enabled           |
+          {7: 20 }line                 │{1: }                       |
+          ==========                ==========              |
+          {3:-- TERMINAL --}                                    |
+        ]])
+        command('set mousescroll=ver:0')
+        feed('<ScrollWheelUp><0,0>')
+        screen:expect_unchanged()
+      end)
+
       it('will lose focus if another window is clicked', function()
         feed('<LeftMouse><5,1>')
         screen:expect([[
