@@ -6,33 +6,39 @@ local command = helpers.command
 local exec_lua = helpers.exec_lua
 local pcall_err = helpers.pcall_err
 local matches = helpers.matches
-local pending_c_parser = helpers.pending_c_parser
+local insert = helpers.insert
 
 before_each(clear)
 
-describe('treesitter API', function()
+describe('treesitter language API', function()
   -- error tests not requiring a parser library
   it('handles missing language', function()
-    eq("Error executing lua: .../language.lua:0: no parser for 'borklang' language, see :help treesitter-parsers",
+    eq(".../language.lua:0: no parser for 'borklang' language, see :help treesitter-parsers",
        pcall_err(exec_lua, "parser = vim.treesitter.get_parser(0, 'borklang')"))
 
     -- actual message depends on platform
-    matches("Error executing lua: Failed to load parser: uv_dlopen: .+",
-       pcall_err(exec_lua, "parser = vim.treesitter.require_language('borklang', 'borkbork.so')"))
+    matches("Failed to load parser for language 'borklang': uv_dlopen: .+",
+       pcall_err(exec_lua, "parser = vim.treesitter.language.add('borklang', { path = 'borkbork.so' })"))
 
-    -- Should not throw an error when silent
-    eq(false, exec_lua("return vim.treesitter.require_language('borklang', nil, true)"))
-    eq(false, exec_lua("return vim.treesitter.require_language('borklang', 'borkbork.so', true)"))
+    eq(false, exec_lua("return pcall(vim.treesitter.language.add, 'borklang')"))
 
-    eq("Error executing lua: .../language.lua:0: no parser for 'borklang' language, see :help treesitter-parsers",
-       pcall_err(exec_lua, "parser = vim.treesitter.inspect_language('borklang')"))
+    eq(false, exec_lua("return pcall(vim.treesitter.language.add, 'borklang', { path = 'borkbork.so' })"))
+
+    eq(".../language.lua:0: no parser for 'borklang' language, see :help treesitter-parsers",
+       pcall_err(exec_lua, "parser = vim.treesitter.language.inspect('borklang')"))
+
+    matches("Failed to load parser: uv_dlsym: .+",
+       pcall_err(exec_lua, 'vim.treesitter.language.add("c", { symbol_name = "borklang" })'))
+  end)
+
+  it('shows error for invalid language name', function()
+    eq(".../language.lua:0: '/foo/' is not a valid language name",
+      pcall_err(exec_lua, 'vim.treesitter.language.add("/foo/")'))
   end)
 
   it('inspects language', function()
-    if pending_c_parser(pending) then return end
-
     local keys, fields, symbols = unpack(exec_lua([[
-      local lang = vim.treesitter.inspect_language('c')
+      local lang = vim.treesitter.language.inspect('c')
       local keys, symbols = {}, {}
       for k,_ in pairs(lang) do
         keys[k] = true
@@ -70,14 +76,41 @@ describe('treesitter API', function()
   end)
 
   it('checks if vim.treesitter.get_parser tries to create a new parser on filetype change', function ()
-    if pending_c_parser(pending) then return end
     command("set filetype=c")
     -- Should not throw an error when filetype is c
     eq('c', exec_lua("return vim.treesitter.get_parser(0):lang()"))
     command("set filetype=borklang")
     -- Should throw an error when filetype changes to borklang
-    eq("Error executing lua: .../language.lua:0: no parser for 'borklang' language, see :help treesitter-parsers",
-       pcall_err(exec_lua, "new_parser = vim.treesitter.get_parser(0)"))
+    eq(".../language.lua:0: no parser for 'borklang' language, see :help treesitter-parsers",
+       pcall_err(exec_lua, "new_parser = vim.treesitter.get_parser(0, 'borklang')"))
+  end)
+
+  it('retrieve the tree given a range', function ()
+    insert([[
+      int main() {
+        int x = 3;
+      }]])
+
+    exec_lua([[
+      langtree = vim.treesitter.get_parser(0, "c")
+      tree = langtree:tree_for_range({1, 3, 1, 3})
+    ]])
+
+    eq('<node translation_unit>', exec_lua('return tostring(tree:root())'))
+  end)
+
+  it('retrieve the node given a range', function ()
+    insert([[
+      int main() {
+        int x = 3;
+      }]])
+
+    exec_lua([[
+      langtree = vim.treesitter.get_parser(0, "c")
+      node = langtree:named_node_for_range({1, 3, 1, 3})
+    ]])
+
+    eq('<node primitive_type>', exec_lua('return tostring(node)'))
   end)
 end)
 
