@@ -4,18 +4,16 @@ local clear = helpers.clear
 local eq = helpers.eq
 local insert = helpers.insert
 local exec_lua = helpers.exec_lua
+local pcall_err = helpers.pcall_err
 local feed = helpers.feed
-local pending_c_parser = helpers.pending_c_parser
+local is_os = helpers.is_os
 
 before_each(clear)
 
 describe('treesitter parser API', function()
   clear()
-  if pending_c_parser(pending) then return end
 
   it('parses buffer', function()
-    if helpers.pending_win32(pending) then return end
-
     insert([[
       int main() {
         int x = 3;
@@ -25,7 +23,7 @@ describe('treesitter parser API', function()
       parser = vim.treesitter.get_parser(0, "c")
       tree = parser:parse()[1]
       root = tree:root()
-      lang = vim.treesitter.inspect_language('c')
+      lang = vim.treesitter.language.inspect('c')
     ]])
 
     eq("<tree>", exec_lua("return tostring(tree)"))
@@ -126,6 +124,18 @@ void ui_refresh(void)
     }, res)
   end)
 
+  it('does not get parser for empty filetype', function()
+    insert(test_text);
+
+    eq('.../treesitter.lua:0: There is no parser available for buffer 1 and one'
+         .. ' could not be created because lang could not be determined. Either'
+         .. ' pass lang or set the buffer filetype',
+      pcall_err(exec_lua, 'vim.treesitter.get_parser(0)'))
+
+    -- Must provide language for buffers with an empty filetype
+    exec_lua("vim.treesitter.get_parser(0, 'c')")
+  end)
+
   it('allows to get a child by field', function()
     insert(test_text);
 
@@ -161,7 +171,7 @@ void ui_refresh(void)
 
   it("supports runtime queries", function()
     local ret = exec_lua [[
-      return require"vim.treesitter.query".get_query("c", "highlights").captures[1]
+      return vim.treesitter.query.get("c", "highlights").captures[1]
     ]]
 
     eq('variable', ret)
@@ -169,30 +179,31 @@ void ui_refresh(void)
 
   it("supports caching queries", function()
     local long_query = query:rep(100)
-    local first_run = exec_lua ([[
-      local before = vim.loop.hrtime()
-      cquery = vim.treesitter.parse_query("c", ...)
-      local after = vim.loop.hrtime()
-      return after - before
-    ]], long_query)
+    local function q(n)
+      return exec_lua ([[
+        local query, n = ...
+        local before = vim.loop.hrtime()
+        for i=1,n,1 do
+          cquery = vim.treesitter.query.parse("c", ...)
+        end
+        local after = vim.loop.hrtime()
+        return after - before
+      ]], long_query, n)
+    end
 
-    local subsequent_runs = exec_lua ([[
-      local before = vim.loop.hrtime()
-      for i=1,100,1 do
-        cquery = vim.treesitter.parse_query("c", ...)
-      end
-      local after = vim.loop.hrtime()
-      return after - before
-    ]], long_query)
+    local firstrun = q(1)
+    local manyruns = q(100)
 
-    assert.True(1000 * subsequent_runs < first_run)
+    -- First run should be at least 200x slower than an 100 subsequent runs.
+    local factor = is_os('win') and 100 or 200
+    assert(factor * manyruns < firstrun, ('firstrun: %f ms, manyruns: %f ms'):format(firstrun / 1e6, manyruns / 1e6))
   end)
 
   it('support query and iter by capture', function()
     insert(test_text)
 
     local res = exec_lua([[
-      cquery = vim.treesitter.parse_query("c", ...)
+      cquery = vim.treesitter.query.parse("c", ...)
       parser = vim.treesitter.get_parser(0, "c")
       tree = parser:parse()[1]
       res = {}
@@ -221,7 +232,7 @@ void ui_refresh(void)
     insert(test_text)
 
     local res = exec_lua([[
-      cquery = vim.treesitter.parse_query("c", ...)
+      cquery = vim.treesitter.query.parse("c", ...)
       parser = vim.treesitter.get_parser(0, "c")
       tree = parser:parse()[1]
       res = {}
@@ -249,7 +260,6 @@ void ui_refresh(void)
   end)
 
   it('supports getting text of multiline node', function()
-    if pending_c_parser(pending) then return end
     insert(test_text)
     local res = exec_lua([[
       local parser = vim.treesitter.get_parser(0, "c")
@@ -266,13 +276,13 @@ void ui_refresh(void)
     eq('void', res2)
   end)
 
-  it('support getting text where start of node is past EOF', function()
+  it('support getting text where start of node is one past EOF', function()
     local text = [[
 def run
   a = <<~E
 end]]
     insert(text)
-    local result = exec_lua([[
+    eq('', exec_lua[[
       local fake_node = {}
       function fake_node:start()
         return 3, 0, 23
@@ -280,9 +290,14 @@ end]]
       function fake_node:end_()
         return 3, 0, 23
       end
-      return vim.treesitter.get_node_text(fake_node, 0) == nil
+      function fake_node:range(bytes)
+        if bytes then
+          return 3, 0, 23, 3, 0, 23
+        end
+        return 3, 0, 3, 0
+      end
+      return vim.treesitter.get_node_text(fake_node, 0)
     ]])
-    eq(true, result)
   end)
 
   it('support getting empty text if node range is zero width', function()
@@ -299,6 +314,9 @@ end]]
       function fake_node:end_()
         return 1, 0, 7
       end
+      function fake_node:range()
+        return 1, 0, 1, 0
+      end
       return vim.treesitter.get_node_text(fake_node, 0) == ''
     ]])
     eq(true, result)
@@ -308,7 +326,7 @@ end]]
     insert('char* astring = "\\n"; (1 + 1) * 2 != 2;')
 
     local res = exec_lua([[
-      cquery = vim.treesitter.parse_query("c", '([_] @plus (#vim-match? @plus "^\\\\+$"))'..
+      cquery = vim.treesitter.query.parse("c", '([_] @plus (#vim-match? @plus "^\\\\+$"))'..
                                                '([_] @times (#vim-match? @times "^\\\\*$"))'..
                                                '([_] @paren (#vim-match? @paren "^\\\\($"))'..
                                                '([_] @escape (#vim-match? @escape "^\\\\\\\\n$"))'..
@@ -358,7 +376,7 @@ end]]
     ]])
     exec_lua([[
       function get_query_result(query_text)
-        cquery = vim.treesitter.parse_query("c", query_text)
+        cquery = vim.treesitter.query.parse("c", query_text)
         parser = vim.treesitter.get_parser(0, "c")
         tree = parser:parse()[1]
         res = {}
@@ -398,7 +416,7 @@ end]]
     insert('char* astring = "Hello World!";')
 
     local res = exec_lua([[
-      cquery = vim.treesitter.parse_query("c", '([_] @quote (#vim-match? @quote "^\\"$")) ([_] @quote (#lua-match? @quote "^\\"$"))')
+      cquery = vim.treesitter.query.parse("c", '([_] @quote (#vim-match? @quote "^\\"$")) ([_] @quote (#lua-match? @quote "^\\"$"))')
       parser = vim.treesitter.get_parser(0, "c")
       tree = parser:parse()[1]
       res = {}
@@ -431,7 +449,7 @@ end]]
     local custom_query = "((identifier) @main (#is-main? @main))"
 
     local res = exec_lua([[
-    local query = require"vim.treesitter.query"
+    local query = vim.treesitter.query
 
     local function is_main(match, pattern, bufnr, predicate)
       local node = match[ predicate[2] ]
@@ -443,7 +461,7 @@ end]]
 
     query.add_predicate("is-main?", is_main)
 
-    local query = query.parse_query("c", ...)
+    local query = query.parse("c", ...)
 
     local nodes = {}
     for _, node in query:iter_captures(parser:parse()[1]:root(), 0) do
@@ -456,7 +474,7 @@ end]]
     eq({{0, 4, 0, 8}}, res)
 
     local res_list = exec_lua[[
-    local query = require'vim.treesitter.query'
+    local query = vim.treesitter.query
 
     local list = query.list_predicates()
 
@@ -515,7 +533,7 @@ end]]
 
     local res = exec_lua [[
     parser = vim.treesitter.get_parser(0, "c")
-    query = vim.treesitter.parse_query("c", "(declaration) @decl")
+    query = vim.treesitter.query.parse("c", "(declaration) @decl")
 
     local nodes = {}
     for _, node in query:iter_captures(parser:parse()[1]:root(), 0) do
@@ -563,7 +581,7 @@ end]]
     local parser = vim.treesitter.get_string_parser(str, "c")
 
     local nodes = {}
-    local query = vim.treesitter.parse_query("c", '((identifier) @id (eq? @id "foo"))')
+    local query = vim.treesitter.query.parse("c", '((identifier) @id (eq? @id "foo"))')
 
     for _, node in query:iter_captures(parser:parse()[1]:root(), str) do
       table.insert(nodes, { node:range() })
@@ -585,7 +603,7 @@ end]]
     local parser = vim.treesitter.get_string_parser(str, "c")
 
     local nodes = {}
-    local query = vim.treesitter.parse_query("c", '((identifier) @foo)')
+    local query = vim.treesitter.query.parse("c", '((identifier) @foo)')
     local first_child = parser:parse()[1]:root():child(1)
 
     for _, node in query:iter_captures(first_child, str) do
@@ -609,8 +627,8 @@ end]]
     before_each(function()
       insert([[
 int x = INT_MAX;
-#define READ_STRING(x, y) (char_u *)read_string((x), (size_t)(y))
-#define READ_STRING_OK(x, y) (char_u *)read_string((x), (size_t)(y))
+#define READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+#define READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
 #define VALUE 123
 #define VALUE1 123
 #define VALUE2 123
@@ -632,8 +650,19 @@ int x = INT_MAX;
           {3, 14, 3, 17}, -- VALUE 123
           {4, 15, 4, 18}, -- VALUE1 123
           {5, 15, 5, 18}, -- VALUE2 123
-          {1, 26, 1, 65}, -- READ_STRING(x, y) (char_u *)read_string((x), (size_t)(y))
-          {2, 29, 2, 68}  -- READ_STRING_OK(x, y) (char_u *)read_string((x), (size_t)(y))
+          {1, 26, 1, 63}, -- READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+          {2, 29, 2, 66}  -- READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
+        }, get_ranges())
+
+        helpers.feed('ggo<esc>')
+        eq(5, exec_lua("return #parser:children().c:trees()"))
+        eq({
+          {0, 0, 8, 0},   -- root tree
+          {4, 14, 4, 17}, -- VALUE 123
+          {5, 15, 5, 18}, -- VALUE1 123
+          {6, 15, 6, 18}, -- VALUE2 123
+          {2, 26, 2, 63}, -- READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+          {3, 29, 3, 66}  -- READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
         }, get_ranges())
       end)
     end)
@@ -653,8 +682,20 @@ int x = INT_MAX;
           {3, 14, 5, 18}, -- VALUE 123
                           -- VALUE1 123
                           -- VALUE2 123
-          {1, 26, 2, 68}  -- READ_STRING(x, y) (char_u *)read_string((x), (size_t)(y))
-                          -- READ_STRING_OK(x, y) (char_u *)read_string((x), (size_t)(y))
+          {1, 26, 2, 66}  -- READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+                          -- READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
+        }, get_ranges())
+
+        helpers.feed('ggo<esc>')
+        eq("table", exec_lua("return type(parser:children().c)"))
+        eq(2, exec_lua("return #parser:children().c:trees()"))
+        eq({
+          {0, 0, 8, 0},   -- root tree
+          {4, 14, 6, 18}, -- VALUE 123
+                          -- VALUE1 123
+                          -- VALUE2 123
+          {2, 26, 3, 66}  -- READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+                          -- READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
         }, get_ranges())
       end)
     end)
@@ -662,7 +703,7 @@ int x = INT_MAX;
     describe("when providing parsing information through a directive", function()
       it("should inject a language", function()
         exec_lua([=[
-        vim.treesitter.add_directive("inject-clang!", function(match, _, _, pred, metadata)
+        vim.treesitter.query.add_directive("inject-clang!", function(match, _, _, pred, metadata)
           metadata.language = "c"
           metadata.combined = true
           metadata.content = pred[2]
@@ -681,15 +722,26 @@ int x = INT_MAX;
           {3, 14, 5, 18}, -- VALUE 123
                           -- VALUE1 123
                           -- VALUE2 123
-          {1, 26, 2, 68}  -- READ_STRING(x, y) (char_u *)read_string((x), (size_t)(y))
-                          -- READ_STRING_OK(x, y) (char_u *)read_string((x), (size_t)(y))
+          {1, 26, 2, 66}  -- READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+                          -- READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
+        }, get_ranges())
+
+        helpers.feed('ggo<esc>')
+        eq("table", exec_lua("return type(parser:children().c)"))
+        eq(2, exec_lua("return #parser:children().c:trees()"))
+        eq({
+          {0, 0, 8, 0},   -- root tree
+          {4, 14, 6, 18}, -- VALUE 123
+                          -- VALUE1 123
+                          -- VALUE2 123
+          {2, 26, 3, 66}  -- READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+                          -- READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
         }, get_ranges())
       end)
 
       it("should not inject bad languages", function()
-        if helpers.pending_win32(pending) then return end
         exec_lua([=[
-        vim.treesitter.add_directive("inject-bad!", function(match, _, _, pred, metadata)
+        vim.treesitter.query.add_directive("inject-bad!", function(match, _, _, pred, metadata)
           metadata.language = "{"
           metadata.combined = true
           metadata.content = pred[2]
@@ -713,16 +765,16 @@ int x = INT_MAX;
         eq("table", exec_lua("return type(parser:children().c)"))
         eq({
           {0, 0, 7, 0},   -- root tree
-          {3, 15, 3, 16}, -- VALUE 123
-          {4, 16, 4, 17}, -- VALUE1 123
-          {5, 16, 5, 17}, -- VALUE2 123
-          {1, 26, 1, 65}, -- READ_STRING(x, y) (char_u *)read_string((x), (size_t)(y))
-          {2, 29, 2, 68}  -- READ_STRING_OK(x, y) (char_u *)read_string((x), (size_t)(y))
+          {3, 16, 3, 16}, -- VALUE 123
+          {4, 17, 4, 17}, -- VALUE1 123
+          {5, 17, 5, 17}, -- VALUE2 123
+          {1, 26, 1, 63}, -- READ_STRING(x, y) (char *)read_string((x), (size_t)(y))
+          {2, 29, 2, 66}  -- READ_STRING_OK(x, y) (char *)read_string((x), (size_t)(y))
         }, get_ranges())
       end)
       it("should list all directives", function()
         local res_list = exec_lua[[
-        local query = require'vim.treesitter.query'
+        local query = vim.treesitter.query
 
         local list = query.list_directives()
 
@@ -731,7 +783,7 @@ int x = INT_MAX;
         return list
         ]]
 
-        eq({ 'offset!', 'set!' }, res_list)
+        eq({ 'gsub!', 'offset!', 'set!' }, res_list)
       end)
     end)
   end)
@@ -768,7 +820,7 @@ int x = INT_MAX;
         local result = exec_lua([[
         local result
 
-        query = vim.treesitter.parse_query("c", '((number_literal) @number (#set! "key" "value"))')
+        query = vim.treesitter.query.parse("c", '((number_literal) @number (#set! "key" "value"))')
         parser = vim.treesitter.get_parser(0, "c")
 
         for pattern, match, metadata in query:iter_matches(parser:parse()[1]:root(), 0) do
@@ -788,10 +840,10 @@ int x = INT_MAX;
           ]])
 
           local result = exec_lua([[
-          local query = require("vim.treesitter.query")
+          local query = vim.treesitter.query
           local value
 
-          query = vim.treesitter.parse_query("c", '((number_literal) @number (#set! @number "key" "value"))')
+          query = vim.treesitter.query.parse("c", '((number_literal) @number (#set! @number "key" "value"))')
           parser = vim.treesitter.get_parser(0, "c")
 
           for pattern, match, metadata in query:iter_matches(parser:parse()[1]:root(), 0) do
@@ -810,10 +862,10 @@ int x = INT_MAX;
           ]])
 
           local result = exec_lua([[
-          local query = require("vim.treesitter.query")
+          local query = vim.treesitter.query
           local result
 
-          query = vim.treesitter.parse_query("c", '((number_literal) @number (#set! @number "key" "value") (#set! @number "key2" "value2"))')
+          query = vim.treesitter.query.parse("c", '((number_literal) @number (#set! @number "key" "value") (#set! @number "key2" "value2"))')
           parser = vim.treesitter.get_parser(0, "c")
 
           for pattern, match, metadata in query:iter_matches(parser:parse()[1]:root(), 0) do
@@ -832,4 +884,146 @@ int x = INT_MAX;
       end)
     end)
   end)
+
+  it("can fold via foldexpr", function()
+    insert(test_text)
+
+    local function get_fold_levels()
+      return exec_lua([[
+        local res = {}
+        for i = 1, vim.api.nvim_buf_line_count(0) do
+          res[i] = vim.treesitter.foldexpr(i)
+        end
+        return res
+      ]])
+    end
+
+    exec_lua([[vim.treesitter.get_parser(0, "c")]])
+
+    eq({
+      [1] = '>1',
+      [2] = '1',
+      [3] = '1',
+      [4] = '1',
+      [5] = '>2',
+      [6] = '2',
+      [7] = '2',
+      [8] = '1',
+      [9] = '1',
+      [10] = '>2',
+      [11] = '2',
+      [12] = '2',
+      [13] = '2',
+      [14] = '2',
+      [15] = '>3',
+      [16] = '3',
+      [17] = '3',
+      [18] = '2',
+      [19] = '1' }, get_fold_levels())
+
+    helpers.command('1,2d')
+    helpers.poke_eventloop()
+
+    exec_lua([[vim.treesitter.get_parser():parse()]])
+
+    helpers.poke_eventloop()
+    helpers.sleep(100)
+
+    eq({
+      [1] = '0',
+      [2] = '0',
+      [3] = '>1',
+      [4] = '1',
+      [5] = '1',
+      [6] = '0',
+      [7] = '0',
+      [8] = '>1',
+      [9] = '1',
+      [10] = '1',
+      [11] = '1',
+      [12] = '1',
+      [13] = '>2',
+      [14] = '2',
+      [15] = '2',
+      [16] = '1',
+      [17] = '0' }, get_fold_levels())
+  end)
+
+  it('tracks the root range properly (#22911)', function()
+    insert([[
+      int main() {
+        int x = 3;
+      }]])
+
+    local query0 = [[
+      (declaration) @declaration
+      (function_definition) @function
+    ]]
+
+    exec_lua([[
+      vim.treesitter.start(0, 'c')
+    ]])
+
+    local function run_query()
+      return exec_lua([[
+      local query = vim.treesitter.query.parse("c", ...)
+      parser = vim.treesitter.get_parser()
+      tree = parser:parse()[1]
+      res = {}
+      for id, node in query:iter_captures(tree:root()) do
+        table.insert(res, {query.captures[id], node:range()})
+      end
+      return res
+      ]], query0)
+    end
+
+    eq({
+      { 'function', 0, 0, 2, 1 },
+      { 'declaration', 1, 2, 1, 12 }
+    }, run_query())
+
+    helpers.command'normal ggO'
+    insert('int a;')
+
+    eq({
+      { 'declaration', 0, 0, 0, 6 },
+      { 'function', 1, 0, 3, 1 },
+      { 'declaration', 2, 2, 2, 12 }
+    }, run_query())
+
+  end)
+
+  it("does not produce empty injection ranges (#23409)", function()
+    insert [[
+      Examples: >lua
+        local a = {}
+<
+    ]]
+
+    -- This is not a valid injection since (code) has children and include-children is not set
+    exec_lua [[
+      parser1 = require('vim.treesitter.languagetree').new(0, "vimdoc", {
+        injections = {
+          vimdoc = "((codeblock (language) @injection.language (code) @injection.content))"
+        }
+      })
+      parser1:parse()
+    ]]
+
+    eq(0, exec_lua("return #vim.tbl_keys(parser1:children())"))
+
+    exec_lua [[
+      parser2 = require('vim.treesitter.languagetree').new(0, "vimdoc", {
+        injections = {
+          vimdoc = "((codeblock (language) @injection.language (code) @injection.content) (#set! injection.include-children))"
+        }
+      })
+      parser2:parse()
+    ]]
+
+    eq(1, exec_lua("return #vim.tbl_keys(parser2:children())"))
+    eq( { { { 1, 0, 21, 2, 0, 42 } } }, exec_lua("return parser2:children().lua:included_regions()"))
+
+  end)
+
 end)
