@@ -303,7 +303,7 @@ describe('Buffer highlighting', function()
         {1:~                                       }|
         {1:~                                       }|
         {1:~                                       }|
-        2 change3; before #3  {MATCH:.*}|
+        2 changes; before #3  {MATCH:.*}|
       ]]}
 
       command('undo')
@@ -564,6 +564,16 @@ describe('Buffer highlighting', function()
     ]]
 
     clear_namespace(id, 0, -1)
+    screen:expect{grid=[[
+      fooba^r                                  |
+      {1:~                                       }|
+      {1:~                                       }|
+      {1:~                                       }|
+      {1:~                                       }|
+      {1:~                                       }|
+      {1:~                                       }|
+                                              |
+    ]]}
 
     set_extmark(id, 0, 0, {
       end_line = 0,
@@ -741,8 +751,8 @@ describe('Buffer highlighting', function()
 
     it('validates contents', function()
       -- this used to leak memory
-      eq('Chunk is not an array', pcall_err(set_virtual_text, id1, 0, {"texty"}, {}))
-      eq('Chunk is not an array', pcall_err(set_virtual_text, id1, 0, {{"very"}, "texty"}, {}))
+      eq("Invalid 'chunk': expected Array, got String", pcall_err(set_virtual_text, id1, 0, {"texty"}, {}))
+      eq("Invalid 'chunk': expected Array, got String", pcall_err(set_virtual_text, id1, 0, {{"very"}, "texty"}, {}))
     end)
 
     it('can be retrieved', function()
@@ -752,10 +762,11 @@ describe('Buffer highlighting', function()
       local s1 = {{'Köttbullar', 'Comment'}, {'Kräuterbutter'}}
       local s2 = {{'こんにちは', 'Comment'}}
 
-      -- TODO: only a virtual text from the same ns curretly overrides
+      -- TODO: only a virtual text from the same ns currently overrides
       -- an existing virtual text. We might add a prioritation system.
       set_virtual_text(id1, 0, s1, {})
       eq({{1, 0, 0, {
+        ns_id = 1,
         priority = 0,
         virt_text = s1,
         -- other details
@@ -768,6 +779,7 @@ describe('Buffer highlighting', function()
       local lastline = line_count()
       set_virtual_text(id1, line_count(), s2, {})
       eq({{3, lastline, 0, {
+        ns_id = 1,
         priority = 0,
         virt_text = s2,
         -- other details
