@@ -4,9 +4,10 @@ local Screen = require('test.functional.ui.screen')
 local clear, insert = helpers.clear, helpers.insert
 local command = helpers.command
 local meths = helpers.meths
-local iswin = helpers.iswin
 local testprg = helpers.testprg
 local thelpers = require('test.functional.terminal.helpers')
+local skip = helpers.skip
+local is_os = helpers.is_os
 
 describe('ext_hlstate detailed highlights', function()
   local screen
@@ -14,6 +15,7 @@ describe('ext_hlstate detailed highlights', function()
   before_each(function()
     clear()
     command('syntax on')
+    command('hi VertSplit gui=reverse')
     screen = Screen.new(40, 8)
     screen:attach({ext_hlstate=true})
   end)
@@ -59,7 +61,7 @@ describe('ext_hlstate detailed highlights', function()
 
   it('work with cleared UI highlights', function()
     screen:set_default_attr_ids({
-      [1] = {{}, {{hi_name = "VertSplit", ui_name = "WinSeparator", kind = "ui"}}},
+      [1] = {{}, {{hi_name = "Normal", ui_name = "WinSeparator", kind = "ui"}}},
       [2] = {{bold = true, foreground = Screen.colors.Blue1},
              {{hi_name = "NonText", ui_name = "EndOfBuffer", kind = "ui"}}},
       [3] = {{bold = true, reverse = true},
@@ -118,13 +120,15 @@ describe('ext_hlstate detailed highlights', function()
         [3] = {{bold = true, reverse = true}, {{hi_name = "StatusLine", ui_name = "StatusLine", kind = "ui"}}},
         [4] = {{reverse = true}, {{hi_name = "StatusLineNC", ui_name = "StatusLineNC", kind = "ui"}}},
         [5] = {{background = Screen.colors.Red, foreground = Screen.colors.Grey100}, {{hi_name = "ErrorMsg", ui_name = "LineNr", kind = "ui"}}},
-        [6] = {{bold = true, reverse = true}, {{hi_name = "MsgSeparator", ui_name = "Normal", kind = "ui"}}},
+        [6] = {{bold = true, reverse = true}, {{hi_name = "Normal", ui_name = "Normal", kind = "ui"}}},
         [7] = {{foreground = Screen.colors.Brown, bold = true, reverse = true}, {6, 1}},
-        [8] = {{foreground = Screen.colors.Blue1, bold = true, reverse = true}, {6, 2}},
-        [9] = {{bold = true, foreground = Screen.colors.Brown}, {{hi_name = "Statement", ui_name = "NormalNC", kind = "ui"}}},
+        [8] = {{foreground = Screen.colors.Blue1, bold = true, reverse = true}, {6, 14}},
+        [9] = {{bold = true, foreground = Screen.colors.Brown}, {{hi_name = "NormalNC", ui_name = "NormalNC", kind = "ui"}}},
         [10] = {{bold = true, foreground = Screen.colors.Brown}, {9, 1}},
-        [11] = {{bold = true, foreground = Screen.colors.Blue1}, {9, 2}},
+        [11] = {{bold = true, foreground = Screen.colors.Blue1}, {9, 14}},
         [12] = {{}, {{hi_name = "MsgArea", ui_name = "MsgArea", kind = "ui"}}},
+        [13] = {{background = Screen.colors.Red1, foreground = Screen.colors.Gray100}, {{ui_name = "LineNr", kind = "ui", hi_name = "LineNr"}}};
+        [14] = {{bold = true, foreground = Screen.colors.Blue}, {{ui_name = "EndOfBuffer", kind = "ui", hi_name = "EndOfBuffer"}}};
     })
 
     command("set number")
@@ -142,16 +146,16 @@ describe('ext_hlstate detailed highlights', function()
     ]])
 
     command("set winhl=LineNr:ErrorMsg")
-    screen:expect([[
-      {5:  1 }^                                    |
-      {2:~                                       }|
-      {2:~                                       }|
+    screen:expect{grid=[[
+      {13:  1 }^                                    |
+      {14:~                                       }|
+      {14:~                                       }|
       {3:[No Name]                               }|
       {1:  1 }                                    |
       {2:~                                       }|
       {4:[No Name]                               }|
       {12:                                        }|
-    ]])
+    ]]}
 
     command("set winhl=Normal:MsgSeparator,NormalNC:Statement")
     screen:expect([[
@@ -179,7 +183,7 @@ describe('ext_hlstate detailed highlights', function()
   end)
 
   it("work with :terminal", function()
-    if helpers.pending_win32(pending) then return end
+    skip(is_os('win'))
 
     screen:set_default_attr_ids({
       [1] = {{}, {{hi_name = "TermCursorNC", ui_name = "TermCursorNC", kind = "ui"}}},
@@ -208,7 +212,7 @@ describe('ext_hlstate detailed highlights', function()
     thelpers.set_bold()
     thelpers.feed_data('z\n')
     -- TODO(bfredl): check if this distinction makes sense
-    if iswin() then
+    if is_os('win') then
       screen:expect([[
         ^tty ready                               |
         x {5:y z}                                   |
@@ -234,7 +238,7 @@ describe('ext_hlstate detailed highlights', function()
 
     thelpers.feed_termcode("[A")
     thelpers.feed_termcode("[2C")
-    if iswin() then
+    if is_os('win') then
       screen:expect([[
         ^tty ready                               |
         x {6:y}{5: z}                                   |
