@@ -9,10 +9,16 @@ local meths = helpers.meths
 local async_meths = helpers.async_meths
 local test_build_dir = helpers.test_build_dir
 local nvim_prog = helpers.nvim_prog
-local iswin = helpers.iswin
+local exec = helpers.exec
+local exec_capture = helpers.exec_capture
 local exc_exec = helpers.exc_exec
 local exec_lua = helpers.exec_lua
 local poke_eventloop = helpers.poke_eventloop
+local assert_alive = helpers.assert_alive
+local is_os = helpers.is_os
+local is_ci = helpers.is_ci
+local funcs = helpers.funcs
+local skip = helpers.skip
 
 describe('ui/ext_messages', function()
   local screen
@@ -31,6 +37,7 @@ describe('ui/ext_messages', function()
       [7] = {background = Screen.colors.Yellow},
       [8] = {foreground = Screen.colors.Red},
       [9] = {special = Screen.colors.Red, undercurl = true},
+      [10] = {foreground = Screen.colors.Brown};
     })
   end)
   after_each(function()
@@ -304,13 +311,25 @@ describe('ui/ext_messages', function()
       {1:~                        }|
       {1:~                        }|
       {1:~                        }|
-    ]], messages={
+    ]], msg_history={
       {kind="echoerr", content={{"raa", 2}}},
       {kind="echoerr", content={{"bork", 2}}},
       {kind="echoerr", content={{"fail", 2}}},
       {kind="echoerr", content={{"extrafail", 2}}},
       {kind="echoerr", content={{"problem", 2}}}
-    }}
+    }, messages={{
+      content = {{ "Press ENTER or type command to continue", 4 }},
+      kind = "return_prompt"
+    }}}
+
+    feed '<cr>'
+    screen:expect{grid=[[
+      ^                         |
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+    ]]}
   end)
 
   it('shortmess-=S', function()
@@ -455,11 +474,13 @@ describe('ui/ext_messages', function()
       alphpabe^t                |
       {1:~                        }|
       {1:~                        }|
-    ]], messages={
-      {kind="echomsg", content={{"stuff"}}},
-    }, showmode={
-      { "-- INSERT --", 3 }
-    }}
+    ]], msg_history={{
+      content = {{ "stuff" }},
+      kind = "echomsg",
+    }}, messages={{
+        content = {{ "Press ENTER or type command to continue", 4}},
+        kind = "return_prompt"
+    }}}
   end)
 
   it('&showmode with macro-recording message', function()
@@ -685,12 +706,15 @@ describe('ui/ext_messages', function()
       {1:~                        }|
       {1:~                        }|
       {1:~                        }|
-    ]], messages={
+    ]], msg_history={
       {kind="echomsg", content={{"howdy"}}},
       {kind="", content={{"Type  :qa  and press <Enter> to exit Nvim"}}},
       {kind="echoerr", content={{"bork", 2}}},
       {kind="emsg", content={{"E117: Unknown function: nosuchfunction", 2}}}
-    }}
+    }, messages={{
+      content = {{ "Press ENTER or type command to continue", 4}},
+      kind = "return_prompt"
+    }}}
   end)
 
   it('implies ext_cmdline and ignores cmdheight', function()
@@ -718,7 +742,6 @@ describe('ui/ext_messages', function()
     ]])
     eq(0, eval('&cmdheight'))
 
-    -- normally this would be an error
     feed(':set cmdheight=0')
     screen:expect{grid=[[
       ^                         |
@@ -840,9 +863,60 @@ stack traceback:
       {1:~                        }|
       {1:~                        }|
     ]]}
-
   end)
 
+  it('supports nvim_echo messages with multiple attrs', function()
+    async_meths.echo({{'wow, ',"Search"}, {"such\n\nvery ", "ErrorMsg"}, {"color", "LineNr"}}, true, {})
+    screen:expect{grid=[[
+      ^                         |
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+    ]], messages={
+      { content = { { "wow, ", 7 }, { "such\n\nvery ", 2 }, { "color", 10 } }, kind = "echomsg" }
+    }}
+
+    feed ':ls<cr>'
+    screen:expect{grid=[[
+      ^                         |
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+    ]], messages={
+      { content = { { '\n  1 %a   "[No Name]"                    line 1' } }, kind = "" }
+    }}
+
+    feed ':messages<cr>'
+    screen:expect{grid=[[
+      ^                         |
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+    ]], messages={
+      { content = { { "Press ENTER or type command to continue", 4 } }, kind = "return_prompt" }
+    }, msg_history={
+      { content = { { "wow, ", 7 }, { "such\n\nvery ", 2 }, { "color", 10 } }, kind = "echomsg" }
+    }}
+
+    feed '<cr>'
+    screen:expect{grid=[[
+      ^                         |
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+      {1:~                        }|
+    ]]}
+  end)
+
+  it('does not truncate messages', function()
+    command('write Xtest')
+    screen:expect({messages={
+      {content = { { '"Xtest" [New] 0L, 0B written' } }, kind = "" }
+    }})
+  end)
 end)
 
 describe('ui/builtin messages', function()
@@ -851,17 +925,19 @@ describe('ui/builtin messages', function()
     clear()
     screen = Screen.new(60, 7)
     screen:attach({rgb=true, ext_popupmenu=true})
-    screen:set_default_attr_ids({
-      [1] = {bold = true, foreground = Screen.colors.Blue1},
-      [2] = {foreground = Screen.colors.Grey100, background = Screen.colors.Red},
-      [3] = {bold = true, reverse = true},
-      [4] = {bold = true, foreground = Screen.colors.SeaGreen4},
-      [5] = {foreground = Screen.colors.Blue1},
-      [6] = {bold = true, foreground = Screen.colors.Magenta},
-      [7] = {background = Screen.colors.Grey20},
-      [8] = {reverse = true},
-      [9] = {background = Screen.colors.LightRed}
-    })
+    screen:set_default_attr_ids  {
+      [1] = {bold = true, foreground = Screen.colors.Blue1};
+      [2] = {foreground = Screen.colors.Grey100, background = Screen.colors.Red};
+      [3] = {bold = true, reverse = true};
+      [4] = {bold = true, foreground = Screen.colors.SeaGreen4};
+      [5] = {foreground = Screen.colors.Blue1};
+      [6] = {bold = true, foreground = Screen.colors.Magenta};
+      [7] = {background = Screen.colors.Grey20};
+      [8] = {reverse = true};
+      [9] = {background = Screen.colors.LightRed};
+      [10] = {background = Screen.colors.Yellow};
+      [11] = {foreground = Screen.colors.Brown};
+    }
   end)
 
   it('supports multiline messages from rpc', function()
@@ -911,7 +987,7 @@ describe('ui/builtin messages', function()
 
     -- screen size doesn't affect internal output #10285
     eq('ErrorMsg       xxx ctermfg=15 ctermbg=1 guifg=White guibg=Red',
-       meths.exec("hi ErrorMsg", true))
+       exec_capture("hi ErrorMsg"))
   end)
 
   it(':syntax list langGroup output', function()
@@ -950,10 +1026,57 @@ vimComment     xxx match /\s"[^\-:.%#=*].*$/ms=s+1,lc=1  excludenl contains=@vim
                    match /\<endif\s\+".*$/ms=s+5,lc=5  contains=@vimCommentGroup,vimCommentString 
                    match /\<else\s\+".*$/ms=s+4,lc=4  contains=@vimCommentGroup,vimCommentString 
                    links to Comment]],
-       meths.exec('syntax list vimComment', true))
+       exec_capture('syntax list vimComment'))
     -- luacheck: pop
   end)
 
+  it('no empty line after :silent #12099', function()
+    exec([[
+      func T1()
+        silent !echo
+        echo "message T1"
+      endfunc
+      func T2()
+        silent lua print("lua message")
+        echo "message T2"
+      endfunc
+      func T3()
+        silent call nvim_out_write("api message\n")
+        echo "message T3"
+      endfunc
+    ]])
+    feed(':call T1()<CR>')
+    screen:expect{grid=[[
+      ^                                                            |
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      message T1                                                  |
+    ]]}
+    feed(':call T2()<CR>')
+    screen:expect{grid=[[
+      ^                                                            |
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      message T2                                                  |
+    ]]}
+    feed(':call T3()<CR>')
+    screen:expect{grid=[[
+      ^                                                            |
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      message T3                                                  |
+    ]]}
+  end)
+
   it('supports ruler with laststatus=0', function()
     command("set ruler laststatus=0")
     screen:expect{grid=[[
@@ -1013,10 +1136,10 @@ vimComment     xxx match /\s"[^\-:.%#=*].*$/ms=s+1,lc=1  excludenl contains=@vim
     ]]}
   end)
 
-  it('redraws NOT_VALID correctly after message', function()
-    -- edge case: only one window was set NOT_VALID. Original report
+  it('redraws UPD_NOT_VALID correctly after message', function()
+    -- edge case: only one window was set UPD_NOT_VALID. Original report
     -- used :make, but fake it using one command to set the current
-    -- window NOT_VALID and another to show a long message.
+    -- window UPD_NOT_VALID and another to show a long message.
     command("set more")
     feed(':new<cr><c-w><c-w>')
     screen:expect{grid=[[
@@ -1097,6 +1220,118 @@ vimComment     xxx match /\s"[^\-:.%#=*].*$/ms=s+1,lc=1  excludenl contains=@vim
       {4:Press ENTER or type command to continue}^                     |
     ]]}
   end)
+
+  it('supports nvim_echo messages with multiple attrs', function()
+    async_meths.echo({{'wow, ',"Search"}, {"such\n\nvery ", "ErrorMsg"}, {"color", "LineNr"}}, true, {})
+    screen:expect{grid=[[
+                                                                  |
+      {1:~                                                           }|
+      {3:                                                            }|
+      {10:wow, }{2:such}                                                   |
+                                                                  |
+      {2:very }{11:color}                                                  |
+      {4:Press ENTER or type command to continue}^                     |
+    ]]}
+
+    feed '<cr>'
+    screen:expect{grid=[[
+      ^                                                            |
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+                                                                  |
+    ]]}
+
+    feed ':messages<cr>'
+    screen:expect{grid=[[
+                                                                  |
+      {1:~                                                           }|
+      {3:                                                            }|
+      {10:wow, }{2:such}                                                   |
+                                                                  |
+      {2:very }{11:color}                                                  |
+      {4:Press ENTER or type command to continue}^                     |
+    ]]}
+  end)
+
+  it('prints lines in Ex mode correctly with a burst of carriage returns #19341', function()
+    command('set number')
+    meths.buf_set_lines(0, 0, 0, true, {'aaa', 'bbb', 'ccc'})
+    feed('gggQ<CR><CR>1<CR><CR>vi')
+    screen:expect([[
+      Entering Ex mode.  Type "visual" to go to Normal mode.      |
+      {11:  2 }bbb                                                     |
+      {11:  3 }ccc                                                     |
+      :1                                                          |
+      {11:  1 }aaa                                                     |
+      {11:  2 }bbb                                                     |
+      :vi^                                                         |
+    ]])
+    feed('<CR>')
+    screen:expect([[
+      {11:  1 }aaa                                                     |
+      {11:  2 }^bbb                                                     |
+      {11:  3 }ccc                                                     |
+      {11:  4 }                                                        |
+      {1:~                                                           }|
+      {1:~                                                           }|
+                                                                  |
+    ]])
+  end)
+
+  it('echo messages are shown correctly when getchar() immediately follows', function()
+    feed([[:echo 'foo' | echo 'bar' | call getchar()<CR>]])
+    screen:expect([[
+                                                                  |
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {3:                                                            }|
+      foo                                                         |
+      bar^                                                         |
+    ]])
+  end)
+
+  it('consecutive calls to win_move_statusline() work after multiline message #21014',function()
+    async_meths.exec([[
+      echo "\n"
+      call win_move_statusline(0, -4)
+      call win_move_statusline(0, 4)
+    ]], false)
+    screen:expect([[
+                                                                  |
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {3:                                                            }|
+                                                                  |
+      {4:Press ENTER or type command to continue}^                     |
+    ]])
+    feed('<CR>')
+    screen:expect([[
+      ^                                                            |
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+      {1:~                                                           }|
+                                                                  |
+    ]])
+    eq(1, meths.get_option('cmdheight'))
+  end)
+end)
+
+it('calling screenstring() after redrawing between messages without UI #20999', function()
+  clear()
+  exec([[
+    echo repeat('a', 100)
+    redraw
+    echo "\n"
+    call screenstring(1, 1)
+  ]])
+  assert_alive()
 end)
 
 describe('ui/ext_messages', function()
@@ -1126,7 +1361,6 @@ describe('ui/ext_messages', function()
       {1:~                                                                               }|
       {1:~                                                                               }|
       {1:~                                                                               }|
-      {1:~                                                                               }|
       {MATCH:.*}|
       {1:~                                                                               }|
       {1:~                 }Nvim is open source and freely distributable{1:                  }|
@@ -1137,9 +1371,10 @@ describe('ui/ext_messages', function()
       {1:~                }type  :q{5:<Enter>}               to exit         {1:                 }|
       {1:~                }type  :help{5:<Enter>}            for help        {1:                 }|
       {1:~                                                                               }|
-      {MATCH:.*}|
-      {MATCH:.*}|
+      {1:~{MATCH: +}}type  :help news{5:<Enter>} to see changes in v{MATCH:%d+%.%d+}{1:{MATCH: +}}|
       {1:~                                                                               }|
+      {MATCH:.*}|
+      {MATCH:.*}|
       {1:~                                                                               }|
       {1:~                                                                               }|
       {1:~                                                                               }|
@@ -1182,7 +1417,6 @@ describe('ui/ext_messages', function()
                                                                                       |
                                                                                       |
                                                                                       |
-                                                                                      |
       {MATCH:.*}|
                                                                                       |
                         Nvim is open source and freely distributable                  |
@@ -1193,9 +1427,10 @@ describe('ui/ext_messages', function()
                        type  :q{5:<Enter>}               to exit                          |
                        type  :help{5:<Enter>}            for help                         |
                                                                                       |
-      {MATCH:.*}|
-      {MATCH:.*}|
+      {MATCH: +}type  :help news{5:<Enter>} to see changes in v{MATCH:%d+%.%d+ +}|
                                                                                       |
+      {MATCH:.*}|
+      {MATCH:.*}|
                                                                                       |
                                                                                       |
                                                                                       |
@@ -1209,7 +1444,6 @@ describe('ui/ext_messages', function()
   it('supports global statusline', function()
     feed(":set laststatus=3<cr>")
     feed(":sp<cr>")
-    feed("<c-l>")
     feed(":set cmdheight<cr>")
     screen:expect({grid=[[
       ^                                                                                |
@@ -1223,7 +1457,7 @@ describe('ui/ext_messages', function()
       {1:~                                                                               }|
       {1:~                                                                               }|
       {1:~                                                                               }|
-      {6:────────────────────────────────────────────────────────────────────────────────}|
+      ────────────────────────────────────────────────────────────────────────────────|
                                                                                       |
       {1:~                                                                               }|
       {1:~                                                                               }|
@@ -1241,8 +1475,7 @@ describe('ui/ext_messages', function()
     }})
 
     feed("<c-w>+")
-    feed("<c-l>")
-    feed(":set cmdheight<cr>")
+    feed(":set laststatus<cr>")
     screen:expect({grid=[[
       ^                                                                                |
       {1:~                                                                               }|
@@ -1256,7 +1489,7 @@ describe('ui/ext_messages', function()
       {1:~                                                                               }|
       {1:~                                                                               }|
       {1:~                                                                               }|
-      {6:────────────────────────────────────────────────────────────────────────────────}|
+      ────────────────────────────────────────────────────────────────────────────────|
                                                                                       |
       {1:~                                                                               }|
       {1:~                                                                               }|
@@ -1269,13 +1502,12 @@ describe('ui/ext_messages', function()
       {1:~                                                                               }|
       {7:[No Name]                                                                       }|
     ]], messages={
-      {content = { { "  cmdheight=0" } }, kind = "" }
+      {content = { { "  laststatus=3" } }, kind = "" }
     }})
 
     feed(":set mouse=a<cr>")
     meths.input_mouse('left', 'press', '', 0, 12, 10)
     poke_eventloop()
-    meths.input_mouse('left', 'drag', '', 0, 12, 10)
     meths.input_mouse('left', 'drag', '', 0, 11, 10)
     feed("<c-l>")
     feed(":set cmdheight<cr>")
@@ -1291,7 +1523,7 @@ describe('ui/ext_messages', function()
       {1:~                                                                               }|
       {1:~                                                                               }|
       {1:~                                                                               }|
-      {6:────────────────────────────────────────────────────────────────────────────────}|
+      ────────────────────────────────────────────────────────────────────────────────|
                                                                                       |
       {1:~                                                                               }|
       {1:~                                                                               }|
@@ -1312,7 +1544,6 @@ end)
 
 describe('ui/msg_puts_printf', function()
   it('output multibyte characters correctly', function()
-    if helpers.pending_win32(pending) then return end
     local screen
     local cmd = ''
     local locale_dir = test_build_dir..'/share/locale/ja/LC_MESSAGES'
@@ -1321,7 +1552,7 @@ describe('ui/msg_puts_printf', function()
     screen = Screen.new(25, 5)
     screen:attach()
 
-    if iswin() then
+    if is_os('win') then
       if os.execute('chcp 932 > NUL 2>&1') ~= 0 then
         pending('missing japanese language features', function() end)
         return
@@ -1332,7 +1563,7 @@ describe('ui/msg_puts_printf', function()
       if (exc_exec('lang ja_JP.UTF-8') ~= 0) then
         pending('Locale ja_JP.UTF-8 not supported', function() end)
         return
-      elseif helpers.isCI() then
+      elseif is_ci() then
         -- Fails non--Windows CI. Message catalog directory issue?
         pending('fails on unix CI', function() end)
         return
@@ -1388,7 +1619,7 @@ ullamco laboris nisi ut
 aliquip ex ea commodo consequat.]])
   end)
 
-  it('can be quit', function()
+  it('can be quit with echon', function()
     screen:try_resize(25,5)
     feed(':echon join(map(range(0, &lines*10), "v:val"), "\\n")<cr>')
     screen:expect{grid=[[
@@ -1408,6 +1639,45 @@ aliquip ex ea commodo consequat.]])
     ]]}
   end)
 
+  it('can be quit with Lua #11224 #16537', function()
+    -- NOTE: adds "4" to message history, although not displayed initially
+    --       (triggered the more prompt).
+    screen:try_resize(40,5)
+    feed(':lua for i=0,10 do print(i) end<cr>')
+    screen:expect{grid=[[
+      0                                       |
+      1                                       |
+      2                                       |
+      3                                       |
+      {4:-- More --}^                              |
+    ]]}
+    feed('q')
+    screen:expect{grid=[[
+      ^                                        |
+      {1:~                                       }|
+      {1:~                                       }|
+      {1:~                                       }|
+                                              |
+    ]]}
+    feed(':mess<cr>')
+    screen:expect{grid=[[
+      0                                       |
+      1                                       |
+      2                                       |
+      3                                       |
+      {4:-- More --}^                              |
+    ]]}
+    feed('j')
+    screen:expect{grid=[[
+      1                                       |
+      2                                       |
+      3                                       |
+      4                                       |
+      {4:Press ENTER or type command to continue}^ |
+    ]]}
+    feed('<cr>')
+  end)
+
   it('handles wrapped lines with line scroll', function()
     feed(':lua error(_G.x)<cr>')
     screen:expect{grid=[[
@@ -1672,7 +1942,7 @@ aliquip ex ea commodo consequat.]])
 
     feed('k')
     screen:expect{grid=[[
-      {7:0}{8:                          }{7:)}{8:       }|
+      {7:0}{8:                                  }|
       {9:1}{10:                                  }|
       {9:2}{10:                                  }|
       {9:3}{10:                                  }|
@@ -1696,6 +1966,7 @@ aliquip ex ea commodo consequat.]])
   end)
 
   it('with :!cmd does not crash on resize', function()
+    skip(funcs.executable('sleep') == 0, 'missing "sleep" command')
     feed(':!sleep 1<cr>')
     screen:expect{grid=[[
                                          |
@@ -1761,6 +2032,7 @@ aliquip ex ea commodo consequat.]])
     -- text is not reflown; existing lines get cut
     screen:try_resize(30, 12)
     screen:expect{grid=[[
+      :lua error(_G.x)              |
       {2:E5108: Error executing lua [st}|
       {2:":lua"]:1: Lorem ipsum dolor s}|
       {2:et, consectetur}               |
@@ -1771,12 +2043,27 @@ aliquip ex ea commodo consequat.]])
                                     |
                                     |
                                     |
-                                    |
       {4:-- More --}^                    |
     ]]}
 
     -- continues in a mostly consistent state, but only new lines are
     -- wrapped at the new screen size.
+    feed('<cr>')
+    screen:expect{grid=[[
+      {2:E5108: Error executing lua [st}|
+      {2:":lua"]:1: Lorem ipsum dolor s}|
+      {2:et, consectetur}               |
+      {2:adipisicing elit, sed do eiusm}|
+      {2:mpore}                         |
+      {2:incididunt ut labore et dolore}|
+      {2:a aliqua.}                     |
+      {2:Ut enim ad minim veniam, quis }|
+      {2:nostrud xercitation}           |
+      {2:ullamco laboris nisi ut}       |
+      {2:aliquip ex ea commodo consequa}|
+      {4:-- More --}^                    |
+    ]]}
+
     feed('<cr>')
     screen:expect{grid=[[
       {2:":lua"]:1: Lorem ipsum dolor s}|
@@ -1809,4 +2096,55 @@ aliquip ex ea commodo consequat.]])
                                     |
     ]]}
   end)
+
+  it('with cmdheight=0 does not crash with g<', function()
+    command('set cmdheight=0')
+    feed(':ls<cr>')
+    screen:expect{grid=[[
+                                         |
+      {1:~                                  }|
+      {12:                                   }|
+      :ls                                |
+        1 %a   "[No Name]"               |
+           line 1                        |
+      {4:Press ENTER or type command to cont}|
+      {4:inue}^                               |
+    ]]}
+
+    feed('<cr>')
+    screen:expect{grid=[[
+      ^                                   |
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+    ]]}
+
+    feed('g<lt>')
+    screen:expect{grid=[[
+                                         |
+      {1:~                                  }|
+      {12:                                   }|
+      :ls                                |
+        1 %a   "[No Name]"               |
+           line 1                        |
+      {4:Press ENTER or type command to cont}|
+      {4:inue}^                               |
+    ]]}
+
+    feed('<cr>')
+    screen:expect{grid=[[
+      ^                                   |
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+      {1:~                                  }|
+    ]]}
+  end)
 end)
