@@ -3,7 +3,9 @@ local Screen = require('test.functional.ui.screen')
 local clear, feed, meths = helpers.clear, helpers.feed, helpers.meths
 local insert, feed_command = helpers.insert, helpers.feed_command
 local eq, funcs = helpers.eq, helpers.funcs
+local poke_eventloop = helpers.poke_eventloop
 local command = helpers.command
+local exec = helpers.exec
 
 describe('ui/mouse/input', function()
   local screen
@@ -31,8 +33,9 @@ describe('ui/mouse/input', function()
       [5] = {bold = true, reverse = true},
       [6] = {foreground = Screen.colors.Grey100, background = Screen.colors.Red},
       [7] = {bold = true, foreground = Screen.colors.SeaGreen4},
+      [8] = {foreground = Screen.colors.Brown},
     })
-    command("set display-=msgsep")
+    command("set mousemodel=extend")
     feed('itesting<cr>mouse<cr>support and selection<esc>')
     screen:expect([[
       testing                  |
@@ -595,61 +598,63 @@ describe('ui/mouse/input', function()
     feed('ifoo\nbar<esc>')
 
     screen:expect{grid=[[
-      testing                   {4:│}testing                   |
-      mouse                     {4:│}mouse                     |
-      support and selection     {4:│}support and selection     |
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│[No Name] [+]             }|
-      {0:~                         }{4:│}foo{0:$}                      |
-      {0:~                         }{4:│}ba^r{0:$}                      |
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
+      testing                   │testing                   |
+      mouse                     │mouse                     |
+      support and selection     │support and selection     |
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{4:[No Name] [+]             }|
+      {0:~                         }│foo{0:$}                      |
+      {0:~                         }│ba^r{0:$}                      |
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
       {4:[No Name] [+]              }{5:[No Name] [+]             }|
                                                            |
     ]]}
 
     meths.input_mouse('left', 'press', '', 0, 6, 27)
     screen:expect{grid=[[
-      testing                   {4:│}testing                   |
-      mouse                     {4:│}mouse                     |
-      support and selection     {4:│}support and selection     |
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│[No Name] [+]             }|
-      {0:~                         }{4:│}^foo{0:$}                      |
-      {0:~                         }{4:│}bar{0:$}                      |
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
+      testing                   │testing                   |
+      mouse                     │mouse                     |
+      support and selection     │support and selection     |
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{4:[No Name] [+]             }|
+      {0:~                         }│^foo{0:$}                      |
+      {0:~                         }│bar{0:$}                      |
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
       {4:[No Name] [+]              }{5:[No Name] [+]             }|
                                                            |
     ]]}
     meths.input_mouse('left', 'drag', '', 0, 7, 30)
 
     screen:expect{grid=[[
-      testing                   {4:│}testing                   |
-      mouse                     {4:│}mouse                     |
-      support and selection     {4:│}support and selection     |
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│[No Name] [+]             }|
-      {0:~                         }{4:│}{1:foo}{3:$}                      |
-      {0:~                         }{4:│}{1:bar}{0:^$}                      |
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
-      {0:~                         }{4:│}{0:~                         }|
+      testing                   │testing                   |
+      mouse                     │mouse                     |
+      support and selection     │support and selection     |
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{4:[No Name] [+]             }|
+      {0:~                         }│{1:foo}{3:$}                      |
+      {0:~                         }│{1:bar}{0:^$}                      |
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
+      {0:~                         }│{0:~                         }|
       {4:[No Name] [+]              }{5:[No Name] [+]             }|
       {2:-- VISUAL --}                                         |
     ]]}
   end)
 
-  it('two clicks will select the word and enter VISUAL', function()
-    feed('<LeftMouse><2,2><LeftMouse><2,2>')
+  it('two clicks will enter VISUAL and dragging selects words', function()
+    feed('<LeftMouse><2,2>')
+    feed('<LeftRelease><2,2>')
+    feed('<LeftMouse><2,2>')
     screen:expect([[
       testing                  |
       mouse                    |
@@ -657,10 +662,38 @@ describe('ui/mouse/input', function()
       {0:~                        }|
       {2:-- VISUAL --}             |
     ]])
+    feed('<LeftDrag><0,1>')
+    screen:expect([[
+      testing                  |
+      ^m{1:ouse}                    |
+      {1:support} and selection    |
+      {0:~                        }|
+      {2:-- VISUAL --}             |
+    ]])
+    feed('<LeftDrag><4,0>')
+    screen:expect([[
+      ^t{1:esting}                  |
+      {1:mouse}                    |
+      {1:support} and selection    |
+      {0:~                        }|
+      {2:-- VISUAL --}             |
+    ]])
+    feed('<LeftDrag><14,2>')
+    screen:expect([[
+      testing                  |
+      mouse                    |
+      {1:support and selectio}^n    |
+      {0:~                        }|
+      {2:-- VISUAL --}             |
+    ]])
   end)
 
-  it('three clicks will select the line and enter VISUAL LINE', function()
-    feed('<LeftMouse><2,2><LeftMouse><2,2><LeftMouse><2,2>')
+  it('three clicks will enter VISUAL LINE and dragging selects lines', function()
+    feed('<LeftMouse><2,2>')
+    feed('<LeftRelease><2,2>')
+    feed('<LeftMouse><2,2>')
+    feed('<LeftRelease><2,2>')
+    feed('<LeftMouse><2,2>')
     screen:expect([[
       testing                  |
       mouse                    |
@@ -668,10 +701,40 @@ describe('ui/mouse/input', function()
       {0:~                        }|
       {2:-- VISUAL LINE --}        |
     ]])
+    feed('<LeftDrag><0,1>')
+    screen:expect([[
+      testing                  |
+      ^m{1:ouse}                    |
+      {1:support and selection}    |
+      {0:~                        }|
+      {2:-- VISUAL LINE --}        |
+    ]])
+    feed('<LeftDrag><4,0>')
+    screen:expect([[
+      {1:test}^i{1:ng}                  |
+      {1:mouse}                    |
+      {1:support and selection}    |
+      {0:~                        }|
+      {2:-- VISUAL LINE --}        |
+    ]])
+    feed('<LeftDrag><14,2>')
+    screen:expect([[
+      testing                  |
+      mouse                    |
+      {1:support and se}^l{1:ection}    |
+      {0:~                        }|
+      {2:-- VISUAL LINE --}        |
+    ]])
   end)
 
-  it('four clicks will enter VISUAL BLOCK', function()
-    feed('<LeftMouse><2,2><LeftMouse><2,2><LeftMouse><2,2><LeftMouse><2,2>')
+  it('four clicks will enter VISUAL BLOCK and dragging selects blockwise', function()
+    feed('<LeftMouse><2,2>')
+    feed('<LeftRelease><2,2>')
+    feed('<LeftMouse><2,2>')
+    feed('<LeftRelease><2,2>')
+    feed('<LeftMouse><2,2>')
+    feed('<LeftRelease><2,2>')
+    feed('<LeftMouse><2,2>')
     screen:expect([[
       testing                  |
       mouse                    |
@@ -679,6 +742,30 @@ describe('ui/mouse/input', function()
       {0:~                        }|
       {2:-- VISUAL BLOCK --}       |
     ]])
+    feed('<LeftDrag><0,1>')
+    screen:expect([[
+      testing                  |
+      ^m{1:ou}se                    |
+      {1:sup}port and selection    |
+      {0:~                        }|
+      {2:-- VISUAL BLOCK --}       |
+    ]])
+    feed('<LeftDrag><4,0>')
+    screen:expect([[
+      te{1:st}^ing                  |
+      mo{1:use}                    |
+      su{1:ppo}rt and selection    |
+      {0:~                        }|
+      {2:-- VISUAL BLOCK --}       |
+    ]])
+    feed('<LeftDrag><14,2>')
+    screen:expect([[
+      testing                  |
+      mouse                    |
+      su{1:pport and se}^lection    |
+      {0:~                        }|
+      {2:-- VISUAL BLOCK --}       |
+    ]])
   end)
 
   it('right click extends visual selection to the clicked location', function()
@@ -713,6 +800,66 @@ describe('ui/mouse/input', function()
     feed('<cr>')
   end)
 
+  it('dragging vertical separator', function()
+    screen:try_resize(45, 5)
+    command('setlocal nowrap')
+    local oldwin = meths.get_current_win().id
+    command('rightbelow vnew')
+    screen:expect([[
+      testing               │{0:^$}                     |
+      mouse                 │{0:~                     }|
+      support and selection │{0:~                     }|
+      {4:[No Name] [+]          }{5:[No Name]             }|
+                                                   |
+    ]])
+    meths.input_mouse('left', 'press', '', 0, 0, 22)
+    poke_eventloop()
+    meths.input_mouse('left', 'drag', '', 0, 1, 12)
+    screen:expect([[
+      testing     │{0:^$}                               |
+      mouse       │{0:~                               }|
+      support and │{0:~                               }|
+      {4:< Name] [+]  }{5:[No Name]                       }|
+                                                   |
+    ]])
+    meths.input_mouse('left', 'drag', '', 0, 2, 2)
+    screen:expect([[
+      te│{0:^$}                                         |
+      mo│{0:~                                         }|
+      su│{0:~                                         }|
+      {4:<  }{5:[No Name]                                 }|
+                                                   |
+    ]])
+    meths.input_mouse('left', 'release', '', 0, 2, 2)
+    meths.set_option_value('statuscolumn', 'foobar', { win = oldwin })
+    screen:expect([[
+      {8:fo}│{0:^$}                                         |
+      {8:fo}│{0:~                                         }|
+      {8:fo}│{0:~                                         }|
+      {4:<  }{5:[No Name]                                 }|
+                                                   |
+    ]])
+    meths.input_mouse('left', 'press', '', 0, 0, 2)
+    poke_eventloop()
+    meths.input_mouse('left', 'drag', '', 0, 1, 12)
+    screen:expect([[
+      {8:foobar}testin│{0:^$}                               |
+      {8:foobar}mouse │{0:~                               }|
+      {8:foobar}suppor│{0:~                               }|
+      {4:< Name] [+]  }{5:[No Name]                       }|
+                                                   |
+    ]])
+    meths.input_mouse('left', 'drag', '', 0, 2, 22)
+    screen:expect([[
+      {8:foobar}testing         │{0:^$}                     |
+      {8:foobar}mouse           │{0:~                     }|
+      {8:foobar}support and sele│{0:~                     }|
+      {4:[No Name] [+]          }{5:[No Name]             }|
+                                                   |
+    ]])
+    meths.input_mouse('left', 'release', '', 0, 2, 22)
+  end)
+
   local function wheel(use_api)
     feed('ggdG')
     insert([[
@@ -729,12 +876,12 @@ describe('ui/mouse/input', function()
     feed('k')
     feed_command('sp', 'vsp')
     screen:expect([[
-      lines                     {4:│}lines                     |
-      to                        {4:│}to                        |
-      test                      {4:│}test                      |
-      ^mouse scrolling           {4:│}mouse scrolling           |
-                                {4:│}                          |
-      {0:~                         }{4:│}{0:~                         }|
+      lines                     │lines                     |
+      to                        │to                        |
+      test                      │test                      |
+      ^mouse scrolling           │mouse scrolling           |
+                                │                          |
+      {0:~                         }│{0:~                         }|
       {5:[No Name] [+]              }{4:[No Name] [+]             }|
       to                                                   |
       test                                                 |
@@ -750,12 +897,12 @@ describe('ui/mouse/input', function()
       feed('<ScrollWheelDown><0,0>')
     end
     screen:expect([[
-      ^mouse scrolling           {4:│}lines                     |
-                                {4:│}to                        |
-      {0:~                         }{4:│}test                      |
-      {0:~                         }{4:│}mouse scrolling           |
-      {0:~                         }{4:│}                          |
-      {0:~                         }{4:│}{0:~                         }|
+      ^mouse scrolling           │lines                     |
+                                │to                        |
+      {0:~                         }│test                      |
+      {0:~                         }│mouse scrolling           |
+      {0:~                         }│                          |
+      {0:~                         }│{0:~                         }|
       {5:[No Name] [+]              }{4:[No Name] [+]             }|
       to                                                   |
       test                                                 |
@@ -771,12 +918,12 @@ describe('ui/mouse/input', function()
       feed('<ScrollWheelUp><27,0>')
     end
     screen:expect([[
-      ^mouse scrolling           {4:│}text                      |
-                                {4:│}with                      |
-      {0:~                         }{4:│}many                      |
-      {0:~                         }{4:│}lines                     |
-      {0:~                         }{4:│}to                        |
-      {0:~                         }{4:│}test                      |
+      ^mouse scrolling           │text                      |
+                                │with                      |
+      {0:~                         }│many                      |
+      {0:~                         }│lines                     |
+      {0:~                         }│to                        |
+      {0:~                         }│test                      |
       {5:[No Name] [+]              }{4:[No Name] [+]             }|
       to                                                   |
       test                                                 |
@@ -793,12 +940,12 @@ describe('ui/mouse/input', function()
       feed('<ScrollWheelUp><27,7><ScrollWheelUp>')
     end
     screen:expect([[
-      ^mouse scrolling           {4:│}text                      |
-                                {4:│}with                      |
-      {0:~                         }{4:│}many                      |
-      {0:~                         }{4:│}lines                     |
-      {0:~                         }{4:│}to                        |
-      {0:~                         }{4:│}test                      |
+      ^mouse scrolling           │text                      |
+                                │with                      |
+      {0:~                         }│many                      |
+      {0:~                         }│lines                     |
+      {0:~                         }│to                        |
+      {0:~                         }│test                      |
       {5:[No Name] [+]              }{4:[No Name] [+]             }|
       Inserting                                            |
       text                                                 |
@@ -883,6 +1030,49 @@ describe('ui/mouse/input', function()
     ]])
   end)
 
+  it("'sidescrolloff' applies to horizontal scrolling", function()
+    command('set nowrap')
+    command('set sidescrolloff=4')
+
+    feed("I <esc>020ib<esc>0")
+    screen:expect([[
+      testing                  |
+      mouse                    |
+      ^bbbbbbbbbbbbbbbbbbbb supp|
+      {0:~                        }|
+                               |
+    ]])
+
+    meths.input_mouse('wheel', 'right', '', 0, 0, 27)
+    screen:expect([[
+      g                        |
+                               |
+      bbbb^bbbbbbbbbb support an|
+      {0:~                        }|
+                               |
+    ]])
+
+    -- window-local 'sidescrolloff' should override global value. #21162
+    command('setlocal sidescrolloff=2')
+    feed('0')
+    screen:expect([[
+      testing                  |
+      mouse                    |
+      ^bbbbbbbbbbbbbbbbbbbb supp|
+      {0:~                        }|
+                               |
+    ]])
+
+    meths.input_mouse('wheel', 'right', '', 0, 0, 27)
+    screen:expect([[
+      g                        |
+                               |
+      bb^bbbbbbbbbbbb support an|
+      {0:~                        }|
+                               |
+    ]])
+  end)
+
   describe('on concealed text', function()
     -- Helpful for reading the test expectations:
     -- :match Error /\^/
@@ -896,13 +1086,15 @@ describe('ui/mouse/input', function()
       })
       feed('ggdG')
 
-      feed_command('set concealcursor=ni')
-      feed_command('set nowrap')
-      feed_command('set shiftwidth=2 tabstop=4 list')
-      feed_command('setl listchars=tab:>-')
-      feed_command('syntax match NonText "\\*" conceal')
-      feed_command('syntax match NonText "cats" conceal cchar=X')
-      feed_command('syntax match NonText "x" conceal cchar=>')
+      command([[setlocal concealcursor=ni nowrap shiftwidth=2 tabstop=4 list listchars=tab:>-]])
+      command([[syntax region X0 matchgroup=X1 start=/\*/ end=/\*/ concealends contains=X2]])
+      command([[syntax match X2 /cats/ conceal cchar=X contained]])
+      -- No heap-use-after-free with multi-line syntax pattern #24317
+      command([[syntax match X3 /\n\@<=x/ conceal cchar=>]])
+      command([[highlight link X0 Normal]])
+      command([[highlight link X1 NonText]])
+      command([[highlight link X2 NonText]])
+      command([[highlight link X3 NonText]])
 
       -- First column is there to retain the tabs.
       insert([[
@@ -1572,22 +1764,177 @@ describe('ui/mouse/input', function()
     meths.input_mouse('left', 'release', '', 0, 0, 0)
   end)
 
-  it('scroll keys are not translated into multiclicks #6211 #6989', function()
+  it('scroll keys are not translated into multiclicks and can be mapped #6211 #6989', function()
     meths.set_var('mouse_up', 0)
     meths.set_var('mouse_up2', 0)
-    meths.set_var('mouse_up3', 0)
-    meths.set_var('mouse_up4', 0)
     command('nnoremap <ScrollWheelUp> <Cmd>let g:mouse_up += 1<CR>')
     command('nnoremap <2-ScrollWheelUp> <Cmd>let g:mouse_up2 += 1<CR>')
-    command('nnoremap <3-ScrollWheelUp> <Cmd>let g:mouse_up3 += 1<CR>')
-    command('nnoremap <4-ScrollWheelUp> <Cmd>let g:mouse_up4 += 1<CR>')
-    meths.input_mouse('wheel', 'up', '', 0, 0, 0)
-    meths.input_mouse('wheel', 'up', '', 0, 0, 0)
+    feed('<ScrollWheelUp><0,0>')
+    feed('<ScrollWheelUp><0,0>')
     meths.input_mouse('wheel', 'up', '', 0, 0, 0)
     meths.input_mouse('wheel', 'up', '', 0, 0, 0)
     eq(4, meths.get_var('mouse_up'))
     eq(0, meths.get_var('mouse_up2'))
-    eq(0, meths.get_var('mouse_up3'))
-    eq(0, meths.get_var('mouse_up4'))
+  end)
+
+  it('<MouseMove> is not translated into multiclicks and can be mapped', function()
+    meths.set_var('mouse_move', 0)
+    meths.set_var('mouse_move2', 0)
+    command('nnoremap <MouseMove> <Cmd>let g:mouse_move += 1<CR>')
+    command('nnoremap <2-MouseMove> <Cmd>let g:mouse_move2 += 1<CR>')
+    feed('<MouseMove><0,0>')
+    feed('<MouseMove><0,0>')
+    meths.input_mouse('move', '', '', 0, 0, 0)
+    meths.input_mouse('move', '', '', 0, 0, 0)
+    eq(4, meths.get_var('mouse_move'))
+    eq(0, meths.get_var('mouse_move2'))
+  end)
+
+  it('feeding <MouseMove> in Normal mode does not use uninitialized memory #19480', function()
+    feed('<MouseMove>')
+    helpers.poke_eventloop()
+    helpers.assert_alive()
+  end)
+
+  it('mousemodel=popup_setpos', function()
+    screen:try_resize(80, 24)
+    exec([[
+      5new
+      call setline(1, ['the dish ran away with the spoon',
+            \ 'the cow jumped over the moon' ])
+
+      set mouse=a mousemodel=popup_setpos
+
+      aunmenu PopUp
+      nmenu PopUp.foo :let g:menustr = 'foo'<CR>
+      nmenu PopUp.bar :let g:menustr = 'bar'<CR>
+      nmenu PopUp.baz :let g:menustr = 'baz'<CR>
+      vmenu PopUp.foo y:<C-U>let g:menustr = 'foo'<CR>
+      vmenu PopUp.bar y:<C-U>let g:menustr = 'bar'<CR>
+      vmenu PopUp.baz y:<C-U>let g:menustr = 'baz'<CR>
+    ]])
+
+    meths.win_set_cursor(0, {1, 0})
+    meths.input_mouse('right', 'press', '', 0, 0, 4)
+    meths.input_mouse('right', 'release', '', 0, 0, 4)
+    feed('<Down><Down><CR>')
+    eq('bar', meths.get_var('menustr'))
+    eq({1, 4}, meths.win_get_cursor(0))
+
+    -- Test for right click in visual mode inside the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 9})
+    feed('vee')
+    meths.input_mouse('right', 'press', '', 0, 0, 11)
+    meths.input_mouse('right', 'release', '', 0, 0, 11)
+    feed('<Down><CR>')
+    eq({1, 9}, meths.win_get_cursor(0))
+    eq('ran away', funcs.getreg('"'))
+
+    -- Test for right click in visual mode right before the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 9})
+    feed('vee')
+    meths.input_mouse('right', 'press', '', 0, 0, 8)
+    meths.input_mouse('right', 'release', '', 0, 0, 8)
+    feed('<Down><CR>')
+    eq({1, 8}, meths.win_get_cursor(0))
+    eq('', funcs.getreg('"'))
+
+    -- Test for right click in visual mode right after the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 9})
+    feed('vee')
+    meths.input_mouse('right', 'press', '', 0, 0, 17)
+    meths.input_mouse('right', 'release', '', 0, 0, 17)
+    feed('<Down><CR>')
+    eq({1, 17}, meths.win_get_cursor(0))
+    eq('', funcs.getreg('"'))
+
+    -- Test for right click in block-wise visual mode inside the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 15})
+    feed('<C-V>j3l')
+    meths.input_mouse('right', 'press', '', 0, 1, 16)
+    meths.input_mouse('right', 'release', '', 0, 1, 16)
+    feed('<Down><CR>')
+    eq({1, 15}, meths.win_get_cursor(0))
+    eq('\0224', funcs.getregtype('"'))
+
+    -- Test for right click in block-wise visual mode outside the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 15})
+    feed('<C-V>j3l')
+    meths.input_mouse('right', 'press', '', 0, 1, 1)
+    meths.input_mouse('right', 'release', '', 0, 1, 1)
+    feed('<Down><CR>')
+    eq({2, 1}, meths.win_get_cursor(0))
+    eq('v', funcs.getregtype('"'))
+    eq('', funcs.getreg('"'))
+
+    -- Test for right click in line-wise visual mode inside the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 15})
+    feed('V')
+    meths.input_mouse('right', 'press', '', 0, 0, 9)
+    meths.input_mouse('right', 'release', '', 0, 0, 9)
+    feed('<Down><CR>')
+    eq({1, 0}, meths.win_get_cursor(0)) -- After yanking, the cursor goes to 1,1
+    eq('V', funcs.getregtype('"'))
+    eq(1, #funcs.getreg('"', 1, true))
+
+    -- Test for right click in multi-line line-wise visual mode inside the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 15})
+    feed('Vj')
+    meths.input_mouse('right', 'press', '', 0, 1, 19)
+    meths.input_mouse('right', 'release', '', 0, 1, 19)
+    feed('<Down><CR>')
+    eq({1, 0}, meths.win_get_cursor(0)) -- After yanking, the cursor goes to 1,1
+    eq('V', funcs.getregtype('"'))
+    eq(2, #funcs.getreg('"', 1, true))
+
+    -- Test for right click in line-wise visual mode outside the selection
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 15})
+    feed('V')
+    meths.input_mouse('right', 'press', '', 0, 1, 9)
+    meths.input_mouse('right', 'release', '', 0, 1, 9)
+    feed('<Down><CR>')
+    eq({2, 9}, meths.win_get_cursor(0))
+    eq('', funcs.getreg('"'))
+
+    -- Try clicking outside the window
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {2, 1})
+    feed('vee')
+    meths.input_mouse('right', 'press', '', 0, 6, 1)
+    meths.input_mouse('right', 'release', '', 0, 6, 1)
+    feed('<Down><CR>')
+    eq(2, funcs.winnr())
+    eq('', funcs.getreg('"'))
+
+    -- Test for right click in visual mode inside the selection with vertical splits
+    command('wincmd t')
+    command('rightbelow vsplit')
+    funcs.setreg('"', '')
+    meths.win_set_cursor(0, {1, 9})
+    feed('vee')
+    meths.input_mouse('right', 'press', '', 0, 0, 52)
+    meths.input_mouse('right', 'release', '', 0, 0, 52)
+    feed('<Down><CR>')
+    eq({1, 9}, meths.win_get_cursor(0))
+    eq('ran away', funcs.getreg('"'))
+
+    -- Test for right click inside visual selection at bottom of window with winbar
+    command('setlocal winbar=WINBAR')
+    feed('2yyP')
+    funcs.setreg('"', '')
+    feed('G$vbb')
+    meths.input_mouse('right', 'press', '', 0, 4, 61)
+    meths.input_mouse('right', 'release', '', 0, 4, 61)
+    feed('<Down><CR>')
+    eq({4, 20}, meths.win_get_cursor(0))
+    eq('the moon', funcs.getreg('"'))
   end)
 end)
