@@ -4,9 +4,7 @@ local spawn, set_session, clear = helpers.spawn, helpers.set_session, helpers.cl
 local feed, command = helpers.feed, helpers.command
 local insert = helpers.insert
 local eq = helpers.eq
-local eval = helpers.eval
-local iswin = helpers.iswin
-local funcs, meths, exec_lua = helpers.funcs, helpers.meths, helpers.exec_lua
+local funcs, meths = helpers.funcs, helpers.meths
 
 describe('screen', function()
   local screen
@@ -61,37 +59,10 @@ local function screen_tests(linegrid)
       [5] = {background = Screen.colors.LightGrey, underline = true, bold = true, foreground = Screen.colors.Fuchsia},
       [6] = {bold = true, foreground = Screen.colors.Fuchsia},
       [7] = {bold = true, foreground = Screen.colors.SeaGreen},
+      [8] = {foreground = Screen.colors.White, background = Screen.colors.Red},
     } )
   end)
 
-  describe(':suspend', function()
-    it('is forwarded to the UI', function()
-      local function check()
-        eq(true, screen.suspended)
-      end
-
-      command('let g:ev = []')
-      command('autocmd VimResume  * :call add(g:ev, "r")')
-      command('autocmd VimSuspend * :call add(g:ev, "s")')
-
-      eq(false, screen.suspended)
-      command('suspend')
-      eq({ 's', 'r' }, eval('g:ev'))
-
-      screen:expect(check)
-      screen.suspended = false
-
-      feed('<c-z>')
-      eq({ 's', 'r', 's', 'r' }, eval('g:ev'))
-
-      screen:expect(check)
-      screen.suspended = false
-
-      command('suspend')
-      eq({ 's', 'r', 's', 'r', 's', 'r' }, eval('g:ev'))
-    end)
-  end)
-
   describe('bell/visual bell', function()
     it('is forwarded to the UI', function()
       feed('<left>')
@@ -117,78 +88,13 @@ local function screen_tests(linegrid)
       screen:expect(function()
         eq(expected, screen.title)
       end)
-    end)
-
-    it('has correct default title with unnamed file', function()
-      local expected = '[No Name] - NVIM'
-      command('set title')
+      screen:detach()
+      screen.title = nil
+      screen:attach()
       screen:expect(function()
         eq(expected, screen.title)
       end)
     end)
-
-    it('has correct default title with named file', function()
-      local expected = (iswin() and 'myfile (C:\\mydir) - NVIM' or 'myfile (/mydir) - NVIM')
-      command('set title')
-      command(iswin() and 'file C:\\mydir\\myfile' or 'file /mydir/myfile')
-      screen:expect(function()
-        eq(expected, screen.title)
-      end)
-    end)
-
-    describe('is not changed by', function()
-      local file1 = iswin() and 'C:\\mydir\\myfile1' or '/mydir/myfile1'
-      local file2 = iswin() and 'C:\\mydir\\myfile2' or '/mydir/myfile2'
-      local expected = (iswin() and 'myfile1 (C:\\mydir) - NVIM' or 'myfile1 (/mydir) - NVIM')
-      local buf2
-
-      before_each(function()
-        command('edit '..file1)
-        buf2 = funcs.bufadd(file2)
-        command('set title')
-      end)
-
-      it('calling setbufvar() to set an option in a hidden buffer from i_CTRL-R', function()
-        command([[inoremap <F2> <C-R>=setbufvar(]]..buf2..[[, '&autoindent', 1) ? '' : ''<CR>]])
-        feed('i<F2><Esc>')
-        command('redraw!')
-        screen:expect(function()
-          eq(expected, screen.title)
-        end)
-      end)
-
-      it('an RPC call to nvim_buf_set_option in a hidden buffer', function()
-        meths.buf_set_option(buf2, 'autoindent', true)
-        command('redraw!')
-        screen:expect(function()
-          eq(expected, screen.title)
-        end)
-      end)
-
-      it('a Lua callback calling nvim_buf_set_option in a hidden buffer', function()
-        exec_lua(string.format([[
-          vim.schedule(function()
-            vim.api.nvim_buf_set_option(%d, 'autoindent', true)
-          end)
-        ]], buf2))
-        command('redraw!')
-        screen:expect(function()
-          eq(expected, screen.title)
-        end)
-      end)
-
-      it('a Lua callback calling nvim_buf_call in a hidden buffer', function()
-        exec_lua(string.format([[
-          vim.schedule(function()
-            vim.api.nvim_buf_call(%d, function() end)
-          end)
-        ]], buf2))
-        command('redraw!')
-        screen:expect(function()
-          eq(expected, screen.title)
-        end)
-      end)
-    end)
   end)
 
   describe(':set icon', function()
@@ -199,6 +105,12 @@ local function screen_tests(linegrid)
       screen:expect(function()
         eq(expected, screen.icon)
       end)
+      screen:detach()
+      screen.icon = nil
+      screen:attach()
+      screen:expect(function()
+        eq(expected, screen.icon)
+      end)
     end)
   end)
 
@@ -326,12 +238,12 @@ local function screen_tests(linegrid)
         command('vsp')
         command('vsp')
         screen:expect([[
-          ^                    {3:│}                {3:│}               |
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
+          ^                    │                │               |
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
           {1:[No Name]            }{3:[No Name]        [No Name]      }|
                                                                |
           {0:~                                                    }|
@@ -343,12 +255,12 @@ local function screen_tests(linegrid)
         ]])
         insert('hello')
         screen:expect([[
-          hell^o               {3:│}hello           {3:│}hello          |
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-          {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
+          hell^o               │hello           │hello          |
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
+          {0:~                   }│{0:~               }│{0:~              }|
           {1:[No Name] [+]        }{3:[No Name] [+]    [No Name] [+]  }|
           hello                                                |
           {0:~                                                    }|
@@ -369,12 +281,12 @@ local function screen_tests(linegrid)
       command('vsp')
       insert('hello')
       screen:expect([[
-        hell^o               {3:│}hello           {3:│}hello          |
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
+        hell^o               │hello           │hello          |
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
         {1:[No Name] [+]        }{3:[No Name] [+]    [No Name] [+]  }|
         hello                                                |
         {0:~                                                    }|
@@ -406,12 +318,12 @@ local function screen_tests(linegrid)
       command('tabprevious')
       screen:expect([[
         {2: }{6:4}{2:+ [No Name] }{4: + [No Name] }{3:                         }{4:X}|
-        hell^o               {3:│}hello           {3:│}hello          |
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
-        {0:~                   }{3:│}{0:~               }{3:│}{0:~              }|
+        hell^o               │hello           │hello          |
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
+        {0:~                   }│{0:~               }│{0:~              }|
         {1:[No Name] [+]        }{3:[No Name] [+]    [No Name] [+]  }|
         hello                                                |
         {0:~                                                    }|
@@ -520,36 +432,36 @@ local function screen_tests(linegrid)
 
       command('vsplit')
       screen:expect([[
-        ^foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
+        ^foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
         {1:[No Name] [+]              }{3:[No Name] [+]             }|
                                                              |
       ]])
 
       feed('<PageDown>')
       screen:expect([[
-        ^foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        foo                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
-        bar                       {3:│}foo                       |
+        ^foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        foo                       │foo                       |
+        bar                       │foo                       |
+        bar                       │foo                       |
+        bar                       │foo                       |
+        bar                       │foo                       |
+        bar                       │foo                       |
+        bar                       │foo                       |
+        bar                       │foo                       |
+        bar                       │foo                       |
         {1:[No Name] [+]              }{3:[No Name] [+]             }|
                                                              |
       ]])
@@ -682,30 +594,7 @@ local function screen_tests(linegrid)
       ]])
     end)
 
-    it('execute command with multi-line output without msgsep', function()
-      command("set display-=msgsep")
-      feed(':ls<cr>')
-      screen:expect([[
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        {0:~                                                    }|
-        :ls                                                  |
-          1 %a   "[No Name]"                    line 1       |
-        {7:Press ENTER or type command to continue}^              |
-      ]])
-      feed('<cr>') --  skip the "Press ENTER..." state or tests will hang
-    end)
-
-    it('execute command with multi-line output and with msgsep', function()
-      command("set display+=msgsep")
+    it('execute command with multi-line output', function()
       feed(':ls<cr>')
       screen:expect([[
                                                              |
@@ -748,12 +637,12 @@ local function screen_tests(linegrid)
       command('vsp')
       command('vsp')
       screen:expect([[
-        and                 {3:│}and             {3:│}and            |
-        clearing            {3:│}clearing        {3:│}clearing       |
-        in                  {3:│}in              {3:│}in             |
-        split               {3:│}split           {3:│}split          |
-        windows             {3:│}windows         {3:│}windows        |
-        ^                    {3:│}                {3:│}               |
+        and                 │and             │and            |
+        clearing            │clearing        │clearing       |
+        in                  │in              │in             |
+        split               │split           │split          |
+        windows             │windows         │windows        |
+        ^                    │                │               |
         {1:[No Name] [+]        }{3:[No Name] [+]    [No Name] [+]  }|
         clearing                                             |
         in                                                   |
@@ -768,12 +657,12 @@ local function screen_tests(linegrid)
     it('only affects the current scroll region', function()
       feed('6k')
       screen:expect([[
-        ^scrolling           {3:│}and             {3:│}and            |
-        and                 {3:│}clearing        {3:│}clearing       |
-        clearing            {3:│}in              {3:│}in             |
-        in                  {3:│}split           {3:│}split          |
-        split               {3:│}windows         {3:│}windows        |
-        windows             {3:│}                {3:│}               |
+        ^scrolling           │and             │and            |
+        and                 │clearing        │clearing       |
+        clearing            │in              │in             |
+        in                  │split           │split          |
+        split               │windows         │windows        |
+        windows             │                │               |
         {1:[No Name] [+]        }{3:[No Name] [+]    [No Name] [+]  }|
         clearing                                             |
         in                                                   |
@@ -785,12 +674,12 @@ local function screen_tests(linegrid)
       ]])
       feed('<c-w>l')
       screen:expect([[
-        scrolling           {3:│}and                 {3:│}and        |
-        and                 {3:│}clearing            {3:│}clearing   |
-        clearing            {3:│}in                  {3:│}in         |
-        in                  {3:│}split               {3:│}split      |
-        split               {3:│}windows             {3:│}windows    |
-        windows             {3:│}^                    {3:│}           |
+        scrolling           │and                 │and        |
+        and                 │clearing            │clearing   |
+        clearing            │in                  │in         |
+        in                  │split               │split      |
+        split               │windows             │windows    |
+        windows             │^                    │           |
         {3:[No Name] [+]        }{1:[No Name] [+]        }{3:<Name] [+] }|
         clearing                                             |
         in                                                   |
@@ -802,12 +691,12 @@ local function screen_tests(linegrid)
       ]])
       feed('gg')
       screen:expect([[
-        scrolling           {3:│}^Inserting           {3:│}and        |
-        and                 {3:│}text                {3:│}clearing   |
-        clearing            {3:│}with                {3:│}in         |
-        in                  {3:│}many                {3:│}split      |
-        split               {3:│}lines               {3:│}windows    |
-        windows             {3:│}to                  {3:│}           |
+        scrolling           │^Inserting           │and        |
+        and                 │text                │clearing   |
+        clearing            │with                │in         |
+        in                  │many                │split      |
+        split               │lines               │windows    |
+        windows             │to                  │           |
         {3:[No Name] [+]        }{1:[No Name] [+]        }{3:<Name] [+] }|
         clearing                                             |
         in                                                   |
@@ -819,12 +708,12 @@ local function screen_tests(linegrid)
       ]])
       feed('7j')
       screen:expect([[
-        scrolling           {3:│}with                {3:│}and        |
-        and                 {3:│}many                {3:│}clearing   |
-        clearing            {3:│}lines               {3:│}in         |
-        in                  {3:│}to                  {3:│}split      |
-        split               {3:│}test                {3:│}windows    |
-        windows             {3:│}^scrolling           {3:│}           |
+        scrolling           │with                │and        |
+        and                 │many                │clearing   |
+        clearing            │lines               │in         |
+        in                  │to                  │split      |
+        split               │test                │windows    |
+        windows             │^scrolling           │           |
         {3:[No Name] [+]        }{1:[No Name] [+]        }{3:<Name] [+] }|
         clearing                                             |
         in                                                   |
@@ -836,12 +725,12 @@ local function screen_tests(linegrid)
       ]])
       feed('2j')
       screen:expect([[
-        scrolling           {3:│}lines               {3:│}and        |
-        and                 {3:│}to                  {3:│}clearing   |
-        clearing            {3:│}test                {3:│}in         |
-        in                  {3:│}scrolling           {3:│}split      |
-        split               {3:│}and                 {3:│}windows    |
-        windows             {3:│}^clearing            {3:│}           |
+        scrolling           │lines               │and        |
+        and                 │to                  │clearing   |
+        clearing            │test                │in         |
+        in                  │scrolling           │split      |
+        split               │and                 │windows    |
+        windows             │^clearing            │           |
         {3:[No Name] [+]        }{1:[No Name] [+]        }{3:<Name] [+] }|
         clearing                                             |
         in                                                   |
@@ -853,12 +742,12 @@ local function screen_tests(linegrid)
       ]])
       feed('5k')
       screen:expect([[
-        scrolling           {3:│}^lines               {3:│}and        |
-        and                 {3:│}to                  {3:│}clearing   |
-        clearing            {3:│}test                {3:│}in         |
-        in                  {3:│}scrolling           {3:│}split      |
-        split               {3:│}and                 {3:│}windows    |
-        windows             {3:│}clearing            {3:│}           |
+        scrolling           │^lines               │and        |
+        and                 │to                  │clearing   |
+        clearing            │test                │in         |
+        in                  │scrolling           │split      |
+        split               │and                 │windows    |
+        windows             │clearing            │           |
         {3:[No Name] [+]        }{1:[No Name] [+]        }{3:<Name] [+] }|
         clearing                                             |
         in                                                   |
@@ -870,12 +759,12 @@ local function screen_tests(linegrid)
       ]])
       feed('k')
       screen:expect([[
-        scrolling           {3:│}^many                {3:│}and        |
-        and                 {3:│}lines               {3:│}clearing   |
-        clearing            {3:│}to                  {3:│}in         |
-        in                  {3:│}test                {3:│}split      |
-        split               {3:│}scrolling           {3:│}windows    |
-        windows             {3:│}and                 {3:│}           |
+        scrolling           │^many                │and        |
+        and                 │lines               │clearing   |
+        clearing            │to                  │in         |
+        in                  │test                │split      |
+        split               │scrolling           │windows    |
+        windows             │and                 │           |
         {3:[No Name] [+]        }{1:[No Name] [+]        }{3:<Name] [+] }|
         clearing                                             |
         in                                                   |
@@ -889,12 +778,9 @@ local function screen_tests(linegrid)
   end)
 
   describe('resize', function()
-    before_each(function()
+    it('rebuilds the whole screen', function()
       screen:try_resize(25, 5)
       feed('iresize')
-    end)
-
-    it('rebuilds the whole screen', function()
       screen:expect([[
         resize^                   |
         {0:~                        }|
@@ -905,6 +791,7 @@ local function screen_tests(linegrid)
     end)
 
     it('has minimum width/height values', function()
+      feed('iresize')
       screen:try_resize(1, 1)
       screen:expect([[
         resize^      |
@@ -917,6 +804,103 @@ local function screen_tests(linegrid)
         :ls^         |
       ]])
     end)
+
+    it('VimResized autocommand does not cause invalid UI events #20692 #20759', function()
+      screen:try_resize(25, 5)
+      feed('iresize<Esc>')
+      command([[autocmd VimResized * redrawtabline]])
+      command([[autocmd VimResized * lua vim.api.nvim_echo({ { 'Hello' } }, false, {})]])
+      command([[autocmd VimResized * let g:echospace = v:echospace]])
+      meths.set_option('showtabline', 2)
+      screen:expect([[
+        {2: + [No Name] }{3:            }|
+        resiz^e                   |
+        {0:~                        }|
+        {0:~                        }|
+                                 |
+      ]])
+      screen:try_resize(30, 6)
+      screen:expect([[
+        {2: + [No Name] }{3:                 }|
+        resiz^e                        |
+        {0:~                             }|
+        {0:~                             }|
+        {0:~                             }|
+                                      |
+      ]])
+      eq(29, meths.get_var('echospace'))
+    end)
+
+    it('messages from the same Ex command as resize are visible #22225', function()
+      feed(':set columns=20 | call<CR>')
+      screen:expect([[
+                            |
+                            |
+                            |
+                            |
+                            |
+                            |
+                            |
+                            |
+                            |
+        {1:                    }|
+        {8:E471: Argument requi}|
+        {8:red}                 |
+        {7:Press ENTER or type }|
+        {7:command to continue}^ |
+      ]])
+      feed('<CR>')
+      screen:expect([[
+        ^                    |
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+        {0:~                   }|
+                            |
+      ]])
+      feed(':set columns=0<CR>')
+      screen:expect([[
+                    |
+                    |
+                    |
+                    |
+                    |
+        {1:            }|
+        {8:E594: Need a}|
+        {8:t least 12 c}|
+        {8:olumns: colu}|
+        {8:mns=0}       |
+        {7:Press ENTER }|
+        {7:or type comm}|
+        {7:and to conti}|
+        {7:nue}^         |
+      ]])
+      feed('<CR>')
+      screen:expect([[
+        ^            |
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+        {0:~           }|
+                    |
+      ]])
+    end)
   end)
 
   describe('press enter', function()
@@ -1051,38 +1035,50 @@ describe('Screen default colors', function()
   end)
 end)
 
+it('CTRL-F or CTRL-B scrolls a page after UI attach/resize #20605', function()
+  clear()
+  local screen = Screen.new(100, 100)
+  screen:attach()
+  eq(100, meths.get_option('lines'))
+  eq(99, meths.get_option('window'))
+  eq(99, meths.win_get_height(0))
+  feed('1000o<Esc>')
+  eq(903, funcs.line('w0'))
+  feed('<C-B>')
+  eq(806, funcs.line('w0'))
+  feed('<C-B>')
+  eq(709, funcs.line('w0'))
+  feed('<C-F>')
+  eq(806, funcs.line('w0'))
+  feed('<C-F>')
+  eq(903, funcs.line('w0'))
+  feed('G')
+  screen:try_resize(50, 50)
+  eq(50, meths.get_option('lines'))
+  eq(49, meths.get_option('window'))
+  eq(49, meths.win_get_height(0))
+  eq(953, funcs.line('w0'))
+  feed('<C-B>')
+  eq(906, funcs.line('w0'))
+  feed('<C-B>')
+  eq(859, funcs.line('w0'))
+  feed('<C-F>')
+  eq(906, funcs.line('w0'))
+  feed('<C-F>')
+  eq(953, funcs.line('w0'))
+end)
 
-describe('screen with msgsep deactivated on startup', function()
-  local screen
-
-  before_each(function()
-    clear('--cmd', 'set display-=msgsep')
-    screen = Screen.new()
-    screen:attach()
-    screen:set_default_attr_ids {
-      [0] = {bold=true, foreground=255};
-      [7] = {bold = true, foreground = Screen.colors.SeaGreen};
-    }
-  end)
-
-  it('execute command with multi-line output', function()
-    feed ':ls<cr>'
-    screen:expect([[
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      {0:~                                                    }|
-      :ls                                                  |
-        1 %a   "[No Name]"                    line 1       |
-      {7:Press ENTER or type command to continue}^              |
-    ]])
-    feed '<cr>'  -- skip the "Press ENTER..." state or tests will hang
-  end)
+it("showcmd doesn't cause empty grid_line with redrawdebug=compositor #22593", function()
+  clear()
+  local screen = Screen.new(30, 2)
+  screen:set_default_attr_ids({
+    [0] = {bold = true, foreground = Screen.colors.Blue},
+  })
+  screen:attach()
+  command('set showcmd redrawdebug=compositor')
+  feed('d')
+  screen:expect{grid=[[
+    ^                              |
+                       d          |
+  ]]}
 end)
