@@ -1,10 +1,14 @@
 local helpers = require('test.functional.helpers')(after_each)
+local assert_log = helpers.assert_log
 local eq, neq, eval = helpers.eq, helpers.neq, helpers.eval
 local clear, funcs, meths = helpers.clear, helpers.funcs, helpers.meths
-local iswin = helpers.iswin
 local ok = helpers.ok
 local matches = helpers.matches
 local pcall_err = helpers.pcall_err
+local mkdir = helpers.mkdir
+local is_os = helpers.is_os
+
+local testlog = 'Xtest-server-log'
 
 local function clear_serverlist()
   for _, server in pairs(funcs.serverlist()) do
@@ -13,9 +17,23 @@ local function clear_serverlist()
 end
 
 describe('server', function()
-  before_each(clear)
+  after_each(function()
+    os.remove(testlog)
+  end)
+
+  it('serverstart() stores sockets in $XDG_RUNTIME_DIR', function()
+    local dir = 'Xtest_xdg_run'
+    mkdir(dir)
+    clear({ env={ XDG_RUNTIME_DIR=dir } })
+    matches(dir, funcs.stdpath('run'))
+    if not is_os('win') then
+      matches(dir, funcs.serverstart())
+    end
+  end)
+
 
   it('serverstart(), serverstop() does not set $NVIM', function()
+    clear()
     local s = eval('serverstart()')
     assert(s ~= nil and s:len() > 0, "serverstart() returned empty")
     eq('', eval('$NVIM'))
@@ -30,10 +48,11 @@ describe('server', function()
     eq('', eval('$NVIM_LISTEN_ADDRESS'))
     local servers = funcs.serverlist()
     eq(1, #servers)
-    ok(string.len(servers[1]) > 4)  -- Like /tmp/nvim…/… or \\.\pipe\…
+    ok(string.len(servers[1]) > 4)  -- "~/.local/state/nvim…/…" or "\\.\pipe\…"
   end)
 
   it('sets v:servername at startup or if all servers were stopped', function()
+    clear()
     local initial_server = meths.get_vvar('servername')
     assert(initial_server ~= nil and initial_server:len() > 0,
            'v:servername was not initialized')
@@ -53,8 +72,8 @@ describe('server', function()
     eq('', meths.get_vvar('servername'))
 
     -- v:servername and $NVIM take the next available server.
-    local servername = (iswin() and [[\\.\pipe\Xtest-functional-server-pipe]]
-                                or 'Xtest-functional-server-socket')
+    local servername = (is_os('win') and [[\\.\pipe\Xtest-functional-server-pipe]]
+                                or './Xtest-functional-server-socket')
     funcs.serverstart(servername)
     eq(servername, meths.get_vvar('servername'))
     -- Not set in the current process, only in children.
@@ -62,11 +81,20 @@ describe('server', function()
   end)
 
   it('serverstop() returns false for invalid input', function()
+    clear{env={
+      NVIM_LOG_FILE=testlog,
+      NVIM_LISTEN_ADDRESS='.',
+    }}
     eq(0, eval("serverstop('')"))
     eq(0, eval("serverstop('bogus-socket-name')"))
+    assert_log('Not listening on bogus%-socket%-name', testlog, 10)
   end)
 
-  it('parses endpoints correctly', function()
+  it('parses endpoints', function()
+    clear{env={
+      NVIM_LOG_FILE=testlog,
+      NVIM_LISTEN_ADDRESS='.',
+    }}
     clear_serverlist()
     eq({}, funcs.serverlist())
 
@@ -90,6 +118,7 @@ describe('server', function()
     if status then
       table.insert(expected, v4)
       pcall(funcs.serverstart, v4)  -- exists already; ignore
+      assert_log('Failed to start server: address already in use: 127%.0%.0%.1', testlog, 10)
     end
 
     local v6 = '::1:12345'
@@ -97,23 +126,29 @@ describe('server', function()
     if status then
       table.insert(expected, v6)
       pcall(funcs.serverstart, v6)  -- exists already; ignore
+      assert_log('Failed to start server: address already in use: ::1', testlog, 10)
     end
     eq(expected, funcs.serverlist())
     clear_serverlist()
 
+    -- Address without slashes is a "name" which is appended to a generated path. #8519
+    matches([[.*[/\\]xtest1%.2%.3%.4[^/\\]*]], funcs.serverstart('xtest1.2.3.4'))
+    clear_serverlist()
+
     eq('Vim:Failed to start server: invalid argument',
       pcall_err(funcs.serverstart, '127.0.0.1:65536'))  -- invalid port
     eq({}, funcs.serverlist())
   end)
 
   it('serverlist() returns the list of servers', function()
+    clear()
     -- There should already be at least one server.
     local n = eval('len(serverlist())')
 
     -- Add some servers.
-    local servs = (iswin()
+    local servs = (is_os('win')
       and { [[\\.\pipe\Xtest-pipe0934]], [[\\.\pipe\Xtest-pipe4324]] }
-      or  { [[Xtest-pipe0934]], [[Xtest-pipe4324]] })
+      or  { [[./Xtest-pipe0934]], [[./Xtest-pipe4324]] })
     for _, s in ipairs(servs) do
       eq(s, eval("serverstart('"..s.."')"))
     end
@@ -145,10 +180,14 @@ describe('startup --listen', function()
   end)
 
   it('sets v:servername, overrides $NVIM_LISTEN_ADDRESS', function()
-    local addr = (iswin() and [[\\.\pipe\Xtest-listen-pipe]]
-                          or 'Xtest-listen-pipe')
-    clear({ env={ NVIM_LISTEN_ADDRESS='Xtest-env-pipe' },
+    local addr = (is_os('win') and [[\\.\pipe\Xtest-listen-pipe]]
+                          or './Xtest-listen-pipe')
+    clear({ env={ NVIM_LISTEN_ADDRESS='./Xtest-env-pipe' },
             args={ '--listen', addr } })
     eq(addr, meths.get_vvar('servername'))
+
+    -- Address without slashes is a "name" which is appended to a generated path. #8519
+    clear({ args={ '--listen', 'test-name' } })
+    matches([[.*[/\\]test%-name[^/\\]*]], meths.get_vvar('servername'))
   end)
 end)
