@@ -1,3 +1,5 @@
+-- Tests for system() and :! shell.
+
 local helpers = require('test.functional.helpers')(after_each)
 
 local assert_alive = helpers.assert_alive
@@ -9,9 +11,9 @@ local command = helpers.command
 local insert = helpers.insert
 local expect = helpers.expect
 local exc_exec = helpers.exc_exec
-local iswin = helpers.iswin
 local os_kill = helpers.os_kill
 local pcall_err = helpers.pcall_err
+local is_os = helpers.is_os
 
 local Screen = require('test.functional.ui.screen')
 
@@ -79,13 +81,13 @@ describe('system()', function()
     end)
 
     it('calls executable in $PATH', function()
-      if 0 == eval("executable('python')") then pending("missing `python`") end
-      eq("foo\n", eval([[system(['python', '-c', 'print("foo")'])]]))
+      if 0 == eval("executable('python3')") then pending("missing `python3`") end
+      eq("foo\n", eval([[system(['python3', '-c', 'print("foo")'])]]))
       eq(0, eval('v:shell_error'))
     end)
 
     it('does NOT run in shell', function()
-      if iswin() then
+      if is_os('win') then
         eq("%PATH%\n", eval("system(['powershell', '-NoProfile', '-NoLogo', '-ExecutionPolicy', 'RemoteSigned', '-Command', 'Write-Output', '%PATH%'])"))
       else
         eq("* $PATH %PATH%\n", eval("system(['echo', '*', '$PATH', '%PATH%'])"))
@@ -94,7 +96,7 @@ describe('system()', function()
   end)
 
   it('sets v:shell_error', function()
-    if iswin() then
+    if is_os('win') then
       eval([[system("cmd.exe /c exit")]])
       eq(0, eval('v:shell_error'))
       eval([[system("cmd.exe /c exit 1")]])
@@ -123,7 +125,7 @@ describe('system()', function()
       screen:attach()
     end)
 
-    if iswin() then
+    if is_os('win') then
       local function test_more()
         eq('root = true', eval([[get(split(system('"more" ".editorconfig"'), "\n"), 0, '')]]))
       end
@@ -175,7 +177,7 @@ describe('system()', function()
 
     it('powershell w/ UTF-8 text #13713', function()
       if not helpers.has_powershell() then
-        pending("not tested; powershell was not found", function() end)
+        pending("powershell not found", function() end)
         return
       end
       helpers.set_shell_powershell()
@@ -184,7 +186,7 @@ describe('system()', function()
       -- * on Windows, expected to default to Western European enc
       -- * on Linux, expected to default to UTF8
       command([[let &shellcmdflag = '-NoLogo -NoProfile -ExecutionPolicy RemoteSigned -Command ']])
-      eq(iswin() and '??\n' or 'ああ\n', eval([[system('Write-Output "ああ"')]]))
+      eq(is_os('win') and '??\n' or 'ああ\n', eval([[system('Write-Output "ああ"')]]))
     end)
 
     it('`echo` and waits for its return', function()
@@ -213,7 +215,7 @@ describe('system()', function()
 
       screen:try_resize(72, 14)
       feed(':4verbose echo system("echo hi")<cr>')
-      if iswin() then
+      if is_os('win') then
         screen:expect{any=[[Executing command: "'fake_shell' 'cmdflag' '"echo hi"'"]]}
       else
         screen:expect{any=[[Executing command: "'fake_shell' 'cmdflag' 'echo hi'"]]}
@@ -243,7 +245,7 @@ describe('system()', function()
     end)
 
     it('`yes` interrupted with CTRL-C', function()
-      feed(':call system("' .. (iswin()
+      feed(':call system("' .. (is_os('win')
         and 'for /L %I in (1,0,2) do @echo y'
         or  'yes') .. '")<cr>')
       screen:expect([[
@@ -260,7 +262,7 @@ describe('system()', function()
         ~                                                    |
         ~                                                    |
         ~                                                    |
-]] .. (iswin()
+]] .. (is_os('win')
         and [[
         :call system("for /L %I in (1,0,2) do @echo y")      |]]
         or  [[
@@ -286,7 +288,7 @@ describe('system()', function()
 
     it('`yes` interrupted with mapped CTRL-C', function()
       command('nnoremap <C-C> i')
-      feed(':call system("' .. (iswin()
+      feed(':call system("' .. (is_os('win')
         and 'for /L %I in (1,0,2) do @echo y'
         or  'yes') .. '")<cr>')
       screen:expect([[
@@ -303,7 +305,7 @@ describe('system()', function()
         ~                                                    |
         ~                                                    |
         ~                                                    |
-]] .. (iswin()
+]] .. (is_os('win')
         and [[
         :call system("for /L %I in (1,0,2) do @echo y")      |]]
         or  [[
@@ -330,7 +332,7 @@ describe('system()', function()
 
   describe('passing no input', function()
     it('returns the program output', function()
-      if iswin() then
+      if is_os('win') then
         eq("echoed\n", eval('system("echo echoed")'))
       else
         eq("echoed", eval('system("echo -n echoed")'))
@@ -438,7 +440,7 @@ describe('systemlist()', function()
   before_each(clear)
 
   it('sets v:shell_error', function()
-    if iswin() then
+    if is_os('win') then
       eval([[systemlist("cmd.exe /c exit")]])
       eq(0, eval('v:shell_error'))
       eval([[systemlist("cmd.exe /c exit 1")]])
@@ -613,16 +615,16 @@ describe('systemlist()', function()
 
   it('powershell w/ UTF-8 text #13713', function()
     if not helpers.has_powershell() then
-      pending("not tested; powershell was not found", function() end)
+      pending("powershell not found", function() end)
       return
     end
     helpers.set_shell_powershell()
-    eq({iswin() and 'あ\r' or 'あ'}, eval([[systemlist('Write-Output あ')]]))
+    eq({is_os('win') and 'あ\r' or 'あ'}, eval([[systemlist('Write-Output あ')]]))
     -- Sanity test w/ default encoding
     -- * on Windows, expected to default to Western European enc
     -- * on Linux, expected to default to UTF8
     command([[let &shellcmdflag = '-NoLogo -NoProfile -ExecutionPolicy RemoteSigned -Command ']])
-    eq({iswin() and '?\r' or 'あ'}, eval([[systemlist('Write-Output あ')]]))
+    eq({is_os('win') and '?\r' or 'あ'}, eval([[systemlist('Write-Output あ')]]))
   end)
 
 end)
@@ -630,7 +632,7 @@ end)
 describe('shell :!', function()
   before_each(clear)
 
-  it(':{range}! with powershell filter/redirect #16271', function()
+  it(':{range}! with powershell filter/redirect #16271 #19250', function()
     local screen = Screen.new(500, 8)
     screen:attach()
     local found = helpers.set_shell_powershell(true)
@@ -639,18 +641,54 @@ describe('shell :!', function()
       1
       4
       2]])
-    feed(':4verbose %!sort<cr>')
-    screen:expect{
-      any=[[Executing command: .?Start%-Process sort %-RedirectStandardInput .* %-RedirectStandardOutput .* %-NoNewWindow %-Wait]]
-    }
+    if is_os('win') then
+      feed(':4verbose %!sort /R<cr>')
+      screen:expect{
+        any=[[Executing command: .?& { Get%-Content .* | & sort /R } 2>&1 | %%{ "$_" } | Out%-File .*; exit $LastExitCode"]]
+      }
+    else
+      feed(':4verbose %!sort -r<cr>')
+      screen:expect{
+        any=[[Executing command: .?& { Get%-Content .* | & sort %-r } 2>&1 | %%{ "$_" } | Out%-File .*; exit $LastExitCode"]]
+      }
+    end
     feed('<CR>')
     if found then
       -- Not using fake powershell, so we can test the result.
       expect([[
-        1
-        2
+        4
         3
-        4]])
+        2
+        1]])
     end
   end)
+
+  it(':{range}! without redirecting to buffer', function()
+    local screen = Screen.new(500, 10)
+    screen:attach()
+    insert([[
+      3
+      1
+      4
+      2]])
+    feed(':4verbose %w !sort<cr>')
+    if is_os('win') then
+      screen:expect{
+        any=[[Executing command: .?sort %< .*]]
+      }
+    else
+      screen:expect{
+        any=[[Executing command: .?%(sort%) %< .*]]
+
+      }
+    end
+    feed('<CR>')
+    helpers.set_shell_powershell(true)
+    feed(':4verbose %w !sort<cr>')
+    screen:expect{
+      any=[[Executing command: .?& { Get%-Content .* | & sort }]]
+    }
+    feed('<CR>')
+    helpers.expect_exit(command, 'qall!')
+  end)
 end)
